/*******************************************************************************
 * Licensed Materials - Property of IBM
 * © Copyright IBM Corporation 2005, 2015. All Rights Reserved. 
 * 
 * Note to U.S. Government Users Restricted Rights:  Use, 
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.filesystem.client;

import java.io.File;
import java.io.IOException;
import java.util.Map;

import com.ibm.team.filesystem.client.internal.EnvironmentVariables;
import com.ibm.team.filesystem.client.internal.ProxyAuthenticator;
import com.ibm.team.filesystem.client.internal.SharingManager;
import com.ibm.team.filesystem.client.internal.componenthierarchy.ComponentHierarchyManager;
import com.ibm.team.filesystem.client.internal.copyfileareas.ICopyFileAreaManager;
import com.ibm.team.repository.client.ITeamRepository;
import com.ibm.team.repository.client.TeamPlatform;

/**
 * The central client class of the Filesystem client component.
 * <p>
 * The Filesystem client builds on the SCM component and provides clients access to 
 * file based SCM related artifacts stored in the Jazz repository. This includes APIs for 
 * storing and fetching files from the repository and performant operations which keep 
 * a repository workspace in-sync with a file area on a users local machine.
 * </p><p>
 * This class cannot be instantiated or subclassed by clients; all functionality is 
 * provided by static methods.
 * </p>
 * @since 0.6
 */
public class FileSystemCore {
    public static final String HOME_CONFIG_DIRECTORY = ".jazz-scm"; //$NON-NLS-1$

    /**
     * Name of the directory created in {@link EnvironmentVariables#LOCAL_APP_DATA} 
     * for the daemon registry. 
     */
    public static final String SCM_DIRECTORY = "jazz-scm"; //$NON-NLS-1$
    
    /**
     * The plug-in id as defined in the MANIFEST.MF file.
     */
    public static final String ID = "com.ibm.team.filesystem.client"; //$NON-NLS-1$
    
    /**
     * Extension point name for work items removers.
     */
    public static final String WORK_ITEMS_LINKS_REMOVER = "workItemLinksRemover"; //$NON-NLS-1$

    /**
     * The system property that contains the build ID of our daemon. 
     */
    public static final String DAEMON_BUILD_ID_PROPERTY = "scm.daemon.buildId"; //$NON-NLS-1$
    
    /**
     * The directory containing user-specific configuration.
     * 
     *  @deprecated Use {@link #getUserConfigDir()}. 
     */
    public static final String USER_DIR = System.getProperty("user.home") + File.separator + HOME_CONFIG_DIRECTORY;  //$NON-NLS-1$
    
    /**
     * Provide access to the user's configuration directory, and how the path
     * was calculated. The configuration directory includes (at a minimum) the 
     * daemon registry that the process is using. In the command line 
     * environment, it will include other files. 
     */
    public static final class UserConfigDir {
        private final boolean isUsingAppData;
        private final File configDirAsFile;
        
        public UserConfigDir(File canonicalConfigDir, boolean localAppDataSet) {
        	this.configDirAsFile = canonicalConfigDir;
            this.isUsingAppData = localAppDataSet;
        }
        
        /**
         * @return the user's configuration directory.  
         */
        public String getConfigDir() {
            return configDirAsFile.getAbsolutePath();
        }
        
        public File getConfigDirAsFile() {
            return configDirAsFile;
        }
        
        /**
         * @return <code>true</code> if the LOCALAPPDATA environment variable 
         *      is the source of the location, <code>false</code> otherwise. 
         */
        public boolean isUsingLocalAppData() {
            return isUsingAppData;
        }
    }
    
    public static UserConfigDir getUserConfigDir() throws IOException {
        return getUserConfigDir(System.getenv());
    }
    
    /**
     * @return The directory the user wants their config data to be written to.
     */
    public static UserConfigDir getUserConfigDir(Map<String, String> env) throws IOException {
    	String toReturn = null;
    	boolean isUsingAppData = false;
    	
        // Pull the Windows local location. 
        {
            String localAppData = env.get(EnvironmentVariables.LOCAL_APP_DATA);
            if (localAppData != null) {
                localAppData = localAppData.trim();
                
                if (localAppData.length() > 0) {
                    toReturn = localAppData + File.separator + SCM_DIRECTORY;
                    isUsingAppData = true;
                }
            }
        }
        
        // See if the user is trying to override their config path
        {
            String reqPath = env.get(EnvironmentVariables.SCM_CONFIG_DIRECTORY);
            if (reqPath != null) {
                reqPath = reqPath.trim();
                
                if (reqPath.length() > 0) {
                    toReturn = reqPath;
                    isUsingAppData = false;
                }
            }
        }
        
        // Default to the user's home directory. 
        if (toReturn == null) {
            toReturn = System.getProperty("user.home") + File.separator + HOME_CONFIG_DIRECTORY;  //$NON-NLS-1$
            isUsingAppData = false;
        }
        
        File file = new File(toReturn);
        
        return new UserConfigDir(file.getCanonicalFile(), isUsingAppData);
    }
    
    /**
     * Returns the sharing manager.  Never <code>null</code>.
     * @return the {@link ISharingManager}
     */
    public static ISharingManager getSharingManager() {
        return SharingManager.getInstance();
    }
    
    /**
     * Returns the file system manager.  Never <code>null</code>.
     * @return the {@link IFileSystemManager}
     */
    public static IFileSystemManager getFileSystemManager(ITeamRepository repository) {
        if (repository == null) {
            throw new IllegalArgumentException();
        }
        
        return (IFileSystemManager) repository.getClientLibrary(IFileSystemManager.class);
    }
    
    /**
     * Returns the file system content manager. Never <code>null</code>.
     * @since 0.7
     */
    public static IFileContentManager getContentManager(ITeamRepository repo) {
        return getFileSystemManager(repo).getContentManager();
    }

    /**
     * Returns a class that provides API related to the "Component Hierarchy" feature.
     *  
     * @return the {@link IComponentHierarchyManager}.
     * 
     * @since 0.11.0 (RTC 6.0)
     */
    public static IComponentHierarchyManager getComponentHierarchyManager() {
        return ComponentHierarchyManager.getInstance();
    }

    /*
     * Flag indicating that the filesystem core should not longer provide 
     * change tracking. When we are shutdown, no change tracking will be 
     * provided.
     */
    private static boolean isShutDown = false;

    private static ProxyAuthenticator headlessAuthenticator;

    /**
     * May be called after {@link #shutDown()} to resume change tracking and 
     * other Jazz filesystem features. Does not need to be called when the 
     * filesystem is initially  started. 
     */
    public static void startUp() {
        isShutDown = false;
    }
    
    /**
     * Must be explicitly called by plain java clients (eg, clients not running as an 
     * Eclipse plug-in) to shutdown Filesystem prior to termination. The shutdown ensures 
     * that metadata associated with the {@link ISharingManager} is stored and closed.
     * 
     * @see {@link TeamPlatform#shutdown()}
     * @throws FileSystemException if an error occurs during shutdown.
     */
    public static void shutDown() throws FileSystemException {
        isShutDown = true;
        SharingManager.conditionalShutDown();
        ICopyFileAreaManager.instance.shutdown(null);
    }
    
    /**
     * Returns <code>true</code> if the core has been {@link #shutDown()}. After 
     * shutdown, {@link ISharingManager} instances will no longer track 
     * changes to local files.
     */
    public static boolean isShutDown() {
        return isShutDown;
    }
    
    // Not meant to be instantiated
    private FileSystemCore() {       
    }

    /**
     * Enable the headless proxy authentication support
     */
    public static void enableHeadlessProxies() {
        headlessAuthenticator = new ProxyAuthenticator();
    }
    
    /**
     * API that can be used by a headless client (e.g. CLI/Daemon) to ensure that proxy authentication can occur.
     * @param url the url being proxied
     * @param proxyHost the proxy host
     * @param proxyPort the proxy port
     * @param proxyUserName the proxy user name
     * @param proxyPassword the proxy password
     */
    public static void addProxyData(String url, String proxyHost, int proxyPort, String proxyUserName, String proxyPassword) {
        if (headlessAuthenticator != null) {
            headlessAuthenticator.addProxyData(url, proxyHost, proxyPort, proxyUserName, proxyPassword);
        } 
    }
}
