/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2005, 2012. All Rights Reserved. 
 * 
 * Note to U.S. Government Users Restricted Rights:  Use, 
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.filesystem.client;

import java.io.PrintStream;
import java.io.PrintWriter;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.MultiStatus;
import org.eclipse.core.runtime.Status;

import com.ibm.team.filesystem.client.internal.Messages;
import com.ibm.team.repository.common.util.NLS;

/**
 * Collection of failures happening in the client file system
 * integration API.  The failures are within a IStatus object
 * <p>
 * @since 3.0
 * @subclassNO Not intended to be subclassed by clients.
 */
public class FileSystemStatusException extends FileSystemException {
    private static final long serialVersionUID = 1L;

    /**
     * Create a FileSystemStatusException that wraps a CoreException
     * The status from the CoreException will be contained within this exception.
     * @param message Message for the exception.  If <code>null</code> the message from the CoreException will be used.
     * @param exception The CoreException to wrap.  Never <code>null</code>
     * @return FileSystemStatusException containing the status from the CoreException cause.
     */
    public static FileSystemStatusException fromCoreException(String message, CoreException exception) {
        if (message == null) {
            message = exception.getMessage();
        }
        return new FileSystemStatusException(new MultiStatus(FileSystemCore.ID, 0, new IStatus[] {exception.getStatus()},
                message, exception));
    }

    public FileSystemStatusException(IStatus status) {
        super(status.getMessage());
        // we store the status in the data so it is accessible to
        // classes (like StatusUtil) that don't have access to this
        // class.
        setExtraData(EXCEPTION_DATA, status);
    }

    public FileSystemStatusException(int severity, String pluginID, String message, Throwable cause) {
        super(message);
        IStatus status = new Status(severity, pluginID, message, this);
        setExtraData(EXCEPTION_DATA, status);
    }
    
    public FileSystemStatusException(Object origin, IStatus status) {
        super(origin, status.getMessage());
        setExtraData(EXCEPTION_DATA, status);
    }
    
    public FileSystemStatusException(String message, Throwable cause) {
        super(message, cause);
        IStatus status = new Status(IStatus.ERROR, FileSystemCore.ID, message, cause);
        setExtraData(EXCEPTION_DATA, status);
    }
    
    public Throwable getCause() {
        return super.getCause() == null ? getStatus().getException() : super.getCause();
    }
    
    @Override
    public String toString() {
        return NLS.bind(Messages.FileSystemClientException_0, new Object[] { getClass().getName(), getStatus().toString() });
    }

    /**
     * Returns the status object for this exception.
     *
     * @return a status object
     */
    public final IStatus getStatus() {
        Object data = getExtraData(EXCEPTION_DATA);
        if (data instanceof IStatus) {
            return (IStatus) data;
        }
        return null;
    }

    /**
     * Prints a stack trace out for the exception, and
     * any nested exception that it may have embedded in
     * its Status object.
     */
    public void printStackTrace() {
        printStackTrace(System.err);
    }

    /**
     * Prints a stack trace out for the exception, and
     * any nested exception that it may have embedded in
     * its Status object.
     * 
     * @param output the stream to write to
     */
    public void printStackTrace(PrintStream output) {
        synchronized (output) {
            super.printStackTrace(output);
            IStatus status = getStatus();
            if (status.getException() != null && status.getException() != getCause()) {
                output.print(getClass().getName() + " : " + status.getMessage() + " [" + status.getCode() + "]"); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
                status.getException().printStackTrace(output);
            }
            printChildrenStatus("Contains : ", status, output); //$NON-NLS-1$
        }
    }

    private void printChildrenStatus(String level, IStatus status, PrintStream output) {
        IStatus[] children = status.getChildren();
        if (children == null || children.length == 0)
            return;
        for (int i = 0; i < children.length; i++) {
            String prefix = level + i;
            output.println(prefix + " "  + children[i].getMessage()); //$NON-NLS-1$
            Throwable exception = children[i].getException();
            if (exception != null) {
                exception.printStackTrace(output);
            }
            output.flush();
            printChildrenStatus(prefix + ".", children[i], output); //$NON-NLS-1$
       }
    }

    /**
     * Prints a stack trace out for the exception, and
     * any nested exception that it may have embedded in
     * its Status object.
     * 
     * @param output the stream to write to
     */
    public void printStackTrace(PrintWriter output) {
        synchronized (output) {
            super.printStackTrace(output);
            IStatus status = getStatus();
            if (status.getException() != null) {
                output.print(getClass().getName() + "[" + status.getCode() + "]: "); //$NON-NLS-1$ //$NON-NLS-2$
                status.getException().printStackTrace(output);
            }
            printChildrenStatus("Contains : ", status, output); //$NON-NLS-1$
        }
    }
    
    private void printChildrenStatus(String level, IStatus status, PrintWriter output) {
        IStatus[] children = status.getChildren();
        if (children == null || children.length == 0)
            return;
        for (int i = 0; i < children.length; i++) {
            String prefix = level + i;
            output.println(prefix + " "  + children[i].getMessage()); //$NON-NLS-1$
            Throwable exception = children[i].getException();
            if (exception != null) {
                exception.printStackTrace(output);
            }
            output.flush();
            printChildrenStatus(prefix + ".", children[i], output); //$NON-NLS-1$
       }
    }

    /**
     * Set the status associated with the receiver. Should not be called 
     * outside of serializers. 
     */
    public void setStatus(IStatus status) {
        setExtraData(EXCEPTION_DATA, status);
    }
}
