/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010, 2012. All Rights Reserved. 
 * 
 * Note to U.S. Government Users Restricted Rights:  Use, 
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.filesystem.client;

import java.util.Collection;

import com.ibm.team.filesystem.client.operations.IFileSystemOperation;
import com.ibm.team.filesystem.common.ILogicalChange;
import com.ibm.team.filesystem.common.ILogicalConflict;
import com.ibm.team.filesystem.common.ILogicalConflictReport;
import com.ibm.team.scm.client.IWorkspaceConnection;
import com.ibm.team.scm.common.IFolderHandle;

/**
 * An operation that supports the resolution of conflicts by applying the
 * accepted changes (the proposed changes) to the workspace configuration.
 * 
 * The specification of the conflicts to be overridden so that the
 * proposed change is selected.
 * 
 * The result is the changes are applied.  If any of the changes are
 * logical conflicts then the problem that prevented the change from 
 * being applied is undone prior to the change being applied.
 * For example:
 * Mike adds src/a.java
 * Zoe adds src/a.java
 * Zoe accepts Mike's change.  She has an add-add conflict on src/a.java
 * The override and update will result in her a.java being deleted and Mike's added.
 * 
 * There are restrictions on the changes that must be overridden and updated together.
 * The are 2 categories of changes/conflicts content and structural.  The content
 * ones are {@link ILogicalConflict#MODIFY_MODIFY}.  We should not have changes that 
 * are unapplied modifies, only the conflict.  All other changes (add, delete, move
 * and the conflicts that go with them) fall into the structural category.  The 
 * {@link ILogicalConflict#MODIFY_MODIFY} changes can be done individually or as
 * a part of a larger group of changes.  The structural changes/conflicts for
 * a component must all be overridden and updated together.
 * 
 * Some conflicts can not just be Overridden.  For example, {@link ILogicalConflict#MODIFY_DELETE}
 * requires that the location the modified content is to be restored to be specified.
 * This can be specified by {@link #setParentForResolution(ILogicalConflict, IFolderHandle, String)}
 * 
 * @since 3.0
 * @NoImplement This interface is not intended to be implemented by clients.
 */
public interface IApplyAcceptedOperation extends IFileSystemOperation {

    /**
     * Set the context of the operation.  That is, what workspace the conflicts
     * are to be resolved by applying the accepted/proposed changes.
     * 
     * @param connection The workspace containing the conflicts to be resolved.
     * Never <code>null</code>
     * 
     * @param conflictReport The conflict report that contains the logicalChanges
     * supplied for overriding.
     * Not <code>null</code>
     */
    public void setContext(IWorkspaceConnection connection, ILogicalConflictReport conflictReport);

    /**
     * Add a collection of changes to be applied.  It is possible
     * to configure the operation with a set of changes to resolve
     * with the proposed (accepted). Then discover additional changes
     * must also be applied and add those additional changes.
     * 
     * @param changes  The logical changes to be applied and the conflicts to be
     * resolved by overriding the current configuration with the unapplied change.  The
     * changes/conflicts must have been retrieved from the conflict report supplied
     * in {@link #setContext(IWorkspaceConnection, ILogicalConflictReport)}
     * Not <code>null</code>
     * @throws FileSystemException If the changes can not be resolved (for example unknown conflicts
     * or missing details on the change)
     */
    public void addChangesToResolve(Collection<? extends ILogicalChange> moreChanges) throws FileSystemException;

    /**
     * When resolving some of the logical conflicts that involve deletion of the
     * parent additional information must be supplied to describe the parent of the
     * item to use when resolving the conflict.
     * <p>
     * For example: <br>
     * {@link ILogicalConflict#MODIFY_DELETE} a file has been modified but it (and
     * potentially its parent folder) has been deleted from the configuration.  In order
     * to override the delete and have the modified file back in the configuration, you
     * must supply where to re-add the modified file.
     * <p>
     * As another example.  Consider {@link ILogicalConflict#ADD_DELETE} a file/folder is
     * being added to a parent nolonger in the configuration.  The new parent has to be
     * supplied.
     * <p> 
     * TODO expand the cases supported to include ADD-ADD (allow a different parent/name to 
     * be specified) Move-Move. 
     * @param conflict the conflict to be resolved by re-instating the item.  Only the 
     * following conflicts types are valid.  Not <code>null</code>.
     * {@link ILogicalConflict#ADD_DELETE}
     * {@link ILogicalConflict#MODIFY_DELETE}
     * {@link ILogicalConflict#MOVE_DELETE}
     * A conflict may only be specified once for the operation. 
     *  
     * @param parentFolder the parent of the item re-instated into the configuration.  
     * Not <code>null</code>.  Parent must exist in the current configuration and not
     * about to be deleted via the override of the other changes. 
     * @param name the name to be given to the item being re-instated into the configuration.
     * Not <code>null</code> or empty string.  Another item must not already exist with that name (or if
     * it is it is going to be deleted via the override of the other changes).
     */
    public void setParentForResolution(ILogicalConflict conflict, IFolderHandle parentFolder, String name);
    

    /**
     * @see #setParentForResolution(ILogicalConflict, IFolderHandle, String)
     * This is the list of changes to be overridden and updated, that currently do not have
     * parent locations specified for them.  Moves and Adds will support other locations and 
     * are included in the list but do not require them.
     * 
     * TODO This will include the Moves and Adds that were blocked by existing content which
     * could also just be deleted.  Do we need a separate interface to distinguish these changes
     * from those that could also be resolved by deleting the content in place?
     * 
     * @return a collection of {@link ILogicalChange}
     */
    public Collection<ILogicalConflict> needParentForResolution();

    /**
     * The conflicts returned can not be resolved until the content in the way is
     * deleted first.  This will be done by the operation.  If you have chosen to
     * resolve it by loading into another location then the change will not be
     * returned.
     * 
     * @return a collection of {@link ILogicalChange}
     */
    public Collection<ILogicalChange> needContentToRemoved();

    /**
     * A group of changes have been specified to be resolved.  The changes
     * may depend upon other changes to also be resolved before they can be
     * resolved.  This will provide the changes that must also be resolved.
     * @return a collection of logical changes which have not been specified
     * to be resolved, but must be resolved in order for the changes so
     * far specified to be able to be resolved.
     */
    public Collection<ILogicalChange> getRequiredChanges();
    
    /**
     * Not all changes may be resolved. This method returns the number of
     *  changes that were actually resolved.
     * 
     * @return The number of changes that were resolved if the operation is run. Otherwise,
     *  the return value is undefined.
     */
    public int numChangesResolved();
}
