/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2005, 2012. All Rights Reserved. 
 * 
 * Note to U.S. Government Users Restricted Rights:  Use, 
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.filesystem.client;

import org.eclipse.core.runtime.IPath;

import com.ibm.team.scm.common.IVersionableHandle;

/**
 * An Event that describes what recently occurred to the files and folders shared
 * in the copy file area mirror.
 * 
 * @see {@link ISharingManager#addListener(ICopyFileAreaListener)
 * @see {@link ISharingManager#removeListener(ICopyFileAreaListener)
 * @see {@link ICopyFileAreaListener}
 * 
 * @since 0.6
 * @NoImplement This interface is not intended to be implemented by clients.
 */
public interface ICopyFileAreaEvent {

    /**
     * File area event reason constant indicating that a project has been
     * added which is shared with the repository.
     * 
     * @see #getReason()
     */
    public static final int SHARING_ADDED = 1;

    /**
     * File area event reason constant indicating that a project has been
     * disconnected from the repository. This doesn't meant that the project
     * has been deleted from the repository, but instead that the client
     * is no longer tracking the connections.
     * 
     * @see #getReason()
     */
    public static final int SHARING_REMOVED = 2;

    /**
     * File area event reason constant indicating that a shared project has
     * had it's sharing metadata updated. See {@link ISharingManager#allShares(IPath)}.
     * 
     * @see #getReason()
     */
    public static final int SHARING_STATE_UPDATED = 6;

    /**
     * File area event reason constant indicating that a component has been loaded 
     * (separate events will be fired for each share added due to the load).
     */
    public static final int COMPONENT_LOADED = 7;

    /**
     * File area event reason constant indicating that a component has been unloaded 
     * (separate events will be fired for each share removed due to the unload).
     */
    public static final int COMPONENT_UNLOADED = 8;
    
    /**
     * File area event reason constant indicating that a Copy File are has been registered.
     * (separate events will be fired for each component loaded due to the registration)
     * (separate events will be fired for each share added due to the component load).
     */
    public static final int COPY_FILE_AREA_REGISTERED = 9;
    
    /**
     * File area event reason constant indicating that a Copy File are has been deregistered.
     * (separate events will be fired for each component unloaded due to the deregistration)
     * (separate events will be fired for each share removed due to the component unload).
     */
    public static final int COPY_FILE_AREA_DEREGISTERED = 10;
    
    /**
     * Sandbox listening has been enabled for all copy file areas. No other information is provided
     *  other than the reason.
     */
    public static final int SANDBOX_LISTENING_ENABLED = 11;
    
    /**
     * Sandbox listening has been disabled for all copy file areas. No other information is provided
     *  other than the reason.
     */
    public static final int SANDBOX_LISTENING_DISABLED = 12;
    
    /**
     * Returns the reason for this event.
     * 
     * @return the reason for this event.
     */
    public int getReason();

    /**
     * @return the copy file area root path of the copy file area affected
     * by the event.
     */
    public ILocation getCopyFileAreaRoot();
    
    /**
     * Returns the copy file area root relative path of the resource affected
     * by this event.
     * 
     * @return the copy file area root relative path of the resource affected
     * by this event.
     * Will be <code>null</code> for 
     * {@link #COPY_FILE_AREA_REGISTERED}, {@link #COPY_FILE_AREA_DEREGISTERED},
     * {@link #COMPONENT_LOADED} and {@link #COMPONENT_UNLOADED} events.
     * events
     */
    public IRelativeLocation getPath();
    
    /**
     * Returns the versionable affected by this event.
     * 
     * @return the versionable affected by this event.
     * Will be <code>null</code> for 
     * {@link #COPY_FILE_AREA_REGISTERED}, {@link #COPY_FILE_AREA_DEREGISTERED},
     * {@link #COMPONENT_LOADED} and {@link #COMPONENT_UNLOADED} events.
     */
    public IVersionableHandle getVersionable();
}
