/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2005, 2012. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.filesystem.client;

import java.io.InputStream;
import java.io.OutputStream;
import java.util.Collection;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.OperationCanceledException;

import com.ibm.team.filesystem.common.FileLineDelimiter;
import com.ibm.team.filesystem.common.IFileContent;
import com.ibm.team.filesystem.common.IFileItemHandle;
import com.ibm.team.repository.client.ITeamRepository;
import com.ibm.team.repository.common.NotLoggedInException;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.scm.client.content.AbstractVersionedContentManagerInputStreamProvider;
import com.ibm.team.scm.common.ContentHash;
import com.ibm.team.scm.common.IChangeSetHandle;

/**
 * Methods for storing and retrieving FileItem contents from the repository.
 * The content manager is closely tied to its {@link ITeamRepository}, and only
 * works when the client is logged in to the repository.
 * <p>
 * The content is then stored in the repository in a two-stage process.
 * <ul>
 * <li>Stage 1: upload the content bytes. Upon
 * successful receiving and storing the uploaded content, the server puts the content
 * on a pending list until it is claimed.
 * <li>Stage 2: store the item state embedding the reference to the content by
 * id. The server checks the embedded references to ensure that the content
 * service really has content with that id. The
 * {@link com.ibm.team.scm.client.IWorkspaceConnection#commit(IChangeSetHandle, Collection)}
 * operation fails if the content service does not have it.
 *
 * To fetch the content:
 * <li>Stage 1: fetch the file item and retrieve the
 * {@link IFileContent} instance from the item. This
 * content description contains the identifier that can be used by the server to
 * identify the contents associated with this attribute.
 * <li>Stage 2: download the contents using
 * {@link #retrieveContent(IFileItemHandle, IFileContent, OutputStream, IProgressMonitor)} using the
 * content description from stage 1.
 * </ul>
 * 
 * </p>
 * @since 0.7
 */
public interface IFileContentManager {

    /**
     * Returns the repository associated with this manager. The client need not
     * be logged in to the repository.
     * 
     * @return the repository, which is never <code>null</code>
     * @ShortOp This is a short operation; it may block only momentarily; safe
     *          to call from a responsive thread.
     */
    public ITeamRepository teamRepository();

    /**
     * Uploads the contents to the server. Uploaded content must first be
     * claimed before it can be retrieved from the server. Claiming of
     * uploaded content is done by embedding the reference to the content
     * {@link IFileContent} within a file item
     * and then saving the item to the server with this new content reference.
     * The server checks the embedded references to ensure that the content
     * service really has content with that id. The
     * {@link com.ibm.team.scm.client.IWorkspaceConnection#commit(IChangeSetHandle, Collection)}
     * operation fails if the content cannot be claimed.
     * <p>
     * The operation has the following preconditions:
     * <ul>
     * <li>The client is logged in to the repository.</li>
     * </li>
     * </ul>
     * The operation fails by throwing <code>TeamRepositoryException</code> if
     * any of its preconditions are not met.
     * </p>

     * @param characterEncoding
     *            the character encoding of the bytes being sent, the format is
     *            specified in {@link IFileContent#getCharacterEncoding()} This value
     *            is used by the HTTP server; it is <em>not</em> verified, and
     *            providing an incorrect value for this <em>may</em> result in
     *            uncaught incorrect behavior.  The character encoding is used
     *            when performing validation/translation of line endings, so it
     *            should be an encoding available on the server.  It
     *            does not need to be supplied when storing binary data (i.e.
     *            {@link FileLineDelimiter#LINE_DELIMITER_NONE}).
     * @param lineDelimiter
     *            the line delimiter preferred when content is retrieved.
     *            With {@link FileLineDelimiter#LINE_DELIMITER_NONE} no conversion is done on
     *            line delimiters (i.e. treat as binary data).  With {@link FileLineDelimiter#LINE_DELIMITER_CR},
     *            {@link FileLineDelimiter#LINE_DELIMITER_CRLF}, {@link FileLineDelimiter#LINE_DELIMITER_LF}
     *            the content later retrieved will always have the line delimiter
     *            setting.  The stream supplied does not have to have same the line delimiter
     *            specified but it must have consistent line delimiters (that is if
     *            {@link FileLineDelimiter#LINE_DELIMITER_CRLF} is specified the stream can contain
     *            line feeds, but only line feeds as line delimiters.
     *            With {@link FileLineDelimiter#LINE_DELIMITER_PLATFORM} the content when
     *            retrieved will have the line delimiter preferred by the platform 
     *            (for example: Carriage return/linefeed on Windows platforms).
     *            The stream supplied must have consistent line delimiters.
     * @param streamProvider
     *            the input stream provider from which to read the bytes to send to the
     *            server. Must <em>not</em> be null.
     * @param predecessorHintHash
     *            the predecessor of the current content, i.e. the content it is
     *            derived from. null if none or unknown.
     * @param monitor
     *            a progress monitor, or <code>null</code> if progress
     *            reporting is not desired
     * @return the content identifier for the uploaded bytes
     * @throws NotLoggedInException
     *             if the operation fails because the client was not logged in
     *             to the repository
     * @throws TeamRepositoryException
     *             if the operation fails for any other reason
     * @throws OperationCanceledException
     *             if the operation is canceled. Cancellation can occur even if
     *             no progress monitor is provided.
     * @LongOp This is a long operation; it may block indefinitely; must not be
     *         called from a responsive thread.
     */
    public IFileContent storeContent(String characterEncoding,
            FileLineDelimiter lineDelimiter,
            AbstractVersionedContentManagerInputStreamProvider streamProvider,
            ContentHash predecessorHintHash, IProgressMonitor monitor)
            throws TeamRepositoryException;

    
    /**
     * Downloads the bytes specified by the content handle and writes them to
     * the specified output stream. The output stream is closed before the
     * method returns. Uploaded content must first be claimed before it can be
     * retrieved from the server. Claiming of uploaded content is done by
     * embedding the reference to the content {@link IFileContent} within a file
     * item and then saving the item to the server with this new content reference.
     * <p>
     * Since the content length is supplied by the server, a
     * <code>TeamRepositoryException</code> will thrown if both length values
     * do not match. A hash mismatch will also result in a
     * <code>TeamRepositoryException</code>.
     * </p>
     * <p>
     * The operation has the following preconditions:
     * <ul>
     * <li>The client is logged in to the repository.</li>
     * </li>
     * </ul>
     * The operation fails by throwing <code>TeamRepositoryException</code> if
     * any of its preconditions are not met.
     * </p>
     * 
     * @param item
     *            A state handle identifying the item state that contains the
     *            supplied content
     * @param content
     *            the content that identified the bytes to download. The content
     *            object must be exactly as retrieved from the server, and must
     *            <em>not</em> have been modified. Modification of any fields
     *            of an IFSContent object before retrieving the content stream
     *            <em>may</em> result in unpredictable behavior.
     * @param outStream
     *            the output stream into which the downloaded bytes are written;
     *            this stream will be closed on any return from this method,
     *            normal or exception.
     * @param monitor
     *            a progress monitor, or <code>null</code> if progress
     *            reporting is not desired
     * @throws NotLoggedInException
     *             if the operation failed because the client was not logged in
     *             to the repository
     * @throws TeamRepositoryException
     *             if the operation fails for any other reason
     * @throws OperationCanceledException
     *             if the operation is canceled. Cancellation can occur even if
     *             no progress monitor is provided.
     * @LongOp This is a long operation; it may block indefinitely; must not be
     *         called from a responsive thread.
     */
    public void retrieveContent(IFileItemHandle item, IFileContent content,
            OutputStream outStream, IProgressMonitor monitor) throws TeamRepositoryException;

    /**
     * Sets up a download connection to the server to retrieve the content
     * bytes, and returns an input stream to be used to read the content.
     * Uploaded content must first be claimed before it can be retrieved from
     * the server. Claiming of uploaded content is done by embedding the
     * reference to the content {@link IFileContent} within an item and then saving
     * the item to the server with this new content reference.
     * 
     * <p>
     * <b>This method should be used with caution by client code.</b> An HTTP
     * connection remains open to the server until the client closes the
     * returned input stream. Clients calling this method should immediately
     * retrieve all content from the stream, and close it as quickly as
     * possible.
     * </p>
     * 
     * <p>
     * The operation has the following preconditions:
     * <ul>
     * <li>The client is logged in to the repository.</li>
     * </li>
     * </ul>
     * The operation fails by throwing <code>TeamRepositoryException</code> if
     * any of its preconditions are not met.
     * </p>
     * 
     * @param item
     *            A state handle identifying the item state that contains the
     *            supplied content
     * @param content
     *            the content that identified the bytes to download. The content
     *            object must be exactly as retrieved from the server, and must
     *            <em>not</em> have been modified. Modification of any fields
     *            of an IVersionedContent object before retrieving the content stream
     *            <em>may</em> result in unpredictable behavior.
     * @param monitor
     *            a progress monitor, or <code>null</code> if progress
     *            reporting is not desired
     * @return an input stream from which the content bytes can be read by
     *         client code.
     * @throws NotLoggedInException
     *             if the operation failed because the client was not logged in
     *             to the repository
     * @throws TeamRepositoryException
     *             if the operation fails for any other reason
     * @throws OperationCanceledException
     *             if the operation is canceled. Cancellation can occur even if
     *             no progress monitor is provided.
     * @LongOp This is a long operation; it may block indefinitely; must not be
     *         called from a responsive thread.
     */
    public InputStream retrieveContentStream(IFileItemHandle item,
            IFileContent content, IProgressMonitor monitor) throws TeamRepositoryException;

    /**
     * Sets up a download connection to the server to retrieve the content
     * bytes, and returns an input stream to be used to read the content.
     * Uploaded content must first be claimed before it can be retrieved from
     * the server. Claiming of uploaded content is done by embedding the
     * reference to the content {@link IFileContent} within an item and then saving
     * the item to the server with this new content reference.
     * 
     * <p>
     * <b>This method should be used with caution by client code.</b> An HTTP
     * connection remains open to the server until the client closes the
     * returned input stream. Clients calling this method should immediately
     * retrieve all content from the stream, and close it as quickly as
     * possible.
     * </p>
     * 
     * <p>
     * The operation has the following preconditions:
     * <ul>
     * <li>The client is logged in to the repository.</li>
     * </li>
     * </ul>
     * The operation fails by throwing <code>TeamRepositoryException</code> if
     * any of its preconditions are not met.
     * </p>
     * 
     * @param item
     *            A state handle identifying the item state that contains the
     *            supplied content
     * @param content
     *            the content that identified the bytes to download. The content
     *            object must be exactly as retrieved from the server, and must
     *            <em>not</em> have been modified. Modification of any fields
     *            of an IFSContent object before retrieving the content stream
     *            <em>may</em> result in unpredictable behavior.
     * @param alternativeEncoding
     *            The character encoding to use when reading the input stream
     *            if performing any line delimiter conversion.  An encoding is
     *            not required if the alternativeDelimiter is {@link FileLineDelimiter#LINE_DELIMITER_NONE}
     *            The caller may choose to supply {@link IFileContent#getCharacterEncoding()}
     *            if there is no need to override.  If the original content had a line delimiter
     *            {@link FileLineDelimiter#LINE_DELIMITER_NONE}, the content may not have an encoding though.
     *            The content is not converted from one encoding to another, it is an override.
     *            Incompatible specification may result in corrupt content returned. 
     * @param alternativeDelimiter
     *            The preferred delimiter of the InputStream.  This will force
     *            the content retrieved to have the line delimiter specified
     *            regardless of the setting specified by content and whether
     *            the line delimiters are consistent.  That is a binary file could
     *            be retrieved with a line delimiter of {@link FileLineDelimiter#LINE_DELIMITER_LF}
     * @param monitor
     *            a progress monitor, or <code>null</code> if progress
     *            reporting is not desired
     * @return an input stream from which the content bytes can be read by
     *         client code.
     * @throws NotLoggedInException
     *             if the operation failed because the client was not logged in
     *             to the repository
     * @throws TeamRepositoryException
     *             if the operation fails for any other reason
     * @throws OperationCanceledException
     *             if the operation is canceled. Cancellation can occur even if
     *             no progress monitor is provided.
     * @LongOp This is a long operation; it may block indefinitely; must not be
     *         called from a responsive thread.
     */
    public InputStream retrieveContentStream(IFileItemHandle item,
            IFileContent content, String alternativeEncoding,
            FileLineDelimiter alternativeDelimiter,
            IProgressMonitor monitor) throws TeamRepositoryException;
    
    /**
     * Creates a content transfer session for asynchronous operation. The
     * returned object allows the caller to perform content transfer on threads
     * other than the caller.
     * 
     * @param sessionName
     *            The name of the session. Passed to
     *            {@link IProgressMonitor#beginTask(String, int)}.
     * 
     * @param allowFailures Whether downloads should continue with other contents
     *            when a failure has been encountered.  <code>true</code> to continue,
     *            <code>false</code> end when a failure is encountered.
     *            
     * @param totalWork
     *            The number of contents to be transferred. May be
     *            <code>IProgressMonitor.UNKNOWN</code> if the transfer size
     *            is unknown.
     * 
     * @param monitor
     *            the progress monitor that tracks how far along the session is
     *            in its set of aggregated downloads.
     * 
     * @ShortOp This is a short operation. It may be called from a responsive
     *          thread.
     */
    public IFileContentManagerSession createSession(String sessionName, boolean allowFailures,
            long totalWork, IProgressMonitor monitor);
    
    /**
     * Creates a content transfer session for asynchronous operation. The
     * returned object allows the caller to perform content transfer on threads
     * other than the caller.
     * 
     * @param maxNumberThreads The maximum number of threads to be used when performing
     *          the content transfer.
     *          
     * @param sessionName
     *            The name of the session. Passed to
     *            {@link IProgressMonitor#beginTask(String, int)}.
     * 
     * @param allowFailures Whether downloads should continue with other contents
     *            when a failure has been encountered.  <code>true</code> to continue,
     *            <code>false</code> end when a failure is encountered.
     *            
     * @param totalWork
     *            The number of contents to be transferred. May be
     *            <code>IProgressMonitor.UNKNOWN</code> if the transfer size
     *            is unknown.
     * 
     * @param monitor
     *            the progress monitor that tracks how far along the session is
     *            in its set of aggregated downloads.
     * 
     * @ShortOp This is a short operation. It may be called from a responsive
     *          thread.
     */
    public IFileContentManagerSession createSession(int maxNumberThreads, boolean allowFailures, String sessionName,
            long totalWork, IProgressMonitor monitor);
    
    /**
     * When content is retrieved, it may require conversion to the platform
     * line delimiters.  If a content will be converted, then use 
     * {@link IFileContent#getEstimatedConvertedLength()} and recalculate the 
     * hash code.
     * @return <code>true</code> if the content will be converted on
     * retrieval.  <code>false</code> the content retrieved from the 
     * repository is untouched.
     */
    public boolean convertDelimitersDuringRetrieval(IFileContent content);
    
    /**
     * Returns the cleanup period for unclaimed content in seconds. This is the minimum length
     * of time that unclaimed content will be kept on the server
     * 
     * @LongOp This is a long operation; it may block indefinitely; must not be
     *         called from a responsive thread.
     */
    public long getUnclaimedContentCleanupPeriod(IProgressMonitor monitor) throws TeamRepositoryException;
    
    /**
     * Deletes the specified content in the repository. The given content must exist in the repository.
     * 
     * @param hash
     *      The hash of the content to delete
     * @param force
     *      When false, delete will fail with VersionedContentClaimedByMultipleItems if more than 1 item
     *      claims the content.
     *      When true, delete will not check how many items claim the content.
     *
     * @param monitor
     *            a progress monitor, or <code>null</code> if progress
     *            reporting is not desired
     *            
     * @throws NotLoggedInException
     *             if the operation failed because the client was not logged in
     *             to the repository
     * @throws TeamRepositoryException
     *             if the operation fails for any other reason
     * @throws OperationCanceledException
     *             if the operation is canceled. Cancellation can occur even if
     *             no progress monitor is provided.
     * @LongOp This is a long operation; it may block indefinitely; must not be
     *         called from a responsive thread.
     */
    public void deleteContent(ContentHash hash, boolean force, IProgressMonitor monitor) throws TeamRepositoryException;
    
    //This should go away when everyone adopts the method above
    @Deprecated
    public void deleteContent(ContentHash hash, IProgressMonitor monitor) throws TeamRepositoryException;
}
