/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2005, 2009. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.filesystem.client;

import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.OperationCanceledException;

import com.ibm.team.filesystem.common.FileLineDelimiter;
import com.ibm.team.filesystem.common.IFileContent;
import com.ibm.team.filesystem.common.IFileItemHandle;
import com.ibm.team.repository.common.TeamRepositoryException;

/**
 * Encapsulates a file transfer session for the {@link IFileContentManager}. A
 * session is a group of zero or more downloads/uploads that are logically related.
 * Downloads are performed with
 * {@link #retrieveContent(IFileContent, FileDownloadHandler)}, which takes an
 * {@link FileDownloadHandler} as an argument. The <code>FSDownloadHandler</code>
 * receives the download (possibly in another thread).
 * 
 * Uploads are performed with
 * {@link #storeContent(FileUploadHandler)}, which takes an
 * {@link FileUploadHandler} as an argument. The <code>FSUploadHandler</code>
 * receives the resulting IFSContent (possibly in another thread).
 * 
 * <p/> If one download or upload is canceled, then implementors should cause
 * all downloads and uploads to be canceled. Any subsequent call to one of
 * {@link IFileContentManagerSession}'s methods will result in an
 * <code>OperationCanceledException</code>.
 * 
 * <p/> Sample usage:
 * 
 * <pre>
 * IFSContentManager manager;
 * IFSContent content1, content2;
 * IFileItemHandle v1, v2;
 * 
 * IFSContentManagerSession session = manager.createSession(&quot;Sample download&quot;, 2,
 * 		null);
 * 
 * FSDownloadHandler dl = new FSDownloadHandler() {
 * 	public void downloadCompleted(IFileItemHandle, IFSContent content) {
 * 		System.out.println(&quot;Download completed&quot;);
 * 	}
 * };
 * 
 * session.retrieveContent(v1, content1, dl);
 * session.retrieveContent(v2, content2, dl);
 * 
 * session.join();
 * 
 * session.getErrorStatus();
 * </pre>
 * @since 0.7
 */
public interface IFileContentManagerSession {

	/**
	 * Add a new download request to the receiver.
	 * @param handle
	 *            A state handle identifying the item state that contains the
	 *            supplied content
	 * @param content
	 *            A content object that identifies the bytes to download. The
	 *            content must be exactly as it was when downloaded from the
	 *            server.
	 * 
	 * @param handler
	 *            A handler that specifies how the received content is to be
	 *            treated.
	 * 
	 * @throws OperationCanceledException
	 *             if one or more of the downloads owned by this session has
	 *             already been canceled.
	 */
	public void retrieveContent(IFileItemHandle handle, IFileContent content,
			FileDownloadHandler handler)
			throws TeamRepositoryException;
	
    /**
     * Add a new download request to the receiver.
     * @param handle
     *            A state handle identifying the item state that contains the
     *            supplied content
     * @param content
     *            A content object that identifies the bytes to download. The
     *            content must be exactly as it was when downloaded from the
     *            server.
     * @param alternativeEncoding
     *            The character encoding to use when reading the input stream
     *            if performing any line delimiter conversion.  An encoding is
     *            not required if the alternativeDelimiter is {@link FileLineDelimiter#LINE_DELIMITER_NONE}
     *            The caller may choose to supply {@link IFileContent#getCharacterEncoding()}
     *            if there is no need to override.  If the original content had a line delimiter
     *            {@link FileLineDelimiter#LINE_DELIMITER_NONE}, the content may not have an encoding though.
     *            The content is not converted from one encoding to another, it is an override.
     *            Incompatible specification may result in corrupt content returned. 
     * @param alternativeDelimiter
     *            The preferred delimiter of the InputStream.  This will force
     *            the content retrieved to have the line delimiter specified
     *            regardless of the setting specified by content and whether
     *            the line delimiters are consistent.  That is a binary file could
     *            be retrieved with a line delimiter of {@link FileLineDelimiter#LINE_DELIMITER_LF}
     * @param handler
     *            A handler that specifies how the received content is to be
     *            treated.
     * 
     * @throws OperationCanceledException
     *             if one or more of the downloads owned by this session has
     *             already been canceled.
     */
    public void retrieveContent(IFileItemHandle handle, IFileContent content,
            String alternativeEncoding, FileLineDelimiter alternativeDelimiter,
            FileDownloadHandler handler) throws TeamRepositoryException;
    
	/**
	 * Add a new upload request to the receiver.
	 * 
	 * @param handler
	 *            A handler that specifies how the sent content is to be
	 *            treated.
	 * 
	 * @throws OperationCanceledException
	 *             if one or more of the transfers owned by this session has
	 *             already been canceled.
	 */
	public void storeContent(FileUploadHandler handler) throws TeamRepositoryException;
	
	/** Test to see if the session has been canceled. 
	 * @return true if the session has received a cancel event, false otherwise.  
	 */
	public boolean isCanceled();
	
	
	/** Cause the calling thread to suspend until all transfers are complete.
	 * 
	 * @throws OperationCanceledException
	 * 				if one or more of the transfers has been canceled.
	 * 
	 * @throws TeamRepositoryException
	 * 				on any error. 
	 */ 
	public void join() throws TeamRepositoryException;
	
	
	/** Tell the receiver to update the progress monitor to indicate that one 
	 * of the transfers requested at initialization is no longer necessary.
	 * 
	 *  @param howMuch The number of transfers to decrement. 
	 */
	public void decrementTransferCount(long howMuch);
	
	
	/**
	 * If all transfers completed successfully (or successfully so far), then 
	 * an OK error status is returned. If one or more of the transfers 
	 * encountered an error or a cancellation, a multistatus containing an  
	 * error/cancellation for each. 
	 */
	public IStatus[] getErrorStatus();
}
