/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2006, 2009. All Rights Reserved. 
 * 
 * Note to U.S. Government Users Restricted Rights:  Use, 
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.filesystem.client;

import org.eclipse.core.runtime.IProgressMonitor;

import com.ibm.team.repository.client.ITeamRepository;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.scm.client.IBaselineConnection;
import com.ibm.team.scm.client.IConnection;
import com.ibm.team.scm.client.IWorkspaceConnection;

/**
 * Methods for managing the file system. The filesystem manager is closely 
 * tied to its {@link ITeamRepository} and a particular {@link IConnection}.
 * <p>
 * Team repository objects are intended to be used by multi-threaded clients.
 * Unless noted, <code>IFileSystemManager</code>, can be safely
 * called from multiple threads concurrently without risk of corruption.
 * </p>
 * 
 * @since 0.6
 * @NoImplement This interface is not intended to be implemented by clients.
 */
public interface IFileSystemManager {

    /**
     * Obtain the IFileSystemView for the specified connection. This will allow 
     * one to perform file related operations on versioned files in a 
     * {@link IWorkspaceConnection} or {@link IBaselineConnection}.
     * 
     * @param connection
     *            The workspace or baseline connection the file system view is
     *            requested for; never <code>null</code>.
     * @return The filesystem view for the connection
     * @LocalOp This is a local operation; the server/repository itself is not
     *          accessed.
     */
    IFileSystemView getFileSystemView(IConnection connection);
    
    /**
     * Obtain the file system content manager. This will allow one to upload
     * and retrieve file contents from the server.
     * 
     * @since 0.7
     * @return The filesystem content manager
     * @LocalOp This is a local operation; the server/repository itself is not
     *          accessed.
     */
    IFileContentManager getContentManager();
    
    /**
     * Determines if the Server has support for Symbolic links.
     * This may be a long operation.  If the server has been upgraded since the client
     * has started, the client may need to be restarted in order to know the server has
     * been upgraded and may now support symbolic links.
     * 
     * @return <code>true</code> if the server has support for symbolic links,
     * <code>false</code> if it does not have symbolic link support (even though
     * the client obviously does). 
     * @throws TeamRepositoryException if something goes wrong
     */
    boolean serverSupportsSymbolicLinks(IProgressMonitor progress) throws TeamRepositoryException;
}
