/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2005, 2014. All Rights Reserved. 
 * 
 * Note to U.S. Government Users Restricted Rights:  Use, 
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.filesystem.client;

import java.io.File;
import java.util.Collection;
import java.util.List;

import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.OperationCanceledException;

import com.ibm.team.filesystem.common.IChangeSetSummary;
import com.ibm.team.filesystem.common.IFileItemHandle;
import com.ibm.team.filesystem.common.ILogicalConflictReport;
import com.ibm.team.filesystem.common.IWorkspaceCompareReport;
import com.ibm.team.repository.common.NotLoggedInException;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.scm.client.IConfiguration;
import com.ibm.team.scm.client.IConnection;
import com.ibm.team.scm.client.IFlowNodeConnection;
import com.ibm.team.scm.client.IWorkspaceConnection;
import com.ibm.team.scm.common.IChange;
import com.ibm.team.scm.common.IChangeSet;
import com.ibm.team.scm.common.IChangeSetHandle;
import com.ibm.team.scm.common.IComponentHandle;
import com.ibm.team.scm.common.IFolderHandle;
import com.ibm.team.scm.common.WorkspaceComparisonFlags;
import com.ibm.team.scm.common.dto.IChangeHistorySyncReport;

/**
 * File system view on a team repository. Provides methods for browsing files in
 * arbitrary configurations (workspace, baseline), and for manipulating files in
 * workspace configurations.
 * <p>
 * The structure of a tree of versionables is as follows:
 * <ul>
 * <li><b>Component</b>: 0, 1, or more components, which must be of type
 * Component.</li>
 * <li><b>Folder root </b>: Under a component, an unnamed folder exists to
 * contain versionables. The component's root is always located at the empty
 * path.</li>
 * <li><b>files and folders </b>: Under a root, any number of versionables and
 * folders of type FileItem or Folder, respectively. The folder named
 * &quot;images&quot; is located at the path &quot;images&quot;.</li>
 * <li><b>Depth 1 versionables and folders </b>: Under any of the folders at
 * the above level, any number of files and folders of type Versionable or
 * Folder, respectively. The file named &quot;a.txt&quot; is located at the path
 * &quot;images/a.txt&quot;.</li>
 * <li>etc.</li>
 * </ul>
 * The client explicitly creates the items at each of these layers, including
 * the component (except the root folder?).
 * </p>
 * <p>
 * Design note: folders may contain versionables other than
 * <code>FileItem</code>
 * <p>
 * Unless noted, the methods on <code>IFileSystemView</code> can be safely
 * called from multiple threads concurrently without risk of corruption.
 * </p>
 * 
 * @since 0.6
 * @NoImplement This interface is not intended to be implemented by clients.
 */
public interface IFileSystemView {

    /**
     * For use in the
     * {@link IFileSystemView#loadLocalFileSystem(File, IFolderHandle, String[], int, boolean)}
     * {@link IFileSystemView#updateLocalFileSystem(File, IFolderHandle, String[], int, int)
     * call to retrieve all the files in the path and below.
     */
    public static final int DEPTH_INFINITY = -1;
    
    /**
     * Returns the connection and associated configuration underlying this view.
     * The connection need not be open.
     * 
     * @return the connection, which is never <code>null</code>
     * @LocalOp This is a local operation; the server/repository itself is not
     *          accessed.
     */
    public IConnection connection();

    /**
     * Given a {@link IFileItemHandle} returns a list of {@link IChangeSet}
     * entries, where entry number n is the {@link IChangeSet} responsible for
     * producing line n (numbering of lines starts at 0)
     * 
     * The returned list MAY contain null entries.  
     * 
     * An entry can be null if blame could not find a change set for the line.
     * This can currently happen due to double cross component moves.  In the
     * future it may happen if we scope how far back blame looks for changes.
     * 
     * 
     * @param component
     *            a component handle; must not be <code>null</code> if the
     *            underlying connection is a workspace
     * @param fileItemHandle
     *            The file whose history is requested. Must not be
     *            <code>null</code>
     * @param changeSetHandle
     *            The change set that resulted in the given version of the file
     *            or null if the configuration is to be searched for the 1st
     *            change set that results in the given file state. This change
     *            set is used as the starting point of the search backwards.
     * @return List of Change sets where change set n is responsible for line n.
     *         (element type: {@link IChangeSetHandle}); Never
     *         <code>null</code>
     * 
     * @throws TeamRepositoryException
     *             if the operation fails for any reason
     * 
     * @ServerOp This operation communicates with the server/repository; assume
     *           it is slow.
     */
    public List getBlame(IComponentHandle component, IFileItemHandle fileItemHandle, IChangeSetHandle changeSetHandle, IProgressMonitor progress) throws TeamRepositoryException;
    
    /**
     * For a collaboration, produce a report which describes the difference between two workspaces, and the 
     * interpretation of potential conflicts that could occur should everything incoming is accepted.      
     * 
     * The operation has the following preconditions:
     * <ul>
     * <li>The client is logged in to the repository.</li>
     * <li>The underlying connection is a <code>IWorkspaceConnection</code></li>
     * </ul>
     *
     * @see {@link IFlowNodeConnection#compareTo(IFlowNodeConnection, int, List, IProgressMonitor)}
     * @see {@link IWorkspaceConnection#accept(int,IWorkspaceConnection, IChangeHistorySyncReport, List, List, IProgressMonitor)}
     * @see {@link IFileSystemView#interpretConflicts(IPath, Collection, IProgressMonitor)}
     *
     * @param other             Another workspace connection, cannot be <code>null</code> 
     * @param compareFlags      Valid compare flags according to {@link WorkspaceComparisonFlags#isValid(int)}
     * @param componentFilter optional component filter : clients can compare only the specified components if necessary; 
     *                                cannot be <code>null</code>, an empty collection means that no filter is to be applied. This collection
     *                                will ignore components which are not part of either workspace.   
     * @param progress          a progress monitor, or <code>null</code> if progress reporting is not desired
     * @return                  a workspace comparison report, never <code>null</code>.   
     *
     * @throws TeamRepositoryException
     *             if the operation fails for any reason
     * @ServerOp This operation communicates with the server/repository; assume
     *           it is slow.
     * @since 0.7
     */
    public IWorkspaceCompareReport compareAsWorkspaces(IWorkspaceConnection other, int compareFlags, List<IComponentHandle> componentFilter, IProgressMonitor progress) throws TeamRepositoryException;
    
    
    /**
     * Interprets the conflicts currently present on the configuration.
     * 
     * The operation has the following preconditions:
     * <ul>
     * <li>The client is logged in to the repository.</li>
     * <li>The underlying connection is a <code>IWorkspaceConnection</code></li>
     * </ul>
     * 
     * @param monitor monitor to show progress on.
     *          
     * @return Analysis of the conflicts.  never <code>null</code>
     * 
     * @throws TeamRepositoryException
     *             if the operation fails for any reason
     * @ServerOp This operation communicates with the server/repository; assume
     *           it is slow.
     */
    public ILogicalConflictReport conflictReport(IProgressMonitor monitor) throws TeamRepositoryException;

    /**
     * Computes and retrieves a summary of the net change to the configuration
     * resulting from applying the specified list of change sets. The operation
     * fails by throwing <code>TeamRepositoryException</code> if the client is
     * not logged in to the repository.
     * 
     * @param changeSets
     *            a list of {@link IChangeSetHandle} to be interpretted
     * @param monitor
     *            a progress monitor, or <code>null</code> if progress
     *            reporting is not desired  
     * @return a list of {@link IChangeSetSummary} describing the net change to
     *         the configuration
     * @throws OperationCanceledException
     *             if the operation is canceled. Cancelation can occur even if
     *             no progress monitor is provided.
     * @throws NotLoggedInException
     *             if the operation fails because the client was not logged in
     *             to the repository
     * @throws TeamRepositoryException
     *             if the operation fails for any other reason
     * @LongOp This is a long operation; it may block indefinitely; must not be
     *         called from a responsive thread. 
     * @deprecated Use the {@link IChange#kind()} to understand the type of change.
     *         Use {@link IConfiguration#determineAncestorsInHistory(List, IProgressMonitor)}
     *         to get path information.
     */
    public List interpretChanges(List changeSets, IProgressMonitor monitor)
            throws TeamRepositoryException;
}
