/*******************************************************************************
 * Licensed Materials - Property of IBM
 * © Copyright IBM Corporation 2006, 2015. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  Use, 
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.filesystem.client;

import java.io.Reader;

import org.eclipse.core.runtime.IProgressMonitor;

import com.ibm.team.filesystem.client.internal.load.LoadRuleFactory;
import com.ibm.team.filesystem.client.operations.ILoadRule2;
import com.ibm.team.filesystem.common.IFileItemHandle;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.scm.client.IBaselineConnection;
import com.ibm.team.scm.client.IWorkspaceConnection;
import com.ibm.team.scm.common.IComponentHandle;
import com.ibm.team.scm.common.VersionablePermissionDeniedException;
import com.ibm.team.scm.common.VersionedContentDeleted;

public interface ILoadRuleFactory {

    /**
     * A factory which can be used to construct load rules
     */
    public static final ILoadRuleFactory loadRuleFactory = LoadRuleFactory.getInstance();

    /**
     * Creates a load rule that can be further configured to describe what should be loaded
     * @param workspaceConnection The workspace that the rule should be applied to
     * @param componentHandle The component context for the rule 
     * @return A load rule.  Never <code>null</code>
     * @deprecated
     */
    public ILoadRule createLoadRule(IWorkspaceConnection workspaceConnection, IComponentHandle componentHandle);
    
    /**
     * If the load rule could be of the old free-style format use this to obtain the load rule. It is fine if the
     * input is xml format.
     * @param connection The connection that provides context for the load rule
     * @param potentialComponentContext If this is an old style load rule, this is the component it applies to.
     * @param input The load rule specification 
     * @param progress Monitor to mark progress on.
     * @return The load rule described by the contents. <code>null</code> if there is no load rule described by the content.
     * @throws TeamRepositoryException If anything goes wrong
     */
    public ILoadRule2 getLoadRule(IWorkspaceConnection connection, IComponentHandle potentialComponentContext, Reader input, IProgressMonitor progress) throws TeamRepositoryException;

    /**
     * If the load rule could be of the old free-style format use this to obtain the load rule. It is fine if the
     * input is xml format.
     * @param connection The connection that provides context for the load rule
     * @param potentialComponentContext If this is an old style load rule, this is the component it applies to.
     * @param input The load rule specification 
     * @param allowsNestedShares <code>true</code> if nested shares are allowed, <code>false</code> otherwise
     * @param progress Monitor to mark progress on.
     * @return The load rule described by the contents. <code>null</code> if there is no load rule described by the content.
     * @throws TeamRepositoryException If anything goes wrong
     */
    public ILoadRule2 getLoadRule(IWorkspaceConnection connection, IComponentHandle potentialComponentContext, Reader input, boolean allowsNestedShares, IProgressMonitor progress) throws TeamRepositoryException;

    /**
     * This supports the old free-form load rule style as well as the new xml load rule style.
     * @param connection The connection that provides context for the load rule as well as holds the load rule.
     * @param component The component containing the versionable with load rule contents.  If this is an old style
     * load rule, this is the component it applies to.
     * @param fileVersionable The file containing the load rule specification 
     * @param progress Monitor to mark progress on.
     * @return The load rule described by the contents. <code>null</code> if there is no load rule described by the content.
     * @throws TeamRepositoryException If anything goes wrong
     * @throws VersionablePermissionDeniedException if the item is not visible
     * @throws VersionedContentDeleted if the content associated with the file has been permanently deleted from the repository
     */
    public ILoadRule2 getLoadRule(IWorkspaceConnection connection, IComponentHandle potentialComponentContext, IFileItemHandle fileVersionable, IProgressMonitor progress) throws TeamRepositoryException;

    /**
     * This supports the old free-form load rule style as well as the new xml load rule style.
     * @param connection The connection that provides context for the load rule as well as holds the load rule.
     * @param component The component containing the versionable with load rule contents.  If this is an old style
     * load rule, this is the component it applies to.
     * @param fileVersionable The file containing the load rule specification 
     * @param allowsNestedShares <code>true</code> if nested shares are allowed, <code>false</code> otherwise
     * @param progress Monitor to mark progress on.
     * @return The load rule described by the contents. <code>null</code> if there is no load rule described by the content.
     * @throws TeamRepositoryException If anything goes wrong
     * @throws VersionablePermissionDeniedException if the item is not visible
     * @throws VersionedContentDeleted if the content associated with the file has been permanently deleted from the repository
     */
    public ILoadRule2 getLoadRule(IWorkspaceConnection connection, IComponentHandle potentialComponentContext, IFileItemHandle fileVersionable, boolean allowsNestedShares, IProgressMonitor progress) throws TeamRepositoryException;

    /**
     * This supports the old free-form load rule style as well as the new xml load rule style.
     * @param connection The connection that provides context for the load rule as well as holds the load rule.
     * @param fileVersionable The file containing the load rule specification 
     * @param allowsNestedShares <code>true</code> if nested shares are allowed, <code>false</code> otherwise
     * @param progress Monitor to mark progress on.
     * @return The load rule described by the contents. <code>null</code> if there is no load rule described by the content.
     * @throws TeamRepositoryException If anything goes wrong
     * @throws VersionablePermissionDeniedException if the item is not visible
     * @throws VersionedContentDeleted if the content associated with the file has been permanently deleted from the repository
     */
    public ILoadRule2 getLoadRule(IBaselineConnection connection, IFileItemHandle fileVersionable, boolean allowNestedShares, IProgressMonitor progress) throws TeamRepositoryException;

    /**
     * If input is in xml format, return the load rules described by the input.
     * @param connection The connection that provides context for the load rule
     * @param input The load rule specification 
     * @param progress Monitor to mark progress on.
     * @return The load rule described by the contents. <code>null</code> if there is no load rule described by the content.
     * @throws TeamRepositoryException If anything goes wrong
     */
    public ILoadRule2 getLoadRule(IWorkspaceConnection connection, Reader input, IProgressMonitor progress) throws TeamRepositoryException;

    /**
     * If input is in xml format, return the load rules described by the input.
     * @param connection The connection that provides context for the load rule
     * @param input The load rule specification 
     * @param allowsNestedShares <code>true</code> if nested shares are allowed, <code>false</code> otherwise
     * @param progress Monitor to mark progress on.
     * @return The load rule described by the contents. <code>null</code> if there is no load rule described by the content.
     * @throws TeamRepositoryException If anything goes wrong
     */
    public ILoadRule2 getLoadRule(IWorkspaceConnection connection, Reader input, boolean allowsNestedShares, IProgressMonitor progress) throws TeamRepositoryException;

    /**
     * The input is examined to see if this is an xml style load rule or the initial version of free form load rule.
     * @param input The input to be examined.  The input will be closed after examined to see the type of load rule.
     * @return <code>true</code> if the contents are expected to contain an xml load rule, <code>false</code> otherwise.
     * This will not parse the input.  It is possible there are errors in the xml which will be caught when parsing. 
     */
    public boolean isXMLLoadRule(Reader input);
}
