/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2006, 2010. All Rights Reserved. 
 * 
 * Note to U.S. Government Users Restricted Rights:  Use, 
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.filesystem.client;

import com.ibm.team.scm.common.IComponentHandle;
import com.ibm.team.scm.common.IContextHandle;
import com.ibm.team.scm.common.IFolderHandle;
import com.ibm.team.scm.common.IVersionableHandle;

/**
 * Describes a local change.  
 * <p>
 * <b>Note:</b> This class/interface is part of an interim API that is still under development and expected to
 * change significantly before reaching stability. It is being made available at this early stage to solicit feedback
 * from pioneering adopters on the understanding that any code that uses this API will almost certainly be broken
 * (repeatedly) as the API evolves.
 * </p>
 * 
 * @see ILocalChangeManager
 * @since 0.6
 * @NoImplement This interface is not intended to be implemented by clients.
 */
public interface ILocalChange {
    
    /**
     * Type constant (bit mask value 0) meaning no change.
     *
     * @see #getType()
     */
    public static final int NO_CHANGE = 0x0;

    /**
     * Type constant (bit mask value 1) which identifies content changes.
     *
     * @see #getType()
     */
    public static final int CONTENT = 0x01;
    
    /**
     * Type constant (bit mask value 2) which identifies additions.
     *
     * @see #getType()
     */
    public static final int ADDITION = 0x02;

    /**
     * Type constant (bit mask value 4) which identifies deletions.
     *
     * @see #getType()
     */
    public static final int DELETION = 0x04;

    /**
     * Type constant (bit mask value 8) which identifies the first part of a move operation.
     *
     * @see #getType()
     */
    public static final int MOVE_TO = 0x08;
    
    /**
     * Type constant (bit mask value 16) which identifies the final part of a move operation.
     *
     * @see #getType()
     */
    public static final int MOVE_FROM = 0x10;
    
    /**
     * Type constant (bit mask value 32) which identifies attribute changes
     * 
     * @see #getType()
     */
    public static final int ATTRIBUTES = 0x20;



    /**
     * Returns the type of this change. The returned value will be a combination of 
     * <code>{@link #CONTENT}</code>, <code>{@link #ATTRIBUTES}</code>,
     * <code>{@link #ADDITION}</code>,
     * <code>{@link #DELETION}</code>, <code>{@link #MOVE_FROM}</code> 
     * and <code>{@link #MOVE_TO}</code>.
     *  
     * @return the type of this change
     */
    public int getType();

    /**
     * @return the sandbox root containing the change. never <code>null</code>
     */
    public ILocation getSandboxRoot();
    
    /**
     * Returns the local path for the target versionable, equivalent to resulting
     * path for non deletion changes and the path where the target would have been
     * if it had not been deleted for deletion changes. never null.
     *  
     * @return the path for the target versionable
     */
    public IRelativeLocation getPath();
    
    /**
     * Returns the resulting path for this target after all current changes were applied.
     *          null if this change represents a deletion
     * 
     * @return the resulting path
     */
    public IRelativeLocation getResultingPath();
        
    /**
     * Returns the original path for this item before any local changes have been applied.
     * 
     * @return the original path for this item before any local changes have been applied.
     *          null if this change represents an addition.
     */
    public IRelativeLocation getOriginalPath();
    
    /**
     * Returns the versionable this change applies to.
     * The state will be present if this is a versionable that exists
     * remotely.
     * 
     * @return the change target
     */
    public IVersionableHandle getTarget();
    
    /**
     * Returns the connection this change relates to.
     * 
     * @return the workspace
     */
    public IContextHandle getConnection();
    
    /**
     * Returns the component this change relates to.
     * 
     * @return the component
     */
    public IComponentHandle getComponent();
    
    /**
     * Returns the shareable this change corresponds to, or <code>null</code> if it cannot be determined.
     * 
     * @return a shareable, or <code>null</code>
     * @throws FileSystemException 
     */
    public IShareable getShareable();
    
    /**
     * Returns whether this change is of the given type.
     * 
     * @param mask
     * @return <code>true</code> if this change is of the given type, <code>false</code> otherwise
     */
    public boolean isType(int mask);
    
    /**
     * Returns the counterpart change, or <code>null</code> if the change does not have a counterpart.
     * 
     * @return the counterpart change, or <code>null</code>
     */
    public ILocalChange getCounterpart();

    /**
     * Returns whether the change has been canceled.
     *  
     * @return <code>true</code> if this change has been cancelled, <code>false</code> otherwise
     */
    public boolean isCanceled();

    /**
     * Returns the parent folder if it is affected by this change (i.e. move, delete), or <code>null</code>
     * if the parent is not affected.
     * 
     * @return the parent versionable, or <code>null</code>
     */
    public IFolderHandle getTargetParent();
}