/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2006, 2014. All Rights Reserved. 
 * 
 * Note to U.S. Government Users Restricted Rights:  Use, 
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.filesystem.client;

import java.util.Collection;

import org.eclipse.core.runtime.IProgressMonitor;

import com.ibm.team.filesystem.client.internal.IRepositoryResolver;
import com.ibm.team.filesystem.client.internal.copyfileareas.ComponentLock;
import com.ibm.team.filesystem.client.operations.UndoDilemmaHandler;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.scm.common.IComponentHandle;
import com.ibm.team.scm.common.IContextHandle;
import com.ibm.team.scm.common.IVersionableHandle;

/**
 * The local change manager is the starting point for all needs related to monitoring 
 * changes to shareables.  
 
 * @see ILocalChange
 * @see ILocalChangeListener
 * @since 0.6
 * @NoImplement This interface is not intended to be implemented by clients.
 */
public interface ILocalChangeManager {
    
    /**
     * Refresh types.
     */
    public enum RefreshType {
        /** This refresh will skip any files/folders that overlap with Eclipse resources. */
        OPTIMIZED_TRAVERSAL,
        /** This refresh will include Eclipse resources in the refresh but assumes that
         *  the Eclipse resource tree is up-to-date. */
        TRAVERSE_ALL_KNOWN,
        /** This refresh performed a full scan just like TRAVERSE_ALL_KNOWN with an added refresh
         *  of any Eclipse resources to ensure that the Eclipse resource tree is up-to-date. */
        TRAVERSE_ALL_WITH_RECOMPUTE_OF_KNOWN
    }

    /**
     * @param connection context of the operation
     * @param component the component
     * @param sandbox the sandbox
     * @return all pending changes for the given component. Note that objects
     *         with <code>getType()</code> {@link ILocalChange#MOVE_FROM} will
     *         not be returned from this method. They must be found by querying
     *         the {@link ILocalChange#getCounterpart()} method of instances of
     *         type {@link ILocalChange#MOVE_TO}.
     *         
     *         <p>
     *         Never null. If there are no pending changes then the array will
     *         be empty.
     */
    public ILocalChange[] getPendingChanges(IContextHandle connection, IComponentHandle component, ISandbox sandbox);

    /**
     * Retuns all pending changes for the given shareables and any of their
     * children.
     * 
     * <p>
     * Note that objects with <code>getType()</code>
     * {@link ILocalChange#MOVE_FROM} will not be returned from this method.
     * They must be found by querying the {@link ILocalChange#getCounterpart()}
     * method of instances of type {@link ILocalChange#MOVE_TO}.
     * 
     * 
     * @param startingPoints sharables
     * @param monitor a progress monitor
     * @return local changes
     * @throws FileSystemException 
     */
    public ILocalChange[] getPendingChanges(IShareable[] startingPoints, IProgressMonitor monitor) throws FileSystemException;

    /**
     * Retuns all pending changes for the given share and any of their children.
     * 
     * Note that objects with <code>getType()</code>
     * {@link ILocalChange#MOVE_FROM} will not be returned from this method.
     * They must be found by querying the {@link ILocalChange#getCounterpart()}
     * method of instances of type {@link ILocalChange#MOVE_TO}.
     * 
     * 
     * @param startingPoints
     * @return local changes
     */
    public ILocalChange[] getPendingChanges(IShare[] startingPoints);

    /**
     * Returns the pending change for the given shareable.
     * @param shareable the shareable object
     * @param monitor a progress monitor
     * @return local changes
     * @throws FileSystemException 
     */
    public ILocalChange getPendingChange(IShareable shareable, IProgressMonitor monitor) throws FileSystemException;

    /**
     * Returns the pending change for the given versionable.
     * 
     * @param share
     * @param versionable
     * @return local changes
     */
    public ILocalChange getPendingChange(IShare share, IVersionableHandle versionable);
    
    /**
     * After this method returns, all the local changes collected up to the time of this call (and maybe later)
     * are computed.
     *  
     * If the caller is holding the remote workspace lock, then they must also be 
     * holding the component lock for this context/component/sandbox: 
     * {@link ComponentLock} or a lock that contains it.
     * @param connection context of the operation
     * @param component the component to be synced
     * @param sandbox the sandbox to be synced
     * @param monitor a progress monitor
     */
    public void syncPendingChanges(IContextHandle connection, IComponentHandle component, ISandbox sandbox, IProgressMonitor monitor);

    /**
     * Registers the given listener. Does nothing if the listener is already registered. 
     * @param listener listener
     */
    public void addLocalChangeListener(ILocalChangeListener listener);

    /**
     *  Unregisters the given listener. Does nothing if the listener is not currently registered. 
     * @param listener listener
     */
    public void removeLocalChangeListener(ILocalChangeListener listener);

    /**
     * Undoes the specified changes.
     * 
     * @param resolver
     *          A resolver that will find the team repository for a given repository uri/id
     *          
     * @param toUndo
     *          The changes to undo
     *          
     * @param problemHandler
     *          Used to handle problems encountered during the undo.
     *          May be <code>null</code> in which case the default handler will be used.
     *          
     * @param monitor
     *          For progress reporting
     * @throws FileSystemException 
     * @throws TeamRepositoryException 
     */
    public void undoChanges(ILocalChange[] toUndo, IRepositoryResolver resolver, UndoDilemmaHandler problemHandler, IProgressMonitor monitor) throws FileSystemException, TeamRepositoryException;
    
    /**
     * Undoes the specified changes.
     * 
     * @param toUndo
     *          The changes to undo
     *          
     * @param resolver
     *          A resolver that will find the team repository for a given repository uri/id
     * 
     * @param autoReleaseLocks
     *          Automatically release locks after undoing the changes.          
     *
     * @param problemHandler
     *          Used to handle problems encountered during the undo.
     *          May be <code>null</code> in which case the default handler will be used.
     *          
     * @param monitor
     *          For progress reporting
     * @throws FileSystemException 
     * @throws TeamRepositoryException 
     */
    public void undoChanges(ILocalChange[] toUndo, IRepositoryResolver resolver, boolean autoReleaseLocks, UndoDilemmaHandler problemHandler, IProgressMonitor monitor) throws FileSystemException, TeamRepositoryException;

    /**
     * Determine if based on metadata if the file is dirty or not. The sandbox for the shareable must already be registered.
     * @param shareable The shareable we want to know is dirty or not.
     * @param monitor For progress reporting.
     * @return <code>true</code> if the file/folder is not the original loaded.
     * @throws FileSystemException if the operation fails
     */
    public boolean isContentDirty(IShareable shareable, IProgressMonitor monitor) throws FileSystemException;

    /**
     * Combine an add/delete into a move. 
     * 
     * <p>
     * Note: the add/delete must be within the same workspace/component pair.
     * @param resolver Repository resolver that will resolve a repository uri/id into a
     * team repository
     * @param deletion The {@link ILocalChange} corresponding to the deletion 
     *          of the item at the old location. 
     *          
     * @param addition The {@link ILocalChange} that adds the item at the new 
     *          location. 
     *          
     * @param monitor For progress reporting.
     *  
     * @throws TeamRepositoryException 
     */
    public void combineDeleteAdd(IRepositoryResolver resolver, ILocalChange deletion, ILocalChange addition, IProgressMonitor monitor) throws TeamRepositoryException;
    
    /**
     * Refresh the local changes within the given sandboxes. The scanning of the local file system
     * occurs in the calling thread but notification of any changes found may occur in a separate
     * thread. The refresh types have the following semantics:
     * <ul>
     * <li>OPTIMIZED_TRAVERSAL: This refresh will skip any files/folders that overlap with Eclipse resources.
     * It is used to discover changes made to files/folders that do not correspond to Eclipse resources
     * without the overhead of scanning the entire sandbox.</li>
     * <li>TRAVERSE_ALL_KNOWN: This refresh will include Eclipse resources in the refresh but assumes that
     * the Eclipse resource tree is up-to-date. It is generally used when content becomes shared either
     * via a share operation or a move or copy.</li>
     * <li>TRAVERSE_ALL_WITH_RECOMPUTE_OF_KNOWN: This refresh performed a full scan just like TRAVERSE_ALL_KNOWN
     * with an added refresh of any Eclipse resources to ensure that the Eclipse resource tree is up-to-date.</li>
     * </ul>
     * @param sandboxes the sandboxes to be refreshed
     * @param type indicates the type of traversal to be performed when refreshing.
     * @param monitor a progress monitor
     * @throws FileSystemException 
     */
    public void refreshChanges(ISandbox[] sandboxes, RefreshType type, IProgressMonitor monitor) throws FileSystemException;
    
    /**
     * Find local changes below the given shareables.  The scanning of the local file system
     * occurs in the calling thread but notification of any changes found may occur in a separate
     * thread. The refresh types have the following semantics:
     * <ul>
     * <li>OPTIMIZED_TRAVERSAL: This refresh will skip any files/folders that overlap with Eclipse resources.
     * It is used to discover changes made to files/folders that do not correspond to Eclipse resources
     * without the overhead of scanning the entire sandbox.</li>
     * <li>TRAVERSE_ALL_KNOWN: This refresh will include Eclipse resources in the refresh but assumes that
     * the Eclipse resource tree is up-to-date. It is generally used when content becomes shared either
     * via a share operation or a move or copy.</li>
     * <li>TRAVERSE_ALL_WITH_RECOMPUTE_OF_KNOWN: This refresh performed a full scan just like TRAVERSE_ALL_KNOWN
     * with an added refresh of any Eclipse resources to ensure that the Eclipse resource tree is up-to-date.</li>
     * </ul>
     * @param shareables the shareables (and their children) to be refreshed
     * @param type indicates the type of traversal to be performed when refreshing.
     * @param monitor a progress monitor
     * @throws FileSystemException 
     */
    public void refreshChanges(Collection<IShareable> shareables, RefreshType type, IProgressMonitor monitor) throws FileSystemException;
    
    /**
     * Find local changes below the given shareables asynchronously. The notification of any changes found will occur 
     * in a separate thread. The refresh types have the following semantics:
     * <ul>
     * <li>OPTIMIZED_TRAVERSAL: This refresh will skip any files/folders that overlap with Eclipse resources.
     * It is used to discover changes made to files/folders that do not correspond to Eclipse resources
     * without the overhead of scanning the entire sandbox.</li>
     * <li>TRAVERSE_ALL_KNOWN: This refresh will include Eclipse resources in the refresh but assumes that
     * the Eclipse resource tree is up-to-date. It is generally used when content becomes shared either
     * via a share operation or a move or copy.</li>
     * <li>TRAVERSE_ALL_WITH_RECOMPUTE_OF_KNOWN: This refresh performed a full scan just like TRAVERSE_ALL_KNOWN
     * with an added refresh of any Eclipse resources to ensure that the Eclipse resource tree is up-to-date.</li>
     * </ul>
     * @param shareables the shareables (and their children) to be refreshed
     * @param type indicates the type of traversal to be performed when refreshing.
     * @throws FileSystemException 
     */
    public void refreshChangesAsync(Collection<IShareable> shareables, RefreshType type) throws FileSystemException;

}
