/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010, 2013. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.filesystem.client;

import org.eclipse.core.runtime.IAdaptable;

import com.ibm.team.filesystem.client.internal.IStorageManager;
import com.ibm.team.filesystem.client.internal.copyfileareas.ISharingMetadataFactory;

/**
 * @remoteProjects This entity is used by remote project integration. 
 *      Downstream teams will need to adopt changes made to it.
 */
public interface ILocation extends IAdaptable {
    
    /**
     * @return The type of location, used to determine how to manage it and its metadata.
     * Used to locate the {@link IStorageManager} and {@link ISharingMetadataFactory}
     * for it.
     */
    public String getStorageId();
	
	/**
	 * @return String representation of the location in a visual representation
	 * for the platform.
	 */
	public String toOSString();
	
	/**
 	 * @return a string representation of this path suitable for purposes of displaying
 	 *  and deserializing the location.
 	 *  
 	 * @see IStorageManager#createLocation(String)
	 */
	public String toString();
	
	/**
	 * @return if the storage represented by this is case-sensitive or not
	 */
	public boolean isCaseSensitive();
	
	/**
	 * @return Whether this is an empty location (i.e. no segments in a path nor drive
	 * or no dataset prefix or dataset name)
	 */
	public boolean isEmpty();
	
	/**
	 * @return the location that holds this item.  May be empty path
	 * if this is at the root of the file system.
	 */
	public ILocation getParent();

	/**
     * Returns the name of the file or directory denoted by this abstract
     * location.  This is just the last name in the pathname's name
     * sequence.  If the pathname's name sequence is empty, then the empty
     * string is returned.
     * 
     * this == this.getParent().append(this.getName))
     * 
     * For a location of 
     * C:/A/B/D the name is "D"
     * /A/B/D the name is "D"
     * A.B.C(D) the name is "D"
     * HIGH.QUALIFIER.DATASET.NAME the name is "DATASET.NAME"
     * HIGH.QUALIFIER the name is "HIGH.QUALIFIER"
     *
     * @return  The name of the file or directory denoted by this abstract
     *          pathname, or the empty string if this pathname's name sequence
     *          is empty
     */
	public String getName();
	
	/**
	 * adds the name supplied to the end of the location specified
	 * and returns the new location.  If the name can not be logically added
	 * to the end of location (for example this is a dataset member)
	 * then null will be returned. 
	 * @param name The name to be appended to the location.
	 * @return The new location represented by this location with the name.
	 * May be <code>null</code> if it is not possible to construct such a location.
	 * 
	 * TODO would it be better to raise an exception and have a method that 
	 * can be queried to see if at the end.  But then that implies we would
	 * really know based on what is on disk.
	 */
	public ILocation append(String name);
	
	/**
	 * adds the location supplied to the end of the location specified
	 * and returns the new location.  If the name can not be logically added
	 * to the end of location (for example this is a dataset member)
	 * then {@link IllegalStateException} will be thrown.
	 * 
	 * The location being appended must be hierarchical if this location is 
	 * hierarchical.  If this is not being hierarchical then the location
	 * being appended may not be hierarchical.
	 * 
	 * @param location The location to be appended to this location.
	 * @return The new location represented by this location with the name.
	 * May be <code>null</code> if it is not possible to construct such a location.
	 * 
	 * TODO would it be better to raise an exception and have a method that 
	 * can be queried to see if at the end.  But then that implies we would
	 * really know based on what is on disk.
	 */
	public ILocation append(IRelativeLocation location);
	
    /**
     * Answers an integer hash code for the receiver. Any two
     * objects which answer <code>true</code> when passed to
     * <code>.equals</code> must answer the same value for this
     * method.
     *
     * @return		int
     *					the receiver's hash.
     *
     * @see			#equals
     */
    public int hashCode();

    /**
     * Compares the argument to the receiver, and answers true
     * if they represent the <em>same</em> object using a class
     * specific comparison. The implementation in Object answers
     * true only if the argument is the exact same object as the
     * receiver (==).
     *
     * @param		o Object
     *					the object to compare with this object.
     * @return		boolean
     *					<code>true</code>
     *						if the object is the same as this object
     *					<code>false</code>
     *						if it is different from this object.
     * @see			#hashCode
     */
    public boolean equals (Object o);

	/**
	 * Returns whether this path is a prefix of the given path.
	 * To be a prefix, this path's segments must
	 * appear in the argument path in the same order,
	 * <p>
	 * An empty location is a prefix of all locations of the same type.
	 * If 2 locations are equal, this path will be a prefix of it.
	 * </p>
	 * Note : You can not assume that if a.isPrefixOf(b) && b.isPrefixOf(a) that
	 * a.equals(b).  The internal semantics may contain additional differences.
	 * 
     * @param anotherPath the other path that we want to know if we are a prefix of
	 * @return <code>true</code> if this path is a prefix of the given path,
	 *    and <code>false</code> otherwise
	 */
	public boolean isPrefixOf(ILocation anotherLocation);

    /**
     * Returns whether this path is a prefix of the given path.
     * To be a prefix, this path's segments must
     * appear in the argument path in the same order,
     * <p>
     * An empty location is a prefix of all locations of the same type.
     * If 2 locations are equal, this path will be a prefix of it.
     * </p>
     * @param anotherPath the other path that we want to know if we are a prefix of
     * @param isCaseSensitive whether or not to perfom a case sensitive comparison or not
     * @return <code>true</code> if this path is a prefix of the given path,
     *    and <code>false</code> otherwise
     */
    public boolean isPrefixOf(ILocation anotherLocation, boolean isCaseSensitive);
	
	/**
	 * Determine if this location matches another location.  isCaseSensitive
	 * will dictate whether to do a case-sensitive comparison or not.
	 * @param anotherLocation The other location to compare against.
	 * @param isCaseSensitive Dictates whether to do a case-sensitive comparison or not.  If
     * true, the compare is equivelant to {@link #equals(Object)}.  If false
     * then a case insensitve compare is done.
	 * @return if the this is the same location as another location
	 */
	public boolean sameLocation(ILocation anotherLocation, boolean isCaseSensitive);
	
	/**
	 * Get canonical form of the location.  The idea is that 2 locations
	 * that represent the same place might have different internal 
	 * representations that result equals != true. But the canonical
	 * form of the 2 locations would result in equals == true.
	 * However obtaining the canonical form can be expensive so we only 
	 * want to do it when necessary.
	 */
	public ILocation getCanonicalForm();

	/**
	 * Helps us to deal with case insensitive platforms better when the locations
	 * in question do not exist and as such we want a constant form to compare against.
	 * 
     * Will convert everything to lower case if the FS is not case sensitive and convertCase
     * is true.
	 */
	ILocation getCanonicalForm(boolean isCaseSensitive, boolean convertCase);

	/**
	 * Determine the location relative to the location supplied.
	 * Will throw an illegalArgumentException if the location supplied
	 * is not a prefix of this location.
	 * @param location - The location to determine a location relative to it.
	 * @return relative location of this location relative to the location supplied
	 */
    public IRelativeLocation getLocationRelativeTo(ILocation location);
}
