/*******************************************************************************
 * Licensed Materials - Property of IBM
 * © Copyright IBM Corporation 2008, 2015. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  Use, 
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.filesystem.client;

import com.ibm.team.filesystem.client.internal.IRepositoryResolver;
import com.ibm.team.filesystem.client.internal.OperationFactory;
import com.ibm.team.filesystem.client.internal.load.loadRules.SandboxExaminerDilemmaHandler;
import com.ibm.team.filesystem.client.internal.operations.IUpdateSubcomponentOperation;
import com.ibm.team.filesystem.client.operations.AbortCurrentPatchDilemmaHandler;
import com.ibm.team.filesystem.client.operations.ApplyAcceptedDilemmaHandler;
import com.ibm.team.filesystem.client.operations.AutoResolveConflictsDilemmaHandler;
import com.ibm.team.filesystem.client.operations.BackupDilemmaHandler;
import com.ibm.team.filesystem.client.operations.ChangeExtendedPropertiesDilemmaHandler;
import com.ibm.team.filesystem.client.operations.ChangePropertiesDilemmaHandler;
import com.ibm.team.filesystem.client.operations.CloseChangeSetsDilemmaHandler;
import com.ibm.team.filesystem.client.operations.CommitDilemmaHandler;
import com.ibm.team.filesystem.client.operations.CompleteCurrentPatchDilemmaHandler;
import com.ibm.team.filesystem.client.operations.CreateBaselineDilemmaHandler;
import com.ibm.team.filesystem.client.operations.CreateCurrentPatchDilemmaHandler;
import com.ibm.team.filesystem.client.operations.CreateSnapshotDilemmaHandler;
import com.ibm.team.filesystem.client.operations.DeliverDilemmaHandler;
import com.ibm.team.filesystem.client.operations.DiscardDilemmaHandler;
import com.ibm.team.filesystem.client.operations.IAbortCurrentPatchOperation;
import com.ibm.team.filesystem.client.operations.IAutoResolveConflictsOperation;
import com.ibm.team.filesystem.client.operations.IAutoResolveLocalConflictsOperation;
import com.ibm.team.filesystem.client.operations.IChangeExtendedPropertiesOperation;
import com.ibm.team.filesystem.client.operations.IChangePropertiesOperation;
import com.ibm.team.filesystem.client.operations.ICheckinOperation;
import com.ibm.team.filesystem.client.operations.ICloseChangeSetsOperation;
import com.ibm.team.filesystem.client.operations.ICompleteCurrentPatchOperation;
import com.ibm.team.filesystem.client.operations.ICreateBaselineOperation;
import com.ibm.team.filesystem.client.operations.ICreateCurrentPatchOperation;
import com.ibm.team.filesystem.client.operations.ICreateSnapshotOperation;
import com.ibm.team.filesystem.client.operations.IDeleteFoldersOperation;
import com.ibm.team.filesystem.client.operations.IDeliverOperation;
import com.ibm.team.filesystem.client.operations.IDiscardOperation;
import com.ibm.team.filesystem.client.operations.ILoadOperation;
import com.ibm.team.filesystem.client.operations.ILockOperation;
import com.ibm.team.filesystem.client.operations.IMetadataRepairOperation;
import com.ibm.team.filesystem.client.operations.IMetadataSynchronizeOperation;
import com.ibm.team.filesystem.client.operations.IMetadataValidateOperation;
import com.ibm.team.filesystem.client.operations.IMoveFoldersOperation;
import com.ibm.team.filesystem.client.operations.IMoveVersionablesOperation;
import com.ibm.team.filesystem.client.operations.IProcessAcceptQueueOperation;
import com.ibm.team.filesystem.client.operations.IResumeOperation;
import com.ibm.team.filesystem.client.operations.ISandboxExaminerOperation;
import com.ibm.team.filesystem.client.operations.IShareOperation;
import com.ibm.team.filesystem.client.operations.ISuspendOperation;
import com.ibm.team.filesystem.client.operations.ITransferLockOperation;
import com.ibm.team.filesystem.client.operations.IUnlockOperation;
import com.ibm.team.filesystem.client.operations.IUnshareOperation;
import com.ibm.team.filesystem.client.operations.IUpdateCurrentPatchOperation;
import com.ibm.team.filesystem.client.operations.IVerifyInSyncOperation;
import com.ibm.team.filesystem.client.operations.IWorkspaceUpdateOperation;
import com.ibm.team.filesystem.client.operations.LoadDilemmaHandler;
import com.ibm.team.filesystem.client.operations.LockDilemmaHandler;
import com.ibm.team.filesystem.client.operations.MarkAsMergedDilemmaHandler;
import com.ibm.team.filesystem.client.operations.MetadataRepairDilemmaHandler;
import com.ibm.team.filesystem.client.operations.MoveFoldersDilemmaHandler;
import com.ibm.team.filesystem.client.operations.MoveVersionablesDilemmaHandler;
import com.ibm.team.filesystem.client.operations.OperationOptions;
import com.ibm.team.filesystem.client.operations.OutOfSyncDilemmaHandler;
import com.ibm.team.filesystem.client.operations.ProcessAcceptQueueDilemmaHandler;
import com.ibm.team.filesystem.client.operations.ResumeDilemmaHandler;
import com.ibm.team.filesystem.client.operations.ShareDilemmaHandler;
import com.ibm.team.filesystem.client.operations.SuspendDilemmaHandler;
import com.ibm.team.filesystem.client.operations.UnshareDilemmaHandler;
import com.ibm.team.filesystem.client.operations.UpdateCurrentPatchDilemmaHandler;
import com.ibm.team.filesystem.client.operations.UpdateDilemmaHandler;
import com.ibm.team.filesystem.client.operations.WorkspaceUpdateDilemmaHandler;

/**
 * Methods for obtaining operations to manipulate workspaces, configurations and
 * the file system. The Operations will ensure content loaded remains in sync
 * with the workspace configurations.
 * <p>
 * Unless noted, <code>IFileSystemManager</code>, can be safely
 * called from multiple threads concurrently without risk of corruption.
 * The operations themselves though may not be safely called from multiple
 * threads unless otherwise noted.
 * </p>
 * 
 * @since 0.7 (RTC 2.0)
 * @NoImplement This interface is not intended to be implemented by clients.
 */
public interface IOperationFactory {

    /**
     * The default options that the operation factory configures
     * qualifying operations with.
     * @return options detailing how an operation should perform its duties.
     * Never <code>null</code>
     */
    public OperationOptions getOperationOptions();

    /**
     * Operation factory to be used to create new operations that can 
     * be configured and run against workspaces and copy file areas.
     */
    public static final IOperationFactory instance = OperationFactory.getInstance();

    /**
     * Provides an operation that can be used to load all or part of a workspace
     * <p>
     * The expectation is that the caller will configure the operation with the
     * items to be loaded and then run the operation. 
     * 
     * @param dilemmaHandler 
     *            The dilemmaHandler that will be called if problems are encountered
     *            during the load; can be <code>null</code> in which case a default
     *            handler will be used.
     * @return A load operation that can be further configured and used to load
     *            the Jazz workspace under the local root.
     */
    public ILoadOperation getLoadOperation(LoadDilemmaHandler dilemmaHandler);
    
    /**
     * Provides an operation that can be used to verify if the copy file area
     * contents are in sync with the contents of the repository.  In sync from
     * the point of view that the configuration in the repository has not been
     * updated while the contents on disk have not.  It is valid for there to 
     * be pending local changes.
     * @param dilemmaHandler 
     *            The dilemmaHandler that will be called if problems are encountered
     *            during the verify; can be <code>null</code> in which case a default
     *            handler will be used.
     * @return An operation that can be used to verify if the shares or configurations
     *            loaded are in sync with the repository workspace.  It is expected
     *            the operation will be run once.  To rerun an operation, retrieve
     *            a new operation and configure it accordingly.
     */
    public IVerifyInSyncOperation getVerifyInSyncOperation(OutOfSyncDilemmaHandler dilemmaHandler);
    
    /**
     * Provides an operation that can be used to share files/folders on disk into
     * workspace components.
     * <p>
     * The expectation is that the caller will configure the operation with the
     * items to be shared and then run the operation.
     * 
     * @param dilemmaHandler 
     *            The dilemmaHandler that will be called if problems are encountered
     *            during the sharing of projects.  Can be <code>null</code> in which
     *            case a default handler will be used.
     * @return A share operation to be used to share/reshare files/folders into
     *          Jazz workspaces.
     */
    public IShareOperation getShareOperation(ShareDilemmaHandler dilemmaHandler);
    
    /**
     * Provides an operation that can be used to unshare share roots and configurations
     * from their sandboxes.
     * <p>
     * The expectation is that the caller will configure the operation with the
     * items to be unshared and then run the operation.
     * 
     * @param dilemmaHandler 
     *            The dilemmaHandler that will be called if problems are encountered
     *            during the unsharing of shares and configurations.  Can be <code>null</code> in which
     *            case a default handler will be used.
     * @param resolver The resolver used when determining the repository from the 
     *            repository id & uri.  The repository maybe used when checking changes in.
     * @return A unshare operation to be used to unshare items that are in the Jazz sandboxes.
     */
    public IUnshareOperation getUnshareOperation(UnshareDilemmaHandler dilemmaHandler, IRepositoryResolver resolver);

    /**
     * Provides an operation that can be used to check in new/modified/deleted
     * files/folders from the local workspace (copy file area/sandbox) into
     * the remote workspace.
     * <p>
     * The expectation is that the caller will configure the operation with the
     * items to be shared and then run the operation once.
     *  
     * @param problemHandler
     *            The dilemmaHandler that will be called if problems are encountered
     *            during the checking of changes.  Can be <code>null</code> in which
     *            case a default handler will be used.
     * @param resolver The resolver used when determining the repository from the 
     *            repository id & uri.  The repository maybe used when checking changes in.
     * @return An check-in operation to be used to check in changes into the
     *          Jazz workspaces.
     */
    public ICheckinOperation getCheckinOperation(CommitDilemmaHandler problemHandler, IRepositoryResolver resolver);

    /**
     * Provides an operation that can be used to deliver changes and components to
     * a target workspace.
     * 
     * @param dilemmaHandler 
     *            The dilemmaHandler that will be called if problems are encountered
     *            during the deliver.  Can be <code>null</code> in which case a default
     *            handler will be used.
     * @return A deliver operation to be used to deliver changes and components to
     *         a start workspace.
     */ 
    public IDeliverOperation getDeliverOperation(DeliverDilemmaHandler dilemmaHandler);

    /**
     * Provides an operation that can be used to suspend changes to a target workspace.
     * 
     * @param dilemmaHandler
     *     The dilemma handler that will be called if problems are encountered
     *     during the suspend. Can be <code>null</code> in which case a
     *     default handler will be used.
     * @return
     *     A suspend operation to be used to suspend changes
     */
    public ISuspendOperation getSuspendOperation(SuspendDilemmaHandler dilemmaHandler);

    /**
     * Provides an operation that can be used to discard change sets from a workspace.
     * 
     * @param dilemmaHandler
     *     The dilemma handler that will be called if problems are encountered
     *     during the discard. Can be <code>null</code> in which case a
     *     default handler will be used.
     * @return
     *     A discard operation to be used to discard changes
     */
    public IDiscardOperation getDiscardOperation(DiscardDilemmaHandler dilemmaHandler);

    /**
     * Provides an operation that can be used to resume change sets into a workspace.
     * 
     * @param dilemmaHandler
     *     The dilemma handler that will be called if problems are encountered
     *     during the resume. Can be <code>null</code> in which case a
     *     default handler will be used.
     * @return
     *     A resume operation to be used to resume changes
     */
    public IResumeOperation getResumeOperation(ResumeDilemmaHandler dilemmaHandler);

    /**
     * Provides an operation that can be used to change the properites of a file/folder/link.
     * This should result in local changes that can then be committed
     * to the repository.  This is the operation that should be used to 
     * change the line delimiter and/or the content type of a file.
     * It can also be used to change the User properties of a file/folder/symbolic link.
     * 
     * The expectation is the caller will configure the operation with the shareables
     * whose line delimiter/content type/user defined properties are to be
     * changed and then run the operation.
     * 
     * @param dilemmaHandler
     *          The dilemmaHandler that will be called if problems are encountered
     *          when altering the line delimiter preference.
     *          
     * @return A Change properties operation to be used to change file properties.
     */
    public IChangePropertiesOperation getChangePropertiesOperation(ChangePropertiesDilemmaHandler dilemmaHandler);
    
    /**
     * Provides an operation that can be used to change the extended properties of a file/folder/link.
     * 
     * The expectation is the caller will configure the operation with the shareables
     * whose extended properties are to be changed and then run the operation.
     * 
     * @param dilemmaHandler
     *          The dilemmaHandler that will be called if problems are encountered
     *          when altering the extended properties. If null, a default dilemma handler will be used.
     *          
     * @return A Change properties operation to be used to change extended properties.
     */
    public IChangeExtendedPropertiesOperation getChangeExtendedPropertiesOperation(ChangeExtendedPropertiesDilemmaHandler dilemmaHandler);
    
    /**
     * Provides an operation that can be used to create a baseline.
     * The operation returned can be configured to specify for which 
     * components the baseline should be created for.
     * 
     * @param dilemmaHandler
     *            The dilemma handler that will be called if problems are encountered.
     *            Can be <code>null</code> in which case a default handler will be used.
     * @return an operation that can be used to create baselines.            
     *            It is expected the operation will be run once.  To rerun an operation, retrieve
     *            a new operation and configure it accordingly.
     */
    public ICreateBaselineOperation getCreateBaselineOperation(CreateBaselineDilemmaHandler dilemmaHandler);
    
    /**
     * Provides an operation that can be used to create a snapshot.
     * The operation returned can be configured to specify for which 
     * workspace the snapshot should be created for.
     *
     * @param dilemmaHandler
     *            The dilemma handler that will be called if problems are encountered.
     *            Can be <code>null</code> in which case a default handler will be used.
     * @return an operation that can be used to create a snapshot.
     *            It is expected the operation will be run once.  To rerun an operation, retrieve
     *            a new operation and configure it accordingly.
     *
     * @since 0.11.0 (RTC 6.0)
     */
    public ICreateSnapshotOperation getCreateSnapshotOperation(CreateSnapshotDilemmaHandler dilemmaHandler);

    /**
     * Provides an operation that can be used to update a workspace (both the Jazz workspace configuration
     * and the Eclipse workspace loaded).  The operation returned can be configured to accept just changesets,
     * to accept components, changesets and baselines from other workspaces, replacing the workspace
     * or its components with components from other workspaces, baselines, snapshots or updating
     * a workspace based on its flow targets.
     * 
     * @param dilemmaHandler
     *            The dilemma handler that will be called if problems are encountered
     *            during the updating of the workspaces; can be <code>null</code> in which
     *            case a default handler will be used.
     * @return An operation that can be used to update a workspace (both the Jazz workspace and
     *            the Eclipse workspace).
     *            It is expected the operation will be run once.  To rerun an operation, retrieve
     *            a new operation and configure it accordingly.
     */
    public IWorkspaceUpdateOperation getWorkspaceUpdateOperation(WorkspaceUpdateDilemmaHandler dilemmaHandler);
    
    /**
     * Provides an operation that can be used to auto-resolve conflicts
     * @param dilemmaHandler 
     *            The dilemmaHandler that will be called if problems are encountered
     *            during the auto-resolve; can be <code>null</code> in which case a default
     *            handler will be used.
     * 
     * @param resolver
     *            A resolver that can be used to find repositories referenced 
     *            by their ID/URI. May not be <code>null</code>.
     *            
     * @return An operation that can be used to auto-resolve conflicts.  It is expected
     *            the operation will be run once.  To rerun an operation, retrieve
     *            a new operation and configure it accordingly.
     */
    public IAutoResolveConflictsOperation getAutoResolveConflictsOperation(AutoResolveConflictsDilemmaHandler dilemmaHandler, IRepositoryResolver resolver);
    
    /**
     * Provides an operation that can be used to auto-resolve conflicts
     * @param dilemmaHandler 
     *            The dilemmaHandler that will be called to backup files while auto-merge is running.
     *
     * @return An operation that can be used to auto-resolve conflicts.  It is expected
     *            the operation will be run once.  To rerun an operation, retrieve
     *            a new operation and configure it accordingly.
     */
    public IAutoResolveLocalConflictsOperation getAutoResolveLocalConflictsOperation(BackupDilemmaHandler dilemmaHandler);

    /**
     * Provides an operation to mark conflicts as having been resolved (merged).
     * 
     * @param dilemmaHandler
     *            The dilemmaHandler that will be called if problems are encountered;
     *            can be <code>null</code> in which case a default
     *            handler will be used.
     * @return An operation that can be used to mark conflicts (or parts of conflicts as
     *            having been merged.  Can be used for both resolve as merged (I changed
     *            the stuff) and resolve with mine (mine is right ignore accepted) by
     *            configuring the dilemma handler.
     */
    public IMarkAsMergedOperation getMarkAsMergedOperation(MarkAsMergedDilemmaHandler dilemmaHandler);
    
    /**
     * Provides an operation to remove a local conflict marker from an item.
     * No dilemma handler is passed since there is never a need to prompt the user
     * during the operation
     *  
     * @return An operation that can be used to remove a local conflict marker from an item.
     *         Used for resolve as merged for a local conflict after an out of sync reload
     */
    public IResolveLocalConflictsAsMergedOperation getResolveLocalConflictsAsMergedOperation();
    
    /**
     * Provides an operation to accept proposed changes and resolve the conflict.
     * 
     * @param dilemmaHandler
     *            The dilemmaHandler that will be called if problems are encountered;
     *            can be <code>null</code> in which case a default
     *            handler will be used.
     * @return An operation that can be used to apply the proposed changes and resolve the
     *            conflicts.  Applying the changes may involve reversing changes the user
     *            has made.
     */
    public IApplyAcceptedOperation getApplyAcceptedOperation(ApplyAcceptedDilemmaHandler dilemmaHandler);

    /**
     * Provides an operation that can be used to move folders from one component to another
     * within a workspace.
     * @param dilemmaHandler
     *            The dilemmaHandler that will be called if the problems are encountered
     *            during the move; can be <code>null</code> in which case a default
     *            handler will be used.
     * @return An operation that can be used to move folders between components within a workspace.
     *            It is expected the operation will be run once.  To rerun an operation, retrieve
     *            a new operation and configure it accordingly.
     */
    public IMoveFoldersOperation getMoveFoldersOperation(MoveFoldersDilemmaHandler dilemmaHandler);
    
    /**
     * Provides an operation that can be used to move versionables from one component to another
     * within a workspace.
     * @param dilemmaHandler
     *            The dilemmaHandler that will be called if the problems are encountered
     *            during the move; can be <code>null</code> in which case a default
     *            handler will be used.
     * @return An operation that can be used to move versionables between components within a workspace.
     *            It is expected the operation will be run once.  To rerun an operation, retrieve
     *            a new operation and configure it accordingly.
     */
    public IMoveVersionablesOperation getMoveVersionablesOperation(MoveVersionablesDilemmaHandler dilemmaHandler);
    
    /**
     * Provide an operation that can be used to delete folders immediately under
     * under the share from the repository.  Folders below the share root should
     * be deleted from the copy file area and then the resulting local changes
     * committed.
     * 
     * @param dilemmaHandler
     *            The dilemma handler that will be called if problems are encountered.
     *            Can be <code>null</code> in which case a default handler will be used.
     * @return an operation that can be used to delete folders immediately under a share root.            
     *            It is expected the operation will be run once.  To rerun an operation, retrieve
     *            a new operation and configure it accordingly.
     */
    public IDeleteFoldersOperation getDeleteFoldersOperation(CommitDilemmaHandler dilemmaHandler);
    
    /**
     * Provides an operation to collect and execute lock requests
     * 
     * @param dilemmaHandler
     *            The dilemmaHandler that will be called if problems are encountered;
     *            can be <code>null</code> in which case a default
     *            handler will be used.
     * @return An operation that can be used to lock resources. It is expected
     *            the operation will be run once.  To rerun an operation, retrieve
     *            a new operation and configure it accordingly.
     */
    public ILockOperation getLockOperation(LockDilemmaHandler dilemmaHandler);

    /**
     * Provides an operation to collect and execute unlock requests
     * 
     * @param dilemmaHandler
     *            The dilemmaHandler that will be called if problems are encountered;
     *            can be <code>null</code> in which case a default
     *            handler will be used.
     * @return An operation that can be used to unlock resources. It is expected
     *            the operation will be run once.  To rerun an operation, retrieve
     *            a new operation and configure it accordingly.
     */
    public IUnlockOperation getUnlockOperation(LockDilemmaHandler dilemmaHandler);

    /**
     * Provides an operation to collect and execute lock transfer requests
     * 
     * @param dilemmaHandler
     *            The dilemmaHandler that will be called if problems are encountered;
     *            can be <code>null</code> in which case a default
     *            handler will be used.
     * @return An operation that can be used transfer locks on resources. It is expected
     *            the operation will be run once.  To rerun an operation, retrieve
     *            a new operation and configure it accordingly.
     */
    public ITransferLockOperation getTransferLockOperation(LockDilemmaHandler dilemmaHandler);
    
    /**
     * Provides an operation which when run will repair the
     * metadata for the copy file area at the supplied root.
     * @param root The root of the sandbox/copy file area to repair
     * @param resolver A repository resolver used to obtain the team repository
     * used when accessing connections for metadata repair. 
     * @return An operation to repair metadata.
     */
    public IMetadataRepairOperation getMetadataRepairOperation(ILocation root, MetadataRepairDilemmaHandler dilemmaHandler, IRepositoryResolver resolver);
    
    /**
     * Provides an operation which when run will synchronize the metadata
     * with the contents of the repository.  
     * @param resolver A repository resolver used to obtain the team repository
     * used when accessing connections for the synchronization. 
     * @return An operation to synchronize metadata.
     */
    public IMetadataSynchronizeOperation getMetadataSynchronizeOperation(IRepositoryResolver resolver);
    
    /**
     * Provides an operation which when run will validate the
     * metadata for the copy file area at the supplied root.
     * @param root The root of the sandbox/copy file area to repair
     * @param writer the writer in which errors or metadata can be dumped
     * @return An operation to validate metadata.
     */
    public IMetadataValidateOperation getMetadataValidateOperation(ILocation root);

    /**
     * Provides an operation which when run will examine that sandbox and
     * describe the shares loaded in it.  The resulting description can
     * then be serialized into XML load rules.
     * @param dilemmaHandler
     *            The dilemmaHandler that will be called if problems are encountered;
     *            can be <code>null</code> in which case a default
     *            handler will be used.
     * @return An operation that can be configured to examine sandbox(es).
     * @since 0.9 (RTC 4.0)
     */
    public ISandboxExaminerOperation getSandboxExaminerOperation(SandboxExaminerDilemmaHandler dilemmaHandler);

    /**
     * Provides an operation that can be used to abort current patches and clear the accept queue
     * for a component in a workspace.
     * 
     * @param dilemmaHandler
     *            The dilemmaHandler that will be called if problems are encountered;
     *            can be <code>null</code> in which case a default
     *            handler will be used.
     *
     * @return An operation to be used to abort current the current patch for a
     *         component in a workspace.
     *
     * @since 0.9.0.5 (RTC 4.0.5)  
     */
    public IAbortCurrentPatchOperation getAbortCurrentPatchOperation(AbortCurrentPatchDilemmaHandler dilemmaHandler);

    /**
     * Provides an operation that can be used to update the current patch for a component
     * in a workspace.
     * 
     * @param dilemmaHandler
     *            The dilemmaHandler that will be called if problems are encountered;
     *            can be <code>null</code> in which case a default
     *            handler will be used.
     *
     * @return An operation to be used to update the current patch for a component
     *         in a workspace.
     *
     * @since 0.9.0.5 (RTC 4.0.5)
     */
    public IUpdateCurrentPatchOperation getUpdateCurrentPortOperation(UpdateCurrentPatchDilemmaHandler dilemmaHandler);

    /**
     * Provides an operation that can be used to complete the current patch for a 
     * component in a workspace.
     * 
     * @param dilemmaHandler
     *            The dilemmaHandler that will be called if problems are encountered;
     *            can be <code>null</code> in which case a default
     *            handler will be used.
     *
     * @return An operation to be used to complete the current patch for a component
     *         in a workspace.
     *
     * @since 0.9.0.5 (RTC 4.0.5)
     */
    public ICompleteCurrentPatchOperation getCompleteCurrentPatchOperation(CompleteCurrentPatchDilemmaHandler dilemmaHandler);

    /**
     * Provides an operation that can be used to begin processing the pending port list
     * for a component in a workspace.
     *
     * @param dilemmaHandler
     *            The dilemmaHandler that will be called if problems are encountered;
     *            can be <code>null</code> in which case a default
     *            handler will be used.
     *
     * @return An operation to be used to begin processing the pending port list for
     *         a component in a workspace.
     *
     * @since 0.9.0.5 (RTC 4.0.5)
     */
    public IProcessAcceptQueueOperation getProcessAcceptQueueOperation(ProcessAcceptQueueDilemmaHandler dilemmaHandler);

    /**
     * Provides an operation that can be used to close change sets
     *
     * @param dilemmaHandler
     *            The dilemmaHandler that will be called if problems are encountered;
     *            can be <code>null</code> in which case a default
     *            handler will be used.
     *
     * @return An operation to be used to close change sets.
     * 
     * @since 0.9.0.5 (RTC 4.0.5)
     */
    public ICloseChangeSetsOperation getCloseChangeSetsOperation(CloseChangeSetsDilemmaHandler dilemmaHandler);

    /**
     * Return an operation that can be used to apply a set of state changes to a workspace/component
     * 
     * @param dilemmaHandler
     *            The dilemmaHandler that will be called if problems are encountered;
     *            can be <code>null</code> in which case a default
     *            handler will be used.
     *
     * @return An operation to be used to apply a set a state changes as a patch.
     * 
     * @since 0.10.0.2 (RTC 5.0.2)
     */
    public ICreateCurrentPatchOperation getCreateCurrentPatchOperation(CreateCurrentPatchDilemmaHandler dilemmaHandler);

    /**
     * Return an operation for adding and removing subcomponents from a parent component.
     * 
     * @param dilemmaHandler
     *            The dilemmaHandler that will be called if problems are encountered;
     *            can be <code>null</code> in which case a default
     *            handler will be used.
     *
     * @return An operation for adding and removing subcomponents from a parent component.
     * 
     * @since 0.11.0 (RTC 6.0)
     */
    public IUpdateSubcomponentOperation getUpdateSubcomponentOperation(UpdateDilemmaHandler dilemmaHandler);
}
