/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2006, 2012. All Rights Reserved. 
 * 
 * Note to U.S. Government Users Restricted Rights:  Use, 
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.filesystem.client;

import java.io.Serializable;

import org.eclipse.core.runtime.IPath;

/**
 * Represents a location relative to another concrete location which is physical location.
 * Typically it is a location relative to a sandbox.
 * 
 * @NoImplement This interface is not intended to be implemented by clients.
 * 
 * @remoteProjects This interface is referenced by the remote project integration. 
 *      Downstream teams may need to adopt changes made to it.
 */
public interface IRelativeLocation extends Serializable {
    
    /**
     * @return a string representation of this path suitable for display purposes.
     */
    public String toString();
    
    /**
     * @return Whether this is an empty location (i.e. no parts to the location)
     */
    public boolean isEmpty();
    
    /**
     * @return the relative location that holds this item.  May be empty path
     * if this is at the root of the relative location.
     */
    public IRelativeLocation getParent();

    /**
     * Returns the name of the file or directory denoted by this abstract
     * relative location.  This is just the last name in the pathname's name
     * sequence.  If the pathname's name sequence is empty, then null
     * is returned.
     * 
     * this == this.getParent().append(this.getName))
     *
     * @return  The name of the file or directory denoted by this abstract
     *          pathname, or null if this pathname's name sequence
     *          is empty
     */
    public String getName();
    
    /**
     * adds the name supplied to the end of the location specified
     * and returns the new location. The relative location has no
     * semantics related to the physical location (only known by {@link ILocation})
     * it is purely a relative path, which might not be valid once appended to a
     * pure location.
     * 
     * @param name The name to be appended to the location.
     * @return The new relative location represented by this location with the name.
     */
    public IRelativeLocation append(String name);
    
    /**
     * adds the relative location supplied to the end of this relative location
     * and returns the new relative location.
     * 
     * @param location The relative location to be appended to this location.
     * @return The new relative location represented by this relative location  
     * further qualified by the relative location appended.
     */
    public IRelativeLocation append(IRelativeLocation location);
    
    /**
     * Answers an integer hash code for the receiver. Any two
     * objects which answer <code>true</code> when passed to
     * <code>.equals</code> must answer the same value for this
     * method.
     *
     * @return      int
     *                  the receiver's hash.
     *
     * @see         #equals
     */
    public int hashCode();

    /**
     * Compares the argument to the receiver, and answers true
     * if they represent the <em>same</em> object using a class
     * specific comparison. The implementation in Object answers
     * true only if the argument is the exact same object as the
     * receiver (==).
     *
     * @param       o Object
     *                  the object to compare with this object.
     * @return      boolean
     *                  <code>true</code>
     *                      if the object is the same as this object
     *                  <code>false</code>
     *                      if it is different from this object.
     * @see         #hashCode
     */
    public boolean equals (Object o);

    /**
     * Returns whether this relative location is a prefix of the given relative location.
     * To be a prefix, this relative location's segments must
     * appear in the argument relative location in the same order. The comparison is
     * case sensitive
     * <p>
     * An empty path is a prefix of all relative locations.
     * </p>
     * @param anotherLocation the other relative location that we want to know if we are a prefix of
     * @return <code>true</code> if this relative location is a prefix of the given relative location,
     *    and <code>false</code> otherwise
     */
    public boolean isPrefixOf(IRelativeLocation anotherLocation);


    /**
     * Returns whether this relative location is a prefix of the given relative location.
     * To be a prefix, this relative location's segments must
     * appear in the argument relative location in the same order.
     * <p>
     * An empty path is a prefix of all relative locations.
     * </p>
     * @param anotherLocation the other relative location that we want to know if we are a prefix of
     * @param isCaseSensitive whether or not to perform a case sensitive comparison or not
     * @return <code>true</code> if this relative location is a prefix of the given relative location,
     *    and <code>false</code> otherwise
     */
    public boolean isPrefixOf(IRelativeLocation anotherLocation, boolean isCaseSensitive);

    /**
     * Helps us to deal with case insensitive platforms better when the locations
     * in question do not exist and as such we want a constant form to compare against.
     * 
     * Will convert everything to lower case if the FS is not case sensitive and convertCase
     * is true.
     */
    public IRelativeLocation getCanonicalForm(boolean isCaseSensitive, boolean convertCase);
    
    /**
     * To efficiently access the pieces of the relative location 
     * without having to do getParent() til empty.
     * @return segments in this relative location.
     */
    public String[] segments();

    /**
     * The number of parts to the relative path
     */
    public int segmentCount();
    
    /**
     * Returns the specified part of this location, or
     * <code>null</code> if the relative location does not have such a part.
     *
     * @param index the 0-based part index
     * @return the specified part, or <code>null</code> 
     */
    public String segment(int index);
    
    /**
     * To conveniently turn the relative path into an IPath
     * @return
     */
    public IPath toPath();
    
    
    /**
     * Determine if this relative location matches another relative location.  isCaseSensitive
     * will dictate whether to do a case-sensitive comparison or not.
     * @param anotherLocation The other relative location to compare against.
     * @param isCaseSensitive Dictates whether to do a case-sensitive comparison or not.  If
     * true, the compare is equivalent to {@link #equals(Object)}.  If false
     * then a case insensitive compare is done.
     * @return if the this is the same relative location as another location
     */
    public boolean sameLocation(IRelativeLocation anotherLocation, boolean isCaseSensitive);
    
    /**
     * Returns a copy of this path with the given number of parts
     * removed from the beginning. 
     * The number must be greater or equal zero.
     * If the count is zero, this relative location is returned.
     * If the number equals or exceeds the number
     * of parts in this path, an empty relative location is returned.
     *
     * @param count the number of parts to remove
     * @return the new relative location
     */
    public IRelativeLocation removeFirstSegments(int count);

    /**
     * Returns a copy of this path with the given number of parts
     * removed from the end.
     * The number must be greater or equal zero.
     * If the count is zero, this relative location is returned.
     * <p>
     * If the number equals or exceeds the number
     * of parts in this path, a path with no parts is returned.
     * </p>
     *
     * @param count the number of segments to remove
     * @return the new relative location
     */
    public IRelativeLocation removeLastSegments(int count);
    

    /**
     * Returns a copy of this relative location truncated after the
     * given number of parts. The number must not be negative.
     * <p>
     * Copying up to part zero simply means making an copy with
     * no parts.
     * </p>
     *
     * @param count the part number at which to truncate the relative location
     * @return the new relative location
     */
    public IRelativeLocation uptoSegment(int count);
}
