/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2006, 2012. All Rights Reserved. 
 * 
 * Note to U.S. Government Users Restricted Rights:  Use, 
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.filesystem.client;

import java.util.Collection;
import java.util.List;

import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IProgressMonitor;

import com.ibm.team.filesystem.client.internal.utils.LoadedConfigurationDescriptor;
import com.ibm.team.scm.common.IComponentHandle;
import com.ibm.team.scm.common.IContextHandle;
import com.ibm.team.scm.common.IVersionableHandle;
import com.ibm.team.scm.common.dto.INameItemPair;

/**
 * Representation of the copy file area intended for use 
 * by clients that do not manipulate the copy file area.
 * @since 0.7
 */
public interface ISandbox {
    
    /**
     * Returns the location at which the copy file area (Sandbox) is anchored.
     * Never <code>null</code>
     */
    public ILocation getRoot();
    
    /**
     * Returns all known shares in this particular copy file area.
     * 
     * @return All the known shares in the the copy file areas.  Never <code>null</code>
     * List may be empty if there are no shares in this sandbox.
     * 
     * @throws FileSystemException if anything goes wrong
     * @LocalOp This is a local operation; the server/repository itself is not
     *          accessed.
     */
    public IShare[] allShares(IProgressMonitor monitor) throws FileSystemException;


    /**
     * Returns the shareable object corresponding to the given connection, component and item id in this copy file
     * area.
     * 
     * @param versionable The versionable to be located in the copy file area.  Never <code>null</code>
     * @param component The component loaded in the copy file area to be examined.  Never <code>null</code>
     * @param connection The context of the connection loaded into the copy file area.  Other contexts
     *            loaded are not examined.  Never <code>null</code>
     *            
     * @return a shareable, or <code>null</code> if no existing shareable is associated with the given
     *          connection, component and item id in this copy file area.
     * @LongOp This is a long operation; it may block indefinitely; must not be
     *         called from a responsive thread. 
     */
    public IShareable findShareable(IContextHandle connection, IComponentHandle component, IVersionableHandle versionable, IProgressMonitor monitor) throws FileSystemException;

    /** 
     * Get the shareable at the given path.  The shareable may or may not exist in the repository.
     * The shareable may or may not exist on disk. The shareable may or may not be shared.
     *      
     * @param path The path to the shareable within this CFA.  Must not be <code>null</code>
     *      The path must have atleast 1 segment. 
     * 
     * @param isFolderHint A hint indicating whether the shareable should be 
     *      considered a folder, file or symbolic link. May not be respected.
     * 
     * @return An IShareable representing the item at the given path.  
     *      Never <code>null</code>  
     * 
     * @LongOp This is a long operation which should not be called from a 
     *      responsive thread. 
     * @deprecated use {@link #findShareable(ILocation, ResourceType)}
     */
    public IShareable findShareableByPath(IPath path, ResourceType typeHint);

    /** 
     * Get the shareable at the given path.  The shareable may or may not exist in the repository.
     * The shareable may or may not exist on disk. The shareable may or may not be shared.
     *      
     * @param path The path to the shareable within this CFA.  Must not be <code>null</code>
     *      The path must have atleast 1 segment. 
     * 
     * @param isFolderHint A hint indicating whether the shareable should be 
     *      considered a folder, file or symbolic link. May not be respected.
     * 
     * @return An IShareable representing the item at the given path.  
     *      Never <code>null</code>  
     * 
     * @LongOp This is a long operation which should not be called from a 
     *      responsive thread. 
     */
    public IShareable findShareable(IRelativeLocation path, ResourceType typeHint);
    
    /**
     * The path is calculated relative to the shares in the copy file area.
     * If the item does not fall under a share, null will be returned.  This will correctly
     * handle trees that do not fall under the component root that have been shared.
     * It is recommended the caller first attempt to find the item using 
     * {@link #findShareable(IContextHandle, IComponentHandle, IVersionableHandle, IProgressMonitor)}.
     * If not found that way, this can be used to determine where it is supposed to be in the
     * copy file area or if ever loaded.
     * </p>
     * @param component The component loaded in this copy file area to be examined.  Never <code>null</code>
     * @param context The context of the connection loaded into this copy file area.  Other contexts
     *            loaded are not examined.  Never <code>null</code>
     * @param ancestorReports A list of The list of item name pairs that describe the path to the 
     * element.  Never <code>null</code>
     * @return a list of paths to the items described by each ancestor report.  Each entry in the resulting
     * list reflects the path relative to the share for the corresponding entry in the ancestor report list.
     * The path in the result may be <code>null</code> if the tree containing the item has not been loaded into the
     * copy file area.
     * 
     * @LongOp This is a long operation; it may block indefinitely; must not be
     *         called from a responsive thread. 
     *            
     * @throws FileSystemException If problems occur reading metadata
     */
    public List<IRelativeLocation> getPathRelativeToShares(IContextHandle context, IComponentHandle component, List<List<INameItemPair>> ancestorReports, IProgressMonitor progress) throws FileSystemException;

    /**
     * @return <code>true</code> if the filesystem at the copy file area root is case sensitive.
     */
    public boolean isCaseSensitive();
    
    /**
     * @return <code>true</code> if the sandbox at this path is currently registered.
     * returns <code>false</code> if the sandbox at this path is not currently registered.
     * Note that the sandbox maybe a valid sandbox, but it is currently not known yet
     */
    public boolean isRegistered();
    
    /**
     * @return <code>true</code> is an existing sandbox.  This means that it is either a
     * registered sandbox, or a previously registered sandbox that could be registered.
     * Only considers previously registered sandboxes for this release.
     * The sandbox may be empty.  Returns <code>false</code> if this just points to a
     * directory that is not a sandbox (i.e. does not have metadata for the current release).
     */
    public boolean isExistingSandbox();
    
    /**
     * @return <code>true</code> if the metadata of the sandbox is known to be corrupt.
     * The sandbox must be registered in order to determine if it is known to be corrupt.
     * The caller should register the sandbox first.
     */
    public boolean isCorrupted(IProgressMonitor monitor) throws FileSystemException;
    
    
    /**
     * @param monitor
     * @return a collection of all the configurations (workspace - component pairs) that
     * have been loaded into this copy file area.  Never <code>null</code>
     * @throws FileSystemException
     */
    public Collection<LoadedConfigurationDescriptor> allLoadedConfigurations(IProgressMonitor monitor) throws FileSystemException;
}
