/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2005, 2013. All Rights Reserved. 
 * 
 * Note to U.S. Government Users Restricted Rights:  Use, 
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.filesystem.client;

import org.eclipse.core.runtime.IAdaptable;
import org.eclipse.core.runtime.IProgressMonitor;

import com.ibm.team.filesystem.client.internal.IShareableVisitor;
import com.ibm.team.filesystem.client.internal.ISharingMetadata;
import com.ibm.team.filesystem.common.FileLineDelimiter;
import com.ibm.team.scm.common.IVersionableHandle;

/**
 * Describes the basic protocol for shareable objects. Shareable objects are
 * usually file system objects, such as directories and files. This interface
 * provides a SCM-oriented view for those objects.
 * <p>
 * The underlying shareable object can be obtained by adapting this object to 
 * the underlying object type, or by obtaining them from {@link ISharingManager}
 * and {@link ISandbox}. 
 * </p>
 * A sandbox is a place on disk that contains files/folders.  The files/folders
 * within can be represented by an IShareable.  The shareable can represent an 
 * item that may or may not exist on disk at that instance in time.  The shareable
 * can represent an item that may or may not be shared with the repository.
 * <p>
 * None of these method contact the server. However, several of the methods do
 * obtain locks on the sandbox. Because of this external clients must not call these methods 
 * while other locks are held. For instance, these methods must not be called from within
 * an Eclipse resource delta listener since a low level Eclipse workspace lock is held
 * by the delta notification thread. If clients need to call this API in response to
 * a resource delta, they must spawn a separate thread or job.
 * </p>
 * @NoImplement This interface is not intended to be implemented by clients.
 * @since 0.6
 */
public interface IShareable extends IAdaptable {

    /**
     * Depth constant indicating all child shareables should be visited.
     * 
     * @see #accept(IShareableVisitor, int, boolean, boolean, IProgressMonitor)
     */
    public static final int DEPTH_INFINITE = ISharingMetadata.DEPTH_INFINITE;

    /**
     * Depth constant indicating this shareable and its immediate children
     * should be visited.
     * 
     * @see #accept(IShareableVisitor, int, boolean, boolean, IProgressMonitor)
     */
    public static final int DEPTH_ONE = ISharingMetadata.DEPTH_ONE;

    /**
     * Depth constant indicating only this shareable should e visited.
     * 
     * @see #accept(IShareableVisitor, int, boolean, boolean, IProgressMonitor)
     */
    public static final int DEPTH_ZERO = ISharingMetadata.DEPTH_ZERO;

    /**
     * State flag constant (bit mask) indicating whether a shareable object is
     * actually shared.
     * 
     * @see #getState()
     */
    public final int SHARED = 0x01;

    /**
     * State flag constant (bit mask) indicating whether a shareable object
     * exists remotely in a repository.
     * 
     * @see #getState()
     */
    public final int REMOTE = 0x02;
    
    /**
     * State flag constant (bit mask) indicating whether a shareable object IS
     * LOCAL (i.e. exists locally).
     * 
     * @see #getState()
     */
    public final int LOCAL = 0x08;

    /**
     * Returns <code>true</code> if the resource should not be committed to
     * the repository. Note that it's quite possible that it has been committed
     * anyway in which case it will this method still returns <code>true</code>.
     * 
     * @return returns <code>true</code> if the resource should not be
     *         committed to the repository, and <code>false</code> otherwise.
     * @throws FileSystemException 
     */
    public boolean shouldBeIgnored(IProgressMonitor monitor) throws FileSystemException;

    /**
     * Returns the share to what this shared object belongs. Returns
     * <code>null</code> if this object does not belong to a share.
     * <p>
     * Note that this method will acquire a lock on the sandbox and must not be called
     * from within a Eclipse resource delta.
     * 
     * @return the share, or <code>null</code>
     * @throws FileSystemException 
     */
    public IShare getShare(IProgressMonitor monitor) throws FileSystemException;
    
    /**
     * Returns true if this shareable is the root of the share.
     * <p>
     * Note that this method will acquire a lock on the sandbox and must not be called
     * from within a Eclipse resource delta.
     * 
     * @return <code>true</code> if this shareable is the share root.
     * <code>false</code> otherwise
     * @throws FileSystemException 
     */
    public boolean isShare(IProgressMonitor monitor) throws FileSystemException;

    /**
     * Returns the remote resource of this object in the repository. Returns
     * <code>null</code> if this shareable does not exist in the repository yet.
     * <p>
     * Note that this method will acquire a lock on the sandbox and must not be called
     * from within a Eclipse resource delta.
     * 
     * @return the corresponding remote file system item handle, or <code>null<code>
     */
    public IVersionableHandle getRemote(IProgressMonitor monitor) throws FileSystemException;
    
    /**
     * Returns the versionable for this shareable. Differently from {@link #getRemote()}, this operation
     * returns a non-null reference regardless this shareable has been checked-in or not. The only exception is for 
     * if this shareable is not supposed to be checked in (should be ignored).
     * <p>
     * Note that this method will acquire a lock on the sandbox and must not be called
     * from within a Eclipse resource delta.
     * 
     * @return the corresponding file system item handle, or <code>null<code>
     */
    public IVersionableHandle getVersionable(IProgressMonitor monitor) throws FileSystemException;

    /**
     * Returns flags which describe in more detail the state of this shareable
     * object.
     * <p>
     * Note that this method will acquire a lock on the sandbox and must not be called
     * from within a Eclipse resource delta.
     * 
     * @see #SHARED
     * @see #REMOTE
     * @see #LOCAL
     * 
     * @return the flags
     */
    public int getState(IProgressMonitor monitor) throws FileSystemException;

    /**
     * Applies the given visitor for every (potentially non-existing) resource
     * for which there is sharing information.
     * 
     * @param visitor
     * @param depth
     * @param includeLocal
     *            whether non-remote resources should be included
     * @param mutable whether the visitor needs make changes to shareable metadata (requires exclusive access to metadata) 
     * @param monitor progress monitor
     * @throws FileSystemClientException
     */
    //public void accept(IShareableVisitor visitor, int depth, boolean includeLocal, boolean mutable, IProgressMonitor monitor) throws FileSystemClientException;
    
    /**
     * Applies the given visitor for every (potentially non-existing) resource
     * for which there is sharing information.
     * 
     * @param visitor
     * @param depth
     * @param includeLocal
     *            whether non-remote resources should be included
     * @param rule a scheduling rule to be used during the accept, or <code>null</code> if the visitor will not make changes to the metadata tree 
     * @param monitor progress monitor
     * @throws FileSystemClientException
     */
    // TODO if used, we should set a flag to say use scheduling rule for this shareable.
    //public void accept(IShareableVisitor visitor, int depth, boolean includeLocal, ISchedulingRule rule, IProgressMonitor monitor) throws FileSystemClientException;

    /**
     * Returns this shareable relative to the sandbox root.
     * 
     * @return the local file/resource's path
     */
    public IRelativeLocation getLocalPath();

    /**
     * @return the full path of the shareable
     * (sandbox root + local path). Never <code>null</code>
     */
    public ILocation getFullPath();
    
    /**
     * Returns the sandbox corresponding to the copy file area at which this
     * shareable is rooted.
     * 
     * @return The sandbox that contains this shareable.
     * @since 0.7
     */
    public ISandbox getSandbox();
    
    /**
     * Returns the pending change corresponding to this shareable object, or <code>null</code>
     * if this object should be ignored or isn't shared. If the shareable is shared but unchanged
     * a {@link ILocalChange#NO_CHANGE} is returned. 
     * <p>
     * Note that this method will acquire a lock on the sandbox and must not be called
     * from within a Eclipse resource delta.
     * 
     * @return a local change object, or <code>null</code> if this object should be ignored or
     * is not shared.
     */
    public ILocalChange getChange(IProgressMonitor progress) throws FileSystemException;
    
    /**
     * Returns a scheduling rule for this shareable.
     * 
     * @return a scheduling rule
     */
    // public ISchedulingRule getRule();
    
    /**
     * Returns the line delimiter of the content when it was loaded. Being different 
     * from {@link #getLineDelimiter()} means that there is a pending change for this 
     * item. Applicable only to file items, not folders.
     * <p>
     * Note that this method will acquire a lock on the sandbox and must not be called
     * from within a Eclipse resource delta.
     * 
     * @return the line delimiter for the content or <code>null</code> for new items
     * and folders.
     */
    public FileLineDelimiter getOriginalLineDelimiter(IProgressMonitor monitor) throws FileSystemException;
    
    /**
     * Returns the line delimiter of the content. It may be <code>null</code> meaning it 
     * will be determined when checked-in.  Typically the line delimiter is decided on 
     * the first check-in and is based upon the content type.  After that it is 
     * changed manually be the use. Applicable only to file 
     * items, not folders.
     * <p>
     * Note that this method will acquire a lock on the sandbox and must not be called
     * from within a Eclipse resource delta.
     * 
     * @return the line delimiter of the content or <code>null</code> if it 
     * will be determined on checked-in or this is a folder.
     */
    public FileLineDelimiter getLineDelimiter(IProgressMonitor monitor) throws FileSystemException;
    
    /**
     * Returns the content type (MIME type) of the content when it was loaded. Being 
     * different from {@link #getContentType()} is indicative of a pending change for 
     * the item. It may be <code>null</code> for new files and is applicable only to 
     * file items, not folders.
     * <p>
     * Note that this method will acquire a lock on the sandbox and must not be called
     * from within a Eclipse resource delta.
     * 
     * @return the content type of the content when it was loaded; or <code>null</code>
     * for new files and folders.
     */
    public String getOriginalContentType(IProgressMonitor monitor) throws FileSystemException;

    /**
     * Returns the current content type preference for this shareable. It
     * may be <code>null</code> meaning it will be determined when checked-in.  
     * Typically the content type is determined on first check-in and then 
     * changed manually by the user after that. Applicable only to file items, not 
     * folders.
     * <p>
     * Note that this method will acquire a lock on the sandbox and must not be called
     * from within a Eclipse resource delta.
     * 
     * @return the content type (MIME type) of the content of this shareable or 
     * <code>null</code> if it will be determined on check-in.
     */
    public String getContentType(IProgressMonitor monitor) throws FileSystemException;
    
    /**
     * Returns all the properties in the meta data.
     * <p>
     * Note that this method will acquire a lock on the sandbox and must not be called
     * from within a Eclipse resource delta.
     * 
     * @param monitor A monitor to mark progress on.
     * @return The properties stored in the meta data.  Never <code>null</code>
     * If this item is not shared, or has been deleted from the sandbox,
     * no properties (no original setting, no deltas) will be returned.
     * @throws FileSystemException Thrown if anything goes wrong
     */
    public IMetadataProperties getMetadataProperties(IProgressMonitor monitor) throws FileSystemException;

    /**
     * Returns whether this shareable has the executable bit set.
     * This is only applicable for files.
     * <p>
     * Note that this method will acquire a lock on the sandbox and must not be called
     * from within a Eclipse resource delta.
     * 
     * @return <code>true</code> if this storage is executable
     */
    public boolean isExecutable(IProgressMonitor monitor) throws FileSystemException;
    
    /**
     * Returns whether this storage actually exists.
     * <p>
     * Note that this method will acquire a lock on the sandbox and must not be called
     * from within a Eclipse resource delta.
     * 
     * @return <code>true</code> if the item exists on disk.  False otherwise.
     * @throws FileSystemException 
     */
    public boolean exists(IProgressMonitor progress) throws FileSystemException;

    /**
     * Return the type of the storage this shareable represents.
     * <p>
     * Note that this method will acquire a lock on the sandbox and must not be called
     * from within a Eclipse resource delta.
     * 
     * @return The type of the storage this shareable represents.  Returns <code>null</code>
     * if the shareable does not physically exist (the type is unknown).
     */
    public ResourceType getResourceType(IProgressMonitor progress) throws FileSystemException;

}
