/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2005, 2010. All Rights Reserved. 
 * 
 * Note to U.S. Government Users Restricted Rights:  Use, 
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.filesystem.client;

import com.ibm.team.repository.common.UUID;
import com.ibm.team.scm.client.IConnection;
import com.ibm.team.scm.common.IComponentHandle;
import com.ibm.team.scm.common.IContextHandle;
import com.ibm.team.scm.common.IVersionableHandle;

/**
 * Represents the connection information from a copy file area to the associated
 * location in the repository.
 * 
 * @since 0.6
 * @NoImplement This interface is not intended to be implemented by clients.
 */
public interface ISharingDescriptor {
    /**
     * @return The Component handle under which the root folder is located
     *         in the repository.
     * 
     * @LocalOp This is a local operation; the server/repository itself is not
     *          accessed.
     */
    public IComponentHandle getComponent();
    
    /**
     * Returns the name for the corresponding Component auditable.
     * Returns an unspecified string if the component name was not available.
     * It should be noted that the component name is persisted locally and can become
     * stale so this value should be treated as a hint only. To get the actual name, clients
     * must fetch the component from the repository.
     * 
     * @return the name; never <code>null</code>
     * @LocalOp This is a local operation; the server/repository itself is not
     *          accessed.
     */
    public String getComponentName();

    /**
     * @return The root versionable that was shared into the copy file area.
     * 
     * @LocalOp This is a local operation; the server/repository itself is not
     *          accessed.
     * @since 0.7
     */
    public IVersionableHandle getRootVersionable();

    /**
     * @return The handle of the connection that this copy file area is
     *         associated with.
     * 
     * @LocalOp This is a local operation; the server/repository itself is not
     *          accessed.
     */
    public IContextHandle getConnectionHandle();
    
    /**
     * @return The ID of the repository that this copy file area is associated
     *         with.
     * 
     * @LocalOp This is a local operation; the server/repository itself is not
     *          accessed.
     */
    public UUID getRepositoryId();

    /**
     * Determines if the copy file area is associated with a workspace
     * connection or a baseline connection.
     * 
     * @return <code>true</code> if the copy file area is shared from a
     *         workspace, <code>false</code> otherwise.
     * 
     * @LocalOp This is a local operation; the server/repository itself is not
     *          accessed.
     */
    public boolean associatedWithWorkspace();

    /**
     * Determines if the copy file area is associated with the connection
     * supplied.
     * 
     * @param connection
     *            The connection we want to know if the copy file area is shared
     *            with.
     * 
     * @return <code>true</code> if the copy file area is shared with the
     *         connection supplied, <code>false</code> otherwise.
     * 
     * @LocalOp This is a local operation; the server/repository itself is not
     *          accessed.
     */
    public boolean isAssociatedWithConnection(IConnection connection);
    
    /**
     * Returns the name for the corresponding connection auditable (either an IBaseline or an IWorkspace).
     * Returns an unspecified string if the connection name was not available.
     * It should be noted that the connection name is persisted locally and can become
     * stale so this value should be treated as a hint only. To get the actual name, clients
     * must fetch the connection from the repository.
     * 
     * @return the name; never <code>null</code>
     * @LocalOp This is a local operation; the server/repository itself is not
     *          accessed.
     */
    public String getConnectionName();
    
    /**
     * {@link #getRootVersionable()} identifies the item shared, but
     * for efficiency sake, that might not be the item to use when loading.
     * A better root for loading is the one identified by the load root.
     * @return the versionable to load to get this item loaded.
     */
    public IVersionableHandle getReloadRoot();
    
    /**
     * Return the client data that is associated with the given namespace
     * or <code>null</code> if the namespace has no data associated with it.
     * @param namespace the namespace of the client data to be returned
     * @return the client data that is associated with the given namespace
     * or <code>null</code>
     */
    public byte[] getClientData(String namespace);
}
