/*******************************************************************************
 * Licensed Materials - Property of IBM
 * © Copyright IBM Corporation 2005, 2015. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  Use, 
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.filesystem.client;

import java.io.IOException;
import java.io.InputStream;
import java.util.Collection;
import java.util.List;
import java.util.Map;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.jobs.IJobManager;
import org.eclipse.core.runtime.jobs.ISchedulingRule;

import com.ibm.team.filesystem.client.internal.localchanges.LocalConflictManager;
import com.ibm.team.filesystem.client.internal.utils.ConfigurationFacade;
import com.ibm.team.filesystem.common.FileLineDelimiter;
import com.ibm.team.filesystem.common.ISymbolicLinkHandle;
import com.ibm.team.scm.common.IComponentHandle;
import com.ibm.team.scm.common.IContextHandle;
import com.ibm.team.scm.common.IFolderHandle;
import com.ibm.team.scm.common.IVersionableHandle;

/**
 * The sharing manager provides access to files and folders which are loaded
 * from a repository workspace into a local file system, an area which we call
 * the copy file area (a.k.a CFA). It tracks the metadata needed to keep the 
 * two in sync. When a folder is loaded from a repository workspace we say
 * that it's shared with the repository and each CFA has a set of top-level
 * shares, which are essentially mount points with an associated remove location
 * in a repository workspace. The sharing configuration is described in a
 * {@link ISharingDescriptor}.
 * <p>
 * The {@link ILocalChangeManager} tracks the changes made to files and folders
 * that have been loaded from the repository and allows a client to access
 * the list of files and folders that have been modified but not checked-in
 * to its associated repository workspace.
 * </p>
 * @NoImplement This interface is not intended to be implemented by clients.
 * @since 0.6
 */
public interface ISharingManager {
    /**
     * The type of rule needed. Note that for moves the caller should combine
     * the delete rule of the source with the create rule of the target.
     */
    public enum RuleKind {
        CREATE,
        DELETE,
        MODIFY,
        REFRESH
    }


    /**
     * Adds a listener to copy file areas. Listeners will
     * be notified when the sharing state of a root folder changes.
     * All registered CFA instances will have the listener added to
     * them. Instances created in future will also have the listener
     * added to them.
     * 
     * @param listener
     *            a change listener.  Never <code>null</code>
     * @LocalOp This is a local operation; the server/repository itself is not
     *          accessed.
     */
    public void addListener(ICopyFileAreaListener listener);

    /**
     * Removes a listener previously registered with the copy
     * file areas. Has no affect if an identical listener is
     * not registered. All registered CFA instances will have the 
     * listener removed from them. Instances created in future
     * will not have the listener added to them. 
     * 
     * @param listener
     *            a synchronizer change listener.  Never 
     *            <code>null</code>
     * @LocalOp This is a local operation; the server/repository itself is not
     *          accessed.
     */
    public void removeListener(ICopyFileAreaListener listener);

    /**
     * Returns all known shares across all copy file areas.
     * 
     * @return All the known shares in all the registered
     * copy file areas.
     * 
     * @throws FileSystemException if anything does wrong
     * @LocalOp This is a local operation; the server/repository itself is not
     *          accessed.
     */
    public IShare[] allShares(IProgressMonitor monitor) throws FileSystemException;

    /**
     * Identifies if there are any shares for this connection in the copy file
     * area identified.
     * 
     * @param copyFileAreaRoot
     *            The path to the copyFileArea root. Must not be
     *            <code>null</code>
     * @param connection
     *            The workspace or baseline connection that we are interested in
     *            knowing if any elements from it have been shared.
     * @return True if there are shares that originate from the supplied
     *         connection in the copy file area. False otherwise.
     * @throws FileSystemClientException
     * @LocalOp This is a local operation; the server/repository itself is not
     *          accessed.
     */
    // public boolean isConnectionShared(IPath copyFileAreaRoot, IContextHandle connection) throws FileSystemClientException;

    /**
     * Identifies if there are any shares for this connection in any of the copy file areas registered.
     * 
     * @param connection
     *            The workspace or baseline connection that we are interested in
     *            knowing if any elements from it have been shared.
     * @return True if there are shares that originate from the supplied
     *         connection in the Eclipse workspace. False otherwise.
     * @throws FileSystemException
     * @LocalOp This is a local operation; the server/repository itself is not
     *          accessed.
     */
    public boolean isConnectionShared(IContextHandle connection) throws FileSystemException;

    /**
     * Returns the list of shareables in the local filesystem corresponding to the given item in the local file system
     * Multiple components and multiple connections can be loaded.
     * It is possible that more than 1 state of the versionable is loaded.  Hence more
     * than 1 shareable can be found.
     * 
     * @param versionable The versionable to be located in the copy file area.  Never <code>null</code>
     * 
     * @return a list of (element type: {@link IShareable}), never <code>null</code>
     * @LongOp This is a long operation; it may block indefinitely; must not be
     *         called from a responsive thread. 
     */
    public Collection<IShareable> findShareables(IVersionableHandle versionable, IProgressMonitor monitor) throws FileSystemException;

    /**
     * Returns the list of shareables in the local filesystem corresponding to the given item from the given
     * copy file area and the given connection.  Versionables can exist in more than 1 component.  Multiple
     * components within the connection may contain the versionable hence more than 1 shareable can be found. 
     * 
     * @param copyFileAreaRoot
     *            The path to the copyFileArea root. Must not be
     *            <code>null</code> 
     * @param versionable The versionable to be located in the copy file area.  Never <code>null</code>
     * @param connection The context of the connection loaded into the copy file area.  Other contexts
     *            loaded are not examined.  Never <code>null</code>
     * 
     * @return a list of (element type: {@link IShareable}), never <code>null</code>
     * @LongOp This is a long operation; it may block indefinitely; must not be
     *         called from a responsive thread. 
     */
    // public List findShareables(IPath copyFileAreaRoot, IVersionableHandle versionable, IContextHandle connection, IProgressMonitor monitor) throws FileSystemClientException;
    
    /**
     * Returns the list of shareables in the local filesystem corresponding to the given item in the given component.
     * Pieces of a component may be loaded from multiple connections.
     * Depending on the state of the components, it is possible that more than one state of a versionable exists
     * in the the copy file area.  Hence more than 1 shareable can be found.
     * 
     * @param component The component loaded to be examined.  Never <code>null</code>
     * @param versionable The versionable to be located in the local file system.  Never <code>null</code>
     * 
     * @return a list of (element type: {@link IShareable}), never <code>null</code>
     * @LongOp This is a long operation; it may block indefinitely; must not be
     *         called from a responsive thread. 
     */
    public Collection<IShareable> findShareables(IComponentHandle component, IVersionableHandle versionable, IProgressMonitor monitor) throws FileSystemException;

    /**
     * Returns the shareable object corresponding to the given connection, component and item id in any of
     * the currently known copy file areas.
     * 
     * @param versionable The versionable to be located in the copy file area.  Never <code>null</code>
     * @param component The component loaded in the copy file area to be examined.  Never <code>null</code>
     * @param connection The context of the connection loaded into the copy file area.  Other contexts
     *            loaded are not examined.  Never <code>null</code>
     *            
     * @return collection of shareables associated with the given
     *          connection, component and item id.  There maybe none in the list if none of the
     *          known copy file areas contain the shareable.  There may be more than one if the 
     *          item has been loaded into more than 1 copy file area.
     * @LongOp This is a long operation; it may block indefinitely; must not be
     *         called from a responsive thread. 
     */
    public Collection<IShareable> findShareables(IContextHandle connection, IComponentHandle component, IVersionableHandle versionable, IProgressMonitor monitor) throws FileSystemException;

    /**
     * Given a full path find the shareable that represents that path.  The 
     * Shareable may or may not exist in the repository.  The shareable
     * may or may not exist on disk.  The shareable may or may not be shared.
     * 
     * @param fullPath The full path to an item.  May not be
     *      <code>null</code>
     *        
     * @param resourceType A hint as to whether the path represents a folder.
     * It may be ignored.
     * 
     * @return The shareable for the path.  Will return null if the path
     * of the item does not fall within a known copy file area.
     */
    public IShareable findShareable(ILocation fullPath, ResourceType isFolderHint);
            
    /**
     * Get the name of the metadata root. This directory stores the internal metadata
     * about loaded repository workspaces and their files and folder versions. The 
     * contents of this directory are not to be modified by a user or an external 
     * application. 
     * 
     * @return the name of the directory which stores the 
     */
    // public String getMetadataRootName();

    /**
     * Returns the local change manager.  Never null
     * 
     * @return the local change manager
     */
    public ILocalChangeManager getLocalChangeManager();
    
    /**
     * Returns the local conflict manager.  Never null
     * 
     * @return the local conflict manager
     */
    public LocalConflictManager getLocalConflictManager();

    /**
     * Returns the number of shares in the given copy file area that are associated with 
     * the given component and context
     * 
     * @param copyFileAreaRoot
     *            The path to the copyFileArea root. Must not be
     *            <code>null</code> 
     * @param component The component loaded in the copy file area to be examined.  Never <code>null</code>
     * @param connection The context of the connection loaded into the copy file area.  Other contexts
     *            loaded are not examined.  Never <code>null</code>
     * 
     * @return the number of shares
     * @LongOp This is a long operation; it may block indefinitely; must not be
     *         called from a responsive thread. 
     */
    //public int getNumShares(IPath copyFileAreaRoot, IComponentHandle component, IContextHandle connection, IProgressMonitor monitor) throws FileSystemClientException;

    /**
     * Returns the number of shares in the default copy file area that are associated with 
     * the given component and context
     * <p>
     * This is a convenience method, fully equivalent to
     * <code>getNumShareables(getDefaultCFARoot(), ...)</code>
     * </p>
     * 
     * @param component The component loaded in the copy file area to be examined.  Never <code>null</code>
     * @param connection The context of the connection loaded into the copy file area.  Other contexts
     *            loaded are not examined.  Never <code>null</code>
     * 
     * @return the number of shares
     * @LongOp This is a long operation; it may block indefinitely; must not be
     *         called from a responsive thread. 
     */
    //public int getNumShares(IComponentHandle component, IContextHandle connection, IProgressMonitor monitor) throws FileSystemClientException;
    
    
    /**
     * Get SharingDescriptors for the given context and component
     * @param copyFileAreaRoot
     *            The path to the copyFileArea root. Must not be
     *            <code>null</code> 
     * @param component The component loaded in the copy file area to be examined.  Never <code>null</code>
     * @param context The context of the connection loaded into the copy file area.  Other contexts
     *            loaded are not examined.  Never <code>null</code>
     * @return The descriptors for all the root items loaded from this component-context.
     *            
     * @throws FileSystemClientException If problems occur reading metadata
     */
    //public Collection<ISharingDescriptor> getShares(IPath copyFileAreaRoot, IContextHandle context, IComponentHandle component, IProgressMonitor monitor) throws FileSystemClientException;
    
    /**
     * Creates a file with the given contents and the given item id and item type.
     * 
     * Fails if the file already exists, or the parent does not exist, or file does
     * not refer to a shared location, or an item with the same item id already exists
     * in a share for the same component/connection.
     * 
     * InputStream will be closed before this method returns.
     * 
     * @param file Where the file is to be created
     * @param in The contents of the file
     * @param lineDelimiterExpected the expected line delimiter of the file
     * @param encodingExpected the encoding expected of the file 
     * @param properties The properties of the file.
     * @param handle The handle of the file being restored.
     * @param monitor Monitor to mark progress on
     */
    public void restoreFile(IShareable file, InputStream in, FileLineDelimiter lineDelimiterExpected, String encodingExpected, Map<String, String> properties, IVersionableHandle handle, IProgressMonitor monitor) throws FileSystemException, IOException;
    
    /**
     * Creates a folder with the given item id.
     * 
     * Fails if the folder already exists, or the parent does not exist, or folder does
     * not refer to a shared location, or an item with the same item id already exists
     * in a share for the same component/connection.
     * 
     * @param folder Where the folder is to be created
     * @param properties The restored properties of the folder.  If null the properties of the folder
     * being restored will be untouched.
     * @param handle The handle of the folder being restored.
     * @param monitor Monitor to mark progress on
     */
    public void restoreFolder(IShareable folder, Map<String, String> properties, IFolderHandle handle, IProgressMonitor monitor) throws FileSystemException;
    
    /**
     * Creates a link with the given item id.
     * 
     * Fails if the shareable already exists, or the parent does not exist, or shareable does
     * not refer to a shared location, or an item with the same item id already exists
     * in a share for the same component/connection.
     * 
     * @param linkShareable Where the link is to be created
     * @param properties The properties of the link.  If null the properties of the link
     * being restored will be untouched.
     * @param handle The handle of the link being restored.
     * @param isDirectory True if the symbolic link is a directory type, false otherwise.
     * @param monitor Monitor to mark progress on
     */
    public void restoreLink(IShareable shareable, String target, boolean isDirectory, Map<String, String> properties, ISymbolicLinkHandle handle, IProgressMonitor monitor) throws FileSystemException;
    
    /**
     * Provides the sandbox representation of all currently registered copy file areas.
     * Never <code>null</code>
     * @since 0.7
     */
    public Collection<ISandbox> getRegisteredSandboxes();

    /**
     * Provides the sandbox representation for the copy file area root path supplied.
     * The caller may request that the sandbox be only returned if it is a registered
     * sandbox (i.e it is currently known).
     * 
     * @param copyFileAreaRoot The root of the sandbox
     * @param registered If <code>true</code> the sandbox must already be known.  If
     * <code>false</code> the sandbox may or may not be known already.
     * @return The sandbox requested.  May be <code>null</code> if registered is
     * <code>true</code> and there is no such registered copy file area.
     * @since 0.7
     */
    public ISandbox getSandbox(ILocation copyFileAreaRoot, boolean registered);
    
    /**
     * Provides the sandboxes that have the part or all of the configuration loaded within it.
     * If you want to know the sandboxes containing a particular item, consider using find shareables
     * and from that you can determine the sandbox.
     * @param configuration The configuration for which we want to the sandboxes that has its contents
     * loaded within it
     * @param monitor Monitor to mark progress on.
     * @return The registered sandboxes in which the configuration is loaded.
     * @throws FileSystemException
     * @since 0.7
     */
    public Collection<ISandbox> getSandboxes(ConfigurationFacade configuration, IProgressMonitor monitor) throws FileSystemException;

    /**
     * Register the given sandbox so that it is tracked by the sharing manager.
     * If the sandbox is already registered, this call will do nothing.
     * 
     * @param sandbox the sandbox to be registered.
     * @param mustExist <code>true</code> if the sandbox must already exist on 
     *      disk (a {@link FileSystemException} is thrown if it's new); or 
     *      <code>false</code> if the sandbox is allowed to be new. 
     * 
     * @throws FileSystemException if <code>mustExist</code> is true, 
     *      and the given sandbox does not already contain metadata (error 
     *      code {@link IFileSystemStatus#COPY_AREA_MISSING}). The error 
     *      code will be different if there is a problem with the subsequent registration of the otherwise legitimate sandbox.
     * @since 0.7
     */
    public void register(ISandbox sandbox, boolean mustExist, IProgressMonitor monitor) throws FileSystemException;
    
    /**
     * Register all of the given sandboxes. Any sandboxes that are already 
     * registered will be ignored.
     * 
     * @param sandbox Zero or more sandboxes to register.
     *
     * @throws CopyFileAreaLockedByOtherProcess if one or more of the sandboxes 
     *      are owned by another process.
     */
    public void register(List<ISandbox> sandbox, IProgressMonitor monitor) throws FileSystemException;
    
    /**
     * Detects whether sandbox listening is currently activated for the provided sandbox. This can 
     * still return false even if {@link #enableSandboxListening(IProgressMonitor)} has been called
     * 
     * 
     * @param sandbox A registered sandbox.
     * @return true if the sandbox has a sandbox listener enabling it to automatically detect pending local
     *   changes, false otherwise.
     */
    public boolean hasSandboxListener(ISandbox sandbox);
    
    /**
     * Returns whether a sandbox listener could work on the provided sandbox provided
     *  that {@link #enableSandboxListening(IProgressMonitor)} is called.
     *  
     * @param sandbox A valid sandbox.
     * 
     * @return true if sandbox listening can be enabled on this sandbox. False, otherwise.
     */
    public boolean canSupportSandboxListener(ISandbox sandbox);

    /**
     * Disables sandbox listening for all registered sandboxes. No further monitoring for
     *  pending local changes will occur on these sandboxes. It will be up to the client and/or user to
     *  refresh periodically in order to find the local changes.
     *  <p>
     *  This should <em>not</em> be called if the thread owns any other type of lock except
     *   for the eclipse workspace lock. Also, it should not be called directly from a UI
     *   thread as it is a potentially long-running operation.
     *  </p>
     */
    void disableSandboxListening(IProgressMonitor monitor) throws FileSystemException;
    
    /**
     * Enables sandbox listening for all registered sandboxes provided that the sandbox
     *  can support a listener. In many cases the client and user no longer have to refresh shareables/shares
     *  to discover local pending changes. Any new sandboxes that are registered after this call will
     *  potentially have sandbox listening enabled (provided that the sandbox and platform support this).
     *  <p>
     *  This should <em>not</em> be called if the thread owns any other type of lock except
     *   for the eclipse workspace lock. Also, it should not be called directly from a UI
     *   thread as it is a potentially long-running operation.
     *  </p>
     */
    public void enableSandboxListening(IProgressMonitor monitor) throws FileSystemException;
    
    /** 
     * Determines if the sharing manager has sandbox listening enabled.
     * 
     * @return true if sandbox listening is enabled on all supported sandboxes, false otherwise.
     */
    public boolean isSandboxListeningEnabled();
    
    /**
     * Deregister the given sandbox so that it is no longer tracked by the sharing manager.
     * Removing a sandbox that is not registered is a no-op.
     * @param sandbox the sandbox to be removed
     * @throws FileSystemException if an error occurred
     * @since 0.7
     */
    public void deregister(ISandbox sandbox, IProgressMonitor progress) throws FileSystemException;
    
    /**
     * Deregister the given sandbox so that it is no longer tracked by the sharing manager.
     * Also erase all metadata associated with the sandbox.  The metadata will be removed even
     * if the sandbox is not registered.  This should only be called under exceptional circumstances.
     * USE CAUTION when calling this as all metadata will be removed and normal operations to accept,
     * checkin, etc will nolonger be appropriate for the sandbox.
     * 
     * @param sandbox the sandbox whose metadata is to be removed
     * @throws FileSystemException if an error occurred
     * @since 3.0
     */
    public void eraseSandboxMetadata(ISandbox sandbox, IProgressMonitor progress) throws FileSystemException;
    
    /**
     * Provides a count of the number of shares loaded for the workspace/component pair
     * 
     * @param component The component loaded in the copy file area to be examined.  Never <code>null</code>
     * @param connection The context of the connection loaded into the copy file area.  Other contexts
     *            loaded are not examined.  Never <code>null</code>
     * 
     * @return the number of shares
     * @LongOp This is a long operation; it may block indefinitely; must not be
     *         called from a responsive thread. 
     * @throws FileSystemException 
     */
    public int getNumShares(IContextHandle context, IComponentHandle component, IProgressMonitor progress) throws FileSystemException;
    
    
    /**
     * Generate a scheduling rule that locks the eclipse workspace 
     * paths in the scope of any running IDE. The returned rule will prevent 
     * access to the given set of paths for any threads using 
     * {@link IJobManager#beginRule(ISchedulingRule, IProgressMonitor)}.
     * 
     * <p>
     * Since the lock is only respected within the IDE, the lock returned will 
     * only operate in that portion of the file system. 
     * 
     * @return A scheduling rule. Possibly <code>null</code>.
     */
    public ISchedulingRule makeSchedulingRuleForIDE();
    
    
    /**
     * Generate a scheduling rule that locks all of the given {@link IShareable} 
     * paths in the scope of any running IDE. The returned rule will prevent 
     * access to the given set of paths for any threads using 
     * {@link IJobManager#beginRule(ISchedulingRule, IProgressMonitor)}.
     * 
     * <p>
     * Since the lock is only respected within the IDE, the lock returned will 
     * only operate in that portion of the file system. 
     * 
     * @param shareables A list of shareables to be locked. 
     *      
     * @return A scheduling rule. Possibly <code>null</code>.
     */
    public ISchedulingRule makeSchedulingRuleForIDE(Collection<IShareable> shareables);
    
    
    /**
     * Generate a scheduling rule that locks all of the given {@link IShareable} 
     * paths in the scope of any running IDE. The returned rule will prevent 
     * access to the given set of paths for any threads using 
     * {@link IJobManager#beginRule(ISchedulingRule, IProgressMonitor)}.
     * 
     * <p>
     * Since the lock is only respected within the IDE, the lock returned will 
     * only operate in that portion of the file system. 
     * 
     * @param shareables A list of shareables to be locked.
     * 
     * @param kind The type of rule to provide. 
     *      
     * @return A scheduling rule. Possibly <code>null</code>.
     */
    public ISchedulingRule makeSchedulingRuleForIDE(Collection<IShareable> shareables, RuleKind kind);
    
    /**
     * Mergers, file wrappers etc. are interested in the properties of a file
     * prior to the initial commit, after being committed and what the properties
     * would be based upon preferences/magic files.  The content examiner is
     * able to examine a file and answer those questions based upon the contents,
     * preferences and current properties.
     * 
     * @param shareable the shareable to be examined
     * 
     * @return A content manager capable of divining the properties of files. 
     */
    public IContentExaminer getContentExaminer(IShareable shareable);
    
    /**
     * For an uncommitted (and potentially changed) item, find the original 
     * state of the data from before an local changes. 
     * 
     * @return An {@link IUnmodifiedInfo} item for each sandbox the item has 
     *      been loaded into. If the item is not loaded, or the metadata is
     *      otherwise unavailable, an empty list will be returned. 
     */
    public Collection<IUnmodifiedInfo> findUnmodifiedInfoFor(IContextHandle ctx, IComponentHandle component, IVersionableHandle item, IProgressMonitor progress) throws FileSystemException;
}
