/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2008, 2014. All Rights Reserved. 
 * 
 * Note to U.S. Government Users Restricted Rights:  Use, 
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.filesystem.client.daemon;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.net.URLDecoder;
import java.nio.charset.Charset;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Locale;
import java.util.Map;
import java.util.Set;

import org.apache.commons.logging.Log;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IConfigurationElement;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IProgressMonitorWithBlocking;
import org.eclipse.core.runtime.RegistryFactory;

import com.ibm.team.filesystem.client.FileSystemStatusException;
import com.ibm.team.filesystem.client.internal.daemon.Daemon;
import com.ibm.team.filesystem.client.internal.daemon.FSDaemon;
import com.ibm.team.filesystem.client.internal.http.HttpContext;
import com.ibm.team.filesystem.client.internal.http.HttpHandler;
import com.ibm.team.filesystem.client.internal.http.HttpRequest;
import com.ibm.team.filesystem.client.internal.http.HttpResponse;
import com.ibm.team.filesystem.client.internal.http.HttpServer;
import com.ibm.team.filesystem.client.internal.http.constants.Header;
import com.ibm.team.filesystem.client.internal.http.constants.HttpMethod;
import com.ibm.team.filesystem.client.internal.http.constants.ResponseCode;
import com.ibm.team.filesystem.client.internal.marshalling.ExceptionHandlingJSONSerializer;
import com.ibm.team.filesystem.client.internal.marshalling.ParameterWrapperDeserializer;
import com.ibm.team.filesystem.client.internal.rest.CommonUtil;
import com.ibm.team.filesystem.client.internal.rest.IValidatingParameterWrapper;
import com.ibm.team.filesystem.client.restproxy.exceptions.RestMarshallingException;
import com.ibm.team.filesystem.client.restproxy.notification.HttpProgressReporter;
import com.ibm.team.filesystem.client.restproxy.notification.IServerNotificationChannel;
import com.ibm.team.internal.repository.rcp.streams.UnsynchronizedByteArrayOutputStream;
import com.ibm.team.repository.common.LogFactory;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.repository.common.internal.marshal.MarshalFactory;
import com.ibm.team.repository.common.internal.marshal.MarshallerType;
import com.ibm.team.repository.common.internal.marshal.MarshallingException;
import com.ibm.team.repository.common.internal.marshal.impl.WebServicesMarshaller;

@SuppressWarnings({ "nls", "restriction" })
public class JSONHandler extends HttpHandler {
	private static final Log log = LogFactory.getLog(JSONHandler.class.getName());
	
	/** Prefix for all service requests. */
	public static final String SERVICE_PREFIX = "service"; //$NON-NLS-1$
	
	/**
	 * Path/method pair. 
	 */
	protected final Map<String, JSONMethod> methods = new HashMap<String, JSONMethod>();
	
	/** The set of instantiated receivers for the methods. */
	protected final Set<JSONReceiver> receivers = new HashSet<JSONReceiver>(); 
	
	/**
	 * Listing of the known services. 
	 */
	protected final Map<String, IServiceDescription> services = new HashMap<String, IServiceDescription>();
	
	protected FSDaemon fsd;
    protected HttpServer server;

	IServerController controller = new IServerController() {
		public void shutdown() throws IOException {
			server.shutdown();
		}
	};

	private final ParameterWrapperDeserializer deserializer;

	/**
	 * Listing of extension points that are deferring their instantiation 
	 * until the first call. Note that all access should be synchronized on 
	 * this variable.
	 */
	private final HashMap<String, IConfigurationElement> lazyExtensionPoints 
			= new HashMap<String, IConfigurationElement>();

	private static class ServiceDescription implements IServiceDescription {
	    final String id;
	    final String version;
	    
	    public ServiceDescription(String id, String version) {
	        this.id = id;
	        this.version = version;
	    }

		public String getServiceId() {
			return id;
		}

		public String getVersion() {
			return version;
		}
	}
	
	
	public JSONHandler() {
		deserializer = new ParameterWrapperDeserializer();
	}
	
	
	@Override
	public void registered(FSDaemon fsd, HttpServer httpServer, Collection<HttpContext> httpContext) {
		this.fsd = fsd;
		this.server = httpServer;
		
        loadHandlers();
	}
	
	protected void loadHandlers() {
        IConfigurationElement[] configs = RegistryFactory.getRegistry().getConfigurationElementsFor(Daemon.PLUGIN_ID,
        		"restservice");
        
        HashMap<String, IConfigurationElement> ids = new HashMap<String, IConfigurationElement>();
        
        for (IConfigurationElement config : configs) {
            // Get the service extension point
            String id = config.getAttribute("id");
            if (ids.containsKey(id)) {
                log.error("Multiple definitions for extension " + Daemon.PLUGIN_ID + ".restservice id=\"" + id 
                        + "\" (using entry contributed by \"" 
                        		+ ids.get(id).getContributor().getName()
                        + "\", discarding entry contributed by \"" + config.getContributor().getName() + "\")");
                continue;
            }
            
            ids.put(id, config); 

            // Install (or defer) the service
            if (isLazyConfig(config)) {
                lazyExtensionPoints.put(id, config);
            } else {
                installService(id, config);
            }
            
            // Record that we know this service
            String version = config.getAttribute("version");
            services.put(id, new ServiceDescription(id, version));
        }
	}
	
	private void installService(String id, IConfigurationElement config) {
		final String INIT_ATTR = "initializer";
		
		// Initialize
		if (null != config.getAttribute(INIT_ATTR)) {
			Object initializerObj;
			try {
				initializerObj = config.createExecutableExtension("initializer");
			} catch (CoreException e) {
			    log.error("Could not create extension for initializer " 
			    		+ config.getAttribute(INIT_ATTR) + " from plugin " 
			            + config.getContributor().getName() 
			            + " for extension point " + Daemon.PLUGIN_ID + ".restservice");
			    return;
			}
			
			IRestServiceInitializer init = (IRestServiceInitializer) initializerObj;
			init.initialize(config);
		}
		
		// Create the handler
        Object receiver;
		try {
			receiver = config.createExecutableExtension("class");
		} catch (CoreException e) {
		    log.error("Could not instantiate class " + id + " from plugin " 
		            + config.getContributor().getName() 
		            + " for extension point " + Daemon.PLUGIN_ID + ".restservice");
		    return;
		}
		
		// Register the handler
        if (!addReceiverMethods(receiver, id)) {
        	log.error("Failed to install methods for " + id + " from plugin " 
		            + config.getContributor().getName());
            return;
        }
        
        if (receiver instanceof JSONReceiver) {
        	receivers.add((JSONReceiver)receiver);
        	
            ((JSONReceiver)receiver).installationNotification(controller, this);
        }
	}

	/** Check to see if the lazy flag is set in the given config. */
	private static boolean isLazyConfig(IConfigurationElement config) {
	    String val = config.getAttribute("lazy_start");
	    
	    if (val == null) {
	        return true;
	    }
	    
	    if ("false".equals(val.trim().toLowerCase(Locale.ENGLISH))) {
	        return false;
	    }
	    
	    return true;
	}
	
	/** Add the appropriate methods from the given object to the tail end of the given path. */
	protected boolean addReceiverMethods(Object receiver, String path) {
	    synchronized (methods) {
    		for (Method m : receiver.getClass().getMethods()) {
    			// Verify we know how to call this method
    			Class<?> [] params = m.getParameterTypes();
    			if (params.length > 2) {
    				continue;
    			}
    			if (params.length == 1) {
    				if (!IValidatingParameterWrapper.class.isAssignableFrom(params[0])
    						&& params[0] != IProgressMonitor.class
    						&& params[0] != IProgressMonitorWithBlocking.class) {
    					continue;
    				}
    			} else if (params.length == 2) {
    				if (!IValidatingParameterWrapper.class.isAssignableFrom(params[0])) {
    					continue;
    				}
    				if (params.length == 2) {
    					if (params[1] != IProgressMonitor.class && params[1] != IProgressMonitorWithBlocking.class) {
    						continue;
    					}
    				}
    			}
    			
    			// Hook the HTTP paths appropriately for each valid method
    			if (m.getName().startsWith(HttpMethod.GET.toString().toLowerCase(Locale.ENGLISH))
    					|| m.getName().startsWith(HttpMethod.POST.toString().toLowerCase(Locale.ENGLISH))) {
    				String key = path + "." + m.getName();
    				
    				if (methods.containsKey(key)) {
                        log.error("Multiple definitions of " + key);
                        return false;
    				}
    				
                    JSONMethod jm = new JSONMethod(receiver, m);
    				methods.put(key, jm);
    			}
    		}
	    }
        
        return true;
	}

	@Override
	public void shutdownNotification() {
		super.shutdownNotification();
		
		synchronized (lazyExtensionPoints) {
			for (JSONReceiver receiver : receivers) {
				try {
					receiver.shutdownNotification();
				} catch (Exception e) {
					log.error("Exception during shutdown", e);
				}
			}
		}
	}
	
	/** Handle a call to one of our services. */
	@Override
	public void handle(HttpRequest rq, HttpResponse rp) throws IOException {
		JSONMethod m = getMethod(rq.getMethod(), rq.getPathSegments());

		if (m == null) {
			sendErrorResponse(rq, rp, ResponseCode.NOT_FOUND.getCode(), "Unknown service: " + rq.getRequestURI().getRawPath()); //$NON-NLS-1$
			return;
		}

		// Read headers
		Parameters parm = buildParameters(rq, rp);
		String progressKey = findProgressKey(rq);

		// Demarshall
		Object[] params;
		try {
			params = demarshallParameters(parm, m, rq, rp, progressKey);
		} catch (RestMarshallingException e) {
			log.error(e.getMessage(), e);
			sendErrorResponse(rq, rp, ResponseCode.BAD_REQUEST.getCode(),
					"Exception during parameter demarshalling", e);
			return;
		}
		
		fsd.queueEvent(new JSONHandlerRequestEvent(rq, rp, params, progressKey));
		
		// Validate the input parameters
		try {
			validate(m.getMethod().getName(), params);
		} catch (Exception e) {
			sendErrorResponse(rq, rp, ResponseCode.INTERNAL_SERVER_ERROR.getCode(),
					"Failure during parameter validation", e);
			return;
		}
		
		// Execute
		Object result;
		try {
			try {
				result = m.getMethod().invoke(m.getReceiver(), params);
			} finally {
				if (params.length != 0 && params[params.length - 1] instanceof HttpProgressReporter) {
					HttpProgressReporter r = (HttpProgressReporter)params[params.length - 1];
					r.finish();
				}
			}
		} catch (InvocationTargetException e) {
			if (e.getCause() instanceof TeamRepositoryException
					|| e.getCause() instanceof IllegalArgumentException) {
				// TREs are due to a calling error by our client
				log.error(e.getMessage(), e);
				
				// log nested IStatus info
				if (e.getCause() instanceof TeamRepositoryException) {
					Throwable causeOfTRE = e.getCause().getCause();
					if (causeOfTRE instanceof FileSystemStatusException) {
						String statusInfo = CommonUtil.getStatusErrors((FileSystemStatusException) causeOfTRE);
						log.error(statusInfo);
					}
				}
			}
			else {
				// Non TREs are due to programming errors in the handler
				log.error(e.getCause().getMessage(), e.getCause());
			}
			
			sendErrorResponse(rq, rp, ResponseCode.INTERNAL_SERVER_ERROR.getCode(),
					"Unexpected exception", e.getCause());
			return;
		} catch (IllegalAccessException e) {
			log.error(e.getMessage(), e);
			
			sendErrorResponse(rq, rp, ResponseCode.INTERNAL_SERVER_ERROR.getCode(),
					"Java reflection exception.", e);
			return;
		}
		
		// Marshall
		UnsynchronizedByteArrayOutputStream bOut = new UnsynchronizedByteArrayOutputStream();
        WebServicesMarshaller webServicesMarshaller = (WebServicesMarshaller)MarshalFactory.eINSTANCE
        													.getMarshaller(MarshallerType.WEB_SERVICES_LITERAL);

        Charset charset = Charset.forName("UTF-8");
        try {
			webServicesMarshaller.marshalServiceResponse(m.getMethod().getName(), m.getMethod().getDeclaringClass().getName(), 
	                m.getReturnType(), result, WebServicesMarshaller.MARSHAL_AS_JSON,
	                bOut, charset);
        } catch (MarshallingException e) {
			log.error(e.getMessage(), e);
			sendErrorResponse(rq, rp, ResponseCode.INTERNAL_SERVER_ERROR.getCode(),
					"Exception while marshalling", e);
			return;
        } catch (Exception e) {
        	log.error(e.getMessage(), e);
        	sendErrorResponse(rq, rp, ResponseCode.INTERNAL_SERVER_ERROR.getCode(), 
        			"Failed to marshal response", new RestMarshallingException("Failed to marshal response", e));
        	return;
        }
        
        // Reply
        rp.setCode(ResponseCode.OK);
        rp.writeHeader(Header.CONTENT_TYPE, "text/json; charset=utf-8");
        int size = bOut.size();
        rp.writeHeader(Header.CONTENT_LENGTH, Integer.toString(size));
        OutputStream out = rp.getResponseStream();
        out.write(bOut.getBuffer(), 0, size);
        out.close();
        
        fsd.queueEvent(new JSONHandlerResponseEvent(rq, rp, charset, bOut.getBuffer(), 0, size));
	}
	
	/**
	 * If the given parameters are IValidatingParameterWrapper, run the 
	 * validate method.
	 * @param methodName 
	 */
	private void validate(String methodName, Object[] params) throws IllegalArgumentException {
		for (Object obj : params) {
			if (obj instanceof IValidatingParameterWrapper) {
				((IValidatingParameterWrapper)obj).validate(methodName);
			}
		}
	}


	private String findProgressKey(HttpRequest rq) {
		return rq.getFirstHeader(Header.REQUEST_PROGRESS);
	}
	
	private Object [] demarshallParameters(Parameters parm, JSONMethod m, HttpRequest rq, HttpResponse resp, String progressKey) {
		// Generate the progress reporter (if appropriate)
		HttpProgressReporter progress = null;
		{
			if (m.takesProgress() && progressKey != null) {
				progress = new HttpProgressReporter(findProgressKey(rq), fsd.getNotificationChannel(), resp);
			}
		}
		
		// Handle zero-arg request
		if (m.getParameterType() == null) {
			if (m.takesProgress()) {
				return new Object[] { progress };
			}
			return new Object[0];
		}

		// Handle request with arg
		IValidatingParameterWrapper wrapper = deserializer.deserialize(m.getReceiver().getClass().getClassLoader(), parm, m.getParameterType());
		
		if (m.takesProgress()) {
			return new Object [] { wrapper, progress };
		}
		return new Object[] { wrapper };
	}

	
	/**
	 * Given a request, normalize the path of the service request 
	 */
	protected String normalizeServicePath(HttpMethod method, String[] path) {
		if (path.length < 1) {
			return null;
		}
		
		if (!SERVICE_PREFIX.equals(path[0])) {
			return null;
		}
		
		// Copy the array, so that we don't overwrite anything important
		String[] myPath = new String[path.length - 1];
		System.arraycopy(path, 1, myPath, 0, path.length - 2);
		
		// Add the method qualifier (get/post) to the name of the method
		myPath[myPath.length - 1] = method.name().toLowerCase(Locale.ENGLISH) + path[path.length-1];

		// Join the array bits together
		StringBuffer buf = new StringBuffer();
		
		for (String s : myPath) {
			if (buf.length() > 0) {
				buf.append('.');
			}
			
			buf.append(s);
		}

		return buf.toString();
	}
	
	
	/**
	 * Gets the JSONMethod for the given request. If the service allows for 
	 * lazy instantiation, that occurs here.
	 */
	public JSONMethod getMethod(HttpMethod method, String[] path) {
		String key = normalizeServicePath(method, path);
		
        // See if this method is already installed
		JSONMethod result;
		synchronized (methods) {
		    result = methods.get(key);
		}
        
        if (result != null) {
            return result;
        }


        // The method is not installed. Try to instantiate the service.
		synchronized (lazyExtensionPoints) {
		    if (lazyExtensionPoints.size() > 0) {
		        // Find the key
    		    int lastDot = key.lastIndexOf(".");
    		    if (lastDot == -1) {
    		    	return null;
                }
    	            
    		    String id = key.substring(0, lastDot);
    
    		    // Instantiate the extension point.
    		    IConfigurationElement config = lazyExtensionPoints.remove(id);
    		    if (config != null) {
    		    	installService(id, config);
    		    }
		    }
	    }
		    
        synchronized (methods) {
            // Last hurrah. We (or another thread) installed the pt or failed.  
            return methods.get(key);
		}
	}
	
	protected Parameters buildParameters(HttpRequest rq, HttpResponse rp) throws IOException {
		Parameters p = new Parameters();
		addParametersFromQuery(rq.getRequestURI().getRawQuery(), p);
		if (rq.getMethod() == HttpMethod.POST) {
			String encoding = rq.getFirstHeader(Header.CONTENT_TYPE);
			if ("application/x-www-form-urlencoded".equals(encoding)) {
				InputStream in = rq.getRequestBody();
				String length = rq.getFirstHeader(Header.CONTENT_LENGTH);
				int len = 32;
				if (length != null) {
					len = Integer.valueOf(length);
				}
				StringBuilder b = new StringBuilder(len);
				while (true) { 
					int c = in.read();
					if (c == -1) {
						break;
					}
					//ISO-8859-1 encoding
					b.append((char)c);
				}
				addParametersFromQuery(b.toString(), p);
			}
		}
		
		return p;
	}
	
	protected void addParametersFromQuery(String q, Parameters p) throws IOException {
		if (q != null) {
			int len = q.length();
			int start = 0;
			int end;
			do {
				end = q.indexOf('&', start);
				if (end == -1) {
					end = len;
				}
				
				String pair = q.substring(start, end);
				int keyEnd = pair.indexOf('=');
				if (keyEnd == -1) {
					p.put(URLDecoder.decode(pair, "UTF-8"), null);
				} else {
					String key = pair.substring(0, keyEnd);
					String value = pair.substring(keyEnd + 1);
					key = URLDecoder.decode(key, "UTF-8");
					value = URLDecoder.decode(value, "UTF-8");
					p.put(key, value);
				}
				
				
				start = end + 1;
			} while (end != len);
		}
	}
	
	protected void sendErrorResponse(HttpRequest rq, HttpResponse rp, int code, String text) throws IOException {
		rp.setCode(code, text);
		
		fsd.queueEvent(new JSONHandlerResponseEvent(rq, rp, null, null));
	}
	
	protected void sendErrorResponse(HttpRequest rq, HttpResponse rp, int code, String text, Throwable e) throws IOException {
		ExceptionHandlingJSONSerializer serializer = new ExceptionHandlingJSONSerializer(true);

        UnsynchronizedByteArrayOutputStream bOut = new UnsynchronizedByteArrayOutputStream();
        
        Charset charset = Charset.forName("UTF-8");
		OutputStreamWriter w = new OutputStreamWriter(bOut, charset);
		try {
			serializer.serializeException(e, code, w);
		} catch (Exception ex) {
			log.error("Failed to serialize exception", ex);
			sendErrorResponse(rq, rp, code, "Failure during exception serialization: " + text);
			return;
		}
        w.close();
        
		rp.setCode(code, text);
		rp.writeHeader(Header.CONTENT_TYPE, Header.CONTENT_TYPE_STACKTRACE + "; charset=utf-8");
		int size = bOut.size();
		rp.writeHeader(Header.CONTENT_LENGTH, Integer.toString(size));
		rp.getResponseStream().write(bOut.getBuffer(), 0, size);
		
		fsd.queueEvent(new JSONHandlerResponseEvent(rq, rp, charset, bOut.getBuffer(), 0, size));
	}

	public IServerNotificationChannel getNotificationChannel() {
		return fsd.getNotificationChannel();
	}
	
	public IHttpServer getHttpServer() {
	    return fsd.getHttpServer();
	}
	
	public FSDaemon getFSDaemon() {
	    return fsd;
	}

	/**
	 * Return a description of the registered services. 
	 */
	public Map<String, IServiceDescription> getServiceDescriptions() {
	    return services;
	}

}
