/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2006, 2011. All Rights Reserved. 
 * 
 * Note to U.S. Government Users Restricted Rights:  Use, 
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.filesystem.client.internal;

import java.io.IOException;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.MultiStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.jobs.Job;

import com.ibm.team.filesystem.client.FileSystemCore;
import com.ibm.team.filesystem.client.internal.copyfileareas.ICopyFileAreaManager;
import com.ibm.team.filesystem.client.internal.utils.PersistentBusyFlag;
import com.ibm.team.internal.repository.rcp.dbhm.DBHMException;
import com.ibm.team.internal.repository.rcp.dbhm.PersistentDiskBackedHashMap;

public abstract class AsyncSavePDBHMJob extends Job {

    private static final int SCHEDULING_DELAY = 200;

    private volatile long latestRequest;
    
    private Object lockObj;

    private final PersistentBusyFlag busyFlag;
    
    /** Object to notify in case we have a problem with a save. */
    private final ICorruptible corruptible;
    
    public AsyncSavePDBHMJob(String name, PersistentBusyFlag busyFlag, Object lockObj, ICorruptible corruptible) {
        super(name);
        this.setUser(false);
        this.setSystem(true);
        this.lockObj = lockObj;
        this.busyFlag = busyFlag;
        this.corruptible = corruptible;
    }

    public void requestSave() {
        latestRequest = System.currentTimeMillis();
        schedule(SCHEDULING_DELAY);
    }
    
    protected abstract PersistentDiskBackedHashMap<?, ?> getMapToPesist();
    
    protected IStatus run(IProgressMonitor monitor) {
        if ((System.currentTimeMillis() - latestRequest) < SCHEDULING_DELAY) {
            schedule(SCHEDULING_DELAY);
            return Status.CANCEL_STATUS;
        }
        MultiStatus result = new MultiStatus(FileSystemCore.ID, IStatus.OK, Messages.AsyncSavePDBHMJob_0, null);
        
        synchronized (lockObj) {
            PersistentDiskBackedHashMap<?, ?> toPersist = getMapToPesist();
            try {
                if (toPersist != null) {
                    toPersist.persist();
                    
                    busyFlag.setComplete(result);
                }
            } catch (DBHMException e) {
                corruptible.setCorrupt(true, "Save failed", e); //$NON-NLS-1$
                result.add(FileSystemStatusUtil.getStatusFor(IStatus.ERROR, Messages.AsyncSavePDBHMJob_1, e));
            } catch (IOException e) {
                result.add(FileSystemStatusUtil.getStatusFor(IStatus.ERROR, Messages.AsyncSavePDBHMJob_2, e));
            }
        }
        
        return result;
    }
    
    @Override
    public boolean belongsTo(Object family) {
        if (ICopyFileAreaManager.instance == family) {
            return true;
        }
        return super.belongsTo(family);
    }
}
