/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2005, 2009. All Rights Reserved. 
 * 
 * Note to U.S. Government Users Restricted Rights:  Use, 
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.filesystem.client.internal;

import java.util.concurrent.locks.AbstractQueuedSynchronizer;

/**
 * A lock that has 2 types of mutually exclusive acquires where
 * both partitions may be held by multiple threads simultaneously.
 */
public class BiPartitionLock {
    
    static final class BiPartitionSync extends AbstractQueuedSynchronizer {
        private static final long serialVersionUID = -3004750439878718463L;

        protected int tryAcquireShared(int type) {
            if (type == 0) {
                while (true) {
                    int cnt = getState();
                    if (cnt < 0) {
                        return -1;
                    }
                    if (compareAndSetState(cnt, cnt + 1)) {
                        return 1;
                    }
                }
            } else {
                while (true) {
                    int cnt = getState();
                    if (cnt > 0) {
                        return -1;
                    }
                    if (compareAndSetState(cnt, cnt - 1)) {
                        return 1;
                    }
                }
            }
        }
        
        protected boolean tryReleaseShared(int type) {
            if (type == 0) {
                while (true) {
                    int cnt = getState();
                    if (cnt <= 0) {
                        throw new IllegalMonitorStateException();
                    }
                    int next = cnt - 1;
                    if (compareAndSetState(cnt, next)) {
                        return next == 0;
                    }
                }
            } else {
                while (true) {
                    int cnt = getState();
                    if (cnt >= 0) {
                        throw new IllegalMonitorStateException();
                    }
                    int next = cnt + 1;
                    if (compareAndSetState(cnt, next)) {
                        return next == 0;
                    }
                }
            }
        }
    }
    
    private final BiPartitionSync sync;
    public BiPartitionLock() {
        sync = new BiPartitionSync();
    }

    public void acquirePart1() {
        sync.acquireShared(0);
    }
    
    public boolean tryAcquirePart1() {
        return sync.tryAcquireShared(0) >= 0;
    }
    
    public void releasePart1() {
        sync.releaseShared(0);
    }
    
    public void acquirePart2() {
        sync.acquireShared(1);
    }
    
    public boolean tryAcquirePart2() {
        return sync.tryAcquireShared(1) >= 0;
    }
    
    public void releasePart2() {
        sync.releaseShared(1);
    }
}
