/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2006, 2011. All Rights Reserved. 
 * 
 * Note to U.S. Government Users Restricted Rights:  Use, 
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.filesystem.client.internal;

import java.io.IOException;
import java.io.OutputStream;
import java.security.MessageDigest;

import org.eclipse.core.runtime.Assert;
import org.eclipse.core.runtime.Status;

import com.ibm.team.filesystem.client.FileSystemCore;
import com.ibm.team.filesystem.client.FileSystemException;
import com.ibm.team.filesystem.client.IRelativeLocation;
import com.ibm.team.filesystem.client.ISharingDescriptor;
import com.ibm.team.filesystem.client.ResourceType;
import com.ibm.team.filesystem.client.internal.copyfileareas.AbstractLock;
import com.ibm.team.filesystem.client.internal.copyfileareas.CFALockUtil;
import com.ibm.team.filesystem.client.internal.copyfileareas.CopyFileArea;
import com.ibm.team.filesystem.client.internal.copyfileareas.CopyFileAreaManager;
import com.ibm.team.filesystem.client.internal.copyfileareas.ICopyFileArea.PropertyUpdate;
import com.ibm.team.filesystem.common.IFileItemHandle;
import com.ibm.team.internal.repository.rcp.streams.DigestComputingOutputStream;
import com.ibm.team.scm.common.ContentHash;


/**
 * Hashes data written into an OutputStream and registers a change if the data 
 * is different from the previous resource content. 
 */
public class ChangeTrackingOutputStream extends DigestComputingOutputStream {

    final ContentHash targetHash;

    final IFileItemHandle handle;

    final private Share share;

    final ManagedFileStore managed;
    
    final long targetSize;

        
    /** Flag to see if we have been closed or not. */
    boolean closed = false;

    
    public ChangeTrackingOutputStream(MessageDigest start, OutputStream out, 
            ManagedFileStore managed,
            IFileItemHandle handle, Share share, 
            ContentHash targetHash, long targetSize) {
        super(start, out);
        
        this.managed = managed;
        this.handle = handle;
        this.share = share;
        this.targetHash = targetHash;
        this.targetSize = targetSize;
    }

    public void close() throws IOException {
        // Prevent multiple calls to close.
        if (closed) {
            return;
        }
        closed = true;
        
        
        // Be a good child.
        super.close();
        


        ISharingDescriptor desc = share.getSharingDescriptor();
        try {
            AbstractLock lock = CFALockUtil.lockExistingForUpdate(share.getSandbox().getRoot(), desc.getConnectionHandle(),
                    desc.getComponent(), null);
            if (lock == null) {
                return;
            }
            try {
                CopyFileArea cfa = CopyFileAreaManager.instance.getExistingCopyFileArea(share.getSandbox().getRoot());
                IRelativeLocation path = cfa.getLocalPathFor(handle, 
                        share.getSharingDescriptor().getComponent(), 
                        share.getSharingDescriptor().getConnectionHandle(), 
                        null);
                
                if (path != null) {
                    FileItemInfo fiInfo = cfa.getItemInfo(path);
                    Assert.isTrue(fiInfo.getVersionableHandle().sameItemId(handle));
                    if (targetSize == getContentSize()) {
                        //need to register a change since the sizes are the same.
                        managed.addDatelessItem(handle, share.getSharingDescriptor().getComponent(),
                                share.getSharingDescriptor().getConnectionHandle());
                    }
                    
                    ContentHash hash = ContentHash.valueOf(getFinalDigest());
                    boolean modified = !hash.equals(targetHash) || targetSize != getContentSize();
                    if (modified != fiInfo.isContentChanged()) {
                        long modStamp = ((Shareable)share.getShareable(path, ResourceType.FILE)).getFileStorage().getModificationStamp();
                        cfa.setItemMetaData(path, new FileItemInfo(fiInfo.getVersionableHandle(), modified,
                                modStamp, fiInfo.getParent(), fiInfo.getName(), fiInfo.isLoadedWithAnotherName(),
                                fiInfo.getHash(), fiInfo.getContentLength(),
                                fiInfo.getOriginalLineDelimiter(), fiInfo.getLineDelimiter(), fiInfo.getOriginalContentType(),
                                fiInfo.getContentType(), fiInfo.getStoredPredecessorHintHash(),
                                fiInfo.getStoredSize(), fiInfo.getStoredEncoding(), fiInfo.getStoredHash(), fiInfo.getStoredNumLineDelimiters(),
                                fiInfo.isExecutable(), fiInfo.isOriginalExecutable(), fiInfo.isDirectoryLink(), fiInfo.isOriginalDirectoryLink()),
                                PropertyUpdate.PRESERVE, null, // leave user property changes as is
                                null);
                    }
                }
            } finally {
                CFALockUtil.endBatching(lock, null);
            }
        } catch (FileSystemException e) {
            LoggingHelper.log(FileSystemStatusUtil.getStatusFor(Status.ERROR, 
                    "com.ibm.team.filesystem.client",  //$NON-NLS-1$
                    "Error aquiring path for shareable", e)); //$NON-NLS-1$
            return;
        }
    }
    
    protected void finalize() throws IOException {
        if (!closed) {
            close();
            LoggingHelper.log(FileSystemCore.ID, new FileSystemException("Stream finalized without beind close()ed")); //$NON-NLS-1$
        } 
    }
}
