/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2006, 2015. All Rights Reserved. 
 * 
 * Note to U.S. Government Users Restricted Rights:  Use, 
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.filesystem.client.internal;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.SubMonitor;

import com.ibm.team.filesystem.common.changemodel.ConfigurationChange;
import com.ibm.team.filesystem.common.changemodel.FileChange;
import com.ibm.team.filesystem.common.changemodel.FileState;
import com.ibm.team.filesystem.common.changemodel.StateNode;
import com.ibm.team.filesystem.common.internal.util.ChangeSetUtil;
import com.ibm.team.repository.client.ITeamRepository;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.repository.common.UUID;
import com.ibm.team.repository.rcp.common.collection.CollectionUtil;
import com.ibm.team.scm.client.internal.RepoFetcher;
import com.ibm.team.scm.common.IChange;
import com.ibm.team.scm.common.IChangeSet;
import com.ibm.team.scm.common.IComponent;
import com.ibm.team.scm.common.IVersionable;
import com.ibm.team.scm.common.internal.util.ItemId;
import com.ibm.team.scm.common.internal.util.SiloedItemId;
import com.ibm.team.scm.common.internal.util.StateId;

public class ClientConfigurationChangeFactory {

    /**
     * Creates a map of change set item ids to their corresponding configuration changes given a collection of change sets.
     * 
     * @param repository the team repository that contains the collection of change sets
     * @param changeSetList the collection of change sets to create configuration changes for
     * @param monitor a progress monitor for reporting progress
     * @return the map of change set item ids to the change set's corresponding configuration change
     * @throws TeamRepositoryException if the operation fails for any reason
     */
    public static Map<ItemId<IChangeSet>, ConfigurationChange> createChanges(ITeamRepository repository, Collection<IChangeSet> changeSetList, IProgressMonitor monitor)
            throws TeamRepositoryException {
        return createChanges(repository, changeSetList, null, monitor);     
    }
    
    public static Map<ItemId<IChangeSet>, ConfigurationChange> createChanges(ITeamRepository repository, Collection<IChangeSet> changeSetList, List<String> changeItemIds, IProgressMonitor monitor)
            throws TeamRepositoryException {
        SubMonitor progress = SubMonitor.convert(monitor, 100);
        ITeamRepository repo = repository;
                
        progress.setWorkRemaining(changeSetList.size());
        
        Set<StateId<IVersionable>> toFetch = new HashSet<StateId<IVersionable>>();
        
        for (IChangeSet nextChangeSet : changeSetList) {
            List<IChange> nextChanges = nextChangeSet.changes();
            
            for (IChange next : nextChanges) {
                toFetch.addAll(ChangeSetUtil.getBeforeStates(next));
                toFetch.add(ChangeSetUtil.getAfterState(next));
            }
        }
        
        Map<ItemId<IChangeSet>, ConfigurationChange> result = new HashMap<ItemId<IChangeSet>, ConfigurationChange>();
        Map<StateId<IVersionable>, IVersionable> fetched = RepoFetcher.fetchItems(repository, toFetch, progress.newChild(50));

        for (IChangeSet nextChangeSet : changeSetList) {
            ConfigurationChange nextChange = createChange(fetched, repo, nextChangeSet, changeItemIds);
            result.put(ItemId.forItem(nextChangeSet), nextChange);
        }        
        
        return result;        
    }
    
    /**
     * Creates a configuration change that encompasses the changes affecting the given versionables of the specified change set.
     * 
     * @param versionables the versionables to collect changes for
     * @param repo the team repository that the change set is from
     * @param changeSet the change set to collect the changes from
     * @param excludeNoOps <tt>true</tt> if changes that do not modify a versionable should be excluded, <tt>false</tt> if they should be preserved
     * @return a configuration change that holds changes to the given versionables defined by the provided change set
     */
    public static ConfigurationChange createChange(Map<StateId<IVersionable>, IVersionable> versionables, ITeamRepository repo, IChangeSet changeSet) {
        return createChange(versionables, repo, changeSet, null);
    }
    
    public static ConfigurationChange createChange(Map<StateId<IVersionable>, IVersionable> versionables, ITeamRepository repo, IChangeSet changeSet,  List<String> changeItemIds) {
        ConfigurationChange nextChange = new ConfigurationChange();
        ItemId<IComponent> component = ChangeSetUtil.getComponent(changeSet);
        List<IChange> nextChanges = changeSet.changes();

        for (IChange next : nextChanges) {
            // look for the item affected by this particular change
            boolean found = false;
            UUID affectedItem = next.item().getItemId();
            for (StateId<IVersionable> versionable : versionables.keySet()) {
                //if ChangeItemId is specified, we need changes only for those files
                //else collect all the changes.
                if (versionable.getItemUUID().equals(affectedItem) && (changeItemIds == null || changeItemIds.contains(affectedItem.getUuidValue()))) {
                    found = true;
                    break;
                }
            }

            if (!found) {
                // the item wasn't something we were interested it, ignore it and process the next item instead 
                continue;
            }

            StateId<IVersionable> beforeState = ChangeSetUtil.getLastMergeState(next);
            StateId<IVersionable> afterState = ChangeSetUtil.getAfterState(next);

            IVersionable afterVersionable = versionables.get(afterState);
            FileState after = ClientFileStateFactory.create(repo, afterState.getItemType(), afterVersionable);

            for (StateId<IVersionable> nextBeforeState : ChangeSetUtil.getBeforeStates(next)) {
                IVersionable beforeVersionable = versionables.get(nextBeforeState);
                FileState before = ClientFileStateFactory.create(repo, beforeState.getItemType(), beforeVersionable);

                FileChange nextFileChange = new FileChange(before, after, SiloedItemId.create(beforeState.getItemId(), component));

                // If this is the primary before state (the one we'll use for generating patches,
                // opening compare editors, etc)
                if (nextBeforeState.equals(beforeState)) {
                    nextChange.addChange(nextFileChange);
                } else {
                    // Otherwise, this is a merge state. We'll just use it for resolving conflicts
                    nextChange.addMerge(nextFileChange);
                }
            }
        }
        return nextChange;
    }

    
    public static ConfigurationChange createChange(ITeamRepository repository, Collection<IChangeSet> changeSetList, List<String> changeItemIds, IProgressMonitor monitor) throws TeamRepositoryException {
        Collection<ConfigurationChange> individualChanges = createChanges(repository, changeSetList, changeItemIds, monitor).values();
        
        ConfigurationChange netChange = ConfigurationChange.merge(individualChanges);
        
        return netChange;
    }
    
    /**
     * Creates and returns a configuration change which encompasses the changes in the provided collection of change sets.
     * 
     * @param repository the team repository that contains the collection of change sets
     * @param changeSetList the collection of change sets to create configuration changes for
     * @param monitor a progress monitor for reporting progress
     * @return a configuration change which contains the changes in the provided collection of change sets
     * @throws TeamRepositoryException if the operation fails for any reason
     */
    public static ConfigurationChange createChange(ITeamRepository repository, Collection<IChangeSet> changeSetList, IProgressMonitor monitor) throws TeamRepositoryException {
        return createChange(repository, changeSetList, null, monitor);
        
    }

    /**
     * 
     * @param repo
     * @param changeSetHandle
     * @param changeItemIds to be specified if changeforhandles for only a set of files in the changeset are needed. 
     * Else null should be passed. 
     * @param monitor
     * @return
     * @throws TeamRepositoryException
     */
    public static ConfigurationChange createChangeForHandles(ITeamRepository repo, List<ItemId<IChangeSet>> changeSetHandle, List<String> changeItemIds, IProgressMonitor monitor) throws TeamRepositoryException {
        SubMonitor progress = SubMonitor.convert(monitor, 100);
        
        Map<ItemId<IChangeSet>, IChangeSet> fetched = RepoFetcher.fetchCurrents(repo, changeSetHandle, true, progress.newChild(10));
        
        List<IChangeSet> ordered = CollectionUtil.reorderValues(fetched, changeSetHandle);
        return createChange(repo, ordered, changeItemIds, progress);
    }
    
    public static ConfigurationChange createChangeForHandles(ITeamRepository repo, List<ItemId<IChangeSet>> changeSetHandle, IProgressMonitor monitor) throws TeamRepositoryException {
        return createChangeForHandles(repo, changeSetHandle, null, monitor); 
    }

    public static final List<FileChange> getChangeForTarget(ITeamRepository repository, List<IChangeSet> changeSetList, SubMonitor monitor, SiloedItemId<IVersionable> target) throws TeamRepositoryException {
        SubMonitor progress = SubMonitor.convert(monitor, 100);
        ITeamRepository repo = repository;
                
        progress.setWorkRemaining(changeSetList.size());
        
        Set<StateId<IVersionable>> toFetch = new HashSet<StateId<IVersionable>>();
        
        for (IChangeSet nextChangeSet : changeSetList) {
            List<IChange> nextChanges = nextChangeSet.changes();
            
            for (IChange next : nextChanges) {
                if( target.getItemUUID().equals(next.item().getItemId()))
                {
                    toFetch.addAll(ChangeSetUtil.getBeforeStates(next));
                    toFetch.add(ChangeSetUtil.getAfterState(next));
                }
            }
        }
        
        Map<ItemId<IChangeSet>, ConfigurationChange> result = new HashMap<ItemId<IChangeSet>, ConfigurationChange>();
        Map<StateId<IVersionable>, IVersionable> fetched = RepoFetcher.fetchItems(repository, toFetch, progress.newChild(50));
        
        for (IChangeSet nextChangeSet : changeSetList) {
            ConfigurationChange nextChange = new ConfigurationChange();
            ItemId<IComponent> component = ChangeSetUtil.getComponent(nextChangeSet);
            List<IChange> nextChanges = nextChangeSet.changes();
            
            for (IChange next : nextChanges) {
                if( target.getItemUUID().equals(next.item().getItemId()))
                {
                    StateId<IVersionable> beforeState = ChangeSetUtil.getLastMergeState(next);
                    StateId<IVersionable> afterState = ChangeSetUtil.getAfterState(next);
                    
                    IVersionable afterVersionable = fetched.get(afterState);
                    FileState after = ClientFileStateFactory.create(repo, afterState.getItemType(), afterVersionable);
                    
                    for (StateId<IVersionable> nextBeforeState : ChangeSetUtil.getBeforeStates(next)) {
                        IVersionable beforeVersionable = fetched.get(nextBeforeState);
                        
                        FileState before = ClientFileStateFactory.create(repo, beforeState.getItemType(), beforeVersionable);
                        
                        FileChange nextFileChange = new FileChange(before, after, SiloedItemId.create(beforeState.getItemId(), component));
                        
                        // If this is the primary before state (the one we'll use for generating patches,
                        // opening compare editors, etc)
                        if (nextBeforeState.equals(beforeState)) {
                            nextChange.addChange(nextFileChange);
                        } else {
                            // Otherwise, this is a merge state. We'll just use it for resolving conflicts
                            nextChange.addMerge(nextFileChange);
                        }
                    }
                }
            }
            
            result.put(ItemId.forItem(nextChangeSet), nextChange);
        }
        Map<ItemId<IChangeSet>, ConfigurationChange> createChanges = result;
        Collection<ConfigurationChange> individualChanges = createChanges.values();
        ConfigurationChange result1 = new ConfigurationChange();
        
        for (ConfigurationChange nextConfig : individualChanges) {
            for (StateNode next : nextConfig.getStateGraph().values()) {
                if( !next.getState().getStateId().getItemUUID().equals(target.getItemUUID()) )
                     continue;   
                for (FileChange nextChange : next.getActiveIncoming()) {
                	result1.addChange(nextChange);
                }
                
                for (FileChange nextMerge : next.getMerges()) {
                	result1.addMerge(nextMerge);
                }
            }
        }
        
        ConfigurationChange netChange = result1;
        
        return netChange.getChanges(target);
    }


}
