/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2006, 2011. All Rights Reserved. 
 * 
 * Note to U.S. Government Users Restricted Rights:  Use, 
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.filesystem.client.internal;

import com.ibm.team.filesystem.common.FileLineDelimiter;
import com.ibm.team.filesystem.common.IFileContent;
import com.ibm.team.filesystem.common.IFileItem;
import com.ibm.team.filesystem.common.ISymbolicLink;
import com.ibm.team.filesystem.common.changemodel.FileState;
import com.ibm.team.filesystem.common.changemodel.VersionablePathSegment;
import com.ibm.team.repository.client.ITeamRepository;
import com.ibm.team.repository.common.IItemType;
import com.ibm.team.scm.common.IFolder;
import com.ibm.team.scm.common.IVersionable;
import com.ibm.team.scm.common.internal.util.StateId;

public class ClientFileStateFactory {
    
    
    /**
     * Returns a FileState for the given versionable. The versionable may be <code>null</code>
     * to indicate a 'deleted' FileState, with the given type.
     * @param repo
     * @param type the item type. Used if the versionable is <code>null</code> 
     * so that the deleted state can reflect the item type properly
     * @param item may be <code>null</code>
     * @return a new file state for the given versionable. Never <code>null</code>.
     */
    public static FileState create(ITeamRepository repo, IItemType type, IVersionable item) {
        if (item == null) {
            return FileState.getDeletedState(type, null);
        }
        if (item instanceof IFileItem) {
            IFileItem fileItem = (IFileItem)item;
            return getFileState(repo, fileItem);
        } else if (item instanceof IFolder) {
            return FileState.create(VersionablePathSegment.pathFor(item), item.modified(), false, item.getUserProperties(), StateId.forItem(item));
        } else if (item instanceof ISymbolicLink) {
            return getSymbolicLinkState(repo, (ISymbolicLink)item);
        } else {
            return FileState.getDeletedState(type, VersionablePathSegment.pathFor(item));
        }
    }

    /**
     * @param item must not be <code>null.</code>.
     * @return a new FileState for an IFileItem
     */
    private static FileState getFileState(ITeamRepository repo, IFileItem item) {
        IFileContent content = item.getContent();
        String mimeType = item.getContentType();
        FileLineDelimiter delimiter = content.getLineDelimiter();
        boolean isText = FileState.isTextType(mimeType);
        FileContentInputStreamProvider inputStreamProvider = new FileContentInputStreamProvider(repo, item, content);
        return FileState.create(VersionablePathSegment.pathFor(item), item.getFileTimestamp(), 
                inputStreamProvider, item.isExecutable(), mimeType,
                delimiter, isText, content.getCharacterEncoding(), false, item.getUserProperties(), StateId.forItem(item));
    }
    
    private static FileState getSymbolicLinkState(ITeamRepository repo, ISymbolicLink item) {
        String target = item.getTarget();
        boolean directoryLink = item.isDirectoryLink();
        return FileState.create(VersionablePathSegment.pathFor(item), item.modified(), false, item.getUserProperties(), StateId.forItem(item), target, directoryLink);
    }

}
