/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2006, 2010. All Rights Reserved. 
 * 
 * Note to U.S. Government Users Restricted Rights:  Use, 
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.filesystem.client.internal;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.OperationCanceledException;

import com.ibm.team.repository.client.ITeamRepository;
import com.ibm.team.repository.client.TeamPlatform;
import com.ibm.team.repository.common.IItem;
import com.ibm.team.repository.common.IItemHandle;

public class ClientRepositoryUtil {


    /**
     * If there is exactly one known repository, this returns that
     * repository. Otherwise, returns null. This can be used to set
     * the default value in UI that asks the user to pick a repository,
     * but should not be relied upon for anything important.
     * 
     * @return
     */
    public static ITeamRepository getDefaultRepository() {
        // If there is only one repository, use it
        ITeamRepository[] repositories = TeamPlatform.getTeamRepositoryService().getTeamRepositories();
        if (repositories.length == 1) {
            return repositories[0];
        }
        
        // If there is only one logged in repository, use it
        boolean found = false;
        ITeamRepository targetRepository = null;
        
        for (int i = 0; i < repositories.length; i++) {
            ITeamRepository teamRepository = repositories[i];
            
            if (teamRepository.loggedIn()) {
                if (found) {
                    targetRepository = null;
                } else {
                    targetRepository = teamRepository;
                    found = true;
                }
            }
        }
        
        return targetRepository;
    }

    /**
     * Returns the ITeamRepository associated with the given 
     * @param sharedItem
     * @return
     * @throws IllegalStateException
     */
    public static ITeamRepository getRepository(IItemHandle sharedItem) throws IllegalArgumentException {
        Object origin = sharedItem.getOrigin();
        if (origin instanceof ITeamRepository) {
            return (ITeamRepository) origin;
        }
        throw new IllegalArgumentException("Argument was not a shared item"); //$NON-NLS-1$
    }

    /**
     * Checks if the given monitor has been cancelled, and throws an InterruptedException if so.
     * 
     * @param monitor
     * @throws InterruptedException
     */
    public static void checkCanceled(IProgressMonitor monitor) throws OperationCanceledException {
        if (monitor != null && monitor.isCanceled()) {
            throw new OperationCanceledException(com.ibm.team.filesystem.client.internal.Messages.ClientRepositoryUtil_0);
        }
    }
    
    /**
     * Returns true iff the given item is a shared item from the given repository.
     * 
     * @param repo
     * @param toTest
     * @return
     */
    public static boolean isSharedItem(ITeamRepository repo, IItem toTest) {
        try {
            return repo.itemManager().getImmutableState(toTest) != null;
        }
        catch (IllegalArgumentException e) {
            return false;
        }
    }
}
