/*******************************************************************************
 * Licensed Materials - Property of IBM
 * © Copyright IBM Corporation 2006, 2015. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  Use, 
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.filesystem.client.internal;

import java.util.ArrayList;
import java.util.Collection;

import org.eclipse.core.runtime.IStatus;

import com.ibm.team.filesystem.client.FileSystemCore;
import com.ibm.team.filesystem.client.ILocalChange;
import com.ibm.team.filesystem.client.IShare;
import com.ibm.team.filesystem.client.IShareable;
import com.ibm.team.filesystem.client.ISharingDescriptor;
import com.ibm.team.filesystem.client.operations.ICollision;
import com.ibm.team.filesystem.client.operations.ILoadLocation;
import com.ibm.team.repository.common.UUID;
import com.ibm.team.repository.common.util.NLS;
import com.ibm.team.scm.common.IComponentHandle;
import com.ibm.team.scm.common.IContextHandle;

public class Collision implements ICollision {

    private final IShareable shareable;
    private final IShare share;
    private final ArrayList<ILoadLocation> locations;
    private final boolean alreadyExists;
    private ILoadLocation loadLocation;
    private Object lock = new Object();
    private IStatus status;
    private final boolean nestedShare;

    public Collision(IShareable shareable, IShare share, ILoadLocation loadLocation, boolean alreadyExists, boolean nestedShare) {
        this.shareable = shareable;
        this.share = share;
        this.locations = new ArrayList<ILoadLocation>();
        this.locations.add(loadLocation);
        this.alreadyExists = alreadyExists;
        this.nestedShare = nestedShare;
    }
    
    public IShareable getShareable() {
        return shareable;
    }

    public IShare getShare() {
        return share;
    }
    
    public IStatus getStatus() {
        if (status == null) {
            String name = shareable.getLocalPath().toString();
            if (share != null) {
                ISharingDescriptor descriptor = share.getSharingDescriptor();
                if (share != null) {
                    name = name + " [" + descriptor.getConnectionName() + " - " + descriptor.getComponentName() + "]"; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
                }
                ILocalChange[] changes = FileSystemCore.getSharingManager().getLocalChangeManager().getPendingChanges(new IShare[] {share});
                if (changes.length > 0) {
                    name = NLS.bind(Messages.Collision_0, new Object[] {name, changes.length});
                }
            }

            if (alreadyExists) {
                if (share == null || share.getSharingDescriptor() == null) {
                    status = FileSystemStatusUtil.getStatusFor(IStatus.WARNING, NLS.bind(Messages.Collision_1, name), null);
                } else {
                    ISharingDescriptor descriptor = share.getSharingDescriptor();
                    UUID descriptorId = descriptor.getRepositoryId();
                    IContextHandle descriptorConnectionHandle = descriptor.getConnectionHandle();
                    IComponentHandle descriptorComponent = descriptor.getComponent();
                    for (ILoadLocation location : locations) {
                        if (location.getConnection().teamRepository().getId().equals(descriptorId)
                                && location.getConnection().getContextHandle().equals(descriptorConnectionHandle)
                                && location.getComponent().equals(descriptorComponent)) {
                            status = FileSystemStatusUtil.getStatusFor(nestedShare ? IStatus.ERROR : IStatus.WARNING, NLS.bind(Messages.Collision_2, name), null);
                            break;
                        }
                    }
                    if (status == null) {
                        status = FileSystemStatusUtil.getStatusFor(IStatus.ERROR, NLS.bind(Messages.Collision_3, name), null);
                    }
                }

            } else if (locations.size() > 1) {
                String firstComponent = locations.get(0).getComponent().getName();
                String secondComponent = locations.get(1).getComponent().getName();
                status = FileSystemStatusUtil.getStatusFor(IStatus.WARNING, NLS.bind(Messages.Collision_4, name, firstComponent, secondComponent), null);
            } else {
                // not sure if possible to get here
                status = FileSystemStatusUtil.getStatusFor(IStatus.WARNING, name, null);
            }
                
        }
        return status;
    }

    public ILoadLocation getLoadLocation() {
        synchronized (lock) {
            return loadLocation;
        }
    }
    public Collection<ILoadLocation> getLocations() {
        synchronized (lock) {
            return locations;
        }
    }
    public void setLoadLocation(ILoadLocation location) {
        synchronized (lock) {
            this.loadLocation = location;
        }
    }

    public boolean collidedWithExistingContent() {
        return alreadyExists;
    }

}
