/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2006, 2012. All Rights Reserved. 
 * 
 * Note to U.S. Government Users Restricted Rights:  Use, 
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.filesystem.client.internal;

import java.io.IOException;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Set;

import org.eclipse.core.runtime.Assert;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.MultiStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.jobs.Job;

import com.ibm.team.filesystem.client.FileSystemCore;
import com.ibm.team.filesystem.client.FileSystemException;
import com.ibm.team.filesystem.client.FileSystemStatusException;
import com.ibm.team.filesystem.client.internal.utils.PersistentBusyFlag;
import com.ibm.team.internal.repository.rcp.dbhm.DBHMException;
import com.ibm.team.repository.common.util.NLS;

public abstract class DiskBackedMapManager<K, V> {
    protected ReadWriteLock lock;
    protected Map<IPath, LockableMap<K, V>> inUse;
    private AsyncSaveJob saveJob;

    /** Set of write-mode maps that are currently open. */
    protected Set<IPath> lockedExclusively = new HashSet<IPath>();
    
    protected final ICorruptible corruptible;
    
    public DiskBackedMapManager(ReadWriteLock lock, PersistentBusyFlag busyFlag, ICorruptible corruptible) {
        this.lock = lock;
        this.inUse = new HashMap<IPath, LockableMap<K, V>>();
        this.saveJob = new AsyncSaveJob(busyFlag);
        this.corruptible = corruptible;
    }
    
    protected AsyncSaveJob getSaveJob() {
        return saveJob;
    }
    
    /**
     * Acquires the map corresponding to the given path.
     * The map has to be released after use.
     * 
     * @param path
     * @param mode 
     * @param creates
     * @return
     * @throws FileSystemException
     * @throws {@link DBHMException} if there is an internal problem during map 
     *      manipulations. The caller should catch it and convert to a 
     *      {@link FileSystemException}.
     */
    public LockableMap<K, V> loadMap(IPath path, boolean exclusive, IPath realPath) throws FileSystemException {
        lock.acquireRead();
        boolean success = false;
        try {
            LockableMap<K, V> map;
            // modifications to the list of maps needs to be sync'd
            synchronized (inUse) {
                map = inUse.get(path);
                try {
                    if (map == null) {
                        // Load the map
                        inUse.put(path, map = getLockableMap(path, realPath));
                     
                        if (exclusive) {
                            setBusy();
                            lockedExclusively.add(path);
                        }
                    }
                } catch (DBHMException e) {
                    corruptible.setCorrupt(true, "Failure during load", e); //$NON-NLS-1$
                    throw e;
                } catch(Exception e) {
                    throw new FileSystemException(Messages.DiskBackedMapManager_1, e);
                }
            }

            saveJob.cancelSave(path);

            map.acquire(exclusive);

            success = true;

            return map;
        } finally {
            if (!success) {
                lock.release();
            }
        }
    }

    /** Update the modification flag. We don't raise an exception in the hope 
     * that we'll be able to recover from this at some point in the future. 
     */
    private void setBusy() {
        MultiStatus multi = new MultiStatus(FileSystemCore.ID, -1, "Exception setting busy flag", null); //$NON-NLS-1$
        saveJob.busyFlag.setBusy(multi);
        
        if (!multi.isOK()) {
            LoggingHelper.log(multi);
        }
    }

    public void releaseMap(LockableMap<K, V> map) throws FileSystemException {
        if (map.release()) {
            saveJob.requestSave(map.getPath());
        }
        lock.release();
    }
    
    /**
     * Ensures all maps are closed. Maps need to be closed before their corresponding 
     * files are moved or deleted. 
     * <p>
     * This assumes the lock has been acquired in exclusive mode.
     * </p>
     *   
     * @throws FileSystemException 
     */
    public void closeAll() throws FileSystemException {
        Assert.isTrue(lock.threadWrites() > 0);
        MultiStatus result = new MultiStatus(FileSystemCore.ID, IStatus.OK, "Errors happening while saving map", null); //$NON-NLS-1$
        for (Iterator<LockableMap<K, V>> i = this.inUse.values().iterator(); i.hasNext();) {
            LockableMap<K, V> current = i.next();
            try {
                i.remove();
                current.close();
            } catch (DBHMException e) {
                corruptible.setCorrupt(true, "Failure during closeAll()", e);  //$NON-NLS-1$
                result.add(FileSystemStatusUtil.getStatusFor(IStatus.ERROR, errorClosing(current), e));
            } catch (IOException e) {
                result.add(FileSystemStatusUtil.getStatusFor(IStatus.ERROR, errorClosing(current), e));
            }
        }
        
        lockedExclusively.clear();
        
        // We allow the busy flag to raise an exception because we're shutting 
        //  down. We know that this is our last chance at modifying the 
        //  filesystem.
        saveJob.busyFlag.setComplete(result);
        
        if (!result.isOK())
            throw new FileSystemStatusException(result);
    }

    /**
     * @param current
     * @return
     */
    private String errorClosing(LockableMap<K, V> current) {
        return NLS.bind("Error closing {0}", current.getPath()); //$NON-NLS-1$
    }

    
    protected abstract LockableMap<K, V> getLockableMap(IPath path, IPath realPath);

    protected class AsyncSaveJob extends Job {
        private static final int SAVE_THRESHOLD = 1024;
        private static final int SCHEDULING_DELAY = 200;

        private Collection<IPath> toSave = new LinkedHashSet<IPath>();
        
        
        private volatile long latestRequest;

        private PersistentBusyFlag busyFlag;

        public AsyncSaveJob(PersistentBusyFlag busyFlag) {
            super("Saving metadata maps"); //$NON-NLS-1$
            this.setUser(false);
            this.setSystem(true);
            this.busyFlag = busyFlag;
        }

        public void requestSave(IPath path) {
            synchronized (toSave) {
                toSave.add(path);
            }
            latestRequest = System.currentTimeMillis();
            schedule(SCHEDULING_DELAY);
        }
        
        private void cancelSave(IPath path) {
            synchronized (toSave) {
                toSave.remove(path);
            }
        }

        public PersistentBusyFlag getBusyFlag() {
            return busyFlag;
        }
        
        protected IStatus run(IProgressMonitor monitor) {
            if ((System.currentTimeMillis() - latestRequest) < SCHEDULING_DELAY) {
                boolean mustSave;
                synchronized (toSave) {
                    mustSave = toSave.size() > SAVE_THRESHOLD;
                }
                if (!mustSave) {
                    schedule(SCHEDULING_DELAY);
                    return Status.CANCEL_STATUS;
                }
            }
            MultiStatus result = new MultiStatus(FileSystemCore.ID, IStatus.OK, "Errors happened while saving map", null); //$NON-NLS-1$
            
            lock.acquireWrite();

            try {
                if (!busyFlag.isBusy() && !lockedExclusively.isEmpty()) {
                    result.add(FileSystemStatusUtil.getStatusFor(IStatus.ERROR, NLS.bind("Expected busyFlag to be set: {0}", busyFlag))); //$NON-NLS-1$
                }
                boolean saved = false;
                for (IPath next; (next = getNext()) != null;) {
                    LockableMap<K, V> toSave = inUse.remove(next);
                    lockedExclusively.remove(next);

                    if (toSave != null)
                        try {
                            toSave.close();
                            saved = true;
                        } catch (DBHMException e) {
                            corruptible.setCorrupt(true, "close() failed", e);  //$NON-NLS-1$
                            result.add(FileSystemStatusUtil.getStatusFor(IStatus.ERROR, NLS.bind("Error closing {0}", next), e)); //$NON-NLS-1$
                        } catch (IOException e) {
                            result.add(FileSystemStatusUtil.getStatusFor(IStatus.ERROR, NLS.bind("Error closing {0}", next), e)); //$NON-NLS-1$
                        }
                }

                if (lockedExclusively.isEmpty() && (saved || !result.isOK())) {
                    busyFlag.setComplete(result);
                }
            } finally {
                lock.release();
            }
            return result;
        }

        private IPath getNext() {
            synchronized (toSave) {
                if (toSave.isEmpty()) {
                    return null;
                } else {
                    Iterator<IPath> i = toSave.iterator();
                    IPath path = i.next();
                    i.remove();
                    return path;
                }
            }
        }
    }
}
