/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2005, 2011. All Rights Reserved. 
 * 
 * Note to U.S. Government Users Restricted Rights:  Use, 
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.filesystem.client.internal;


import java.util.Collections;

import org.eclipse.core.runtime.Assert;

import com.ibm.team.filesystem.common.FileLineDelimiter;
import com.ibm.team.filesystem.common.IFileContent;
import com.ibm.team.filesystem.common.IFileItemHandle;
import com.ibm.team.filesystem.common.ISymbolicLinkHandle;
import com.ibm.team.scm.common.ContentHash;
import com.ibm.team.scm.common.IFolderHandle;
import com.ibm.team.scm.common.IVersionableHandle;

/**
 * Stores information related to a file/directory living on the server and/or on the
 * client file system. Also used to maintain the information required to
 * synchronize a local file resource with its mirror residing on a workspace.
 * 
 * <p>
 * FileItemInfo may have a number of states. 
 * <ul>
 *   <li>In general: {@link #handle} will never be null. If the item is not 
 *          present locally, then calling {@link IVersionableHandle#getStateId()} 
 *          will return null.
 *          
 *   <li>The share root may have a null name and parent if it is the component
 *          root folder or meta data prior to 2.0.  The remote name and loaded
 *          name may be different if the root was loaded as an alternative name.
 *   
 *   <li>If the info represents an uncommitted file:
 *      <ul>
 *         <li>the {@link #handle} will be set to non-null
 *         <li>the {@link #lastModification} will be NULL_STAMP
 *         <li>the {@link #parent} will be <code>null</code>
 *         <li>the {@link #name} will be <code>null</code>
 *         <li>the {@link #contentLength} will be NULL_SIZE
 *         <li>the {@link #hash} will be <code>null</code> 
 *      </ul>
 *      
 *   <li>If the info represents a committed file:
 *      <ul>
 *         <li>the {@link #handle} will be set to non-null
 *         <li>the {@link #lastModification} will be the timestamp at last commit, 
 *              OR the timestamp at last touch. The timestamp is returned by 
 *              {@link IFileStorage#getModificationStamp()}. 
 *         <li>the {@link #parent} will be the parent at last commit
 *         <li>the {@link #name} will be the name at last commit
 *         <li>the {@link #contentLength} will be the on-disk size at last commit 
 *         <li>the {@link #hash} will be the hash of the on-disk content at last commit 
 *      </ul>
 *
 *      
 *   <li>If the info represents a directory:
 *      <ul>
 *         <li>the {@link #lastModification} will always be NULL_STAMP
 *         <li>the {@link #contentLength} will always be NULL_SIZE
 *         <li>the {@link #hash} will always be <code>null</code>
 *      </ul>
 *   
 *   <li>If the info represents an uncommitted directory:
 *      <ul>
 *         <li>the {@link #handle} will be set to non-null
 *         <li>the {@link #parent} will be set to <code>null</code>
 *         <li>the {@link #name} will be <code>null</code>
 *      </ul>
 *      
 *   <li>The the info represents a committed directory:
 *      <ul>
 *         <li>the {@link #handle} will be set to non-null
 *         <li>the {@link #parent} will be the parent at last commit
 *         <li>the {@link #name} will be the name at last commit
 *      </ul>
 *      
 *      
 *   <li>If the info represents a symbolic link:
 *      <ul>
 *         <li>the {@link #lastModification} will always be NULL_STAMP
 *         <li>the {@link #contentLength} will always be NULL_SIZE
 *      </ul>
 *      
 *   <li>If the info represents an uncommitted symbolic link:
 *      <ul>
 *         <li>the {@link #handle} will be set to non-null
 *         <li>the {@link #parent} will be set to <code>null</code>
 *         <li>the {@link #name} will be <code>null</code>
 *         <li>the {@link #hash} will be <code>null</code> 
 *      </ul>
 *      
 *   <li>The the info represents a committed symbolic link:
 *      <ul>
 *         <li>the {@link #handle} will be set to non-null
 *         <li>the {@link #parent} will be the parent at last commit
 *         <li>the {@link #name} will be the name at last commit
 *         <li>the {@link #hash} will be the hash of the on-disk content at last commit 
 *      </ul>
 *
 * </ul>
 */
public class FileItemInfo {
    /**
     * A null time stamp is used in the event it is not known/applicable.
     */
    public static final long NULL_STAMP = -1;
    public static final long NULL_DELIMITER_COUNT = -1;
    


    final private IVersionableHandle handle;

    /**
     * The timestamp of the file when it was last checked for content modification,
     * if unknown or it was never checked then the value is {@link #NULL_STAMP}.
     */
    final private long lastContentChangeCheckStamp;
    
    final private boolean contentChanged;
    
    final private IFolderHandle parent;
    
    final private String name;
    
    /**
     * <code>true</code> when the file or folder was loaded under an alternative name.
     * For new files/folders this will be <code>false</code>
     */
    final private boolean loadedWithAnotherName;
    
    /**
     * A null size for files. To be used for directories or files whose size 
     * is unknown. 
     */
    public static final long NULL_SIZE = -1;
    
    /**
     * The size of the file in bytes. May be <code>NULL_SIZE</code> if this 
     * file has not been committed. 
     */
    final private long contentLength;
    
    /**
     * The hashcode of the file. If we're dealing with a folder, then the 
     * hash must be null. 
     */
    final private ContentHash hash;

    /**
     * The preferred line delimiter of the file on retrieval.  If we're dealing with
     * a folder then the line delimiter will be <code>null</code>.  If the file has 
     * not yet been created in the repository (new file addition)/or had the property assigned,
     * it may be <code>null</code>.  If <code>null</code> then it will be decided when
     * committed. This is the current preference.
     */
    final private FileLineDelimiter lineDelimiter;

    /**
     * The preferred line delimiter of the file on retrieval.  If we're dealing with
     * a folder then the line delimiter will be <code>null</code>.
     * This is the original setting when the content was retrieved.
     * For new files this will be <code>null</code>  It will
     * be determined on the first commit.
     */
    final private FileLineDelimiter originalLineDelimiter;

    /**
     * The content type of the file.  If we're dealing with a folder
     * then the content type will be empty string.
     * This is the current preference.  For new files this will be <code>null</code>.
     * It will be determined on the first commit.
     */
    final private String contentType;

    /**
     * The content type of the file.  If we're dealing with a folder
     * then the content type will be empty string.  This is the original setting
     * when the content was retrieved.  For new files this will be <code>null</code>.
     * It will be determined on the first commit.
     */
    final private String originalContentType;
    
    
    /* original content fields */
    private final ContentHash predecessorHintHash;
    private final long sizeInRepo;
    private final String encoding;
    private final ContentHash storedHash;
    private final long numLineDelimiters;


    /**
     * True if the file is marked executable on the server, false otherwise. False for folders or symbolic links.
     */
    private final boolean originalExecutable;
    
    /**
     * True if the file is executable, false otherwise. False for folders or symbolic links.
     */
    private final boolean executable;
    
    /**
     * True if the symbolic link has a directory type on the server, false otherwise. False for folders or files.
     */
    private final boolean originalDirectoryLink;
    
    /**
     * True if the symbolic link has a directory type, false otherwise. False for folders or files.
     */
    private final boolean directoryLink;

    /**
     * @param handle must not be null, must not be a full state. If this item is present remotely
     *          it must have a state id. If this item is not present remotely then it must not have
     *          a state id.
     * @param lastModification the modification time of the file at which time it had the same
     *          content as the content in the repo. Should be NULL_STAMP if unknown (that is
     *          it currently does not have the same contents as what's in repo) or for newly
     *          added files.
     * @param parent The remote parent. Must be null for share root or if the item is not present remotely.
     *                  Must be not null otherwise. Must not be a full state.
     * @param name The remote name. Must be null for share root or if the item is not present remotely.
     *              Must be not null otherwise.
     * @param hash The hash of the contents of the file when it was retrieved and stored on disk
     *          This may be different from the hash in the {@link IFileContent#getRawHashCode()} because
     *          of differences in line delimiters.
     * @param size The size of the file when it was retrieved and stored on disk.  This may different
     *          from the size in the repo due to differences in line delimiters.
     * @param originalLineDelimiter The preferred line delimiter the content had when originally
     *                  retrieved.
     * @param lineDelimiter The preferred line delimiter should have when the content is next
     *                  committed.                      
     */
    public FileItemInfo(IVersionableHandle handle, boolean contentChanged, long lastContentChangeCheckStamp, IFolderHandle parent,
            String name, boolean loadedAsAlternativeName, ContentHash hash, long size,
            FileLineDelimiter originalLineDelimiter, FileLineDelimiter lineDelimiter,
            String originalContentType, String contentType,
            ContentHash contentPredecessorHintHash, long contentSizeInRepo, String originalEncoding,
            ContentHash hashInRepo, long lineDelimiterCountInRepo, boolean executable, boolean originalExecutable, boolean directoryLink, boolean originalDirectoryLink) {
        validateFileItemInfo(handle, parent, name, loadedAsAlternativeName);
        
        if (handle instanceof IFileItemHandle) {
            if (contentType == null && originalContentType != null) {
                throw new IllegalArgumentException();
            }
            if (lineDelimiter == null && originalLineDelimiter != null) {
                throw new IllegalArgumentException();
            }
        } else {
            if (originalContentType != null || contentType != null) {
                throw new IllegalArgumentException();
            } 
            if (originalLineDelimiter != null || lineDelimiter != null) {
                throw new IllegalArgumentException();
            }
            //executable is always ignored for folders or symbolic links, set it to false
            executable = false;
            originalExecutable = false;
        }
        
        if (!(handle instanceof ISymbolicLinkHandle)) {
            directoryLink = false;
            originalDirectoryLink = false;
        }
        
        if (!handle.hasStateId() || handle instanceof IFolderHandle) {
            if (hash != null) {
                throw new IllegalArgumentException();
            }
            if (lastContentChangeCheckStamp != NULL_STAMP) {
                throw new IllegalArgumentException();
            }
            if (contentChanged) {
                throw new IllegalArgumentException();
            }
            if (size != NULL_SIZE) {
                throw new IllegalArgumentException();
            }
            if (originalLineDelimiter != null) {
                throw new IllegalArgumentException();
            }
            if (originalContentType != null) {
                throw new IllegalArgumentException();
            }
            if (contentPredecessorHintHash != null) {
                throw new IllegalArgumentException();
            }
            if (contentSizeInRepo != NULL_SIZE) {
                throw new IllegalArgumentException();
            }
            if (originalEncoding != null) {
                throw new IllegalArgumentException();
            }
            if (hashInRepo != null) {
                throw new IllegalArgumentException();
            }
            if (lineDelimiterCountInRepo != NULL_DELIMITER_COUNT) {
                throw new IllegalArgumentException();
            }
        } else if (handle.hasStateId()) {
            if (hash == null) {
                throw new IllegalArgumentException();
            }
            if (hashInRepo == null) {
                throw new IllegalArgumentException();
            }
            if (handle instanceof IFileItemHandle) {
                if (size == NULL_SIZE) {
                    throw new IllegalArgumentException();
                }
            } else {
                if (size != NULL_SIZE) {
                    throw new IllegalArgumentException();
                }
            }
        }
        
        if (hashInRepo == null) {
            Assert.isTrue(originalLineDelimiter == null);
            Assert.isTrue(originalContentType == null);
            Assert.isTrue(contentPredecessorHintHash == null);
            Assert.isTrue(contentSizeInRepo == NULL_SIZE);
            Assert.isTrue(originalEncoding == null);
            Assert.isTrue(lineDelimiterCountInRepo == NULL_DELIMITER_COUNT);
        } else {
            Assert.isTrue(handle.hasStateId());
            Assert.isTrue(!(handle instanceof IFolderHandle));
            if (handle instanceof IFileItemHandle) {
                Assert.isTrue(contentSizeInRepo != NULL_SIZE);
                Assert.isTrue(originalLineDelimiter != null);
                Assert.isTrue(originalContentType != null);
            } else {
                Assert.isTrue(contentSizeInRepo == NULL_SIZE);
                Assert.isTrue(originalLineDelimiter == null);
                Assert.isTrue(originalContentType == null);
            }
        }
        
        this.handle = handle;
        this.parent = parent;
        this.name = name;
        this.loadedWithAnotherName = loadedAsAlternativeName;
        this.lastContentChangeCheckStamp = lastContentChangeCheckStamp;
        this.contentChanged = contentChanged;
        this.hash = hash;
        this.contentLength = size;
        this.lineDelimiter = lineDelimiter;
        this.originalLineDelimiter = originalLineDelimiter;
        this.originalContentType = originalContentType;
        this.contentType = contentType;
        
        this.predecessorHintHash = contentPredecessorHintHash;
        this.sizeInRepo = contentSizeInRepo;
        this.encoding = originalEncoding;
        this.storedHash = hashInRepo;
        this.numLineDelimiters = lineDelimiterCountInRepo;
        
        this.executable = executable;
        this.originalExecutable = originalExecutable;
        
        this.directoryLink = directoryLink;
        this.originalDirectoryLink = originalDirectoryLink;
    }
    
    /**
     * hash and modification information is not applicable for folders So
     * this constructor allows you to omit it.
     * 
     * @param handle must not be null, must not be a full state. If this item is present remotely
     *          it must have a state id. If this item is not present remotely then it must not have
     *          a state id.
     * @param parent The remote parent. Must be null for share root or if the item is not present remotely.
     *                  Must be not null otherwise. Must not be a full state.
     * @param name The remote name. Must be null for share root or if the item is not present remotely.
     *              Must be not null otherwise.
     */
    public FileItemInfo(IFolderHandle handle, IFolderHandle parent, String name, boolean loadedWithAlternativeName) {
        validateFileItemInfo((IVersionableHandle)handle, parent, name, loadedWithAlternativeName);
        this.handle = handle;
        this.parent = parent;
        this.name = name;
        this.loadedWithAnotherName = loadedWithAlternativeName;
        this.lastContentChangeCheckStamp = NULL_STAMP;
        this.contentChanged = false;
        this.contentLength = NULL_SIZE;
        this.originalLineDelimiter = null;
        this.lineDelimiter = null;
        this.hash = null;
        this.originalContentType = null;
        this.contentType = null;
        
        this.predecessorHintHash = null;
        this.sizeInRepo = NULL_SIZE;
        this.encoding = null;
        this.storedHash = null;
        this.numLineDelimiters = NULL_DELIMITER_COUNT;
        
        this.originalExecutable = false;
        this.executable = false;
        
        this.directoryLink = false;
        this.originalDirectoryLink = false;
    }
    
    /**
     * Special constructor specifically for symbolic links.
     * 
     * @param handle must not be null, must not be a full state. If this item is present remotely
     *          it must have a state id. If this item is not present remotely then it must not have
     *          a state id.
     * @param parent The remote parent. Must be null for share root or if the item is not present remotely.
     *                  Must be not null otherwise. Must not be a full state.
     * @param name The remote name. Must be null for share root or if the item is not present remotely.
     *              Must be not null otherwise.
     * @param hash The hash of the target of the symbolic link in local form last time it was updated from the repo.
     * @param hashInRepo The hash of the target of the symbolic link when it was last updated from the repo.
     * @param directoryLink True if the symbolic link has a directory type, false otherwise.
     * @param originalDirectoryLink True if the symbolic link has a directory type in the repo, false otherwise.
     */
    public FileItemInfo(ISymbolicLinkHandle handle, boolean contentChanged, IFolderHandle parent,
            String name, boolean loadedAsAlternativeName, ContentHash hash,
            ContentHash hashInRepo, boolean directoryLink, boolean originalDirectoryLink) {
        this(handle, contentChanged, NULL_STAMP, parent, name, loadedAsAlternativeName, hash, 
                NULL_SIZE, null, null, null, null, null, NULL_SIZE, null, 
                hashInRepo, NULL_DELIMITER_COUNT, false, false, directoryLink, originalDirectoryLink);
    }
    
    private void validateFileItemInfo(IVersionableHandle handle, IFolderHandle parent, String name, boolean loadedWithAlternativeName) {
        Assert.isNotNull(handle);
        Assert.isTrue(!handle.hasFullState());
        Assert.isTrue(parent == null || !parent.hasFullState());
        if (name != null && name.length() == 0) {
            // The transport layer doesn't differentiate between null and the empty string so lets treat them the same here
            name = null;
        }
        Assert.isTrue((parent == null) == (name == null));
        Assert.isTrue(parent == null || handle.hasStateId());
        // if its remote (has state id) and is component root (parent is null) then it must be loaded with an alternative name.
        // if not remote then it must not be considered loaded with an alternative name
        if (handle.hasStateId() && parent == null) {
            Assert.isTrue(loadedWithAlternativeName);
        } else if (!handle.hasStateId()) {
            Assert.isTrue(!loadedWithAlternativeName);
        }
        //disallow non-folders with state id but no remote parent
        Assert.isTrue(parent != null || !handle.hasStateId() || handle instanceof IFolderHandle);
    }

    public IVersionableHandle getVersionableHandle() {
        return handle;
    }

    /**
     * Returns whether the handle associated with this info is a folder.
     * If this method returns false, callers should consult either the isFile or
     * getItemType methods to determine the handle type.
     * @return whether the handle associated with this info is a folder
     */
    public boolean isFolder() {
        return handle instanceof IFolderHandle;
    }
    
    /**
     * Returns whether the handle associated with this info is a file.
     * If this method returns false, callers should consult either the isFolder or
     * getItemType methods to determine the handle type.
     * @return whether the handle associated with this info is a file
     */
    public boolean isFile() {
        return handle instanceof IFileItemHandle;
    }
    
    /**
     * Returns whether the handle associated with this info is a symbolic link.
     * @return whether the handle associated with this info is a symbolic link
     */
    public boolean isSymbolicLink() {
        return handle instanceof ISymbolicLinkHandle;
    }

    /**
     * Returns the timestamp of the file when it was last checked for content modification,
     * if unknown or it was never checked then the value is {@link #NULL_STAMP}.
     */
    public long getLastContentChangeCheckStamp() {
        return lastContentChangeCheckStamp;
    }
    
    /**
     * Returns true if the content for this item on disk has been changed from what is
     * in the repo. Returns false if this is not a file or the content is the same.
     */
    public boolean isContentChanged() {
        return contentChanged;
    }
    
    /**
     * Returns the remote parent of this item.
     */
    public IFolderHandle getParent() {
        return parent;
    }
    
    /**
     * Returns the remote name of this item.
     */
    public String getName() {
        return name;
    }
    
    /**
     * Returns whether this item was loaded under an alternative name.
     */
    public boolean isLoadedWithAnotherName() {
        return loadedWithAnotherName;
    }
    
    /**
     * @return A hashcode, or <code>null</code> if:
     *  <ul>
     *    <li/>The file has not been committed.
     *    <li/>The receiver represents a folder. 
     *  </ul>
     */
    public ContentHash getHash() {
        return hash;
    }
    
    
    /**
     * @return The length of the file in bytes, or <code>NULL_SIZE</code> if: 
     *      the file has not been committed; or the receiver represents a 
     *      folder.
     */
    public long getContentLength() {
        return contentLength;
    }
    
    /**
     * @return The line delimiter preferred for retrieval of content.  If this is
     * a folder <code>null</code> will be returned.  If this is for a file that is not
     * remote, the value will be <code>null</code>.
     */
    public FileLineDelimiter getOriginalLineDelimiter() {
        return originalLineDelimiter;
    }
   
    /**
     * @return The line delimiter preferred for retrieval of content.  If this is
     * a folder <code>null</code> will be returned.  If this is for a file that is not
     * remote, the value will be <code>null</code>.
     */
    public FileLineDelimiter getLineDelimiter() {
        return lineDelimiter;
    }
    
    /**
     * @return The original Media type setting for the content.  If this is
     * a folder <code>null</code> will be returned.  If this is for a file that is not
     * remote, the value will be <code>null</code>.
     */
    public String getOriginalContentType() {
        return originalContentType;
    }
   
    /**
     * @return The current preferred Media type setting for the content.  If this is
     * a folder <code>null</code> will be returned.  If <code>null</code> is returned,
     * it will be inferred.
     */
    public String getContentType() {
        return contentType;
    }
    

    public ContentHash getStoredPredecessorHintHash() {
        return predecessorHintHash;
    }

    public long getStoredSize() {
        return sizeInRepo;
    }

    public String getStoredEncoding() {
        return encoding;
    }

    public ContentHash getStoredHash() {
        return storedHash;
    }

    public long getStoredNumLineDelimiters() {
        return numLineDelimiters;
    }
    
    public boolean isExecutable() {
        return this.executable;
    }
    
    public boolean isOriginalExecutable() {
        return this.originalExecutable;
    }
    
    public boolean isDirectoryLink() {
        return this.directoryLink;
    }
    
    public boolean isOriginalDirectoryLink() {
        return this.originalDirectoryLink;
    }
    
    /**
     * Return a new inverse file item info for this item which doesn't have a remote
     * counterpart.
     * @param localParent the local parent
     * @param localName the local name
     * @return a new inverse file item info for this item
     */
    public InverseFileItemInfo createInverseFileItemInfoForLocalItem(IFolderHandle localParent, String localName) {
        return new InverseFileItemInfo(getVersionableHandle(), false, FileItemInfo.NULL_STAMP, null,
                null, false, Collections.EMPTY_MAP, localParent, localName,
                null, FileItemInfo.NULL_SIZE, null, getLineDelimiter(),
                null, getContentType(), null,
                FileItemInfo.NULL_SIZE, null, null, FileItemInfo.NULL_DELIMITER_COUNT,
                isExecutable(), false, 
                isDirectoryLink(), false, Collections.EMPTY_MAP, Collections.EMPTY_SET, Collections.EMPTY_MAP);
    }
}
