/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2006, 2011. All Rights Reserved. 
 * 
 * Note to U.S. Government Users Restricted Rights:  Use, 
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.filesystem.client.internal;

import java.io.IOException;

import com.ibm.team.filesystem.client.FileSystemException;
import com.ibm.team.filesystem.client.internal.copyfileareas.CopyFileArea;
import com.ibm.team.filesystem.common.FileLineDelimiter;
import com.ibm.team.filesystem.common.IFileItem;
import com.ibm.team.filesystem.common.ISymbolicLink;
import com.ibm.team.filesystem.common.ISymbolicLinkHandle;
import com.ibm.team.filesystem.common.internal.FileContent;
import com.ibm.team.repository.common.util.NLS;
import com.ibm.team.scm.common.ContentHash;
import com.ibm.team.scm.common.ContentHashAlgorithmException;
import com.ibm.team.scm.common.IFolderHandle;
import com.ibm.team.scm.common.IVersionable;
import com.ibm.team.scm.common.IVersionableHandle;

public class FileItemInfoProxy {

    private IVersionableHandle handle;

    /**
     * The timestamp of the file when it was last checked for content modification,
     * if unknown or it was never checked then the value is {@link #NULL_STAMP}.
     */
    private long lastContentChangeCheckStamp = FileItemInfo.NULL_STAMP;
    
    private boolean contentChanged = false;
    
    private IFolderHandle parent;
    
    private String name;
    
    private boolean loadedWithAnotherName = false;
    
    /**
     * The size of the file in bytes. May be <code>NULL_SIZE</code> if this 
     * file has not been committed. 
     */
    private long contentLength = FileItemInfo.NULL_SIZE;
    
    /**
     * The hashcode of the file or symbolic link. If we're dealing with a folder, then the 
     * hash must be null. 
     */
    private ContentHash hash = null;

    /**
     * The preferred line delimiter of the file on retrieval.  If we're dealing with
     * a folder then the line delimiter will be <code>null</code>.  If the file has 
     * not yet been created in the repository (new file addition)/or had the property assigned,
     * it may be <code>null</code>.  If <code>null</code> then it will be decided when
     * committed. This is the current preference.
     */
    private FileLineDelimiter lineDelimiter = null;

    /**
     * The preferred line delimiter of the file on retrieval.  If we're dealing with
     * a folder then the line delimiter will be <code>null</code>.
     * This is the original setting when the content was retrieved.
     * For new files this will be <code>null</code>  It will
     * be determined on the first commit.
     */
    private FileLineDelimiter originalLineDelimiter = null;

    /**
     * The content type of the file.  If we're dealing with a folder
     * then the content type will be empty string.
     * This is the current preference.  For new files this will be <code>null</code>.
     * It will be determined on the first commit.
     */
    private String contentType = null;

    /**
     * The content type of the file.  If we're dealing with a folder
     * then the content type will be empty string.  This is the original setting
     * when the content was retrieved.  For new files this will be <code>null</code>.
     * It will be determined on the first commit.
     */
    private String originalContentType = null;
    
    
    /* original content fields */
    ContentHash predecessorHintHash = null;
    long sizeInRepo = FileItemInfo.NULL_SIZE;
    String encoding = null;
    ContentHash hashInRepo = null;
    long numLineDelimiters = FileItemInfo.NULL_DELIMITER_COUNT;


    /**
     * True if the file is marked executable on the server, false otherwise. False for folders.
     */
    private boolean originalExecutable = false;
    
    /**
     * True if the file is executable, false otherwise. False for folders.
     */
    private boolean executable = false;
    
    /**
     * True if the symbolic link has a directory type on the server, false otherwise. False for folders or files.
     */
    private boolean originalDirectoryLink = false;
    
    /**
     * True if the symbolic link has a directory type, false otherwise. False for folders or files.
     */
    private boolean directoryLink = false;
    
    /**
     * If this is a folder then no other settings are required.  If a file or symbolic link you will want
     * to set potentially some of the other content settings.
     * 
     * @param handle must not be null, must not be a full state. If this item is present remotely
     *          it must have a state id. If this item is not present remotely then it must not have
     *          a state id.
     * @param parent The remote parent. Must be null for share root or if the item is not present remotely.
     *                  Must be not null otherwise. Must not be a full state.
     * @param name The remote name. Must be null for share root or if the item is not present remotely.
     *              Must be not null otherwise.
     * @param loadedWithAnotherName indication of whether this item is loaded as a different name.
     */
    public FileItemInfoProxy(IVersionableHandle handle, IFolderHandle parent, String name, boolean loadedWithAnotherName) {
        this.handle = handle;
        this.parent = parent;
        if (name == null || name.length() == 0) {
            this.name = null;
        } else {
            this.name = name;
        }
        this.loadedWithAnotherName = loadedWithAnotherName;
    }

    /**
     * If this is a folder then no other settings are required.  If a file, you will want
     * to set potentially some of the other content settings.
     * 
     * @param handle must not be null, must not be a full state. If this item is present remotely
     *          it must have a state id. If this item is not present remotely then it must not have
     *          a state id.
     * @param parent The remote parent. Must be null for share root or if the item is not present remotely.
     *                  Must be not null otherwise. Must not be a full state.
     * @param name The remote name. Must be null for share root or if the item is not present remotely.
     *              Must be not null otherwise.
     */
    public FileItemInfoProxy(IVersionableHandle handle, IFolderHandle parent, String name) {
        this.handle = handle;
        this.parent = parent;
        if (name == null || name.length() == 0) {
            this.name = null;
        } else {
            this.name = name;
        }
        this.loadedWithAnotherName = false;
    }

    /**
     * Create a copy of the old info so that the desired fields can be further updated.
     * @param oldInfo The old file item info to clone.
     */
    public FileItemInfoProxy(FileItemInfo oldInfo) {
        this.handle = oldInfo.getVersionableHandle();
        this.hashInRepo = oldInfo.getStoredHash();
        this.contentChanged = oldInfo.isContentChanged();
        this.contentLength = oldInfo.getContentLength();
        this.contentType = oldInfo.getContentType();
        this.predecessorHintHash = oldInfo.getStoredPredecessorHintHash();
        this.encoding = oldInfo.getStoredEncoding();
        this.executable = oldInfo.isExecutable();
        this.directoryLink = oldInfo.isDirectoryLink();
        this.hash = oldInfo.getHash();
        this.lastContentChangeCheckStamp = oldInfo.getLastContentChangeCheckStamp();
        this.lineDelimiter = oldInfo.getLineDelimiter();
        this.loadedWithAnotherName = oldInfo.isLoadedWithAnotherName();
        this.name = oldInfo.getName();
        this.numLineDelimiters = oldInfo.getStoredNumLineDelimiters();
        this.originalContentType = oldInfo.getOriginalContentType();
        this.originalExecutable = oldInfo.isOriginalExecutable();
        this.originalDirectoryLink = oldInfo.isOriginalDirectoryLink();
        this.originalLineDelimiter = oldInfo.getOriginalLineDelimiter();
        this.parent = oldInfo.getParent();
        this.sizeInRepo = oldInfo.getStoredSize();
    }

    public FileItemInfo getFileItemInfo() {
        if (handle instanceof IFolderHandle) {
            return new FileItemInfo((IFolderHandle) handle, parent, name, loadedWithAnotherName);
        } else if (handle instanceof ISymbolicLinkHandle) {
            return new FileItemInfo((ISymbolicLinkHandle) handle, contentChanged, parent, name, loadedWithAnotherName, hash, hashInRepo, directoryLink, originalDirectoryLink);
        } else {
            return new FileItemInfo(handle, contentChanged, lastContentChangeCheckStamp,
                    parent, name, loadedWithAnotherName, hash, contentLength,
                    originalLineDelimiter, lineDelimiter,
                    originalContentType, contentType,
                    predecessorHintHash, sizeInRepo, encoding, hashInRepo, numLineDelimiters,
                    executable, originalExecutable, false, false);
        }
    }

    /**
     * Set the File's content information.  This will setup the information as if the
     * file's content is completely in sync with the remote content.  The non-content
     * properties must be set separately.  See {@link #setProperties(IVersionable)}
     * 
     * @param modificationStamp The file's local modification time stamp
     * @param hashCode The hash code of the file locally (can be different from the content's hash code due to line
     *  delimiter translation.
     * @param contentSize The size of the file locally.
     * @param content The remote information about the content
     */
    public void setContentInfo(long modificationStamp, ContentHash hashCode, long contentSize, FileContent content) {
        this.contentChanged = false;
        this.lastContentChangeCheckStamp = modificationStamp;
        this.hash = hashCode;
        this.contentLength = contentSize;
        this.originalLineDelimiter = content.getLineDelimiter();
        this.lineDelimiter = content.getLineDelimiter();
        this.predecessorHintHash = content.getPredecessorHintHash();
        this.sizeInRepo = content.getSize();
        this.encoding = content.getCharacterEncoding();
        this.hashInRepo = content.getHash();
        // this should expand to include other content properties if necessary
        this.numLineDelimiters = content.getLineDelimiterCount();
    }

    /**
     * Set whether the file is to be considered changed from the remote state.
     * @param contentChanged a boolean indicating whether the file is changed or not
     * @param modificationStamp the timestamp of the file
     */
    public void setContentChanged(boolean contentChanged, long modificationStamp) {
        this.contentChanged = contentChanged;
        this.lastContentChangeCheckStamp = modificationStamp;
    }

    /**
     * After a commit or accept, update the properties of the item committed
     * This means the before and after values of the properties
     * (i.e. executable, contentType) will be the same
     * @param versionable The newly updated state.
     */
    public void setProperties(IVersionable versionable) {
        if (versionable instanceof IFileItem) {
            IFileItem fileItem = (IFileItem) versionable;
            this.originalExecutable = fileItem.isExecutable();
            this.executable = fileItem.isExecutable();
            this.originalContentType = fileItem.getContentType();
            this.contentType = fileItem.getContentType();
        } else if (versionable instanceof ISymbolicLink) {
            ISymbolicLink link = (ISymbolicLink)versionable;
            this.originalDirectoryLink =  link.isDirectoryLink();
            this.directoryLink = link.isDirectoryLink();
        } // else a folder and no properties right now
    }

    public void setContentType(String contentType) {
        this.originalContentType = contentType;
        this.contentType = contentType;
    }

    public void setExecutable(boolean originalExecutable, boolean executable) {
        this.originalExecutable = originalExecutable;
        this.executable = executable;
    }

    public void setItemState(IVersionableHandle itemState) {
        this.handle = itemState;
    }

    public void changeIdentity(IFolderHandle newParent, String newName) {
        this.parent = newParent;
        this.name = newName;
    }

    /**
     * Change the property information.  This is in response to user
     * initiated local changes.  The before setting if applicable remains.
     * 
     * @param contentChanged
     * @param lastContentChangeCheckStamp
     * @param currentLineDelimiter
     * @param currentContentType
     * @param currentExecutable
     */
    public void changeProperties(boolean contentChanged, long lastContentChangeCheckStamp,
            FileLineDelimiter currentLineDelimiter, String currentContentType, boolean currentExecutable) {
        
        this.contentChanged = contentChanged;
        this.lastContentChangeCheckStamp = lastContentChangeCheckStamp;
        this.lineDelimiter = currentLineDelimiter;
        this.contentType = currentContentType;
        this.executable = currentExecutable;
    }
    
    public void setTarget(long modificationStamp, String target, String targetInLocalForm, boolean directoryLink) throws FileSystemException {
        this.contentChanged = false;
        this.lastContentChangeCheckStamp = modificationStamp;
        this.hash = getLinkTargetHash(targetInLocalForm);
        this.hashInRepo = getLinkTargetHash(target);
        this.directoryLink = directoryLink;
        this.originalDirectoryLink = directoryLink;
    }
    
    public void setDirectoryLink(boolean directoryLink) {
        this.directoryLink = directoryLink;
    }
    
    public static ContentHash getLinkTargetHash(String linkTarget) throws FileSystemException {
        try {
            return CopyFileArea.getLinkHash(linkTarget);
        } catch (ContentHashAlgorithmException e) {
            throw new FileSystemException(NLS.bind(Messages.FileItemInfoProxy_0, linkTarget), e);
        } catch (IOException e) {
            throw new FileSystemException(e);
        }
    }
}
