/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010, 2011. All Rights Reserved. 
 * 
 * Note to U.S. Government Users Restricted Rights:  Use, 
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.filesystem.client.internal;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.SubMonitor;

import com.ibm.team.filesystem.client.FileSystemException;
import com.ibm.team.filesystem.common.FileLineDelimiter;
import com.ibm.team.filesystem.common.IFileItem;

public class FileOptions implements IFileOptions {

    private final boolean force;
    private final String encoding;
    private final FileLineDelimiter lineDelimiter;
    private final Map<String, String> properties;
 
    /**
     * @param force Whether to force the get/update/creation of the file
     * @param fileItem The versionable describing the file to be read/updated/created
     */
    public FileOptions(boolean force, IFileItem fileItem) {
        super();
        this.force = force;
        this.encoding = nonNullEncoding(fileItem.getContent().getCharacterEncoding());
        this.lineDelimiter = nonNullDelimiter(fileItem.getContent().getLineDelimiter());
        this.properties = new HashMap(fileItem.getUserProperties());
    }

    /**
     * @param force Whether to force the get/update/creation of the file
     * @param lineDelimiter The line delimiter that the file is expected to have.  It may have a different delimiter though.
     * @param characterEncoding The encoding the file is expected to have.  If line delimiter none, then the encoding can be null.
     * @param properties Additional properties of the file.
     */
    public FileOptions(boolean force, FileLineDelimiter lineDelimiter, String characterEncoding, Map<String, String> properties) {
        this.force = force;
        this.encoding = nonNullEncoding(characterEncoding);
        this.lineDelimiter = nonNullDelimiter(lineDelimiter);
        
        if (properties == null) {
            this.properties = Collections.EMPTY_MAP;
        } else {
            this.properties = Collections.unmodifiableMap(new HashMap<String, String>(properties));
        }
    }
    
    /**
     * @param force Whether to force the get/update/creation of the file
     * @param fileItemInfo The metadata describing the file to be read/updated/created
     * @param properties Additional properties of the file.
     */
    public FileOptions(boolean force, FileItemInfo fileItemInfo, Map<String, String> properties) {
        this.force = force;
        if (fileItemInfo != null) {
            this.lineDelimiter = nonNullDelimiter(fileItemInfo.getLineDelimiter());
            this.encoding = nonNullEncoding(fileItemInfo.getStoredEncoding());
        } else {
            this.encoding = ""; //$NON-NLS-1$
            this.lineDelimiter = FileLineDelimiter.LINE_DELIMITER_NONE;
        }
        if (properties == null) {
            this.properties = Collections.EMPTY_MAP;
        } else {
            this.properties = Collections.unmodifiableMap(new HashMap<String, String>(properties));
        }
    }
    
    private FileLineDelimiter nonNullDelimiter(FileLineDelimiter delimiter) {
        return delimiter == null ? FileLineDelimiter.LINE_DELIMITER_NONE : delimiter;
    }
    
    private String nonNullEncoding(String characterEncoding) {
        return characterEncoding == null ? "" : characterEncoding; //$NON-NLS-1$
    }
    
    /**
     * @param force Whether to force the get/update/creation of the file
     * @param shareable The shareable representing the file to be read/updated/created
     * @param progress
     * @throws FileSystemException
     */
    public FileOptions(boolean force, IShareableInternal shareable, IProgressMonitor progress) throws FileSystemException {
        super();
        
        this.force = force;

        SubMonitor monitor = SubMonitor.convert(progress, 2);
        if (shareable != null) {
            FileItemInfo info = shareable.getFileItemInfo(monitor.newChild(1));
            if (info != null) {
                this.encoding = info.getStoredEncoding();
                this.lineDelimiter = info.getLineDelimiter();
            } else {
                this.encoding = ""; //$NON-NLS-1$
                this.lineDelimiter = FileLineDelimiter.LINE_DELIMITER_NONE;
            }
            this.properties = shareable.getMetadataProperties(monitor.newChild(1)).getCurrentProperties();
        } else {
            this.encoding = ""; //$NON-NLS-1$
            this.lineDelimiter = FileLineDelimiter.LINE_DELIMITER_NONE;
            this.properties = Collections.EMPTY_MAP;
        }
    }

    public boolean force() {
        return force;
    }

    public String getCharacterEncoding() {
        return encoding;
    }

    public FileLineDelimiter getLineDelimiter() {
        return lineDelimiter;
    }

    public Map<String, String> getProperties() {
        return properties;
    }
}
