/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2006, 2014. All Rights Reserved. 
 * 
 * Note to U.S. Government Users Restricted Rights:  Use, 
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.filesystem.client.internal;

import java.io.InputStream;
import java.util.Collection;
import java.util.Locale;

import org.eclipse.core.filesystem.IFileStore;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.jobs.ISchedulingRule;

import com.ibm.team.filesystem.client.FileSystemException;
import com.ibm.team.filesystem.client.IRelativeLocation;
import com.ibm.team.filesystem.client.ISandbox;
import com.ibm.team.filesystem.client.ISharingManager.RuleKind;
import com.ibm.team.filesystem.client.ResourceType;
import com.ibm.team.filesystem.client.internal.api.storage.ContentMeta;
import com.ibm.team.filesystem.client.internal.api.storage.IBackupHandler;
import com.ibm.team.filesystem.client.internal.api.storage.IRetryRequest;
import com.ibm.team.filesystem.client.internal.exceptions.ContentRetrievalFailure;
import com.ibm.team.filesystem.client.operations.BackupDilemmaHandler;
import com.ibm.team.internal.repository.rcp.streams.DigestComputingInputStream;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.scm.common.ContentHash;

/**
 * Abstraction for accessing the file system
 * 
 * @remoteProjects This class is extended by the remote project integration. 
 *    Downstream teams may need to adopt changes made to it.
 */
public abstract class FileStorage implements INonEFS {

	// TODO need to exclude MVS platform 
    public static final boolean PLATFORM_SUPPORTS_EXEC_BIT = !System.getProperty("os.name") //$NON-NLS-1$
        .toLowerCase(Locale.ENGLISH).contains("windows"); //$NON-NLS-1$

    private final IFileStorage storage;
    
    public FileStorage(IFileStorage storage) {
        this.storage = storage;
        ((FileStorageWrapper) storage).setUnderlyingStorage(this);
    }
    
    protected final ISandbox getSandbox() {
        return storage.getShareable().getSandbox();
    }

    protected final IRelativeLocation getLocalPath() {
        return storage.getShareable().getLocalPath();
    }

    public final IFileStorage getStorage() {
        return storage;
    }
    
    public void lock(IProgressMonitor mon) {
    }
    
    public void unlock() {
    }
    
    /** 
     * For backing up Files, Folders and Symbolic links
     */
    public abstract void backup(IBackupHandler shed, BackupDilemmaHandler dilemmaHandler, IProgressMonitor monitor) throws FileSystemException;

    /**
     * Create a folder.  This method should not access metadata since it may be locked by the calling thread
     * @param createParent option directs the parents be created if necessary
     * @param additionalInfo Additional information that can be used to indicate how to create the folder.
     * @param monitor Monitor to mark progress on.
     */
    public abstract void create(boolean createParents, Object additionaInfo, IProgressMonitor progress) throws FileSystemException;
    
    /**
     * Deletes this storage object plus any children. The order in which the children are deleted
     *  (before the parent or after the parent) is implementation dependent.
     */
    public abstract void delete(IProgressMonitor monitor) throws FileSystemException;
    
    public abstract void deregisterRepositoryProvider(IProgressMonitor monitor) throws FileSystemException;
    
    public abstract String getActualName();

    /**
     * resolves the FileStorage that should be associated with the childStorage.
     * the child is a child of this storage with the name supplied.  The type is
     * just a hint. The overriding class is responsible for connecting
     * childStorage to the underlying resource.
     */
    public abstract void resolveChildStorage(FileStorageWrapper childStorage, String name, ResourceType type);

    /**
     * If this is a file or a symbolic link or does not exist, it should return empty collection as the children.
     * @param refreshAllowed In the event that we detect the resource is out of sync with whats on disk
     * when determining the children of it, whether or not a refresh can be done to determine what are
     * the children.  In the case of a Eclipse move, while in the middle of it before the tree is refeshed
     * it would not be good to update the children.  In other situations like updating the copy file area
     * it is a good idea to allow for a refresh.
     */
    public abstract Collection<IFileStorage> getChildren(boolean refreshAllowed, IProgressMonitor progress) throws FileSystemException;

    public abstract IRelativeLocation getIDEPath();

    public abstract long getLocalTimeStamp() throws FileSystemException;

    public abstract long getModificationStamp() throws FileSystemException;

    public abstract IFileStorage getParent();

    public abstract ISchedulingRule getResourceRuleForIDE(RuleKind kind);

    public abstract boolean isExecutable(IProgressMonitor monitor) throws FileSystemException;

    public abstract boolean isRepositoryProviderRegistered(IProgressMonitor monitor) throws FileSystemException;

    public abstract void move(FileStorage target, IProgressMonitor monitor) throws FileSystemException;
    
    public abstract void preserveHistory(IProgressMonitor monitor) throws FileSystemException;
    
    public abstract void refreshCachedSubTree(int depth, IProgressMonitor monitor) throws FileSystemException;
    
    public abstract void registerRepositorProvider(IProgressMonitor monitor) throws FileSystemException;
    
    public boolean requiresBackup(IProgressMonitor monitor) throws FileSystemException {
        return false;
    }

    public abstract boolean setExecutable(boolean exec, IProgressMonitor monitor);

    public abstract boolean storageExists(IProgressMonitor progress) throws FileSystemException;

    public abstract boolean supportsExecBit();
    
    public boolean supportsSymbolicLinks() {
        return false;
    }

    public abstract boolean contains(ISchedulingRule rule);
    
    public abstract boolean isConflicting(ISchedulingRule rule);

    public abstract Object getAdapter(Class adapter);

    public abstract boolean shouldBeIgnored(IProgressMonitor monitor);
    
    protected abstract IFileStore getFileStore();

    /**
     * make this storage writable or read only.
     * @param writable Whether the storage should be made writable.
     * @param progress Monitor to mark progress on.
     * @return <code>true</code> if the setting of the property succeeded.
     * <code>false</code> otherwise.
     */
    public abstract boolean setWritable(boolean writable, IProgressMonitor progress);

    
	// For non-EFS storage they must implement, otherwise we expect that if called its invalid
	public IFileStorage getChild(String name, IProgressMonitor newChild)
			throws FileSystemException {
		
		// Must be overwritten if the storage does not support EFS
		throw new IllegalStateException();
	}

	public ResourceType getResourceType(IProgressMonitor progress)
			throws FileSystemException {
		
		// Must be overwritten if the storage does not support EFS
		throw new IllegalStateException();
	}

	public long getSize(IProgressMonitor progress) throws FileSystemException {
		
		// Must be overwritten if the storage does not support EFS
		throw new IllegalStateException();
	}

	public boolean isFile(IProgressMonitor progress) throws FileSystemException {
		
		// Must be overwritten if the storage does not support EFS
		throw new IllegalStateException();
	}

	public boolean isFolder(IProgressMonitor progress)
			throws FileSystemException {
		
		// Must be overwritten if the storage does not support EFS
		throw new IllegalStateException();
	}

	public boolean isReadOnly(IProgressMonitor progress)
			throws FileSystemException {
		
		// Must be overwritten if the storage does not support EFS
		throw new IllegalStateException();
	}

	public boolean isSymbolicLink(IProgressMonitor monitor)
			throws FileSystemException {
		
		// Must be overwritten if the storage does not support EFS
		throw new IllegalStateException();
	}
	
	/**
	 * {@inheritDoc}
	 */
    public boolean isBrokenSymbolicLink(IProgressMonitor monitor) throws FileSystemException {
        throw new IllegalStateException();
    }
    
    /**
     * {@inheritDoc}
     */
    public boolean verifySymbolicLinkTargetType(boolean directory, IProgressMonitor monitor) throws FileSystemException {
        return true;
    }

    /**
     * {@inheritDoc}
     */
    public String convertIntoLocalForm(String target) {
        // Default is to simply return the target as-is regardless of whether there is symbolic link support
        //  here.
        return target;
    }
	
	/**
	 * {@inheritDoc}
	 */
	public void create(String targetPath, IProgressMonitor monitor) throws FileSystemException {
	    // Must be overwritten if the storage does not support EFS and provides a SymLinkHelper library
	    throw new IllegalStateException();
	}

	public void setSymLinkTarget(String targetPath, IProgressMonitor newChild)
			throws FileSystemException {
		
		// Must be overwritten if the storage does not support EFS
		throw new IllegalStateException();
	}

	public String getSymLinkTarget(IProgressMonitor monitor)
			throws FileSystemException {
		
		// Must be overwritten if the storage does not support EFS
		throw new IllegalStateException();
	}
	
    public void setLocalTimeStamp(long timeStamp, IProgressMonitor monitor) 
            throws FileSystemException {
        
        // Must be overwritten if the storage does not support EFS
        throw new IllegalStateException();
    }
    
    /**
     * Retrieve the contents of the file.  This method should not access metadata since it may be locked by the calling thread
     * @param fileOptions options which can affect how the contents are retrieved.
     * @param monitor Monitor to mark progress on.
     * @return input stream containing the contents
     * @throws ContentRetrievalFailure
     */
    public abstract InputStream getContents(IFileOptions fileOptions) throws ContentRetrievalFailure;

    /**
     * Create a file.  This method should not access metadata since it may be locked by the calling thread
     * @param fileOptions options which can affect how the file is created in the underlying storage.
     * @param contents The contents to be put in the file.
     * @param monitor Monitor to mark progress on.
     */
    public abstract void create(IFileOptions fileOptions, InputStream contents, IProgressMonitor progress) throws FileSystemException;
    
    /**
     * Set the contents of a File in the storage.  This should not access metadata since it may already be locked by the calling thread.
     * @param fileOptions Options which can affect how the file is created in the underlying storage.
     * @param inputStream The contents to be put in the file
     * @param monitor Monitor to mark progress on.
     */
    public abstract void setContents(IFileOptions fileOptions, InputStream inputStream, Shed backupShed, IProgressMonitor monitor) throws FileSystemException;
    
    /**
     * Determine how to handle an upload failure. 
     * 
     * @return <code>null</code> indicating that the default behaviour should 
     *          be used; or a retry handler that will be called in future
     *          to perform the retry (accessible files will be uploaded, the 
     *          retry handlers will be run, and then the commits will be 
     *          tried again).
     */
    public IRetryRequest recoverFromUploadFailure(Exception e, IProgressMonitor progress) {
        return null;
    }

    /**
     * Retrieve the contents of the file ready for metadata hash calculation.  This method should not access metadata since
     * it may be locked by the calling thread.  The idea is that its possible that there is transcoding
     * on an input stream might be done the storage implementation.  The inputStream supplied should NOT be wrapped by the
     * conversion is done by this storage implementation. 
     * 
     * @param fileOptions options which can affect how the contents are retrieved.
     * @param monitor Monitor to mark progress on.
     * @return input stream containing the contents
     * @throws ContentRetrievalFailure
     */
    public InputStream getLocalChangeDetectionStream(IFileOptions fileOptions) throws ContentRetrievalFailure {
        return getContents(fileOptions);
    }
    
    /**
     * Retrieve the contents of the file ready for storage in the repo given an input stream equivelant to 
     * {@link #getLocalChangeDetectionStream(IFileOptions)}.  This method should not access metadata since
     * it may be locked by the calling thread.  The idea is that its possible that there is transcoding
     * on an input stream might be done the storage implementation.  The inputStream supplied should be wrapped by the
     * conversion is done by this storage implementation.  This method would normally be declared static, but
     * it would be difficult to call the storage extension's override.
     * 
     * @param fileOptions options which can affect how the contents are retrieved.
     * @param inputStream The input stream that represents the Local Change detection stream.
     * @param monitor Monitor to mark progress on.
     * @return input stream containing the contents
     * @throws ContentRetrievalFailure
     */
    public InputStream getContents(IFileOptions fileOptions, InputStream inputStream) throws ContentRetrievalFailure {
        return inputStream;
    }

    public ContentMeta setContentsCalcMeta(IFileOptions optionsForWrite, InputStream in, Shed backupShed, IProgressMonitor progress) throws TeamRepositoryException {
        DigestComputingInputStream hashedIn = ContentHash.getDigestComputingStream(in);
        setContents(optionsForWrite, hashedIn, backupShed, progress);
        return new ContentMeta(ContentHash.valueOf(hashedIn.getFinalDigest()), hashedIn.getContentSize());
    }

    /**
     * Return the actual name of the item as it appears on disk. This method
     * may need to do some work to determine the actual name.
     * @param monitor a progress monitor
     * @return the actual name of the item as it appears on disk
     */
    public String getActualNameOnDisk(IProgressMonitor monitor) {
        // We need a default implementation since we have clients that
        // subclass this class
        return getActualName();
    }
}
