/*******************************************************************************
 * Licensed Materials - Property of IBM 
 * (c) Copyright IBM Corporation 2005, 2014. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights: Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 ******************************************************************************/
package com.ibm.team.filesystem.client.workitems;

import java.sql.Timestamp;
import java.util.List;

import org.eclipse.core.runtime.IProgressMonitor;

import com.ibm.team.filesystem.client.operations.IDeliverOperation;
import com.ibm.team.links.client.ILinkEvent;
import com.ibm.team.links.common.ILink;
import com.ibm.team.process.common.advice.IOperationReport;
import com.ibm.team.process.common.advice.TeamOperationCanceledException;
import com.ibm.team.repository.client.ITeamRepository;
import com.ibm.team.repository.common.IContributorHandle;
import com.ibm.team.repository.common.PermissionDeniedException;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.scm.client.IWorkspaceConnection;
import com.ibm.team.scm.common.IChangeSetHandle;
import com.ibm.team.scm.common.IWorkspaceHandle;
import com.ibm.team.scm.common.dto.IChangeHistorySyncReport;
import com.ibm.team.scm.common.dto.IComponentLockReport;
import com.ibm.team.workitem.common.model.IWorkItem;
import com.ibm.team.workitem.common.model.IWorkItemHandle;

/**
 * Methods for managing the features combining file system, scm, and work items. The 
 * filesystem workitem manager is closely tied to its {@link ITeamRepository}.
 * <p>
 * Team repository objects are intended to be used by multi-threaded clients.
 * Unless noted, <code>IFileSystemWorkItemManager</code>, can be safely
 * called from multiple threads concurrently without risk of corruption.
 * </p>
 * 
 * @since 0.6
 * @NoImplement This interface is not intended to be implemented by clients.
 */
public interface IFileSystemWorkItemManager {
    
    /**
     * Delivers from the source workspace to the target workspace the given outgoing 
     * change sets. In addition updates the provided work items with reviewers, comments
     * and resolution state. All this happens atomically on the server.
     * <p>
     * This operation is process enabled and runs as the 
     * <code>com.ibm.team.scm.client.deliver</code> operation thus all client side
     * advisors are allowed to participate.
     * </p><p>
     * All regular client side notifications are sent as if you would of saved the 
     * work items or delivered using their respective client libraries.
     * </p>
     *  
     * @param syncReport
     *            the change history sync report to use as the foundation for
     *            identifying the target workspace, and for verifying the change
     *            sets to be delivered as outgoing to that target
     * @param source
     *            the source workspace containing the change sets to deliver;
     *            never <code>null</code>
     * @param target
     *            the target workspace to be delivered to; never <code>null</code>
     * @param changeSets
     *            the outgoing change sets indentified in the sync report as outgoing;
     *            never <code>null</code>
     * @param locksToRelease
     *            a list to populate during the call with locks on versionables in the outgoing change sets;
     *            never <code>null</code>.
     * @param workItems
     *            the work items to update; never <code>null</code>
     * @param shouldResolve
     *            indicates if the work item should have it's state transitioned to resolved, 
     *            even if the work item is not resolve, it can be updated with reviewers and
     *            comments
     * @param comment
     *            the comment to append to the discussions in the provided work items or
     *            <code>null</code> if no comment is to be added.
     * @param addReview
     *            indicates if the work item should have a review/approval added with the 
     *            provided subject. A review can be left empty without reviewers.
     * @param subject
     *            the subject of the review
     * @param reviewers
     *            an optional list of reviewers to be added to the provided work items
     * @param dueDate
     *            an optional due date to be added to the provided work items
     * @param deliverOperation
     *            the client side deliver operation to be used to run the client 
     *            pre-condition checks.  
     * @param monitor
     *            a progress monitor, or <code>null</code> if progress
     *            reporting is not desired
     * @return the process operation report returned from the server
     * @throws TeamRepositoryException
     *             if the operation fails for any other reason
     * @throws TeamOperationCanceledException
     *             if the operation fails because of a process precondition isn't satisfied
     * @throws PermissionDeniedException
     *             if the operation fails because the user doesn't have permissions                        
     * @LongOp This is a long operation; it may block indefinitely; must not be
     *         called from a responsive thread.
     * @see {@link IWorkspaceConnection#deliver(IWorkspaceConnection, IChangeHistorySyncReport, List, List, IProgressMonitor)}
     * @since 0.7         
     */
	public IOperationReport deliverAndResolve(
			IChangeHistorySyncReport syncReport, 
			IWorkspaceConnection source, IWorkspaceConnection target, 
			List<IChangeSetHandle> changeSets,
			List<IComponentLockReport> locksToRelease,
			List<IWorkItem> workItems,
			boolean shouldResolve,
			String comment, 
			boolean addReview, 
			String subject, 
			List<IContributorHandle> reviewers,
			Timestamp dueDate,
			IDeliverOperation deliverOperation,
			IProgressMonitor monitor) throws TeamRepositoryException;
	
	/**
	 * Creates a link from a change set to a set of work items. An event for this
	 * creation will be sent via the {@link ILinkEvent#LINK_SAVED_EVENT_TYPE} event.
	 * 
	 * @param workspace optional workspace that contains the change set being linked.
	 * @param changeSet the change set being linked; never <code>null</code>
	 * @param workItems the work items being linked to; never <code>null</code>
     * @param monitor
     *            a progress monitor, or <code>null</code> if progress
     *            reporting is not desired
     * @return the links that were created
	 * @throws TeamRepositoryException
	 * 			if the operation fails for any reason
	 */
	public List<ILink> createLink(IWorkspaceHandle workspace, IChangeSetHandle changeSet,
			IWorkItemHandle[] workItems, IProgressMonitor monitor)
			throws TeamRepositoryException;
	
	/**
	 * Deletes a set of change set to work items links. An event for this
	 * deletion will be sent via the {@link ILinkEvent#LINK_DELETED_EVENT_TYPE} event.
	 * 
	 * @param links the links being deleted; never <code>null</code>
     * @param monitor
     *            a progress monitor, or <code>null</code> if progress
     *            reporting is not desired
     * @return the links that were created
	 * @throws TeamRepositoryException
	 * 			if the operation fails for any reason
	 */
	public void deleteLinks(List<ILink> links, IProgressMonitor monitor)
			throws TeamRepositoryException;
}
