/*******************************************************************************
 * Licensed Materials - Property of IBM 
 * (c) Copyright IBM Corporation 2005, 2014. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights: Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 ******************************************************************************/
package com.ibm.team.filesystem.client.workitems.internal;

import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.NullProgressMonitor;
import org.eclipse.core.runtime.SubMonitor;

import com.ibm.team.filesystem.client.FileSystemException;
import com.ibm.team.filesystem.client.operations.IDeliverOperation;
import com.ibm.team.filesystem.client.workitems.IFileSystemWorkItemManager;
import com.ibm.team.filesystem.common.workitems.internal.IFilesystemWorkItemService;
import com.ibm.team.filesystem.common.workitems.internal.dto.DeliverAndResolveDTO;
import com.ibm.team.filesystem.common.workitems.internal.dto.FilesystemWorkItemsDTOFactory;
import com.ibm.team.links.client.ILinkManager;
import com.ibm.team.links.common.ILink;
import com.ibm.team.process.common.advice.IObjectsResponse;
import com.ibm.team.process.common.advice.IOperationReport;
import com.ibm.team.repository.client.IItemManager;
import com.ibm.team.repository.client.ITeamRepository;
import com.ibm.team.repository.client.util.IClientLibraryContext;
import com.ibm.team.repository.common.IContributorHandle;
import com.ibm.team.repository.common.StaleDataException;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.scm.client.IWorkspaceConnection;
import com.ibm.team.scm.client.internal.ScmClientLibraryContext;
import com.ibm.team.scm.client.internal.WorkspaceConnection;
import com.ibm.team.scm.client.internal.isv.IWorkspaceDeliverRunnable;
import com.ibm.team.scm.client.internal.isv.IWorkspaceOperationRunner;
import com.ibm.team.scm.client.internal.process.ClientAdvisableOperation;
import com.ibm.team.scm.client.internal.process.ProcessUtils;
import com.ibm.team.scm.common.IBaselineHandle;
import com.ibm.team.scm.common.IChangeSetHandle;
import com.ibm.team.scm.common.IRepositoryProgressMonitor;
import com.ibm.team.scm.common.IScmProcessConfiguration;
import com.ibm.team.scm.common.IWorkspaceHandle;
import com.ibm.team.scm.common.dto.IChangeHistorySyncReport;
import com.ibm.team.scm.common.dto.IComponentLockReport;
import com.ibm.team.scm.common.process.DeliverOperationData;
import com.ibm.team.scm.common.process.RepositoryRunnable;
import com.ibm.team.scm.common.process.RunnableAdvisableOperation;
import com.ibm.team.workitem.client.IWorkItemClient;
import com.ibm.team.workitem.client.IWorkItemWorkingCopyManager;
import com.ibm.team.workitem.common.model.IWorkItem;
import com.ibm.team.workitem.common.model.IWorkItemHandle;
import com.ibm.team.workitem.common.model.WorkItemApprovals;

public class FileSystemWorkItemManager implements IFileSystemWorkItemManager {
	private final IFilesystemWorkItemService service;
	private final ITeamRepository repo;
	private ILinkManager linkManager;

    public FileSystemWorkItemManager(IClientLibraryContext context) {
        context = new ScmClientLibraryContext(context);
        this.repo = context.teamRepository();
		this.service = (IFilesystemWorkItemService) context.getServiceInterface(IFilesystemWorkItemService.class);
		this.linkManager = (ILinkManager)repo.getClientLibrary(ILinkManager.class);
    }

	@Override
    public IOperationReport deliverAndResolve(
    		final IChangeHistorySyncReport syncReport, final IWorkspaceConnection source, final IWorkspaceConnection target, 
			List<IChangeSetHandle> changeSets, final List<IComponentLockReport> toRelease, final List<IWorkItem> workItems, 
			boolean shouldResolve, final String comment, final boolean addReview, final String subject, final List<IContributorHandle> reviewers,
			final Timestamp dueDate, final IDeliverOperation deliverOperation, IProgressMonitor monitor) throws TeamRepositoryException {
    	
    	final DeliverAndResolveDTO data = FilesystemWorkItemsDTOFactory.eINSTANCE.createDeliverAndResolveDTO();
    	data.setShouldResolveWorkItem(shouldResolve);
    	
    	if(changeSets == null || changeSets.isEmpty())
    		throw new IllegalArgumentException();
    	
    	if(workItems == null || workItems.isEmpty())
    		throw new IllegalArgumentException();

    	if(toRelease == null)
    		throw new IllegalArgumentException();

    	// TODO: Move checkAgainstReport into somewhere else more public after it has been refactored
    	// so that it is not a public static method on the concrete WorkspaceConnection
    	final List<IChangeSetHandle> theChangeSets = WorkspaceConnection.checkAgainstReport(syncReport, false, changeSets, source.teamRepository());
    	
    	final IWorkItem[] workItemHandles = workItems.toArray(new IWorkItem[0]);
    	DeliverOperationData deliverData = new DeliverOperationData(
        		new ProcessUtils.ClientProcessProxy(repo),
        		source.getResolvedWorkspace(),
        		target.getResolvedWorkspace(),
        		Collections.EMPTY_LIST, changeSets);
    	SubMonitor progress = SubMonitor.convert(monitor, 100);
        RunnableAdvisableOperation operation = new ClientAdvisableOperation(
       			repo,
       			IScmProcessConfiguration.DELIVER_CLIENT_OPERATION_ID,
       			deliverData,
       			// It's really important that the deliver process use the context of
       			// the deliver target and not the source.
       			target.getResolvedWorkspace(), progress.newChild(1));
        operation.setIgnoredExceptions(false);
        operation.setBlock(new RepositoryRunnable() {
    		public Object run(IProgressMonitor monitor) throws TeamRepositoryException {
    			
    		    // check the deliver client-side pre-conditions
    			preDeliverConditions(deliverOperation, monitor);
    							
				// update the work items and convert to working copies that will be saved
				// on the server
    			final List<IWorkItem> workingCopies = new ArrayList<IWorkItem>(workItems.size());
				List<IWorkItem> updatedWorkItems = repo.itemManager().fetchCompleteItems(workItems, IItemManager.REFRESH, monitor); // need to ensure we have the latest, otherwise we will lose our edits
                List<IWorkItem> refreshedWorkItems = new ArrayList<IWorkItem>(workItems.size());
				for (int i = 0; i < updatedWorkItems.size(); i++) {
					IWorkItem workItem = updatedWorkItems.get(i);
					if(workItem == null)
					    refreshedWorkItems.add(workItems.get(i));
					else
					    refreshedWorkItems.add(workItem);
				}
				
				for (IWorkItem workItem : refreshedWorkItems) {
                	workingCopies.add(
                		WorkItemUtils.createReviewTask(repo, workItem, 
                			WorkItemApprovals.VERIFICATION_TYPE.getIdentifier(), subject, 
                			reviewers, addReview, comment, false, dueDate, monitor));
                }
    							
				// run the deliver operation via the workspace connection operation running such that
				// locking and notifications 
				
    	    	IOperationReport report = ((IWorkspaceOperationRunner)source).runDeliver(new IWorkspaceDeliverRunnable() {
    			
    				public IObjectsResponse deliver(IWorkspaceConnection workspace,
    						IWorkspaceConnection targetWorkspace, IBaselineHandle[] baseline,
    						IChangeSetHandle[] outgoingChangeSets,
    						IProgressMonitor monitor) throws TeamRepositoryException {
    					return service.deliverAndResolve((IWorkspaceHandle)workspace.getResolvedWorkspace(), 
    							(IWorkspaceHandle)targetWorkspace.getResolvedWorkspace(), 
    							outgoingChangeSets, workingCopies.toArray(new IWorkItem[0]), data,
    							IRepositoryProgressMonitor.ITEM_FACTORY.createItem(monitor));
    				}
    			
    			}, target, syncReport, Collections.EMPTY_LIST, theChangeSets, toRelease, monitor);
    	    	
    	    	// update the client caches so that listeners know that there are new
    	    	// work item states available.
    	    	IWorkItemWorkingCopyManager wcManager = ((IWorkItemClient) (repo).getClientLibrary(IWorkItemClient.class)).getWorkItemWorkingCopyManager();
    	    	for (IWorkItem workItem : workItemHandles) {
    	    	    try {
    	    	    	wcManager.connect(workItem, IWorkItem.FULL_PROFILE, monitor);
    	    	    	wcManager.refreshWithCurrent(workItem, monitor);
    	    	    } finally {
    	    	        wcManager.disconnect(workItem);
    	    	    }
    			}
    	    	return report;
    		}
    	});
       	return operation.processExec(progress.newChild(99));
    }
    
    private void preDeliverConditions(final IDeliverOperation deliverOperation, IProgressMonitor monitor) throws TeamRepositoryException {
		try {
			if(deliverOperation != null) {
				deliverOperation.runPreConditionsOnly(monitor);
			}
		} catch (FileSystemException e) {
			throw new TeamRepositoryException(e);
		}
	}
    
    public List<ILink> createLink(IWorkspaceHandle workspace, IChangeSetHandle changeSet,
            IWorkItemHandle[] workItems, IProgressMonitor monitor)
            throws TeamRepositoryException {
        
        // arg checking
        if(changeSet == null)
            throw new IllegalArgumentException();
        
        if(workItems == null || workItems.length == 0)
            throw new IllegalArgumentException();
        
        monitor = monitor == null ? new NullProgressMonitor() : monitor;
        try {
            monitor.worked(5);            
            ILink[] linkItems = service.createLink(workspace, changeSet, workItems, IRepositoryProgressMonitor.ITEM_FACTORY.createItem(monitor));
            List<ILink> links = new ArrayList<ILink>(linkItems.length);
            links.addAll(Arrays.asList(linkItems));
            monitor.worked(75);
            repo.itemManager().applyItemUpdates(links);
            linkManager.applySavedLinks(links);
            return links;
        } catch (StaleDataException sde) {
            throw sde;
        } finally {
            monitor.done();
        }
    }
    
    public void deleteLinks(List<ILink> links, IProgressMonitor monitor)
            throws TeamRepositoryException {
        
        // arg checking
        if(links == null || links.size()  == 0)
            throw new IllegalArgumentException();
        
        monitor = monitor == null ? new NullProgressMonitor() : monitor;
        try {
            monitor.worked(5);            
            service.deleteLinks(links.toArray(new ILink[0]), IRepositoryProgressMonitor.ITEM_FACTORY.createItem(monitor));
            monitor.worked(75);
            linkManager.applyDeletedLinks(links);
        } catch (StaleDataException sde) {
            throw sde;
        } finally {
            monitor.done();
        }
    }
}
