/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2011, 2012. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp. 
 *******************************************************************************/

package com.ibm.team.filesystem.client.workitems.internal;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.eclipse.core.runtime.IAdaptable;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.OperationCanceledException;
import org.eclipse.core.runtime.PlatformObject;
import org.eclipse.core.runtime.SubMonitor;

import com.ibm.team.foundation.rcp.core.IModelAccessor;
import com.ibm.team.links.client.ILinkManager;
import com.ibm.team.links.common.IItemReference;
import com.ibm.team.links.common.ILink;
import com.ibm.team.links.common.ILinkCollection;
import com.ibm.team.links.common.ILinkQueryPage;
import com.ibm.team.links.common.IReference;
import com.ibm.team.links.common.factory.IReferenceFactory;
import com.ibm.team.repository.client.IItemManager;
import com.ibm.team.repository.client.ITeamRepository;
import com.ibm.team.repository.common.IFetchResult;
import com.ibm.team.repository.common.IItemHandle;
import com.ibm.team.repository.common.ItemNotFoundException;
import com.ibm.team.repository.common.PermissionDeniedException;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.repository.common.UUID;
import com.ibm.team.scm.common.IChangeSet;
import com.ibm.team.scm.common.IChangeSetHandle;
import com.ibm.team.scm.common.links.ILinkConstants;
import com.ibm.team.workitem.common.model.IWorkItem;
import com.ibm.team.workitem.common.model.IWorkItemHandle;
import com.ibm.team.workitem.common.model.WorkItemLinkTypes;

/**
 * Provide the tree information displayed in the change sets section of the
 * "Locate Change Sets" editor
 * 
 * @since 0.9 (RTC 4.0)
 */
public class LocateChangeSetsUtil {

	public static final String PARENT_WORKITEM_LINKTYPE_ID = WorkItemLinkTypes.PARENT_WORK_ITEM;

	public static final int UNBOUNDED = -1;

	/**
	 * Given a selection of work items and change sets from a given repository,
	 * return an hierarchical representation of that selection.
	 * 
	 * Example 1: selection [Change Set 1], includeChildWorkItems
	 * <code>true</code> Result:
	 * 
	 * <pre>
	 * + Work Item 1
	 *  + Change Set 1 [x]
	 *  + Change Set 2
	 *  + Child Work Item 1
	 *    + Change Set 3
	 * </pre>
	 * 
	 * Note how Change Set 1 node will have the property isChecked to
	 * <code>true</code>.
	 * 
	 * Example 2: selection [Change Set 1], includeChildWorkItems
	 * <code>false</code> Result:
	 * 
	 * <pre>
	 * + Work Item 1
	 *  + Change Set 1 [x]
	 *  + Change Set 2
	 * </pre>
	 * 
	 * Example 3: selection [Work Item 1], includeChildWorkItems
	 * <code>true</code> Result:
	 * 
	 * <pre>
	 * + Work Item 1 [x]
	 *  + Change Set 1 [x]
	 *  + Change Set 2 [x]
	 *  + Child Work Item 1 [x]
	 *    + Change Set 3 [x]
	 * </pre>
	 * 
	 * Note how selected work item and its children are all checked.
	 * 
	 * Example 4: selection [Change Set 1] where Change Set 1 has no parent work
	 * item Result:
	 * 
	 * <pre>
	 * + Unassigned Work Item
	 *  + Change Set 1 [x]
	 * </pre>
	 * 
	 * Example 5: selection [change set 1] where change set 1 is associated to
	 * multiple work items Result:
	 * 
	 * <pre>
	 * + Work Item 1
	 *  + Change Set 1 [x]
	 * + Work Item 2
	 *  + Change Set 1 [x]
	 * </pre>
	 * 
	 * Note the same change set appears multiple times in the tree under
	 * different parents. UI can provide some visual cue when a change set shows
	 * up multiple times under different checked states if desired, but this
	 * isn't really a pb.
	 * 
	 * @see LcsEditorEntry for the base element type
	 * @see LcsEditorUnassignedWorkItemEntry for the parent type whose children
	 *      are change sets with no work item association
	 * @see LcsEditorWorkItemEntry for the parent type representing a work item
	 *      from a particular repository
	 * @see LcsEditorChangeSetEntry for the type representing a change set from
	 *      a particular repository
	 * 
	 * @param parms
	 *            The input parameters (ex: which change sets or work items to
	 *            search for).
	 * @param monitor
	 *            The monitor to use to track progress.
	 * @param monitor
	 * @return A results object providing result details (ex: the set of root
	 *         nodes).
	 * @throws TeamRepositoryException
	 */
	public static LcsEditorChangeSetsResult getLcsEditorChangeSets(
			ParmsGetLcsEditorChangeSets parms, IProgressMonitor monitor)
			throws TeamRepositoryException {
		
		SubMonitor progress = SubMonitor.convert(monitor, 100);
		
		if (parms == null) {
			throw new IllegalArgumentException();
		}
		if (parms.repo == null) {
			throw new IllegalArgumentException();
		}
		if ((parms.maxChildWorkItemsDepth < 0 && parms.maxChildWorkItemsDepth != UNBOUNDED)
				|| (parms.maxWorkItems < 0 && parms.maxWorkItems != UNBOUNDED)
				|| (parms.maxChangeSets < 0 && parms.maxChangeSets != UNBOUNDED)) {
			throw new IllegalArgumentException();
		}

		ITeamRepository repo = parms.repo;
		LcsEditorChangeSetsResult result = new LcsEditorChangeSetsResult();
		List<LcsEditorChangeSetEntry> changeSetRoots = Collections.EMPTY_LIST;
		if (parms.changeSets != null) {
			changeSetRoots = createChangeSetEntries(repo, parms.changeSets,
					true, progress.newChild(10));
		}

		progress.setWorkRemaining(90);
		if (progress.isCanceled()) {
			throw new OperationCanceledException();
		}
		
		List<LcsEditorWorkItemEntry> workItemRoots1 = Collections.EMPTY_LIST;
		if (parms.workItems != null) {
			workItemRoots1 = createWorkItemEntries(repo, parms.workItems, true,
					progress.newChild(10));
		}
		
		progress.setWorkRemaining(80);
		if (progress.isCanceled()) {
			throw new OperationCanceledException();
		}
		
		ChangeSetsParentWorkItemsResult parentsResult = fetchChangeSetsParentWorkItems(
				repo, changeSetRoots, progress.newChild(10));
		result.unassignedRoot = parentsResult.unassignedParent;
		List<LcsEditorWorkItemEntry> workItemRoots2 = parentsResult.parents;

		// The top work items come from the selected work items and the parent
		// work items of the selected change sets
		List<LcsEditorWorkItemEntry> workItemRoots = mergeWorkItemEntries(
				workItemRoots2, workItemRoots1);
		result.args = parms;
		result.workItemRoots = workItemRoots;
		int depth = 0;
		List<LcsEditorWorkItemEntry> parents = workItemRoots;
		int workItemsCount = parents.size();
		int changeSetsCount = 0;
		
		SubMonitor loopProgress = progress.newChild(70).setWorkRemaining(parents.size() * 2);
		while (parents.size() > 0) {
			if (loopProgress.isCanceled()) {
				throw new OperationCanceledException();
			}
			
			List<LcsEditorChangeSetEntry> csEntries = fetchWorkItemChangeSets(
					repo, parents, loopProgress.newChild(1));
			if (depth == 0 && parms.changeSets != null) {
				setChecked(csEntries, parms.changeSets);
			}
			changeSetsCount += csEntries.size();
			if (parms.maxChangeSets != UNBOUNDED && changeSetsCount > parms.maxChangeSets) {
				result.maxChangeSetsTooSmall = true;
				break;
			}
			if (parms.maxChildWorkItemsDepth == UNBOUNDED
					|| depth < parms.maxChildWorkItemsDepth) {
				parents = fetchWorkItemChildren(repo, parents, loopProgress.newChild(1));
				workItemsCount += parents.size();
				if (parms.maxWorkItems != UNBOUNDED && workItemsCount > parms.maxWorkItems) {
					result.maxWorkItemsTooSmall = true;
					break;
				}
			} else {
				parents = new ArrayList<LcsEditorWorkItemEntry>();
			}
			if (depth == parms.maxChildWorkItemsDepth || (parms.maxChildWorkItemsDepth == UNBOUNDED && parents.size() == 0)) {
				if (parents.size() > 0) {
					result.maxChildWorkItemsDepthTooSmall = true;
				}
				break;
			}
			depth++;
		}
		loopProgress.done();
		progress.done();
		return result;
	}

	// **************************************************************************************************
	// Class: ParmsGetLcsEditorChangeSets
	// **************************************************************************************************
	/**
	 * @see LocateChangeSetsUtil#getLcsEditorChangeSets(ParmsGetLcsEditorChangeSets,
	 *      IProgressMonitor)
	 */
	public static class ParmsGetLcsEditorChangeSets {
		/**
		 * The repo the selection belongs to. If an actual user selection
		 * crosses repositories, it is up to the caller to break it up into
		 * separate calls.
		 */
		public ITeamRepository repo;

		/**
		 * Max requested nested calculation of child work items. 0: no child
		 * work items returned 1: direct child work items returned etc...
		 * 
		 * @see LcsEditorChangeSetsResult#maxChildWorkItemsDepthTooSmall
		 */
		public int maxChildWorkItemsDepth;

		/**
		 * Max requested work items.
		 * 
		 * @see LcsEditorChangeSetsResult#maxWorkItemsTooSmall
		 */
		public int maxWorkItems;

		/**
		 * Max requested change sets.
		 * 
		 * @see LcsEditorChangeSetsResult#maxChangeSetsTooSmall
		 */
		public int maxChangeSets;

		/**
		 * User initial selection of work items. They MUST belong to the given
		 * {@link #repo}. Can be <cude>null</code>.
		 */
		public List<IWorkItemHandle> workItems;

		/**
		 * User initial selection of change sets. They MUST belong to the given
		 * {@link #repo}; Can be <code>null</code>
		 */
		public List<IChangeSetHandle> changeSets;
	}

	// ***************************************************************************************************************************************
	// Class: LcsEditorChangeSetsResult
	// ***************************************************************************************************************************************
	/**
	 * Represents the hierarchical tree of work items and change sets.
	 */
	public static class LcsEditorChangeSetsResult {

		private Collection<LcsEditorWorkItemEntry> workItemRoots;
		private LcsEditorUnassignedWorkItemEntry unassignedRoot;
		private boolean maxChildWorkItemsDepthTooSmall;
		private boolean maxWorkItemsTooSmall;
		private boolean maxChangeSetsTooSmall;

		private ParmsGetLcsEditorChangeSets args;

		/**
		 * Gets the top parent work item nodes. Never <code>null</code>.
		 */
		public Collection<LcsEditorWorkItemEntry> getRoots() {
			return workItemRoots;
		}

		/**
		 * Sets the top parent work item nodes.
		 */
		public void setRoots(Collection<LcsEditorWorkItemEntry> workItemRoots) {
			if (workItemRoots == null) {
				this.workItemRoots = Collections.EMPTY_LIST;
			} else {
				this.workItemRoots = workItemRoots;
			}
		}

		/**
		 * The top parent node that represents change sets with no known work
		 * item association. May be <code>null</code>.
		 */
		public LcsEditorUnassignedWorkItemEntry getUnassignedRoot() {
			return unassignedRoot;
		}

		/**
		 * The top parent node that represents change sets with no known work
		 * item association. May be <code>null</code>.
		 */
		public void setUnassignedRoot(
				LcsEditorUnassignedWorkItemEntry unassignedRoot) {
			this.unassignedRoot = unassignedRoot;
		}

		/**
		 * <code>true</code> if
		 * {@link ParmsGetLcsEditorChangeSets#maxChildWorkItemsDepth} is too
		 * small on the given selection. <code>false</code> otherwise.
		 * 
		 * Caller can provide a visual cue to the user to inform they are not
		 * looking at the whole data set. User can decide to see more (call
		 * again with a larger value).
		 */
		public boolean isMaxChildWorkItemsDepthTooSmall() {
			return maxChildWorkItemsDepthTooSmall;
		}

		/**
		 * <code>true</code> if
		 * {@link ParmsGetLcsEditorChangeSets#maxChildWorkItemsDepth} is too
		 * small on the given selection. <code>false</code> otherwise.
		 * 
		 * Caller can provide a visual cue to the user to inform they are not
		 * looking at the whole data set. User can decide to see more (call
		 * again with a larger value).
		 */
		public void setMaxChildWorkItemsDepthTooSmall(
				boolean maxChildWorkItemsDepthTooSmall) {
			this.maxChildWorkItemsDepthTooSmall = maxChildWorkItemsDepthTooSmall;
		}

		/**
		 * <code>true</code> if {@link ParmsGetLcsEditorChangeSets#maxWorkItems}
		 * is too small, aka if the user will be missing some work items in the
		 * results.
		 */
		public boolean isMaxWorkItemsTooSmall() {
			return maxWorkItemsTooSmall;
		}

		/**
		 * <code>true</code> if {@link ParmsGetLcsEditorChangeSets#maxWorkItems}
		 * is too small, aka if the user will be missing some work items in the
		 * results.
		 */
		public void setMaxWorkItemsTooSmall(boolean maxWorkItemsTooSmall) {
			this.maxWorkItemsTooSmall = maxWorkItemsTooSmall;
		}

		/**
		 * <code>true</code> if
		 * {@link ParmsGetLcsEditorChangeSets#maxChangeSets} is too small, aka
		 * if the user will be missing some change sets in the results.
		 */
		public boolean isMaxChangeSetsTooSmall() {
			return maxChangeSetsTooSmall;
		}

		/**
		 * <code>true</code> if
		 * {@link ParmsGetLcsEditorChangeSets#maxChangeSets} is too small, aka
		 * if the user will be missing some change sets in the results.
		 */
		public void setMaxChangeSetsTooSmall(boolean maxChangeSetsTooSmall) {
			this.maxChangeSetsTooSmall = maxChangeSetsTooSmall;
		}

		/**
		 * The initial arguments This is provided as a convenience to the caller
		 * - e.g. if they wish to refresh the roots.
		 */
		public ParmsGetLcsEditorChangeSets getArgs() {
			return args;
		}

		/**
		 * The initial arguments This is provided as a convenience to the caller
		 * - e.g. if they wish to refresh the roots.
		 */
		public void setArgs(ParmsGetLcsEditorChangeSets args) {
			this.args = args;
		}
	}

	// **************************************************************************************************
	// Class: ChangeSetsParentWorkItemsResult
	// **************************************************************************************************
	public static class ChangeSetsParentWorkItemsResult {
		/**
		 * Work items the change sets are associated with. Never
		 * <code>null</code>
		 */
		List<LcsEditorWorkItemEntry> parents;

		/**
		 * Parent for change sets that have no work item association.
		 * <code>null</code> if each change set has a parent work item.
		 */
		LcsEditorUnassignedWorkItemEntry unassignedParent;
	}

	/**
	 * Retrieve the direct children of the given elements. Children and parent
	 * nodes correctly point to each other.
	 */
	private static List<LcsEditorWorkItemEntry> fetchWorkItemChildren(
			ITeamRepository repo, List<LcsEditorWorkItemEntry> parents,
			IProgressMonitor monitor) throws TeamRepositoryException {
		
		SubMonitor progress = SubMonitor.convert(monitor, 100);
		
		ILinkManager linkManager = (ILinkManager) repo
				.getClientLibrary(ILinkManager.class);
		List<LcsEditorWorkItemEntry> result = new ArrayList<LcsEditorWorkItemEntry>();
		final int count = parents.size();
		IReference[] workItemReferences = new IReference[count];
		int i = 0;
		Map<UUID, LcsEditorWorkItemEntry> map = new HashMap<UUID, LcsEditorWorkItemEntry>();
		for (LcsEditorWorkItemEntry entry : parents) {
			IWorkItemHandle wi = entry.fWorkItem;
			IItemReference ref = IReferenceFactory.INSTANCE
					.createReferenceToItem(wi);
			workItemReferences[i] = ref;
			map.put(wi.getItemId(), entry);
			i++;
		}
		if (workItemReferences.length > 0) {
			ILinkQueryPage queryPage = linkManager.findLinksByTarget(
					new String[] { PARENT_WORKITEM_LINKTYPE_ID },
					workItemReferences, progress.newChild(50));
			ILinkCollection linkCollection = queryPage.getAllLinksFromHereOn();
			
			SubMonitor loopProgress = progress.newChild(50).setWorkRemaining(linkCollection.size());
			for (ILink link : linkCollection) {
				if (loopProgress.isCanceled()) {
					throw new OperationCanceledException();
				}
				
				IReference ref = link.getSourceRef();
				if (!ref.isItemReference())
					continue;
				IItemReference itemReference = (IItemReference) ref;
				IItemHandle itemHandle = itemReference.getReferencedItem();
				if (!(itemHandle instanceof IWorkItemHandle))
					continue;
				IWorkItemHandle wiSourceHandle = (IWorkItemHandle) itemHandle;
				ref = link.getTargetRef();
				if (!ref.isItemReference())
					continue;
				itemReference = (IItemReference) ref;
				itemHandle = itemReference.getReferencedItem();
				if (!(itemHandle instanceof IWorkItemHandle))
					continue;
				IWorkItemHandle wiTargetHandle = (IWorkItemHandle) itemHandle;
				IWorkItem wiSource = (IWorkItem) repo.itemManager()
						.fetchCompleteItem(wiSourceHandle,
								IItemManager.DEFAULT, loopProgress.newChild(1));
				LcsEditorWorkItemEntry wiParentEntry = map.get(wiTargetHandle
						.getItemId());
				if (wiParentEntry == null) {
					continue;
				}
				LcsEditorWorkItemEntry wiChildEntry = new LcsEditorWorkItemEntry(
						repo, wiSource, wiParentEntry,
						new ArrayList<LcsEditorEntry>(),
						wiParentEntry.isChecked());
				wiParentEntry.getChildren().add(wiChildEntry);
				result.add(wiChildEntry);
			}
			loopProgress.done();
		}
		progress.done();
		return result;
	}

	/**
	 * Retrieve the change sets associated to the parent work items
	 */
	private static List<LcsEditorChangeSetEntry> fetchWorkItemChangeSets(
			ITeamRepository repo, List<LcsEditorWorkItemEntry> parents,
			IProgressMonitor monitor) throws TeamRepositoryException {
		
		SubMonitor progress = SubMonitor.convert(monitor, 100);
		
		ILinkManager linkManager = (ILinkManager) repo
				.getClientLibrary(ILinkManager.class);
		List<LcsEditorChangeSetEntry> result = new ArrayList<LcsEditorChangeSetEntry>();
		final int count = parents.size();
		IReference[] workItemReferences = new IReference[count];
		int i = 0;
		Map<UUID, LcsEditorWorkItemEntry> map = new HashMap<UUID, LcsEditorWorkItemEntry>();
		for (LcsEditorWorkItemEntry entry : parents) {
			IWorkItemHandle wi = entry.fWorkItem;
			IItemReference ref = IReferenceFactory.INSTANCE
					.createReferenceToItem(wi);
			workItemReferences[i] = ref;
			map.put(wi.getItemId(), entry);
			i++;
		}
		if (workItemReferences.length > 0) {
			ILinkQueryPage queryPage = linkManager
					.findLinksByTarget(
							new String[] { ILinkConstants.CHANGESET_WORKITEM_LINKTYPE_ID },
							workItemReferences, progress.newChild(50));
			ILinkCollection linkCollection = queryPage.getAllLinksFromHereOn();
			
			SubMonitor loopProgress = progress.newChild(50).setWorkRemaining(linkCollection.size());
			for (ILink link : linkCollection) {
				IReference ref = link.getSourceRef();
				if (!ref.isItemReference())
					continue;
				IItemReference itemReference = (IItemReference) ref;
				IItemHandle itemHandle = itemReference.getReferencedItem();
				if (!(itemHandle instanceof IChangeSetHandle))
					continue;
				IChangeSetHandle csHandle = (IChangeSetHandle) itemHandle;
				ref = link.getTargetRef();
				if (!ref.isItemReference())
					continue;
				itemReference = (IItemReference) ref;
				itemHandle = itemReference.getReferencedItem();
				if (!(itemHandle instanceof IWorkItemHandle))
					continue;
				IWorkItemHandle wiTargetHandle = (IWorkItemHandle) itemHandle;
				IChangeSet csSource = null;
				try {
					csSource = (IChangeSet) repo.itemManager()
							.fetchCompleteItem(csHandle, IItemManager.DEFAULT,
									loopProgress.newChild(1));
				} catch (PermissionDeniedException e) {
					// This should not be an error case
					// We just wont include the change set
					continue;
				} catch (ItemNotFoundException e) {
					// This should not be an error case
					// We just wont include the change set
					continue;
				}
				LcsEditorWorkItemEntry wiParentEntry = map.get(wiTargetHandle
						.getItemId());
				if (wiParentEntry == null) {
					continue;
				}
				LcsEditorChangeSetEntry csChildEntry = new LcsEditorChangeSetEntry(
						repo, csSource, wiParentEntry, Collections.EMPTY_LIST,
						wiParentEntry.isChecked());
				wiParentEntry.getChildren().add(csChildEntry);
				result.add(csChildEntry);
			}
			loopProgress.done();
		}
		progress.done();
		return result;
	}

	/**
	 * Retrieve the work items the change sets are associated with.
	 */
	private static ChangeSetsParentWorkItemsResult fetchChangeSetsParentWorkItems(
			ITeamRepository repo, List<LcsEditorChangeSetEntry> entries,
			IProgressMonitor monitor) throws TeamRepositoryException {
		
		SubMonitor progress = SubMonitor.convert(monitor, 100);
		
		ILinkManager linkManager = (ILinkManager) repo
				.getClientLibrary(ILinkManager.class);
		ChangeSetsParentWorkItemsResult result = new ChangeSetsParentWorkItemsResult();
		result.parents = new ArrayList<LcsEditorWorkItemEntry>();

		LcsEditorUnassignedWorkItemEntry unassignedParent = new LcsEditorUnassignedWorkItemEntry();
		unassignedParent.setChildren(new ArrayList<LcsEditorEntry>());
		unassignedParent.setChecked(false);
		unassignedParent.setParent(null);
		unassignedParent.setRepository(repo);
		final int count = entries.size();
		IReference[] changeSetReferences = new IReference[count];
		int i = 0;
		Map<UUID, LcsEditorChangeSetEntry> map = new HashMap<UUID, LcsEditorChangeSetEntry>();
		for (LcsEditorChangeSetEntry entry : entries) {
			IChangeSetHandle cs = entry.fChangeSet;
			IItemReference ref = IReferenceFactory.INSTANCE
					.createReferenceToItem(cs);
			changeSetReferences[i] = ref;
			map.put(cs.getItemId(), entry);
			i++;
		}
		Set<UUID> changeSetsWithParents = new HashSet<UUID>();
		if (changeSetReferences.length > 0) {
			ILinkQueryPage queryPage = linkManager
					.findLinksBySource(
							new String[] { ILinkConstants.CHANGESET_WORKITEM_LINKTYPE_ID },
							changeSetReferences, progress.newChild(50));
			ILinkCollection linkCollection = queryPage.getAllLinksFromHereOn();
			
			SubMonitor loopProgress = progress.newChild(50).setWorkRemaining(linkCollection.size()); 
			for (ILink link : linkCollection) {
				if (loopProgress.isCanceled()) {
					throw new OperationCanceledException();
				}
				
				IReference ref = link.getSourceRef();
				if (!ref.isItemReference())
					continue;
				IItemReference itemReference = (IItemReference) ref;
				IItemHandle itemHandle = itemReference.getReferencedItem();
				if (!(itemHandle instanceof IChangeSetHandle))
					continue;
				IChangeSetHandle csHandle = (IChangeSetHandle) itemHandle;
				ref = link.getTargetRef();
				if (!ref.isItemReference())
					continue;
				itemReference = (IItemReference) ref;
				itemHandle = itemReference.getReferencedItem();
				if (!(itemHandle instanceof IWorkItemHandle))
					continue;
				IWorkItem wi = (IWorkItem) repo.itemManager()
						.fetchCompleteItem(itemHandle, IItemManager.DEFAULT,
								loopProgress.newChild(1));
				LcsEditorChangeSetEntry csEntry = map.get(csHandle.getItemId());
				if (csEntry == null) {
					continue;
				}
				/**
				 * Note. Some work items may show up multiple times here. We
				 * should use a map to avoid duplicates.
				 */
				LcsEditorWorkItemEntry wiParentEntry = new LcsEditorWorkItemEntry(
						repo, wi, null, new ArrayList<LcsEditorEntry>(), false);
				result.parents.add(wiParentEntry);
				changeSetsWithParents.add(csHandle.getItemId());
			}
			loopProgress.done();
		}
		progress.done();
		
		/**
		 * Add each change set that doesn't belong to a work item
		 */
		for (LcsEditorChangeSetEntry entry : entries) {
			if (!changeSetsWithParents.contains(entry.fChangeSet.getItemId())) {
				unassignedParent.getChildren().add(entry);
				entry.setParent(unassignedParent);
			}
		}
		if (unassignedParent.getChildren().size() > 0)
			result.unassignedParent = unassignedParent;
		return result;
	}

	private static List<LcsEditorWorkItemEntry> createWorkItemEntries(
			ITeamRepository repo, List<IWorkItemHandle> workItems,
			boolean isChecked, IProgressMonitor monitor)
			throws TeamRepositoryException {

		List<LcsEditorWorkItemEntry> result = new ArrayList<LcsEditorWorkItemEntry>();

		IFetchResult fetchResult = repo.itemManager()
				.fetchCompleteItemsPermissionAware(workItems,
						IItemManager.DEFAULT, monitor);

		if (fetchResult != null && fetchResult.getRetrievedItems() != null) {
			for (Object item : fetchResult.getRetrievedItems()) {
				if (item instanceof IWorkItem) {
					IWorkItem wi = (IWorkItem) item;
					result.add(new LcsEditorWorkItemEntry(repo, wi, null,
							new ArrayList<LcsEditorEntry>(), isChecked));
				}
			}
		}
		return result;
	}

	private static List<LcsEditorChangeSetEntry> createChangeSetEntries(
			ITeamRepository repo, List<IChangeSetHandle> changeSets,
			boolean isChecked, IProgressMonitor monitor)
			throws TeamRepositoryException {

		List<LcsEditorChangeSetEntry> result = new ArrayList<LcsEditorChangeSetEntry>();

		IFetchResult fetchResult = repo.itemManager()
				.fetchCompleteItemsPermissionAware(changeSets,
						IItemManager.DEFAULT, monitor);

		if (fetchResult != null && fetchResult.getRetrievedItems() != null) {
			for (Object item : fetchResult.getRetrievedItems())
				if (item instanceof IChangeSet) {
					IChangeSet cs = (IChangeSet) item;
					result.add(new LcsEditorChangeSetEntry(repo, cs, null,
							Collections.<LcsEditorEntry> emptyList(), isChecked));
				}
		}
		return result;
	}

	private static Map<UUID, LcsEditorWorkItemEntry> createMap(
			List<LcsEditorWorkItemEntry> entries) {
		HashMap<UUID, LcsEditorWorkItemEntry> result = new HashMap<UUID, LcsEditorWorkItemEntry>();
		for (LcsEditorWorkItemEntry entry : entries)
			result.put(entry.fWorkItem.getItemId(), entry);
		return result;
	}

	/**
	 * Returns a list that contains all entries from main plus all entries from
	 * additions that don't have the same item id as entries in main.
	 */
	private static List<LcsEditorWorkItemEntry> mergeWorkItemEntries(
			List<LcsEditorWorkItemEntry> main,
			List<LcsEditorWorkItemEntry> additions) {
		Map<UUID, LcsEditorWorkItemEntry> workItemsMap = createMap(main);
		for (LcsEditorWorkItemEntry entry : additions) {
			if (!workItemsMap.containsKey(entry.fWorkItem.getItemId())) {
				workItemsMap.put(entry.fWorkItem.getItemId(), entry);
			}
		}
		return new ArrayList<LcsEditorWorkItemEntry>(workItemsMap.values());
	}

	/**
	 * Set the isChecked attribute if the given entry matches a given change set
	 * handle. If the isChecked attribute is already set, it is left as is.
	 * 
	 * @see LcsEditorEntry#fIsChecked
	 */
	private static void setChecked(List<LcsEditorChangeSetEntry> entries,
			List<IChangeSetHandle> changeSets) {
		Set<UUID> uuids = new HashSet<UUID>();
		for (IChangeSetHandle cs : changeSets)
			uuids.add(cs.getItemId());
		for (LcsEditorChangeSetEntry entry : entries) {
			if (uuids.contains(entry.fChangeSet.getItemId()))
				entry.setChecked(true);
		}
	}

	// ***************************************************************************************************************************************
	// Class: LcsEditorEntry
	// ***************************************************************************************************************************************
	/**
	 * Base element
	 */
	public static abstract class LcsEditorEntry extends PlatformObject
			implements IModelAccessor, IAdaptable {

		volatile private ITeamRepository fRepo;
		volatile private LcsEditorEntry fParent;
		volatile private Collection<LcsEditorEntry> fChildren;
		volatile private boolean fIsChecked;

		/**
		 * Gets the repository this element belongs to.
		 */
		public ITeamRepository getRepository() {
			return fRepo;
		}

		/**
		 * Sets the repository this element belongs to.
		 */
		public void setRepository(ITeamRepository repo) {
			this.fRepo = repo;
		}

		/**
		 * Gets the parent element. <code>null</code> if this node is top level.
		 */
		public LcsEditorEntry getParent() {
			return fParent;
		}

		/**
		 * Sets the parent element.
		 */
		public void setParent(LcsEditorEntry parent) {
			this.fParent = parent;
		}

		/**
		 * Gets the child elements. Never <code>null</code>.
		 */
		public Collection<LcsEditorEntry> getChildren() {
			return fChildren;
		}

		/**
		 * Sets the child elements.
		 */
		public void setChildren(Collection<LcsEditorEntry> children) {
			if (children == null) {
				this.fChildren = Collections.emptyList();
			} else {
				this.fChildren = children;
			}
		}

		/**
		 * <code>true</code> if this element is included in the user's selection
		 * (directly or because one of its parents is in the user's selection).
		 */
		public boolean isChecked() {
			return fIsChecked;
		}

		public void setChecked(boolean isChecked) {
			this.fIsChecked = isChecked;
		}

		/**
		 * Checks if the given object is semantically equivalent to the this
		 * object.
		 */
		public abstract boolean isEquivalentTo(Object other);
	}

	// ***************************************************************************************************************************************
	// Class: LcsEditorWorkItemEntry
	// ***************************************************************************************************************************************
	/**
	 * Represents a work item.
	 * 
	 * Children are either {@link LcsEditorChangeSetEntry} or
	 * {@link LcsEditorWorkItemEntry}'s instances. Parent is either
	 * {@link LcsEditorWorkItemEntry} or <code>null</code> if this is part of
	 * the roots.
	 */
	public static class LcsEditorWorkItemEntry extends LcsEditorEntry {
		private IWorkItem fWorkItem;

		public LcsEditorWorkItemEntry(ITeamRepository repo, IWorkItem workItem,
				LcsEditorEntry parent, Collection<LcsEditorEntry> children,
				boolean isChecked) {
			this.fWorkItem = workItem;
			setRepository(repo);
			setParent(parent);
			setChildren(children);
			setChecked(isChecked);
		}

		public IWorkItem getWorkItem() {
			return fWorkItem;
		}

		public Object getModelInstance() {
			return fWorkItem;
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see
		 * org.eclipse.core.runtime.PlatformObject#getAdapter(java.lang.Class)
		 */
		public Object getAdapter(Class adapter) {
			if (adapter == IWorkItem.class) {
				return fWorkItem;
			} else if (adapter == IWorkItemHandle.class) {
				return (IWorkItemHandle) fWorkItem.getItemHandle();
			} else if (adapter == IItemHandle.class) {
				return fWorkItem.getItemHandle();
			} else {
				return null;
			}
		}

		/**
		 * Checks if the given object is semantically equivalent to the this
		 * object. This is <code>true</code> if the other object is:<br/>
		 * <ol>
		 * <li>Is of the same class</li>
		 * <li>Has the same repository</li>
		 * <li>Has the same work item handle</li>
		 * <li>Has the same parent</li>
		 * </ol>
		 */
		public boolean isEquivalentTo(Object other) {
			boolean equivalent = true;

			if (!(other instanceof LcsEditorWorkItemEntry)) {
				equivalent = false;
			} else {
				LcsEditorWorkItemEntry otherWorkItem = (LcsEditorWorkItemEntry) other;
				if (!otherWorkItem.getRepository().equals(getRepository())) {
					equivalent = false;
				}
				if (!otherWorkItem.fWorkItem.equals(fWorkItem)) {
					equivalent = false;
				}
				if ((otherWorkItem.getParent() == null && getParent() != null)
						|| (otherWorkItem.getParent() != null && getParent() == null)) {
					equivalent = false;
				}
				if (otherWorkItem.getParent() != null && getParent() != null
						&& !otherWorkItem.getParent().equals(getParent())) {
					equivalent = false;
				}
			}
			return equivalent;
		}
	}

	// ***************************************************************************************************************************************
	// Class: LcsEditorChangeSetEntry
	// ***************************************************************************************************************************************
	/**
	 * Represents a change set.
	 * 
	 * Has no children. Parent is either a {@link LcsEditorWorkItemEntry} or a
	 * {@link LcsEditorUnassignedWorkItemEntry}.
	 */
	public static class LcsEditorChangeSetEntry extends LcsEditorEntry {
		private IChangeSet fChangeSet;

		public LcsEditorChangeSetEntry(ITeamRepository repo,
				IChangeSet changeSet, LcsEditorEntry parent,
				Collection<LcsEditorEntry> children, boolean isChecked) {
			this.fChangeSet = changeSet;
			setRepository(repo);
			setParent(parent);
			setChildren(children);
			setChecked(isChecked);
		}

		public IChangeSet getChangeSet() {
			return fChangeSet;
		}

		public Object getModelInstance() {
			return fChangeSet;
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see
		 * org.eclipse.core.runtime.PlatformObject#getAdapter(java.lang.Class)
		 */
		public Object getAdapter(Class adapter) {
			if (adapter == IChangeSet.class) {
				return fChangeSet;
			} else if (adapter == IItemHandle.class) {
				return fChangeSet;
			} else {
				return super.getAdapter(adapter);
			}
		}

		/**
		 * Checks if the given object is semantically equivalent to the this
		 * object. This is <code>true</code> if the other object is:<br/>
		 * <ol>
		 * <li>Is of the same class</li>
		 * <li>Has the same repository</li>
		 * <li>Has the same change set handle</li>
		 * <li>Has the same parent</li>
		 * </ol>
		 */
		public boolean isEquivalentTo(Object other) {
			boolean equivalent = true;

			if (!(other instanceof LcsEditorChangeSetEntry)) {
				equivalent = false;
			} else {
				LcsEditorChangeSetEntry otherChangeSet = (LcsEditorChangeSetEntry) other;
				if (!otherChangeSet.getRepository().equals(getRepository())) {
					equivalent = false;
				}
				if (!otherChangeSet.fChangeSet.equals(fChangeSet)) {
					equivalent = false;
				}
				if ((otherChangeSet.getParent() == null && getParent() != null)
						|| (otherChangeSet.getParent() != null && getParent() == null)) {
					equivalent = false;
				}
				if (otherChangeSet.getParent() != null && getParent() != null
						&& !otherChangeSet.getParent().equals(getParent())) {
					equivalent = false;
				}
			}
			return equivalent;
		}
	}

	// ***************************************************************************************************************************************
	// Class: LcsEditorUnassignedWorkItemEntry
	// ***************************************************************************************************************************************
	/**
	 * Parent node for change sets that have no work item association.
	 * 
	 * Children are {@link LcsEditorChangeSetEntry}'s instances. They all belong
	 * to the same repo represented by this node. Parent is <code>null</code>.
	 */
	public static class LcsEditorUnassignedWorkItemEntry extends LcsEditorEntry {

		/**
		 * Checks if the given object is semantically equivalent to the this
		 * object. This is <code>true</code> if the other object is:<br/>
		 * <ol>
		 * <li>Is of the same class</li>
		 * <li>Has the same repository</li>
		 * <li>Has the same parent</li>
		 * </ol>
		 */
		public boolean isEquivalentTo(Object other) {
			boolean equivalent = true;

			if (!(other instanceof LcsEditorUnassignedWorkItemEntry)) {
				equivalent = false;
			} else {
				LcsEditorUnassignedWorkItemEntry otherChangeSet = (LcsEditorUnassignedWorkItemEntry) other;
				if (!otherChangeSet.getRepository().equals(getRepository())) {
					equivalent = false;
				}
				if ((otherChangeSet.getParent() == null && getParent() != null)
						|| (otherChangeSet.getParent() != null && getParent() == null)) {
					equivalent = false;
				}
				if (otherChangeSet.getParent() != null && getParent() != null
						&& !otherChangeSet.getParent().equals(getParent())) {
					equivalent = false;
				}
			}
			return equivalent;
		}

		@Override
		public Object getModelInstance() {
			return null;
		}
	}
}
