/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2006, 2012. All Rights Reserved. 
 * 
 * Note to U.S. Government Users Restricted Rights:  Use, 
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.filesystem.client.workitems.internal;

import java.sql.Timestamp;
import java.util.Collection;
import java.util.List;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.SubProgressMonitor;

import com.ibm.team.foundation.common.text.XMLString;
import com.ibm.team.links.common.ILink;
import com.ibm.team.process.common.IDevelopmentLine;
import com.ibm.team.process.common.IIteration;
import com.ibm.team.process.common.IProjectAreaHandle;
import com.ibm.team.repository.client.ITeamRepository;
import com.ibm.team.repository.common.IContributorHandle;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.repository.common.UUID;
import com.ibm.team.scm.common.IChangeSetHandle;
import com.ibm.team.scm.common.links.ChangeSetLinks;
import com.ibm.team.scm.common.links.ILinkConstants;
import com.ibm.team.scm.common.providers.ProviderFactory;
import com.ibm.team.workitem.client.IWorkItemClient;
import com.ibm.team.workitem.client.IWorkItemWorkingCopyManager;
import com.ibm.team.workitem.client.WorkItemWorkingCopy;
import com.ibm.team.workitem.client.WorkflowUtilities;
import com.ibm.team.workitem.common.model.IApproval;
import com.ibm.team.workitem.common.model.IApprovalDescriptor;
import com.ibm.team.workitem.common.model.IApprovals;
import com.ibm.team.workitem.common.model.IComment;
import com.ibm.team.workitem.common.model.IWorkItem;
import com.ibm.team.workitem.common.model.IWorkItemHandle;
import com.ibm.team.workitem.common.model.Identifier;
import com.ibm.team.workitem.common.workflow.IWorkflowAction;
import com.ibm.team.workitem.common.workflow.IWorkflowInfo;

/**
 * Class which annotates work items with links to change sets.
 * 
 * In theory all this should happen server side, but if it did then a work item editor which was open wouldn't be updated. It's the 
 * link changes the item or is the link modeled outside of the item. 
 */
public class WorkItemUtils {

    private static IWorkItemWorkingCopyManager getWorkItemWorkingCopyManager(ITeamRepository repo) {
        return ((IWorkItemClient) (repo).getClientLibrary(IWorkItemClient.class)).getWorkItemWorkingCopyManager();
    }
    
    public static IWorkItemHandle getWorkItemHandle(ITeamRepository repo, String workItemId, IProgressMonitor monitor) throws TeamRepositoryException {
    	return (IWorkItemHandle) IWorkItem.ITEM_TYPE.createItemHandle(repo, UUID.valueOf(workItemId), null);
    }
	
    public static IWorkItem createReviewTask(
    		ITeamRepository repo,
    		IWorkItem workItem,
    		String approvalType,
    		String approvalTypeLabel,
    		List<IContributorHandle> verifiers,
    		boolean addReview,
    		String comment,
    		boolean save,
    		Timestamp dueDate,
    		IProgressMonitor monitor) throws TeamRepositoryException {
        IWorkItemWorkingCopyManager manager = getWorkItemWorkingCopyManager(repo);
        WorkItemWorkingCopy copy = null;
        if(save) {
    		manager.connect(workItem, IWorkItem.FULL_PROFILE, null);
    		copy = manager.getWorkingCopy(workItem);
            workItem = copy.getWorkItem();
    	} else {
    	    workItem = (IWorkItem) workItem.getWorkingCopy();
    	}
        try {
	        if(verifiers.size() > 0 || addReview) {
	        	IApprovals approvals= workItem.getApprovals();
	        	IApprovalDescriptor descriptor= approvals.createDescriptor(approvalType, approvalTypeLabel);
	        	if (dueDate != null) {
	        		descriptor.setDueDate(dueDate);
	        	}
	        	for (IContributorHandle reviewer: verifiers) {
	        		IApproval approval= approvals.createApproval(descriptor, reviewer);
	        		approvals.add(approval);
	        		workItem.getSubscriptions().add(reviewer);
	        	}
	        	if(verifiers.isEmpty()) {
	    			approvals.add(descriptor);
	    		}
	        }
	        if(comment != null && comment.length() > 0) {
	        	IComment wiComment = workItem.getComments().createComment(
	        			repo.loggedInContributor(), XMLString.createFromPlainText(comment));
	        	workItem.getComments().append(wiComment);            
	        }
	        if(save)
	        	copy.save(new SubProgressMonitor(monitor, 1));
	        return workItem;
        } finally {
        	if(save)
        		manager.disconnect(workItem);
        }
    }
    
    /**
     * Mark the given list of work items as resolved
     */
    public static void closeWorkItems(ITeamRepository repo, Collection<IWorkItem> workItems, String comment, boolean shouldResolve) throws TeamRepositoryException {
        IWorkItemWorkingCopyManager manager = getWorkItemWorkingCopyManager(repo);
        for (IWorkItem workItem : workItems) {
            try {
                manager.connect(workItem, IWorkItem.FULL_PROFILE, null);
                WorkItemWorkingCopy copy = manager.getWorkingCopy(workItem);
                boolean shouldSave = !copy.isDirty();
                if(comment != null && comment.length() > 0) {
                    IComment wiComment = copy.getWorkItem().getComments().createComment(
                    		repo.loggedInContributor(), XMLString.createFromPlainText(comment));
                    copy.getWorkItem().getComments().append(wiComment);
                }
                if(shouldResolve) {
                	IWorkflowInfo workflowInfo = WorkflowUtilities
							.findWorkflowInfo(copy.getWorkItem(), null);
					if (workflowInfo != null) {
						// check if already closed
						if (workflowInfo.getStateGroup(copy.getWorkItem()
								.getState2()) != IWorkflowInfo.CLOSED_STATES) {
    						Identifier<IWorkflowAction> resolveActionId = workflowInfo
    								.getResolveActionId();
    						if (resolveActionId != null) {
    							copy.setWorkflowAction(resolveActionId
    									.getStringIdentifier());
    						}
						}
					}
				}
                if (shouldSave)
                    copy.save(null);
            } finally {
                manager.disconnect(workItem);
            }
        }
    }
    
    public static void fillInDefaults(IWorkItemHandle createdWorkItem, IProgressMonitor monitor) throws TeamRepositoryException {
        ITeamRepository repo = (ITeamRepository) createdWorkItem.getOrigin();
        IWorkItemClient service = (IWorkItemClient) repo.getClientLibrary(IWorkItemClient.class);
        IWorkItem workItem = null;
            try {
                service.getWorkItemWorkingCopyManager().connect(createdWorkItem, IWorkItem.FULL_PROFILE, null);
                WorkItemWorkingCopy wc= service.getWorkItemWorkingCopyManager().getWorkingCopy(createdWorkItem);                   
                workItem = wc.getWorkItem();
                workItem.setCreator(repo.loggedInContributor());
                // issue, the owner may not be on the same team as the originating work item
                workItem.setOwner(repo.loggedInContributor()); 
                IProjectAreaHandle projectArea = workItem.getProjectArea();
                IDevelopmentLine developmentLine= service.getAuditableCommon().findDefaultDevelopmentLine(projectArea, null);
                if(developmentLine != null) {
                	// TODO This is a workaround for defect 29575. Process doesn't have the APIs to find the current milestone 
                	List<IIteration> iterations= service.getAuditableCommon().findCurrentIterations(developmentLine, null);
                	if(iterations.size() > 1)
                		workItem.setTarget(iterations.get(1));
                }
                wc.save(new SubProgressMonitor(monitor, 1));
            } finally {
                if(workItem != null) {
                    service.getWorkItemWorkingCopyManager().disconnect(workItem);                    
                }
            }
    }

    /**
	 * Checks if the given work item is associated with the change set
	 * 
     * @param wi the work item possibly linked to the change set
     * @param repo the change set's repository
     * @param csh the change set possibly linked to the work item
     * @return <code>true</code> if the work item is associated with the change set
     * @throws TeamRepositoryException
     */
	public static boolean isWorkItemAssociated(IWorkItemHandle wi, ITeamRepository repo, IChangeSetHandle csh) throws TeamRepositoryException {
		ProviderFactory providerFactory = (ProviderFactory) repo.getClientLibrary(ProviderFactory.class);
		List<ILink> links = ChangeSetLinks.findLinks(providerFactory, csh, new String[] {ILinkConstants.CHANGESET_WORKITEM_LINKTYPE_ID }, null);

		for (ILink link : links) {
			Object resolved = link.getTargetRef().resolve();
			if (resolved instanceof IWorkItemHandle) {
				IWorkItemHandle wiHandle = (IWorkItemHandle) resolved;
				if (wiHandle.sameItemId(wi))
					return true;
			}
		}
		
		return false;
	}
}
