/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2007, 2011. All Rights Reserved. 
 * 
 * Note to U.S. Government Users Restricted Rights:  Use, 
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.filesystem.client.workitems.internal.operations;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.SubMonitor;
import org.eclipse.osgi.util.NLS;

import com.ibm.team.filesystem.client.FileSystemException;
import com.ibm.team.filesystem.client.ILocalChange;
import com.ibm.team.filesystem.client.IOperationFactory;
import com.ibm.team.filesystem.client.internal.IRepositoryResolver;
import com.ibm.team.filesystem.client.internal.operations.FileSystemOperation;
import com.ibm.team.filesystem.client.operations.DeliverDilemmaHandler;
import com.ibm.team.filesystem.client.operations.ICheckinOperation;
import com.ibm.team.filesystem.client.operations.ICheckinOptions;
import com.ibm.team.filesystem.client.operations.IDeliverOperation;
import com.ibm.team.filesystem.client.workitems.IFileSystemWorkItemManager;
import com.ibm.team.filesystem.client.workitems.internal.Messages;
import com.ibm.team.filesystem.client.workitems.operations.ICheckinAndDeliverOperation;
import com.ibm.team.repository.client.IItemManager;
import com.ibm.team.repository.client.ITeamRepository;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.repository.common.UUID;
import com.ibm.team.scm.client.IWorkspaceConnection;
import com.ibm.team.scm.client.IWorkspaceManager;
import com.ibm.team.scm.client.SCMPlatform;
import com.ibm.team.scm.common.IChangeSet;
import com.ibm.team.scm.common.IChangeSetHandle;
import com.ibm.team.scm.common.IComponentHandle;
import com.ibm.team.scm.common.WorkspaceComparisonFlags;
import com.ibm.team.scm.common.dto.IChangeHistorySyncReport;
import com.ibm.team.workitem.common.model.IWorkItemHandle;

@SuppressWarnings("restriction")
// TODO put this lower into operations
public class CheckinAndDeliverOperation extends FileSystemOperation implements ICheckinAndDeliverOperation {

	private static class CheckInAndDeliverRequest {
		final public IWorkspaceConnection source;
		final public IWorkspaceConnection target;
		final public List<ILocalChange> localChanges;
		final public IChangeHistorySyncReport report;
		final public IComponentHandle component;
		final public IRepositoryResolver resolver;
		
		public CheckInAndDeliverRequest(
			IWorkspaceConnection sourceConnection,
			IWorkspaceConnection targetConnection,
			List<ILocalChange> changes,
			IChangeHistorySyncReport report,
			IComponentHandle component) {
			
			this.source = sourceConnection;
			this.target = targetConnection;
			this.localChanges = new ArrayList<ILocalChange>(changes.size());
			for (ILocalChange change : changes) {
				if (!change.getConnection().sameItemId(source.getContextHandle())) {
					throw new IllegalArgumentException();
				}
				this.localChanges.add(change);
			}
			this.report = report;
			this.component = component;
			this.resolver = new IRepositoryResolver() {
				
				public ITeamRepository getRepoFor(String uri, UUID id) throws TeamRepositoryException {
					if (id.equals(source.teamRepository().getId())) {
						return source.teamRepository();
					} else if (id.equals(target.teamRepository().getId())) {
						return target.teamRepository();
					}
					throw new TeamRepositoryException(NLS.bind(Messages.CheckinAndDeliverOperation_0,
							source.teamRepository().getRepositoryURI(), target.teamRepository().getRepositoryURI()));
				}
			};
		}
	}
	
	private IWorkItemHandle workItem;
	private final DeliverDilemmaHandler dilemmaHandler;
	private String changeSetComment;
	private List<CheckInAndDeliverRequest> requests = new ArrayList<CheckInAndDeliverRequest>();
	private boolean refreshBeforeRun;
	private int nonAtomicCommitFileUploadLimit = ICheckinOptions.NON_ATOMIC_COMMIT_FILE_UPDATE_DEFAULT_LIMIT;
	private boolean nonAtomicCommitIsAnOption = ICheckinOptions.NON_ATOMIC_COMMIT_DEFAULT_ENABLEMENT;


	public CheckinAndDeliverOperation(DeliverDilemmaHandler dilemmaHandler) {
		super(dilemmaHandler == null ? DeliverDilemmaHandler.getDefault() : dilemmaHandler);
		this.dilemmaHandler = dilemmaHandler == null ? DeliverDilemmaHandler.getDefault() : dilemmaHandler;		
	}

    public void changeSetLinkage(IWorkItemHandle workItem,
            String changeSetComment) {
        this.workItem = workItem;
        this.changeSetComment = changeSetComment;
    }
	
	public void add(IWorkspaceConnection source, IWorkspaceConnection target, IChangeHistorySyncReport report, 
			IComponentHandle component, List<ILocalChange> localChanges) {
		CheckInAndDeliverRequest request = new CheckInAndDeliverRequest(source, target, localChanges, report, component);
		requests.add(request);
	}
	
	/**
	 * 2 round trip per component (compare, check-in) for check-in and 1 deliver per workspace
	 * Would be nice to make this more optimal, but my guess is that most uses of this
	 * operation will be at the single component level.
	 */
	public void execute(IProgressMonitor monitor) throws FileSystemException, TeamRepositoryException {
		SubMonitor progress = SubMonitor.convert(monitor, Messages.CheckinAndDeliverOperation_PROGRESS_DELIVER, 100);
		IDeliverOperation deliverOperation = IOperationFactory.instance.getDeliverOperation(dilemmaHandler);
		configureCheckinOptions(deliverOperation);
		deliverOperation.setRefreshBeforeRun(refreshBeforeRun);
		SubMonitor loopProgress = progress.newChild(75).setWorkRemaining(requests.size());
		for (CheckInAndDeliverRequest request : requests) {
			checkIn(request, deliverOperation, loopProgress.newChild(1));
		}
		deliverOperation.run(progress.newChild(25));
	}
	
	private void configureCheckinOptions(ICheckinOptions operation) {
		if (nonAtomicCommitIsAnOption) {
			operation.enableNonAtomicCommit(nonAtomicCommitFileUploadLimit);
		} else {
			operation.disableNonAtomicCommit();
		}
	}

	private void checkIn(CheckInAndDeliverRequest request, IDeliverOperation deliverOperation, IProgressMonitor monitor) throws TeamRepositoryException, FileSystemException {
		SubMonitor progress = SubMonitor.convert(monitor, Messages.CheckinAndDeliverOperation_PROGRESS_MESSAGE, 100);
		
		IWorkspaceConnection source = request.source;
		IWorkspaceConnection target = request.target;
		List<ILocalChange> localChanges = request.localChanges;
		IChangeHistorySyncReport report = request.report;		
		IComponentHandle component = request.component;
		
		// There are no outgoing change sets before the checkin happens
		ICheckinOperation checkInOperation = IOperationFactory.instance.getCheckinOperation(
			dilemmaHandler.getCommitDilemmaHandler(), request.resolver);
		configureCheckinOptions(checkInOperation);
		checkInOperation.setRefreshBeforeRun(refreshBeforeRun);
		checkInOperation.requestCheckin(
	                localChanges.toArray(new ILocalChange[0]), 
	                null, 
	                changeSetComment, progress.newChild(5));
		
		checkInOperation.run(progress.newChild(55));
		
		Collection<IChangeSetHandle> committedChangeSets = checkInOperation.getCommittedChangeSets();

		report = source.compareTo(target, WorkspaceComparisonFlags.INCLUDE_BASELINE_INFO, Collections.EMPTY_LIST, progress.newChild(10));		
		
		progress.setTaskName(Messages.CheckinAndDeliverOperation_PROGRESS_MESSAGE_CHECKIN);
		IFileSystemWorkItemManager workItemManager = (IFileSystemWorkItemManager) source.teamRepository().getClientLibrary(IFileSystemWorkItemManager.class);
		List<IChangeSetHandle> outgoingHandles = new ArrayList<IChangeSetHandle>(committedChangeSets);
		List<IChangeSet> outgoing = source.teamRepository().itemManager().fetchCompleteItems(outgoingHandles, IItemManager.DEFAULT, progress.newChild(5));
		SubMonitor loopProgress = progress.newChild(10).setWorkRemaining(outgoing.size() * 2);
		IWorkspaceManager manager = SCMPlatform.getWorkspaceManager(source.teamRepository());
		for (IChangeSet changeSet : outgoing) {
			if(changeSet != null && changeSet.isActive()) {
				String trimmedComment = changeSetComment.trim();
				if(trimmedComment.length() > 0) {
					manager.setComment(changeSet, changeSetComment, loopProgress.newChild(1));
				}
				if(workItem != null) {
					workItemManager.createLink(
							source.getResolvedWorkspace(), 
							changeSet, new IWorkItemHandle[] {workItem}, loopProgress.newChild(1));
				}
			}
		}
		
		deliverOperation.deliver(source, target, report, component);
	}

	public void disableNonAtomicCommit() {
		nonAtomicCommitIsAnOption  = false;
	}

	public void enableNonAtomicCommit(int numberUploads) {
		nonAtomicCommitIsAnOption = true;
		nonAtomicCommitFileUploadLimit = numberUploads;
	}

	public boolean getRefreshBeforeRun() {
		return refreshBeforeRun;
	}

	public void setRefreshBeforeRun(boolean newValue) {
		refreshBeforeRun = newValue;
	}
}
