/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2007, 2014. All Rights Reserved. 
 * 
 * Note to U.S. Government Users Restricted Rights:  Use, 
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.filesystem.client.workitems.internal.operations;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.OperationCanceledException;
import org.eclipse.core.runtime.SubMonitor;

import com.ibm.team.filesystem.client.FileSystemException;
import com.ibm.team.filesystem.client.IOperationFactory;
import com.ibm.team.filesystem.client.internal.operations.FileSystemOperation;
import com.ibm.team.filesystem.client.internal.operations.OperationRequest;
import com.ibm.team.filesystem.client.internal.operations.UnlockOperation;
import com.ibm.team.filesystem.client.internal.operations.DeliverOperation.DeliverDescriptor;
import com.ibm.team.filesystem.client.internal.operations.DeliverOperation.DeliverOperationRequest;
import com.ibm.team.filesystem.client.internal.utils.ConfigurationFacade;
import com.ibm.team.filesystem.client.internal.utils.ExclusiveFileLockPatternUtil;
import com.ibm.team.filesystem.client.internal.utils.FlowNodeLock;
import com.ibm.team.filesystem.client.internal.utils.WorkspaceLockUtil;
import com.ibm.team.filesystem.client.operations.DeliverDilemmaHandler;
import com.ibm.team.filesystem.client.operations.ICheckinOptions;
import com.ibm.team.filesystem.client.operations.IDeliverOperation;
import com.ibm.team.filesystem.client.operations.IDilemmaHandler;
import com.ibm.team.filesystem.client.operations.IUnlockOperation;
import com.ibm.team.filesystem.client.operations.LockEntry;
import com.ibm.team.filesystem.client.workitems.IFileSystemWorkItemManager;
import com.ibm.team.filesystem.client.workitems.internal.Messages;
import com.ibm.team.filesystem.client.workitems.internal.WorkItemUtils;
import com.ibm.team.filesystem.client.workitems.operations.DeliverAndFixDilemmaHandler;
import com.ibm.team.filesystem.client.workitems.operations.DeliverAndFixParameter;
import com.ibm.team.filesystem.client.workitems.operations.IDeliverAndFixOperation;
import com.ibm.team.repository.client.IItemManager;
import com.ibm.team.repository.client.ITeamRepository;
import com.ibm.team.repository.common.IAuditable;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.scm.client.IWorkspaceConnection;
import com.ibm.team.scm.client.IFlowNodeConnection.IComponentOp;
import com.ibm.team.scm.common.IBaselineHandle;
import com.ibm.team.scm.common.IChangeSet;
import com.ibm.team.scm.common.IChangeSetHandle;
import com.ibm.team.scm.common.IComponentHandle;
import com.ibm.team.scm.common.IContextHandle;
import com.ibm.team.scm.common.WorkspaceComparisonFlags;
import com.ibm.team.scm.common.dto.IChangeHistorySyncReport;
import com.ibm.team.scm.common.dto.IComponentLockReport;
import com.ibm.team.scm.common.dto.IStreamLockReport;
import com.ibm.team.workitem.common.model.IWorkItem;
import com.ibm.team.workitem.common.model.IWorkItemHandle;

public class DeliverAndFixOperation extends FileSystemOperation implements IDeliverAndFixOperation {

	private DeliverAndFixParameter deliverAndFixParameters;
	private DeliverAndFixDilemmaHandler dilemmaHandler;
	private boolean refreshBeforeRun = false;
	private boolean nonAtomicCommitIsAnOption = ICheckinOptions.NON_ATOMIC_COMMIT_DEFAULT_ENABLEMENT;
	private int nonAtomicCommitFileUploadLimit = ICheckinOptions.NON_ATOMIC_COMMIT_FILE_UPDATE_DEFAULT_LIMIT;
	private boolean autoReleaseLocks = false;

	public DeliverAndFixOperation(DeliverAndFixDilemmaHandler dilemmaHandler) {
		super(dilemmaHandler == null ? DeliverAndFixDilemmaHandler.getDefault() : dilemmaHandler);
	    this.dilemmaHandler = dilemmaHandler == null ? DeliverAndFixDilemmaHandler.getDefault(): dilemmaHandler;
    }

    public void deliverAndFix(DeliverAndFixParameter parms) {
    	deliverAndFixParameters = parms;
    }

    public void execute(IProgressMonitor monitor) throws FileSystemException, TeamRepositoryException {
    	SubMonitor loopProgress = SubMonitor.convert(monitor, 100);
    	
    	ITeamRepository repo = deliverAndFixParameters.getTarget().teamRepository();
    	if (!repo.equals(deliverAndFixParameters.getSource().teamRepository())) {
    		throw new TeamRepositoryException(Messages.DeliverAndFixOperation_0);
    	}
    	IFileSystemWorkItemManager workItemManager = (IFileSystemWorkItemManager) repo.getClientLibrary(IFileSystemWorkItemManager.class);
    	List<IComponentLockReport> locks = new ArrayList<IComponentLockReport>();
    	
    	List<IWorkItemHandle> workItems = new ArrayList<IWorkItemHandle>(deliverAndFixParameters.getWorkItems());
    	
    	Iterator<IWorkItemHandle> i = workItems.iterator();
    	while (i.hasNext()) {
    		if (i.next() == null)
				i.remove();
    	}

    	// get full work items
    	List<IWorkItem> fullWorkItems = repo.itemManager().fetchCompleteItems(workItems, IItemManager.REFRESH, loopProgress.newChild(20));
    	List<IWorkItemHandle> inaccessibleWorkItems = new ArrayList<IWorkItemHandle>();

    	for (int j = 0; j < workItems.size(); j++) {
    		if (fullWorkItems.get(j) == null) {
    			inaccessibleWorkItems.add(workItems.get(j));
    		}
    	}
    	
    	processInaccessibleWorkItems(inaccessibleWorkItems, fullWorkItems, monitor);

    	// Assign work item to change sets if not associated to any of the given change sets
		for (IWorkItemHandle workItemToAssign : fullWorkItems) {
			boolean associated = false;
			
    		for (IChangeSetHandle changeSet : deliverAndFixParameters.getChangeSets()) {
    			if (WorkItemUtils.isWorkItemAssociated(workItemToAssign, repo, changeSet)) {
    				associated = true;
    				break;
    			}
    		}
    		
    		if (!associated) {
    			for (IChangeSetHandle changeSet : deliverAndFixParameters.getChangeSets()) {
    				workItemManager.createLink(deliverAndFixParameters.getTarget().getResolvedWorkspace(), changeSet,
    						new IWorkItemHandle[] { workItemToAssign }, loopProgress.newChild(1));
    			}
    		}
		}
		
    	IDeliverOperation deliverOp = IOperationFactory.instance.getDeliverOperation(dilemmaHandler.getDeliverDilemmaHandler());
    	
    	deliverOp.setRefreshBeforeRun(refreshBeforeRun);
    	if (nonAtomicCommitIsAnOption) {
    		deliverOp.enableNonAtomicCommit(nonAtomicCommitFileUploadLimit);
    	} else {
    		deliverOp.disableNonAtomicCommit();
    	}
    	deliverOp.deliver(deliverAndFixParameters.getSource(), deliverAndFixParameters.getTarget(), deliverAndFixParameters.getSyncReport(),
    			new ArrayList<IChangeSetHandle>(deliverAndFixParameters.getChangeSets()));
    	
    	if (fullWorkItems.size() > 0) {
	    	boolean addReview = deliverAndFixParameters.getReviewers().length > 0;
			workItemManager.deliverAndResolve(deliverAndFixParameters.getSyncReport(),
	    			deliverAndFixParameters.getSource(), deliverAndFixParameters.getTarget(), new ArrayList<IChangeSetHandle>(deliverAndFixParameters.getChangeSets()),
	    			locks,
	    			new ArrayList<IWorkItem>(fullWorkItems),
	    			deliverAndFixParameters.isShouldResolve(), deliverAndFixParameters.getComment(), addReview,
	    			deliverAndFixParameters.getSubject(), Arrays.asList(deliverAndFixParameters.getReviewers()),
	    			deliverAndFixParameters.getDueDate(), deliverOp, monitor);
			
			DeliverOperationRequest deliverRequest = new DeliverOperationRequest(deliverAndFixParameters.getSource(), deliverAndFixParameters.getTarget(),
					deliverAndFixParameters.getSyncReport(), new ArrayList<IBaselineHandle>(), new ArrayList<IChangeSetHandle>(),
					new ArrayList<IComponentHandle>());
			deliverRequest.addChangeSet(new ArrayList(deliverAndFixParameters.getChangeSets()));
			
			// release the locks
			DeliverDilemmaHandler deliverDilemmaHandler = (DeliverDilemmaHandler) dilemmaHandler
					.getDeliverDilemmaHandler();
			
			IChangeHistorySyncReport updatedReport = deliverRequest.getWorkspace().compareTo(deliverRequest.getTargetWorkspace(),
                    WorkspaceComparisonFlags.CHANGE_SET_COMPARISON_ONLY, Collections.EMPTY_LIST, SubMonitor.convert(monitor, 1));
			
			if (locks.size() > 0) {
				List<IChangeSetHandle> outgoingChangeSetHandles = updatedReport.outgoingChangeSets();
                if (outgoingChangeSetHandles.size() != 0) {				
                	locks = ExclusiveFileLockPatternUtil.updateComponentLocks(locks, deliverRequest.getWorkspace().teamRepository(), outgoingChangeSetHandles, SubMonitor.convert(monitor, 1));
                }
			}

			if (locks.size() > 0) {
				IWorkspaceConnection target = deliverAndFixParameters.getTarget();
		        Map<IStreamLockReport, IWorkspaceConnection> locksToRelease = new HashMap<IStreamLockReport, IWorkspaceConnection>();
				locksToRelease.put(IStreamLockReport.FACTORY.createInstance(target.getResolvedWorkspace(), locks), target);
	            Map<IStreamLockReport, IWorkspaceConnection> locksToRelease2;
	            if (this.autoReleaseLocks) {
	            	locksToRelease2 = locksToRelease;
	            } else {
	            	locksToRelease2 = deliverDilemmaHandler.locksToRelease(locksToRelease);
	            }
				if (locksToRelease2.size() > 0) {
	                IUnlockOperation unlockOperation = new UnlockOperation(deliverDilemmaHandler.getLockDilemmaHandler());
	                for (Map.Entry<IStreamLockReport, IWorkspaceConnection> entry : locksToRelease2.entrySet()) {
	                    List<LockEntry> locksEntries = LockEntry.createEntries(entry.getValue().teamRepository(), entry.getKey());
	                    for (LockEntry lockEntry : locksEntries) {
	                        unlockOperation.unlock(
	                                lockEntry.getWorkspaceConnection(SubMonitor.convert(monitor, 1)),
	                                lockEntry.getComponent(),
	                                lockEntry.getVersionable(),
	                                lockEntry.getContributor());
	                    }
	                }
	                unlockOperation.run(SubMonitor.convert(monitor, 100));
	            }
			}
			
			if (ExclusiveFileLockPatternUtil.getPatternsToMatch(deliverAndFixParameters.getSource(), SubMonitor.convert(monitor, 1)).size() != 0) {
				// acquire workspace lock
				List<IAuditable> determinants = new ArrayList<IAuditable>();
				Collection<ConfigurationFacade> affectedConfigurations = getAffectedConfigurations(Collections.singletonList(deliverRequest), determinants, SubMonitor.convert(monitor, 1));
				List<IContextHandle> targets = new ArrayList<IContextHandle>(2);
				targets.add(deliverAndFixParameters.getSource().getResolvedWorkspace());
				targets.add(deliverAndFixParameters.getTarget().getResolvedWorkspace());
				FlowNodeLock workspaceLock = WorkspaceLockUtil.acquireWriteForConfigurations(affectedConfigurations, targets, determinants, SubMonitor.convert(monitor, 1));
	
				// update file attributes
				try {
					Map<IWorkspaceConnection, List<IChangeSet>> workspaceToDeliveredChangeSets = new HashMap<IWorkspaceConnection, List<IChangeSet>>();
					Map<IWorkspaceConnection, List<IComponentOp>> workspaceComponentOps = new HashMap<IWorkspaceConnection, List<IComponentOp>>();
					
					// get the change sets
			        IItemManager itemManager = deliverAndFixParameters.getSource().teamRepository().itemManager();
			        List<IChangeSet> changeSets = null;
			        try {
			            changeSets = itemManager.fetchCompleteItems(new ArrayList(deliverAndFixParameters.getChangeSets()), IItemManager.DEFAULT, SubMonitor.convert(monitor, 10));
						workspaceToDeliveredChangeSets.put(deliverAndFixParameters.getSource(), changeSets);
						ExclusiveFileLockPatternUtil.refreshFileAttributes(affectedConfigurations, workspaceToDeliveredChangeSets, workspaceComponentOps, true, SubMonitor.convert(monitor, 10));
			        } catch (TeamRepositoryException e) {
			            // ignore
			        }
				} finally {
					WorkspaceLockUtil.release(workspaceLock);
				}
			}
		} else {
    		SubMonitor progress = SubMonitor.convert(monitor, 100);
    		deliverOp.run(progress);
    	}
    	
		monitor.done();
    }

	private void processInaccessibleWorkItems(List<IWorkItemHandle> inaccessibleWorkItems, List<IWorkItem> fullWorkItems, IProgressMonitor monitor) throws TeamRepositoryException {
		if (inaccessibleWorkItems.size() == 0)
			return;
		
		int action = this.dilemmaHandler.workItemsUnreachable(inaccessibleWorkItems.size(), fullWorkItems.size());
		
		if (action == IDilemmaHandler.CONTINUE) {
			// for now the work items are ignored
			for (Iterator<IWorkItem> i = fullWorkItems.iterator(); i.hasNext(); ) {
				IWorkItem wi = i.next();
				if (wi == null)
					i.remove();
			}
		} else if (action == IDilemmaHandler.FAIL) {
			throw new TeamRepositoryException(Messages.DeliverAndFixOperation_1);
		} else {
			throw new OperationCanceledException();
		}
	}

	public boolean getRefreshBeforeRun() {
		return refreshBeforeRun;
	}

	public void setRefreshBeforeRun(boolean newValue) {
		this.refreshBeforeRun = newValue;
	}

	public void disableNonAtomicCommit() {
		this.nonAtomicCommitIsAnOption = false;
	}

	public void enableNonAtomicCommit(int numberUploads) {
		this.nonAtomicCommitIsAnOption = true;
		this.nonAtomicCommitFileUploadLimit = numberUploads;
	}
	
    protected Set<ConfigurationFacade> getAffectedConfigurations(Collection<? extends OperationRequest> acceptRequests, List<IAuditable> determinants, IProgressMonitor progress) throws TeamRepositoryException {
        SubMonitor monitor = SubMonitor.convert(progress, acceptRequests.size());
        Set<ConfigurationFacade> configurations = new HashSet<ConfigurationFacade>();
        for (OperationRequest acceptRequest : acceptRequests) {
            configurations.addAll(acceptRequest.getAffectedConfigurations(determinants, monitor.newChild(1)));
        }
        return configurations;
    }
    
    public void setAutoReleaseLocks(boolean value) {
        this.autoReleaseLocks = value;
    }
}
