/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2007, 2012. All Rights Reserved. 
 * 
 * Note to U.S. Government Users Restricted Rights:  Use, 
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.filesystem.client.workitems.internal.operations;

import java.lang.reflect.InvocationTargetException;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.OperationCanceledException;
import org.eclipse.core.runtime.SubMonitor;

import com.ibm.team.filesystem.client.FileSystemException;
import com.ibm.team.filesystem.client.IOperationFactory;
import com.ibm.team.filesystem.client.ISandbox;
import com.ibm.team.filesystem.client.internal.copyfileareas.ICopyFileAreaLockRequest;
import com.ibm.team.filesystem.client.internal.copyfileareas.ICopyFileAreaManager;
import com.ibm.team.filesystem.client.internal.copyfileareas.ICopyFileAreasLock;
import com.ibm.team.filesystem.client.internal.operations.ChangeHistoryChangeSetsOperation;
import com.ibm.team.filesystem.client.internal.operations.IFilesystemRefresher;
import com.ibm.team.filesystem.client.internal.utils.ChangeSetRefreshUtils;
import com.ibm.team.filesystem.client.internal.utils.ConfigurationFacade;
import com.ibm.team.filesystem.client.internal.utils.FlowNodeLock;
import com.ibm.team.filesystem.client.internal.utils.IRunnableWithProgress;
import com.ibm.team.filesystem.client.internal.utils.WorkspaceLockUtil;
import com.ibm.team.filesystem.client.operations.ICheckinOptions;
import com.ibm.team.filesystem.client.operations.IDilemmaHandler;
import com.ibm.team.filesystem.client.operations.ISuspendOperation;
import com.ibm.team.filesystem.client.operations.IVerifyInSyncOperation;
import com.ibm.team.filesystem.client.workitems.IFileSystemWorkItemManager;
import com.ibm.team.filesystem.client.workitems.internal.Messages;
import com.ibm.team.filesystem.client.workitems.internal.WorkItemUtils;
import com.ibm.team.filesystem.client.workitems.operations.ISubmitForReviewOperation;
import com.ibm.team.filesystem.client.workitems.operations.SubmitForReviewDilemmaHandler;
import com.ibm.team.filesystem.client.workitems.operations.SubmitForReviewParameter;
import com.ibm.team.filesystem.client.workitems.operations.SubmitForReviewParameter.ChangeSetReview;
import com.ibm.team.repository.client.IItemManager;
import com.ibm.team.repository.client.ITeamRepository;
import com.ibm.team.repository.client.util.ThreadCheck;
import com.ibm.team.repository.common.IAuditable;
import com.ibm.team.repository.common.IContributorHandle;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.scm.client.IWorkspaceConnection;
import com.ibm.team.scm.common.IChangeSet;
import com.ibm.team.scm.common.IChangeSetHandle;
import com.ibm.team.workitem.common.model.IWorkItem;
import com.ibm.team.workitem.common.model.IWorkItemHandle;
import com.ibm.team.workitem.common.model.WorkItemApprovals;

public class SubmitForReviewOperation extends ChangeHistoryChangeSetsOperation implements ISubmitForReviewOperation {

    private SubmitForReviewParameter submitForReviewParameter;
    private final Map<IWorkspaceConnection, List<IChangeSet>> sourceToChangeSets;
    private SubmitForReviewDilemmaHandler baseProblemHandler;
	private boolean nonAtomicCommitIsAnOption = ICheckinOptions.NON_ATOMIC_COMMIT_DEFAULT_ENABLEMENT;
	private int nonAtomicCommitFileUploadLimit = ICheckinOptions.NON_ATOMIC_COMMIT_FILE_UPDATE_DEFAULT_LIMIT;

	/**
     * Creates a new submit for review operation.
     * @param handler suspend dilemma handler if the change sets will be suspended 
     */
    public SubmitForReviewOperation(SubmitForReviewDilemmaHandler handler) {
    	super(handler == null ? SubmitForReviewDilemmaHandler.getDefault().getSuspendDilemmaHandler() : handler.getSuspendDilemmaHandler());
        baseProblemHandler = handler == null ? SubmitForReviewDilemmaHandler.getDefault() : handler;
        sourceToChangeSets = new HashMap<IWorkspaceConnection, List<IChangeSet>>();
    }
    
    public void execute(IProgressMonitor monitor) throws FileSystemException, TeamRepositoryException {

        ThreadCheck.checkLongOpsAllowed();

        SubMonitor progress = SubMonitor.convert(monitor, 100);
    	ITeamRepository repo = submitForReviewParameter.getRepo();
    		
    	Map<IChangeSetHandle, IWorkspaceConnection> changeSetToSource = new HashMap<IChangeSetHandle, IWorkspaceConnection>();
    	
    	// map change set handle to connection
    	for (ChangeSetReview changeSetToReview : submitForReviewParameter.getChangeSetReviews()) {
    		if (!changeSetToReview.getSource().teamRepository().equals(repo))
    			throw new TeamRepositoryException("Repository for work item is not the same as for the change sets."); //$NON-NLS-1$
    		changeSetToSource.put(changeSetToReview.getChangeSet(), changeSetToReview.getSource());
    	}
    	
    	// use change set handles to get full change set in a batch call
    	ArrayList<IChangeSetHandle> itemHandles = new ArrayList<IChangeSetHandle>(changeSetToSource.keySet());
    	@SuppressWarnings("unchecked")
		List<IChangeSet> fullChangeSets = repo.itemManager().fetchCompleteItems(itemHandles, IItemManager.DEFAULT, progress.newChild(1));
        
        // Locking order is :  
        // Jazz Workspace lock - obtained here
        // Eclipse Workspace lock - obtained here (we may checkin which may also result in updates)
        // CFA lock - obtained here (we may checkin and suspend needs them)
        final Set<ConfigurationFacade> affectedConfigurations = new HashSet<ConfigurationFacade>();
        Collection<IAuditable> determinants = new ArrayList<IAuditable>();
        
    	int i = 0;
        for (IChangeSet changeSet : fullChangeSets) {
    		// #fetchCompleteItems(*) preserves order of IChangeSet to the given list of IChangeSetHandle
        	IWorkspaceConnection connection = changeSetToSource.get(itemHandles.get(i));
            affectedConfigurations.add(new ConfigurationFacade(connection, changeSet.getComponent()));
    		
    		List<IChangeSet> changeSets = sourceToChangeSets.get(connection);
    		if (changeSets == null) {
    			changeSets = new ArrayList<IChangeSet>();
        		sourceToChangeSets.put(connection, changeSets);
    		}
    		
    		changeSets.add(changeSet);
            i++;
        }
        
        // lock jazz workspace for all configurations related to all suspending change sets
        FlowNodeLock workspaceLock = WorkspaceLockUtil.acquireWrite(affectedConfigurations, determinants, progress.newChild(1));
        
        // runnable to lock eclipse workspace
        try {
            IRunnableWithProgress runnableWithProgress = new IRunnableWithProgress() {
                public void run(IProgressMonitor monitor) throws InvocationTargetException, InterruptedException {

                    try {
                    	submitForReview(affectedConfigurations, monitor);
                    } catch (FileSystemException e) {
                        throw new InvocationTargetException(e);
                    } catch (TeamRepositoryException e) {
                        throw new InvocationTargetException(e);
                    }

    	
                }
            };
            runWithinFileSystemLock(runnableWithProgress, Messages.SubmitForReviewOperation_1, progress.newChild(98));
        } finally {
            WorkspaceLockUtil.release(workspaceLock);
        }
    }
    
    private void submitForReview(Set<ConfigurationFacade> affectedConfigurations, IProgressMonitor progress) throws TeamRepositoryException, FileSystemException {
    	
    	SubMonitor monitor = SubMonitor.convert(progress, 100);
    	
        if (verifyInSyncEnabled() && !baseProblemHandler.getSuspendDilemmaHandler().getOutOfSyncDilemmaHandler().willIgnoreAllSharesOutOfSync()) {
        	SubMonitor iterMonitor = monitor.newChild(1);
        	iterMonitor.setWorkRemaining(affectedConfigurations.size());
        	
        	IVerifyInSyncOperation verifyOp = IOperationFactory.instance.getVerifyInSyncOperation(baseProblemHandler.getSuspendDilemmaHandler().getOutOfSyncDilemmaHandler());
	    	for (ConfigurationFacade descriptor : affectedConfigurations) {
	    		verifyOp.addToVerify(descriptor.getConnection(iterMonitor.newChild(1)), descriptor.getComponentHandle());
	        }
	    	verifyOp.run(monitor.newChild(1));
        }
        
        if (progress.isCanceled()) {
            throw new OperationCanceledException();
        }
        
        ICopyFileAreaLockRequest lockRequest = ICopyFileAreaManager.instance.lockRequestFactory().getLockRequest(affectedConfigurations, true);
        ICopyFileAreasLock copyFileAreaLock = ICopyFileAreaManager.instance.lock(Collections.singleton(lockRequest), monitor.newChild(1));

        try {
        	if (!submitForReviewParameter.isSuspend()) {
        		// check-in any pending changes. If the changes are going to be suspended, then that will prompt to checkin the changes.
        		checkinPendingChanges(affectedConfigurations, baseProblemHandler.getSuspendDilemmaHandler(), true, monitor.newChild(15));
        	}
    
	    	List<IWorkItemHandle> workItems = new ArrayList<IWorkItemHandle>(submitForReviewParameter.getWorkItems());
	
	    	Iterator<IWorkItemHandle> i = workItems.iterator();
	    	while (i.hasNext()) {
	    		if (i.next() == null)
					i.remove();
	    	}
	    	
	    	// get full work items
	    	ITeamRepository repo = submitForReviewParameter.getRepo();
	    	@SuppressWarnings("unchecked")
			List<IWorkItem> fullWorkItems = repo.itemManager().fetchCompleteItems(workItems, IItemManager.REFRESH, monitor.newChild(1));

	    	List<IWorkItemHandle> inaccessibleWorkItems = new ArrayList<IWorkItemHandle>();

	    	Set<IWorkItem> uniqueWorkItems = new HashSet<IWorkItem>(); 
	    	for (int j = 0; j < workItems.size(); j++) {
	    		IWorkItem currentWorkItem = fullWorkItems.get(j);
				if (currentWorkItem == null) {
	    			inaccessibleWorkItems.add(workItems.get(j));
	    		} else {
	    			uniqueWorkItems.add(currentWorkItem);
	    		}
	    	}
	    	
	    	processInaccessibleWorkItems(inaccessibleWorkItems, fullWorkItems, monitor);
	    	
	    	// Assign work item to change sets if not associated to any of the given change sets
	    	SubMonitor loopProgress = SubMonitor.convert(monitor.newChild(15), uniqueWorkItems.size() * submitForReviewParameter.getChangeSetReviews().size());
	    	for (IWorkItemHandle workItemToAssign : uniqueWorkItems) {
	    		for (ChangeSetReview changeSetReview : submitForReviewParameter.getChangeSetReviews()) {
	    			if (!WorkItemUtils.isWorkItemAssociated(workItemToAssign, repo, changeSetReview.getChangeSet())) {

	    				IFileSystemWorkItemManager workItemManager = (IFileSystemWorkItemManager) repo.getClientLibrary(IFileSystemWorkItemManager.class);
    					workItemManager.createLink(changeSetReview.getSource().getResolvedWorkspace(), changeSetReview.getChangeSet(),
    							new IWorkItemHandle[] { workItemToAssign }, loopProgress.newChild(1));
	    			}
	    		}
	  	    }

	    	//  Collect change sets to close for review and suspend if specified
	    	ISuspendOperation op = IOperationFactory.instance.getSuspendOperation(baseProblemHandler.getSuspendDilemmaHandler());
			if(submitForReviewParameter.isSuspend()) {
	    		op.setRefreshBeforeRun(getRefreshBeforeRun());
	        	if (nonAtomicCommitIsAnOption) {
	        		op.enableNonAtomicCommit(nonAtomicCommitFileUploadLimit);
	        	} else {
	        		op.disableNonAtomicCommit();
	        	}
	        	// we already did the verify
		    	disableVerifyInSync(op);
			}
			
    		if(submitForReviewParameter.isSuspend()) {
    			for (Map.Entry<IWorkspaceConnection, List<IChangeSet>> entry : sourceToChangeSets.entrySet()) {
					op.suspend(entry.getKey(), entry.getValue());
	    		}
	    	}
	
	    	// Add approvals to work items
	    	submitChangesForReview(new ArrayList<IWorkItem>(uniqueWorkItems),
	    			submitForReviewParameter.getReviewers(),
	    			submitForReviewParameter.getComment(),
	    			submitForReviewParameter.getSubject(),
	    			submitForReviewParameter.getDueDate(),
	    			monitor.newChild(15));
	    	
	    	// suspend change sets
	    	if(submitForReviewParameter.isSuspend()) {
	    		op.run(monitor.newChild(15));
	    	}
	    	
	    	// close all change sets by source
	    	loopProgress = SubMonitor.convert(monitor.newChild(15));
	    	loopProgress.setWorkRemaining(sourceToChangeSets.size());
	    	for (Map.Entry<IWorkspaceConnection, List<IChangeSet>> entry : sourceToChangeSets.entrySet()) {
	    		entry.getKey().ensureClosed(entry.getValue(), loopProgress.newChild(1));
	    	}
        } finally {
            copyFileAreaLock.release(monitor.newChild(1));
        }
    }
    
	private void processInaccessibleWorkItems(List<IWorkItemHandle> inaccessibleWorkItems, List<IWorkItem> fullWorkItems, IProgressMonitor monitor) throws TeamRepositoryException {
		if (inaccessibleWorkItems.size() == 0)
			return;
		
		int action = this.baseProblemHandler.workItemsUnreachable(inaccessibleWorkItems.size(), fullWorkItems.size());
		
		if (action == IDilemmaHandler.CONTINUE) {
			// for now the work items are ignored
			for (Iterator<IWorkItem> i = fullWorkItems.iterator(); i.hasNext(); ) {
				IWorkItem wi = i.next();
				if (wi == null)
					i.remove();
			}
		} else if (action == IDilemmaHandler.FAIL) {
			throw new TeamRepositoryException(Messages.SubmitForReviewOperation_2);
		} else {
			throw new OperationCanceledException();
		}
	}

    private void submitChangesForReview(
    		List<IWorkItem> workItems,
    		IContributorHandle[] reviewers,
    		String comment,
    		String subject,
    		Timestamp dueDate,
            IProgressMonitor monitor) throws TeamRepositoryException {
        try {
            monitor.beginTask(Messages.SubmitForReviewOperation_0, 3);
            for (IWorkItem workItem : workItems) {
            	if (workItem == null )
            		continue;
            	WorkItemUtils.createReviewTask((ITeamRepository)workItem.getOrigin(), workItem, 
            			WorkItemApprovals.REVIEW_TYPE.getIdentifier(), subject, 
            			Arrays.asList(reviewers), true, comment, true, dueDate, monitor);
            }
        } finally {
            monitor.done();
        }
    }

	public void submitForReview(SubmitForReviewParameter reviewParameter) {
		this.submitForReviewParameter = reviewParameter;
	}

	public void disableNonAtomicCommit() {
		this.nonAtomicCommitIsAnOption = false;
		super.disableNonAtomicCommit();
	}

	public void enableNonAtomicCommit(int numberUploads) {
		this.nonAtomicCommitIsAnOption = true;
		this.nonAtomicCommitFileUploadLimit = numberUploads;
		super.enableNonAtomicCommit(numberUploads);
	}

	@Override
	protected void refreshLocal(IFilesystemRefresher refresher,
			Map<ISandbox, Set<ConfigurationFacade>> affectedAreas,
			IProgressMonitor progress) throws FileSystemException,
			TeamRepositoryException {

		for (Map.Entry<IWorkspaceConnection, List<IChangeSet>> entry : sourceToChangeSets.entrySet()) {
			ChangeSetRefreshUtils.refreshLocalForRemoval(refresher, affectedAreas, entry.getKey(), entry.getValue(), progress);
		}
	}
}
