/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2009, 2012. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  Use,
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.filesystem.client.workitems.operations;

import java.sql.Timestamp;
import java.util.Collection;
import java.util.Collections;

import com.ibm.team.repository.common.IContributorHandle;
import com.ibm.team.scm.client.IWorkspaceConnection;
import com.ibm.team.scm.common.IChangeSetHandle;
import com.ibm.team.scm.common.dto.IChangeHistorySyncReport;
import com.ibm.team.workitem.common.model.IWorkItemHandle;

/**
 * A class serving as a structure of parameters for {@link IDeliverAndFixOperation}.
 * The fields are final once the parameter has been constructed.
 */
public class DeliverAndFixParameter {
	final private String comment;
	final private String subject;
	final private IContributorHandle[] reviewers;
	final private IWorkspaceConnection source;
	final private IWorkspaceConnection target;
	final private Collection<IWorkItemHandle> workItems;
	final private IChangeHistorySyncReport syncReport;
	final private Collection<IChangeSetHandle> changeSets;
	final private boolean shouldResolve;
	final private Timestamp dueDate;

	/**
	 * @param changeSets
	 *            - Cannot be <code>null</code>. Unmodifiable after it is set.
	 * @param comment
	 *            - Comment for the work item.
	 * @param subject
	 *            - Review subject. May be <code>null</code>.
	 * @param reviewers
	 *            - The reviewers for <code>workItemToAssign</code>. May be
	 *            <code>null</code> if <code>addReview</code> is
	 *            <code>false</code>.
	 * @param workItems
	 *            - A {@link Collection} of work items that may be associated to
	 *            any of the change sets in <code>deliverAndResolveRequests</code>. 
	 *            These work items will be resolved if <code>shouldResolve</code> 
	 *            is <code>true</code>. Work items not associated with any of 
	 *            the change sets will be associated with all the change sets.
	 *            Cannot be <code>null</code>. Unmodifiable after it is set.
	 * @param syncReport
	 *            - Cannot be <code>null</code>.
	 * @param source
	 *            - source workspace changes are from
	 * @param target
	 *            - target workspace changes are going to
	 * @param shouldResolve - resolve the work item on deliver
	 * @param dueDate - the due date for the reviewers
     */
	public DeliverAndFixParameter(Collection<IChangeSetHandle> changeSets, String comment, String subject,
			IContributorHandle[] reviewers,
			Collection<IWorkItemHandle> workItems, IChangeHistorySyncReport syncReport,
			IWorkspaceConnection source, IWorkspaceConnection target, boolean shouldResolve,
			Timestamp dueDate) {
		super();
		this.changeSets = Collections.unmodifiableCollection(changeSets);
		this.comment = comment;
		this.subject = subject;
		this.reviewers = reviewers;
		this.workItems = Collections.unmodifiableCollection(workItems);
		this.syncReport = syncReport;
		this.source = source;
		this.target = target;
		this.shouldResolve = shouldResolve;
		this.dueDate = dueDate;
	}

	public String getComment() {
		return comment;
	}

	public String getSubject() {
		return subject;
	}

	public IContributorHandle[] getReviewers() {
		return reviewers;
	}

	public IWorkspaceConnection getSource() {
		return source;
	}

	public IWorkspaceConnection getTarget() {
		return target;
	}

	public Collection<IWorkItemHandle> getWorkItems() {
		return workItems;
	}

	public IChangeHistorySyncReport getSyncReport() {
		return syncReport;
	}

	public Collection<IChangeSetHandle> getChangeSets() {
		return changeSets;
	}

	public boolean isShouldResolve() {
		return shouldResolve;
	}

	public Timestamp getDueDate() {
		return this.dueDate;
	}
}
