/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2009, 2012. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  Use,
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.filesystem.client.workitems.operations;

import java.sql.Timestamp;
import java.util.Collection;

import com.ibm.team.repository.client.ITeamRepository;
import com.ibm.team.repository.common.IContributorHandle;
import com.ibm.team.scm.client.IWorkspaceConnection;
import com.ibm.team.scm.common.IChangeSetHandle;
import com.ibm.team.workitem.common.model.IWorkItemHandle;

/**
 * A class serving as a structure of parameters for {@link ISubmitForReviewOperation}.
 * The fields are final once the parameter has been constructed.
 */
public class SubmitForReviewParameter {
	private final String comment;
	private final Timestamp dueDate;
	private final IContributorHandle[] reviewers;
	private final boolean suspend;      
	private final String subject;
	private final Collection<IWorkItemHandle> workItems;
	private final Collection<ChangeSetReview> changeSetReviews;
	private final ITeamRepository repo;

	/**
	 * A change set to be submitted for review. 
	 */
	public static class ChangeSetReview {
		private final IChangeSetHandle changeSet;
		private final IWorkspaceConnection source;

		/**
		 * @param changeSet the change set for review, required
		 * @param source the source workspace of the change set, required
		 */
		public ChangeSetReview(IChangeSetHandle changeSet,
				IWorkspaceConnection source) {
			super();
			this.changeSet = changeSet;
			this.source = source;
		}

		public IChangeSetHandle getChangeSet() {
			return changeSet;
		}
		public IWorkspaceConnection getSource() {
			return source;
		}
	}

	/**
	 * @param comment The comment to add to the work items.
	 * @param reviewers The reviewers to add to the approval. Cannot be <code>null</code>.
	 * @param suspend Suspend the change sets for review.
	 * @param subject The approval subject. Cannot be <code>null</code>.
	 * @param workItemsToAssign The work items to associate with the change sets. Cannot be <code>null</code>.
	 * @param workItems The work items already associated with the change sets. Cannot be <code>null</code>.
	 * @param changeSetReviews The change sets for review with their source and component. Cannot be <code>null</code>.
	 * @param repo the given repository
	 * @param dueDate the given due date in milliseconds since midnight January 1, 1970 UTC
	 */
	public SubmitForReviewParameter(String comment,
			IContributorHandle[] reviewers, boolean suspend, String subject,
			Collection<IWorkItemHandle> workItems,
			Collection<ChangeSetReview> changeSetReviews, ITeamRepository repo,
			Timestamp dueDate) {
		super();
		this.comment = comment;
		this.reviewers = reviewers;
		this.suspend = suspend;
		this.subject = subject;
		this.workItems = workItems;
		this.changeSetReviews = changeSetReviews;
		this.repo = repo;
		this.dueDate = dueDate;
	}

	public String getComment() {
		return comment;
	}

	public IContributorHandle[] getReviewers() {
		return reviewers;
	}

	public boolean isSuspend() {
		return suspend;
	}

	public String getSubject() {
		return subject;
	}

	public Collection<IWorkItemHandle> getWorkItems() {
		return workItems;
	}

	public Collection<ChangeSetReview> getChangeSetReviews() {
		return changeSetReviews;
	}

	public ITeamRepository getRepo() {
		return repo;
	}

	public Timestamp getDueDate() {
		return dueDate;
	}
}
