/*******************************************************************************
 * Licensed Materials - Property of IBM 
 * (c) Copyright IBM Corporation 2009, 2012. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights: Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 ******************************************************************************/
package com.ibm.team.filesystem.client.workitems.rest.parameters;

import com.ibm.team.filesystem.client.internal.rest.IValidatingParameterWrapper;
import com.ibm.team.filesystem.client.internal.rest.ParmValidation;
import com.ibm.team.filesystem.client.rest.parameters.ParmsUpdateDilemmaHandler;
import com.ibm.team.filesystem.client.rest.parameters.ParmsContributorList;
import com.ibm.team.filesystem.client.rest.parameters.ParmsDeliverDilemmaHandler;
import com.ibm.team.filesystem.client.rest.parameters.ParmsOutOfSyncInstructions;
import com.ibm.team.filesystem.client.rest.parameters.ParmsPendingChangesDilemmaHandler;
import com.ibm.team.filesystem.client.rest.parameters.ParmsPreOperationRefresh;
import com.ibm.team.filesystem.client.rest.parameters.ParmsSandboxUpdateDilemmaHandler;
import com.ibm.team.filesystem.client.rest.parameters.ParmsWorkspace;

public final class ParmsPostDeliverAndResolve implements IValidatingParameterWrapper {
    
    /**
     * The change set id to which the work item will be associated. Required.
     */
    public String[] changeSets;

	/**
	 * Delivery source. Required.
	 */
    public ParmsWorkspace source;
    
	/**
	 * Delivery target. Required.
	 */
	public ParmsWorkspace target;
	
	/**
	 * The work items associated with the change sets. If a work item in this
	 * list is not associated with any of the change sets, it will be associated
	 * with all change sets (i.e. a new work item may be added to this list).
	 * Required.
	 */
	public ParmsWorkItem[] workItemIds;
	
	/**
	 * The reviewers to be assigned to the work items. May be <code>null</code>.
	 */
	public ParmsContributorList reviewers; 
	
	/**
	 * Adds a comment to the work items. Specify or empty string 
	 * <code>null</code> to post no comment. A non-zero length string will
	 * post the string to the work items.
	 */
	public String workItemComment;
	
	/**
	 * The subject added to the work item approval. An approval is added if
	 * <code>reviewers</code> is populated. Required if an approval will be created.
	 */
	public String approvalSubject;
	
    /**
     * Tells the receiver how to refresh the filesystem before the op runs.
     */
	public ParmsPreOperationRefresh preOpRefresh;

	// Dilemma handlers
    /**
     * Directions on what to do if projects/components that will be affected by the
     * operation are out of sync.  If not supplied the default
     * out of sync handler will be used (it will fail if problems detected).
     */
	public ParmsOutOfSyncInstructions outOfSyncHandler;

	/**
     * Directions on what to do component flows are detected 
     * during the deliver.
     */
	public ParmsDeliverDilemmaHandler deliverHandler;

    /**
     * Directions on what to do if when potential problems are detected when the 
     * history of the configuration is about to change.  If not 
     * supplied, the default handler will be used (will fail if problems encountered).
     */ 
    public ParmsPendingChangesDilemmaHandler pendingChangesHandler;
    
    /**
     * Instructions on how to update the sandbox when it is
     * being updated/deleted by the load/incidental conflict resolution
     * Optional.  If not supplied the default will be used.
     */
    public ParmsSandboxUpdateDilemmaHandler sandboxUpdateDilemmaHandler;

    /**
     * Directions on what to do when questions arise related to updating
     * the sandbox in response to the deliver re-ordering
     * as well as any resolved incidental conflicts if a check in is done.
     * If not supplied the defaults will be used.
     */
    public ParmsUpdateDilemmaHandler updateDilemmaHandler;
    
    /**
     * Directions on how to handle work items that are not reachable.
     * If not supplied, the default handler will be used (will fail if problems are encountered).
     */
    public ParmsWorkItemDilemmaHandler workItemHandler;
    
    /**
     * Should resolve all the given work items after delivery. Default is <code>true</code>.
     */
    public Boolean shouldResolve;
	
    /**
     * Sets the due date for the change sets to review. The due date is measured in milliseconds, between
     * the current time and midnight, January 1, 1970 UTC.
     * <p>If not set, the value is <code>null</code>.</p>
     */
    public Long dueDate;

    public void validate(String requestName, Object... path) {
		ParmValidation.requiredValue(source, requestName, path, "source"); //$NON-NLS-1$
		ParmValidation.requiredValue(target, requestName, path, "target"); //$NON-NLS-1$
		ParmValidation.requiredArray(changeSets, requestName, path, "changeSets"); //$NON-NLS-1$
		ParmValidation.requiredArray(workItemIds, requestName, path, "workItemIds"); //$NON-NLS-1$
		
		if (workItemComment == null) {
			workItemComment = ""; //$NON-NLS-1$
		}

		for (int i = 0; i < workItemIds.length; i++) {
			workItemIds[i].validate(requestName, path, "workItemIds"); //$NON-NLS-1$
		}
		
		source.validate(requestName, path, "source"); //$NON-NLS-1$
		target.validate(requestName, path, "target"); //$NON-NLS-1$
		
		if (!source.repositoryUrl.equalsIgnoreCase(target.repositoryUrl)){
			throw new IllegalArgumentException("Distributed flow not supported with deliver and resolve."); //$NON-NLS-1$
		}
		
		if (reviewers != null) {
			reviewers.validate(requestName, path, "reviewers"); //$NON-NLS-1$
			
			if (approvalSubject.length() == 0) {
				throw new IllegalArgumentException("Review subject cannot be zero length when adding a review."); //$NON-NLS-1$
			}
		}
		
		if (changeSets.length < 1) {
			throw new IllegalArgumentException("At least one change set needs to be set for review."); //$NON-NLS-1$
		}

		if (deliverHandler != null) {
			deliverHandler.validate(requestName, path, "deliverHandler"); //$NON-NLS-1$
		}
		
		if (outOfSyncHandler != null) {
			outOfSyncHandler.validate(requestName, path, "outOfSyncHandler"); //$NON-NLS-1$
		}
		
        if (pendingChangesHandler != null) {
            pendingChangesHandler.validate(requestName, path, "pendingChangesHandler"); //$NON-NLS-1$
        }
        
        if (sandboxUpdateDilemmaHandler != null) {
            sandboxUpdateDilemmaHandler.validate(requestName, path, "sandboxUpdateDilemmaHandler"); //$NON-NLS-1$
        }

        if (updateDilemmaHandler != null) {
            updateDilemmaHandler.validate(requestName, path, "updateDilemmaHandler"); //$NON-NLS-1$
        }

        if (preOpRefresh != null) {
        	preOpRefresh.validate(requestName, path, "preOpRefresh"); //$NON-NLS-1$
        }
        
        if (shouldResolve == null)
        	shouldResolve = Boolean.TRUE;
        
        if (dueDate != null) {
        	if (dueDate.longValue() <= 0) {
        		throw new IllegalArgumentException("The given due date value must be a positive value."); //$NON-NLS-1$
        	}
        }
	}
}