/*******************************************************************************
 * Licensed Materials - Property of IBM 
 * (c) Copyright IBM Corporation 2009, 2012. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights: Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 ******************************************************************************/
package com.ibm.team.filesystem.client.workitems.rest.parameters;

import java.util.Collection;
import java.util.HashSet;

import com.ibm.team.filesystem.client.internal.rest.IValidatingParameterWrapper;
import com.ibm.team.filesystem.client.internal.rest.ParmValidation;
import com.ibm.team.filesystem.client.rest.parameters.ParmsUpdateDilemmaHandler;
import com.ibm.team.filesystem.client.rest.parameters.ParmsContributorList;
import com.ibm.team.filesystem.client.rest.parameters.ParmsOutOfSyncInstructions;
import com.ibm.team.filesystem.client.rest.parameters.ParmsPendingChangesDilemmaHandler;
import com.ibm.team.filesystem.client.rest.parameters.ParmsPreOperationRefresh;
import com.ibm.team.filesystem.client.rest.parameters.ParmsSandboxUpdateDilemmaHandler;

public final class ParmsPostSubmitForReview implements IValidatingParameterWrapper {

	/**
	 * List of reviewers to add to the work item. Required.
	 */
	public ParmsContributorList reviewers;

	/**
	 * List of change sets to review grouped by the source workspace connection
	 * where they exist.
	 */
	public ParmsChangeSetsToSource[] changeSetsToReview;
	
	/**
	 * The work items associated with the change sets. If a work item in this
	 * list is not associated with any of the change sets, it will be associated
	 * with all change sets (i.e. a new work item may be added to this list).
	 * Required.
	 */
	public ParmsWorkItem[] workItemIds;

	/**
	 * Suspend the change sets as part of the operation. Default is <code>false</code>.
	 */
	public Boolean suspendChangeSets;

	/**
	 * Adds a comment to the work items. Specify empty string or 
	 * <code>null</code> to post no comment. A non-zero length string will
	 * post the string to the work items.
	 */
	public String workItemComment;
	
	/**
	 * The subject added to the work item approval. Required.
	 */
	public String approvalSubject;
	
    /**
     * Tells the receiver how to refresh the filesystem before the op runs.
     */
	public ParmsPreOperationRefresh preOpRefresh;
	
	// Dilemma handlers
	
    /**
     * Directions on what to do if projects/components that will be affected by the
     * operation are out of sync.  If not supplied the default
     * out of sync handler will be used (it will fail if problems detected).
     */
    public ParmsOutOfSyncInstructions outOfSyncInstructions;

    /**
     * Directions on what to do if when potential problems are detected when the 
     * history of the configuration is about to change.  If not 
     * supplied, the default handler will be used (will fail if problems encountered).
     */ 
    public ParmsPendingChangesDilemmaHandler pendingChangesHandler;
    
    /**
     * Instructions on how to update the sandbox when it is
     * being updated/deleted by the load/incidental conflict resolution
     * Optional.  If not supplied the default will be used.
     */
    public ParmsSandboxUpdateDilemmaHandler sandboxUpdateDilemmaHandler;

    /**
     * Directions on what to do when questions arise related to updating
     * the sandbox in response to the suspend
     * as well as any resolved incidental conflicts if a check in is done.
     * If not supplied the defaults will be used.
     */
    public ParmsUpdateDilemmaHandler updateDilemmaHandler;

    /**
     * Instructions on how to handle work items that are not reachable.
     * If not supplied, the default handler will be used (will fail if problems are encountered).
     */
    public ParmsWorkItemDilemmaHandler workItemHandler;
	
    /**
     * Sets the due date for the change sets to review. The due date is measured in milliseconds, between
     * the current time and midnight, January 1, 1970 UTC.
     * <p>If not set, the value is <code>null</code>.</p>
     */
    public Long dueDate;

    public void validate(String requestName, Object... path) {
		ParmValidation.requiredValue(reviewers, requestName, path, "reviewers"); //$NON-NLS-1$
		ParmValidation.requiredValue(approvalSubject, requestName, path, "approvalSubject"); //$NON-NLS-1$
		ParmValidation.requiredArray(workItemIds, requestName, path, "workItemIds"); //$NON-NLS-1$
		ParmValidation.requiredArray(changeSetsToReview, requestName, path, "changeSetsToReview"); //$NON-NLS-1$
		
		Collection<String> repoList = new HashSet<String>();
		
		for (int i = 0; i < workItemIds.length; i++) {
			workItemIds[i].validate(requestName, path, "workItemIds"); //$NON-NLS-1$
			repoList.add(workItemIds[i].repositoryUri);
		}
		
		for (int i = 0; i < changeSetsToReview.length; i++) {
			changeSetsToReview[i].validate(requestName, path, "changeSetsToReview"); //$NON-NLS-1$
			repoList.add(changeSetsToReview[i].source.repositoryUrl);
		}
		
		if (repoList.size() > 1) {
			throw new IllegalArgumentException("Submit for review supports only one repository for work items and change sets."); //$NON-NLS-1$
		}
		
		if (workItemComment == null) {
			workItemComment = ""; //$NON-NLS-1$
		}
		
		reviewers.validate(requestName, path, "reviewers"); //$NON-NLS-1$
		
		if (reviewers.contributorItemIds.length < 1) {
			throw new IllegalArgumentException("At least one reviewer is required."); //$NON-NLS-1$
		}
		if (approvalSubject.length() == 0) {
			throw new IllegalArgumentException("Review subject cannot be zero length when adding a review."); //$NON-NLS-1$
		}
		if (changeSetsToReview.length < 1) {
			throw new IllegalArgumentException("At least one change set needs to be set for review."); //$NON-NLS-1$
		}
		if (workItemIds.length < 1) {
			throw new IllegalArgumentException("At least one work item needs to be set for review."); //$NON-NLS-1$
		}
        
        if (pendingChangesHandler != null) {
            pendingChangesHandler.validate(requestName, path, "pendingChangesHandler"); //$NON-NLS-1$
        }
        
        if (outOfSyncInstructions != null) {
            outOfSyncInstructions.validate(requestName, path, "outOfSyncInstructions"); //$NON-NLS-1$
        }
        
        if (sandboxUpdateDilemmaHandler != null) {
            sandboxUpdateDilemmaHandler.validate(requestName, path, "sandboxUpdateDilemmaHandler"); //$NON-NLS-1$
        }

        if (updateDilemmaHandler != null) {
            updateDilemmaHandler.validate(requestName, path, "updateDilemmaHandler"); //$NON-NLS-1$
        }

        if (preOpRefresh != null) {
        	preOpRefresh.validate(requestName, path, "preOpRefresh"); //$NON-NLS-1$
        }
        
        if (suspendChangeSets == null)
        	suspendChangeSets = Boolean.FALSE;
        
        if (dueDate != null) {
        	if (dueDate.longValue() <= 0) {
        		throw new IllegalArgumentException("The given due date value must be a positive value."); //$NON-NLS-1$
        	}
        }
	}
	
}
