/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2006, 2012. All Rights Reserved. 
 * 
 * Note to U.S. Government Users Restricted Rights:  Use, 
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.filesystem.client.internal;

import java.io.File;
import java.io.IOError;
import java.io.IOException;
import java.nio.file.FileSystem;
import java.nio.file.FileSystems;
import java.nio.file.Files;
import java.nio.file.InvalidPathException;
import java.nio.file.LinkOption;
import java.nio.file.Path;
import java.nio.file.attribute.BasicFileAttributeView;
import java.nio.file.attribute.BasicFileAttributes;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;

import com.ibm.team.filesystem.client.FileSystemCore;
import com.ibm.team.filesystem.client.internal.SymLinkHelper;

class Java17SymLinkHelper extends SymLinkHelper {
    private static final String altSeparator = File.separator.equals("\\") ? "/" : null;  //$NON-NLS-1$ //$NON-NLS-2$
    final FileSystem fs;
    final boolean supportsLinks;
    
    static Java17SymLinkHelper newInstance() {
        return new Java17SymLinkHelper();
    }
    
    private Java17SymLinkHelper() {
        fs = FileSystems.getDefault();
        boolean supported = true;
        try {
            Files.readSymbolicLink(fs.getPath(System.getProperty("user.dir"))); //$NON-NLS-1$
        } catch (UnsupportedOperationException e) {
            supported = false;
        } catch (IOException e) {
            //ignore
        }
        supportsLinks = supported;
    }
    
    @Override
    public void createSymlink(String path, String targetPath) throws CoreException {
        try {
            Files.createSymbolicLink(fs.getPath(path), fs.getPath(targetPath));
        } catch (IOException e) {
            throw new CoreException(new Status(IStatus.ERROR, FileSystemCore.ID, e.getMessage(), e));
        } catch (UnsupportedOperationException e) {
            throw new CoreException(new Status(IStatus.ERROR, FileSystemCore.ID, e.getMessage(), e));
        } catch (InvalidPathException e) {
        	throw new CoreException(new Status(IStatus.ERROR, FileSystemCore.ID, e.getMessage(), e));
        }
    }

    @Override
    public String getResolvedPath(String path) throws CoreException {
        if (path.length() == 0) {
            throw new CoreException(new Status(IStatus.ERROR, FileSystemCore.ID, "The empty path is not valid")); //$NON-NLS-1$
        }
        
    	try {
	        Path p = fs.getPath(path).toAbsolutePath();
	        
	        Path parent = p.getParent();
	        Path existingParent = null;
	        while (parent != null) {
	            BasicFileAttributeView v = Files.getFileAttributeView(parent, BasicFileAttributeView.class, LinkOption.NOFOLLOW_LINKS);
	            BasicFileAttributes a;
	            try {
	                a = v.readAttributes();
	                if (!a.isDirectory()) {
	                    return null;
	                }
	                if (existingParent == null) {
	                    existingParent = parent;
	                }
	            } catch (IOException e) {
	                // probably doesn't exist, continue
	            }
	            parent = parent.getParent();
	        }
	        
	        
	        if (existingParent != null) {
	            Path relativePart = existingParent.relativize(p); 
	            
	            try {
	                existingParent = existingParent.toRealPath(LinkOption.NOFOLLOW_LINKS);
	            } catch (IOException e) {
	                throw new CoreException(new Status(IStatus.ERROR, FileSystemCore.ID, e.getMessage(), e));
	            }
	            p = existingParent.resolve(relativePart);
	        }
	        p = p.normalize();
	        
	        String result = p.toString();
	        if (path.endsWith(File.separator) || (altSeparator != null && path.endsWith(altSeparator))) {
	            BasicFileAttributeView v = Files.getFileAttributeView(p, BasicFileAttributeView.class, LinkOption.NOFOLLOW_LINKS);
	            BasicFileAttributes a;
	            try {
	                a = v.readAttributes();
	                if (!a.isDirectory()) {
	                    return null;
	                }
	            } catch (IOException e) {
	                // probably doesn't exist, continue
	            }
	            if (!result.endsWith(File.separator)) {
	                result += File.separator;
	            }
	        }
	        
	        return result;
    	} catch (IOError e) {
        	throw new CoreException(new Status(IStatus.ERROR, FileSystemCore.ID, e.getMessage(), e));
        } catch (InvalidPathException e) {
        	throw new CoreException(new Status(IStatus.ERROR, FileSystemCore.ID, e.getMessage(), e));
    	}
    }

    @Override
    public boolean supportsSymLinks() {
        return supportsLinks;
    }

	@Override
	public boolean hasEmbeddedTargetType() {
		// Java API's don't have anything for symbolic link types
		// Better hope that we don't need this on windows
		return false;
	}

	@Override
	public void createSymlink(String path, String targetPath,
			boolean isDirectory) throws CoreException {
		// Java API's don't have anything for symbolic link types
		// Better hope that we don't need this symlink helper on windows platforms
		createSymlink(path, targetPath);
	}

	@Override
	public String normalizeSymlinkTarget(String targetPath)
			throws CoreException {
		try {
			return FileSystems.getDefault().getPath(targetPath).toString();
		} catch (InvalidPathException e) {
			throw new CoreException(new Status(IStatus.ERROR, FileSystemCore.ID, e.getMessage(), e));
		}
	}
}
