/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <sys/stat.h>
#ifdef __APPLE__
#include <CoreFoundation/CoreFoundation.h>
#include <CoreServices/CoreServices.h>
#endif
#include "fsnatives.h"

#ifdef __MVS__
/* z/OS doesn't have dual mode large file support, it's all or nothing with _LARGE_FILES */
#define stat64 stat
#define lstat64 lstat
#endif

static jclass convertClass;
static jmethodID fromPlatformBytes;
#ifndef __APPLE__
static jmethodID toPlatformBytes;
#endif

static void handleError(JNIEnv *env);
static void throwCoreException(JNIEnv *env, char *message, int isModifiedUTF8);

JNIEXPORT jint JNICALL JNI_OnLoad(JavaVM *vm, void *reserved) {
  JNIEnv *env;
  jclass fileClass;
  jfieldID sepCharFieldId;

  if ((*vm)->GetEnv(vm, (void **)&env, JNI_VERSION_1_2) != JNI_OK) {
    return JNI_VERSION_1_2;
  }

#ifdef __MVS__
#pragma convert(819)
#endif
  fileClass = (*env)->FindClass(env, "java/io/File");
#ifdef __MVS__
#pragma convert(pop)
#endif
  if (fileClass == NULL) {
    return JNI_VERSION_1_2;
  }

#ifdef __MVS__
#pragma convert(819)
#endif
  sepCharFieldId = (*env)->GetStaticFieldID(env, fileClass, "separatorChar", "C");
#ifdef __MVS__
#pragma convert(pop)
#endif
  if (sepCharFieldId == NULL) {
    return JNI_VERSION_1_2;
  }

#ifdef __MVS__
#pragma convert(819)
#endif
  if ((*env)->GetStaticCharField(env, fileClass, sepCharFieldId) != '/') {
    throwCoreException(env, "Name separator char is not /", 1);
#ifdef __MVS__
#pragma convert(pop)
#endif
    return JNI_VERSION_1_2;
  }

#ifdef __MVS__
#pragma convert(819)
#endif
  convertClass = (*env)->FindClass(env, "org/eclipse/core/internal/filesystem/local/Convert");
#ifdef __MVS__
#pragma convert(pop)
#endif
  if (convertClass == NULL) {
    return JNI_VERSION_1_2;
  }

  convertClass = (*env)->NewGlobalRef(env, convertClass);
  if (convertClass == NULL) {
    return JNI_VERSION_1_2;
  }

#ifdef __MVS__
#pragma convert(819)
#endif
  fromPlatformBytes = (*env)->GetStaticMethodID(env, convertClass, "fromPlatformBytes", "([B)Ljava/lang/String;");
#ifdef __MVS__
#pragma convert(pop)
#endif
  if (fromPlatformBytes == NULL) {
    goto cleanup;
  }

#ifndef __APPLE__
#ifdef __MVS__
#pragma convert(819)
#endif
  toPlatformBytes = (*env)->GetStaticMethodID(env, convertClass, "toPlatformBytes", "(Ljava/lang/String;)[B");
#ifdef __MVS__
#pragma convert(pop)
#endif
  if (toPlatformBytes == NULL) {
    return JNI_VERSION_1_2;
  }
#endif
  return JNI_VERSION_1_2;

cleanup:
  (*env)->DeleteGlobalRef(env, convertClass);
  convertClass = NULL;
  fromPlatformBytes = NULL;
#ifndef __APPLE__
  toPlatformBytes = NULL;
#endif
  return JNI_VERSION_1_2;
}

JNIEXPORT void JNICALL JNI_OnUnload(JavaVM *vm, void *reserved) {
  JNIEnv *env;
  if (convertClass == NULL || (*vm)->GetEnv(vm, (void **)&env, JNI_VERSION_1_2) != JNI_OK) {
    return;
  }

  (*env)->DeleteGlobalRef(env, convertClass);
  convertClass = NULL;
  fromPlatformBytes = NULL;
#ifndef __APPLE__
  toPlatformBytes = NULL;
#endif
}


static jstring fromPlatformString(JNIEnv *env, char *str) {
  jstring result;
  jsize len;
  jbyteArray bytes;

  len = strlen(str);
  bytes = (*env)->NewByteArray(env, len);
  if (bytes == NULL) {
    return NULL;
  }

  (*env)->SetByteArrayRegion(env, bytes, 0, len, (jbyte *)str);
  if ((*env)->ExceptionOccurred(env) != NULL) {
    return NULL;
  }
  
  result = (*env)->CallStaticObjectMethod(env, convertClass, fromPlatformBytes, bytes);
  if ((*env)->ExceptionOccurred(env) != NULL) {
    return NULL;
  }

  return result;
}


#ifdef __APPLE__
static jstring getStringFromPlatformPath(JNIEnv *env, char *path) {
  CFStringRef cfsr = CFStringCreateWithFileSystemRepresentation(NULL, path);
  CFIndex len;
  int allocated;
  jchar *chars;
  jstring result;

  if (cfsr == NULL) {
    throwCoreException(env, "Failed to create CFString", 1);
    return NULL;
  }

  len = CFStringGetLength(cfsr);
  chars = (jchar *)CFStringGetCharactersPtr(cfsr);
  if (chars == NULL) {
    allocated = 1;
    chars = malloc(len * sizeof(jchar));
    CFStringGetCharacters(cfsr, CFRangeMake(0, len), chars);
  } else {
    allocated = 0;
  }

  result = (*env)->NewString(env, chars, len);
  if (allocated) {
    free(chars);
  }
  CFRelease(cfsr);

  return result;
}
#else
#define getStringFromPlatformPath(env, path) fromPlatformString((env), (path))
#endif

static char * getPathAsPlatformString(JNIEnv *env, jstring jpath) {
#ifdef __APPLE__
  jsize len;
  CFIndex resultSize;
  CFStringRef cfsr;
  char *result;
  const jchar * chars = (*env)->GetStringChars(env, jpath, NULL);
  if (chars == NULL) {
    return NULL;
  }
  len = (*env)->GetStringLength(env, jpath);
  cfsr = CFStringCreateWithCharacters(NULL, chars, len);
  if (cfsr == NULL) {
    (*env)->ReleaseStringChars(env, jpath, chars);
    throwCoreException(env, "Failed to create CFString", 1);
    return NULL;
  }
  resultSize = CFStringGetMaximumSizeOfFileSystemRepresentation(cfsr);
  result = malloc(resultSize);
  if (!CFStringGetFileSystemRepresentation(cfsr, result, resultSize)) {
    free(result);
    CFRelease(cfsr);
    (*env)->ReleaseStringChars(env, jpath, chars);
    throwCoreException(env, "Failed to convert to filesystem representation" , 1);
    return NULL;
  }
  CFRelease(cfsr);
  (*env)->ReleaseStringChars(env, jpath, chars);
  return result;
#else
  jbyteArray bytes;
  jsize len;
  char *result;
  bytes = (*env)->CallStaticObjectMethod(env, convertClass, toPlatformBytes, jpath);
  if ((*env)->ExceptionOccurred(env) != NULL) {
    return NULL;
  }
  len = (*env)->GetArrayLength(env, bytes);
  result = malloc(len + 1L);
  (*env)->GetByteArrayRegion(env, bytes, 0, len, (jbyte *)result);
  if ((*env)->ExceptionOccurred(env) != NULL) {
    free(result);
    return NULL;
  }
  result[len] = '\0';
  return result;
#endif
}

void JNICALL Java_com_ibm_team_filesystem_client_internal_NativeSymLinkHelper_createLink
    (JNIEnv *env, jclass cls, jstring joldPath, jstring jnewPath) {

  char * oldPath = getPathAsPlatformString(env, joldPath);
  char * newPath;
  int result;

  if (oldPath == NULL) {
    return;
  }

  newPath = getPathAsPlatformString(env, jnewPath);
  if (newPath == NULL) {
    goto cleanupOldPath;
  }

  result = symlink(oldPath, newPath);
  if (result != 0) {
    handleError(env);
  }

  free(newPath);
cleanupOldPath:
  free(oldPath);
}


static void handleError(JNIEnv *env) {
    int err = errno;
    size_t len = 256;
    char *errStr;
    int isUTF8 = 0;

    errStr = malloc(len);
    

    while (1) {
      int r = strerror_r(err, errStr, len);
      if (r == 0) break;

      if (errno == ERANGE) {
        len *= 2;
        errStr = realloc(errStr, len);
      } else {
#ifdef __MVS__
#pragma convert(819)
#endif
        sprintf(errStr, "Error number %d occured", err);
#ifdef __MVS__
#pragma convert(pop)
#endif
        isUTF8 = 1;
        break;
      }
    }

    throwCoreException(env, errStr, isUTF8);

    free(errStr);
}

#define ERROR_SEVERITY 4

static void throwCoreException(JNIEnv *env, char *message, int isModifiedUTF8) {
    jclass coreClass;
    jclass statusCls;
    jmethodID statusConstructor;
    jobject statusObj;
    jmethodID coreConstructor;
    jobject coreObj;
    jstring msg, pluginId;

#ifdef __MVS__
#pragma convert(819)
#endif
    coreClass = (*env)->FindClass(env, "org/eclipse/core/runtime/CoreException");
#ifdef __MVS__
#pragma convert(pop)
#endif
    if (coreClass == NULL) {
      return;
    }
#ifdef __MVS__
#pragma convert(819)
#endif
    statusCls = (*env)->FindClass(env, "org/eclipse/core/runtime/Status");
#ifdef __MVS__
#pragma convert(pop)
#endif
    if (statusCls == NULL) {
      return;
    }

#ifdef __MVS__
#pragma convert(819)
#endif
    statusConstructor = (*env)->GetMethodID(env, statusCls, "<init>", "(ILjava/lang/String;Ljava/lang/String;)V");
#ifdef __MVS__
#pragma convert(pop)
#endif
    if (statusConstructor == NULL) {
      return;
    }

    if (isModifiedUTF8) {
      msg = (*env)->NewStringUTF(env, message);
    } else {
      msg = fromPlatformString(env, message);
    }
    if (msg == NULL) {
      return;
    }

#ifdef __MVS__
#pragma convert(819)
#endif
    pluginId = (*env)->NewStringUTF(env, "com.ibm.team.filesystem.client");
#ifdef __MVS__
#pragma convert(pop)
#endif

    if (pluginId == NULL) {
      return;
    }

    statusObj = (*env)->NewObject(env, statusCls, statusConstructor, ERROR_SEVERITY, pluginId, msg);
    if (statusObj == NULL) {
      return;
    }

#ifdef __MVS__
#pragma convert(819)
#endif
    coreConstructor = (*env)->GetMethodID(env, coreClass, "<init>", "(Lorg/eclipse/core/runtime/IStatus;)V");
#ifdef __MVS__
#pragma convert(pop)
#endif
    if (coreConstructor == NULL) {
      return;
    }

    coreObj = (*env)->NewObject(env, coreClass, coreConstructor, statusObj);
    if (coreObj == NULL) {
      return;
    }

    (*env)->Throw(env, coreObj);
}


JNIEXPORT jstring JNICALL Java_com_ibm_team_filesystem_client_internal_NativeSymLinkHelper_getRealPath
  (JNIEnv * env, jclass cls, jstring jpath) {
  char *converted;
  char *path = getPathAsPlatformString(env, jpath);
  size_t pathlen;
  size_t srcIdx, dstIdx;
  int lastSegWasDirectory;
  int nonExistingSegments;
  jstring result;
#ifdef __APPLE__
  size_t lastExistingSegmentIdx;
#endif
  

  if (path == NULL) {
    return NULL;
  }

  pathlen = strlen(path);

  if (pathlen == 0) {
    free(path);
#ifdef __MVS__
#pragma convert(819)
#endif
    throwCoreException(env, "The empty path is not valid", 1);
#ifdef __MVS__
#pragma convert(pop)
#endif
    return NULL;
  }

  
  if (*path != '/') {
    size_t fullpathsz = 4096 > pathlen + 1 ? 4096 : pathlen + 4096;
    size_t fullpathlen;
    char *fullpath = malloc(fullpathsz);

    while (1) {
       char * result = getcwd(fullpath, fullpathsz);
       if (result != NULL) {
         break;
       }
       if (errno == ERANGE) {
         fullpathsz *= 2;
         fullpath = realloc(fullpath, fullpathsz);
       } else {
         handleError(env);
         free(fullpath);
         free(path);
         return NULL;
       }
    }

    fullpathlen = strlen(fullpath);
    if (fullpathlen + 2 + pathlen > fullpathsz) {
      fullpath = realloc(fullpath, fullpathlen + 2 + pathlen);
    }

    fullpath[fullpathlen] = '/';
    memcpy(fullpath + fullpathlen + 1, path, pathlen + 1);
    free(path);
    path = fullpath;
    pathlen += fullpathlen + 1;
  }


  converted = malloc(pathlen + 1);

  srcIdx = 0UL;
  dstIdx = 0UL;
  lastSegWasDirectory = 1;
  nonExistingSegments = 0;
#ifdef __APPLE__
  lastExistingSegmentIdx = 0UL;
#endif

  while (1) {
    size_t nextSeg;
    size_t seglen;
    if (srcIdx == pathlen) {
      break;
    }
    if (!lastSegWasDirectory) {
      free(path);
      free(converted);
      return NULL;
    }

    do {
      srcIdx++;
    } while (path[srcIdx] == '/');

    if (srcIdx == pathlen) {
      /* trailing separator */
      converted[dstIdx] = '/';
      dstIdx++;
      break;
    }

    for (nextSeg = srcIdx + 1; path[nextSeg] != '\0' && path[nextSeg] != '/'; nextSeg++);

    seglen = nextSeg - srcIdx;

    if (seglen == 2 && path[srcIdx] == '.' && path[srcIdx + 1] == '.') {
      if (dstIdx != 0UL) {
        while (1) {
          dstIdx--;
          if (converted[dstIdx] == '/') {
            break;
          }
        }
        if (nonExistingSegments != 0) {
          nonExistingSegments--;
        }
#ifdef __APPLE__
        if (dstIdx < lastExistingSegmentIdx) {
          lastExistingSegmentIdx = dstIdx;
        }
#endif
      }
    } else if (seglen != 1 || path[srcIdx] != '.') {
      memcpy(converted + dstIdx, path + srcIdx - 1, seglen + 1);
      dstIdx += seglen + 1;
      if (nonExistingSegments != 0) {
        nonExistingSegments++;
      } else {
        struct stat64 stbuf;

        converted[dstIdx] = '\0';
        while (1) {
          if (lstat64(converted, &stbuf) != 0) {
            int err = errno;
            if (err == ENOENT) {
              nonExistingSegments++;
              break;
            } else if (err != EINTR) {
              handleError(env);
              free(converted);
              free(path);
              return NULL;
            }
          } else {
            lastSegWasDirectory = S_ISDIR(stbuf.st_mode);
#ifdef __APPLE__
            lastExistingSegmentIdx = dstIdx;
#endif
            break;
          }
        }
      }
    }

    srcIdx = nextSeg;
  }


  free(path);

  if (dstIdx == 0) {
    converted[0] = '/';
    converted[1] = '\0';
  } else {
    converted[dstIdx] = '\0';
  }

#ifdef __APPLE__
  if (lastExistingSegmentIdx != 0UL) {
    FSRef fsr;
    OSStatus r;
    converted[lastExistingSegmentIdx] = '\0';
    r = FSPathMakeRefWithOptions((UInt8*)converted, kFSPathMakeRefDoNotFollowLeafSymlink, &fsr, NULL);
    if (r == noErr) {
      size_t newlen;
      size_t cclen = dstIdx + 1;
      char * casecorrect = malloc(cclen);
      while (1) {
        r = FSRefMakePath(&fsr, (UInt8*)casecorrect, cclen);
        if (r == noErr) {
          break;
        }
        if (r == pathTooLongErr) {
          cclen *= 2;
          casecorrect = realloc(casecorrect, cclen);
        } else {
          free(casecorrect);
          goto error_path;
        }
      }
      newlen = strlen(casecorrect);
      if (newlen + dstIdx - lastExistingSegmentIdx > pathlen) {
        converted = realloc(converted, newlen + dstIdx - lastExistingSegmentIdx + 1);
      }
      if (dstIdx != lastExistingSegmentIdx) {
        memmove(converted + newlen + 1, converted + lastExistingSegmentIdx + 1,
                 dstIdx - lastExistingSegmentIdx);
        converted[newlen] = '/';
      }
      memcpy(converted, casecorrect, newlen);
      free(casecorrect);
    } else {
error_path:
      /*some error creating the reference, ignore it and return the
        original path we found*/
      if (dstIdx != lastExistingSegmentIdx) {
        converted[lastExistingSegmentIdx] = '/';
      }
    }
  }
#endif

  result = getStringFromPlatformPath(env, converted);
  
  free(converted);
  return result;
}


JNIEXPORT jboolean JNICALL Java_com_ibm_team_filesystem_client_internal_NativeSymLinkHelper_nativeSupportsSymlinks
  (JNIEnv * env, jclass cls) {
  return JNI_TRUE;
}

