/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2012. All Rights Reserved.
 *
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/inotify.h>
#include <dirent.h>
#include <unistd.h>
#include <errno.h>
#include <pthread.h>
#include <time.h>

#include "listener.h"

#define MAX_EVENTS_PER_BUFFER 100
#define ADD_NOTIFY_TIMEOUT_MICRO 10000

#define EVENT_TYPE_SIMULATED 0x20
static pthread_mutex_t event_queue_mutex;
static struct evt* event_queue;

/* Data necessary to reconstruct fully-qualified paths from
   inotify events */
struct pathinfo {
	int wd;
	char* path;
	struct pathinfo* next;
} pathinfo_t;

static pthread_mutex_t directories_mutex;
static struct pathinfo* directories;
static struct pathinfo* get_pathinfo(int wd);

/* The primary file descriptor for reading events from inotify */
static int inotify_fd;

static struct evt* process_subdirectory_added(const char* subdirectory, struct evt* event_chain);
static void process_subdirectory_deleted(char *subdirectory);

/* Main functions for various threads */
static void* process_inotify_events(void* id);

/* Utility functions */
static int strprefix(const char* prefix, const char* instance) {
	int i=0;
	
	while(1) {
		if (prefix[i] != instance[i]) {
			return 0;
		}
		
		if (prefix[i] == 0) {
			return 1;
		}
		
		i++;
	}
	
	return 1;
}

void init() {
	pthread_t *reader = malloc(sizeof(pthread_t));
	pthread_mutexattr_t *mutexattr = malloc(sizeof(pthread_mutexattr_t));
	pthread_mutexattr_init(mutexattr);
	pthread_mutexattr_settype(mutexattr,PTHREAD_MUTEX_DEFAULT);
	pthread_mutex_init(&directories_mutex,mutexattr);
	directories = NULL;
	
	mutexattr = malloc(sizeof(pthread_mutexattr_t));
	pthread_mutexattr_init(mutexattr);
	pthread_mutexattr_settype(mutexattr,PTHREAD_MUTEX_DEFAULT);
	pthread_mutex_init(&event_queue_mutex,mutexattr);
	event_queue = NULL;
	 
	inotify_fd = inotify_init();
	
	/* Create the reader thread that will process any incoming notifications
	   from inotify and build up the event queue */
	pthread_create(reader,NULL,process_inotify_events,NULL);
}

static void* process_inotify_events(void* id) {
	int read_buffer_size = sizeof(struct inotify_event)*MAX_EVENTS_PER_BUFFER;
	char* buffer = malloc(read_buffer_size);
	struct inotify_event* event = NULL;
		
	while (1) {	
		int read_size = read(inotify_fd, buffer, read_buffer_size);
		int pos = 0;
		
		if (read_size <= 0) {
			/* Weird that this read failed, sleep and give it another try */
			printf("Read failed!\n");
			sleep(1);
			continue;
		} else {
			while (pos < read_size && pos < read_buffer_size) {
				struct pathinfo* parent;
				event = (struct inotify_event*)(buffer+pos);
		
				parent = get_pathinfo(event->wd);
				
				if (parent == NULL) {
					printf("Could not acquire parent information. Skipping this event for %s\n",event->name);
				} else if (event->len == 0) {
					if (!(event->mask & IN_IGNORED)) {
						printf("Not an ignored event and also no file name. Weird...\n");
					}
				} else {
					char* childpath = NULL;
					char type = 0x00;
					
					asprintf(&childpath,"%s/%s",parent->path,event->name);
					free(parent->path);
					free(parent);
		
					if (event->mask & IN_CREATE
						|| event->mask & IN_MOVED_TO) {
						type = EVENT_TYPE_ADD;
					} else if (event->mask & IN_MODIFY) {
						type = EVENT_TYPE_MODIFY;
					} else if (event->mask & IN_DELETE
							|| event->mask & IN_MOVED_FROM) {
						type = EVENT_TYPE_DELETE;
					}
					
					if (event->mask & IN_ISDIR) {
						type |= FILE_TYPE_DIRECTORY;
					} else {
						type |= FILE_TYPE_FILE;
					}
					
					/* Push this new path into the event queue. */
					pthread_mutex_lock(&event_queue_mutex);
					if (event_queue == NULL) {
						event_queue = malloc(sizeof(struct evt));
						event_queue->path = childpath;
						event_queue->type = type;
						event_queue->next = NULL;	
					} else {
						/* TODO Perhaps we should try to bound the size of this event queue?*/
						struct evt* i = event_queue;
						while (i != NULL) {
							if (i->next == NULL) {
								i->next = malloc(sizeof(struct evt));
								i->next->path = childpath;
								i->next->type = type;
								i->next->next = NULL;
								break;
							}
							
							i = i->next;
						}
					}
					pthread_mutex_unlock(&event_queue_mutex);
				}
				
				/* Skip over this event in the buffer and after the characters at the end
				    that form the file name in the event */
				pos += sizeof(struct inotify_event)+event->len;
			}
		}
    }
    
    return NULL;
}

struct evt* getNextChange() {
	struct evt* ret = NULL;
	
    pthread_mutex_lock(&event_queue_mutex);
    
    if (event_queue != NULL) {
    	ret = event_queue;
    	event_queue = event_queue->next;
    }
    
    pthread_mutex_unlock(&event_queue_mutex);
    
    /** TODO Find a better way to create a synchronized queue using pthreads */
    /* Sleep for a bit if we found no events so that we don't spin at 100% of CPU */
    if (ret == NULL) {
    	sleep(1);
    } else if ((ret->type & FILE_TYPE_DIRECTORY) && (ret->type & EVENT_TYPE_SIMULATED) == 0) {
    	/* If this is a directory change then we have some work to do */
    	if (ret->type & EVENT_TYPE_ADD) {
    		struct evt *root = malloc(sizeof(struct evt));
    		struct evt *last = NULL;
    		root->next = NULL;
    		root->type = 0x00;
    		root->path = NULL;
    		last = process_subdirectory_added(ret->path,root);
    		
    		if (root != last) {
    			struct evt* i = NULL;
    			pthread_mutex_lock(&event_queue_mutex);
    			if (event_queue == NULL) {
    				event_queue = root->next;
    			} else {
    				i = event_queue;
    				while (i->next != NULL) {
    					i = i->next;
    				}
    				i->next = root->next;
    			}
    			pthread_mutex_unlock(&event_queue_mutex);
    		}
    	} else if (ret->type & EVENT_TYPE_DELETE) {
    		process_subdirectory_deleted(ret->path);
    	}
    }
    
    return ret;
}

static struct pathinfo* get_pathinfo(int wd) {
    struct pathinfo* i = NULL;
    struct pathinfo* match = NULL;
    struct pathinfo* ret = NULL;
    
    pthread_mutex_lock(&directories_mutex);
    
    i = directories;
    while (i != NULL) {
       if (i->wd == wd) {
          match = i;
          break;
       }

       i = i->next;
    }
    
    /* Copy-construct a new path info to be freed by the caller to avoid
        cases where the entry is freed by another thread after this method is
        exited */
    if (match != NULL) {
    	ret = malloc(sizeof(struct pathinfo));
    	ret->path = strdup(match->path);
    	ret->wd = match->wd;
    }
    
    pthread_mutex_unlock(&directories_mutex);

    return ret;
}

static struct evt* process_subdirectory_added(const char* subdirectory, struct evt* event_chain) {
	int wd;
	time_t snapshot;
	DIR* dir;
	struct pathinfo* newinfo = malloc(sizeof(struct pathinfo));
	
	pthread_mutex_lock(&directories_mutex);
	wd = inotify_add_watch(inotify_fd, subdirectory, IN_MODIFY | IN_CREATE | IN_DELETE | IN_MOVE);
	
	newinfo->wd = wd;
	newinfo->path = strdup(subdirectory);
	newinfo->next = NULL;
	if (directories == NULL) {
		directories = newinfo;
	} else {
		struct pathinfo* i = directories;
		struct pathinfo* prev = NULL;
		int exists = 0;
		while (i != NULL) {
			if (i->wd == wd) {
				exists = 1;
				break;
			}
			prev = i;
			i = i->next;
		}
	
		if (!exists) {	
			prev->next = newinfo;
		} else {
			free(newinfo);
		}
	}
	
	pthread_mutex_unlock(&directories_mutex);

	/* Only if we are intercepting events do we need to sleep */
	if (event_chain != NULL) {
		/* Sleep to allow the inotify watch to start up */
		usleep(ADD_NOTIFY_TIMEOUT_MICRO);
		snapshot = time(NULL);
	}
	
	dir = opendir(subdirectory);
	
	if (dir != NULL) {
		struct dirent* child = readdir(dir);
		
		while (child != NULL) {
			if (child->d_name[0] != '.'
				|| (child->d_name[1] != 0 && child->d_name[1] != '.')) {
				
				char* childpath;
				struct stat* childinfo = malloc(sizeof(struct stat));
				char type = 0x00;
				int stat_result = 0;
				asprintf(&childpath,"%s/%s",subdirectory,child->d_name);
				stat_result = stat(childpath,childinfo);
				if (stat_result != -1) { /* File or directory disappeared on us */
					if (S_ISDIR(childinfo->st_mode)) {
						type = FILE_TYPE_DIRECTORY | EVENT_TYPE_ADD | EVENT_TYPE_SIMULATED;
						event_chain = process_subdirectory_added(childpath,event_chain);
					} else {
						type = FILE_TYPE_FILE | EVENT_TYPE_ADD | EVENT_TYPE_SIMULATED;
					}
					
					if (event_chain != NULL && difftime(childinfo->st_ctime, snapshot) <= 0.0) {
						struct evt* new_event = malloc(sizeof(struct evt));
						new_event->path = childpath;
						new_event->type = type;
						new_event->next = NULL;
						event_chain->next = new_event;
						event_chain = new_event;
					} else {
						free(childpath);
					}
				}
				
				free(childinfo);
			}
			
			child = readdir(dir);
		}
	
		closedir(dir);
	}
	
	return event_chain;
}

static void process_subdirectory_deleted(char *subdirectory) {
	struct pathinfo* i = NULL;
	struct pathinfo* prev = NULL;
	char* prefix_to_match;
	
	asprintf(&prefix_to_match,"%s/",subdirectory);
	
	pthread_mutex_lock(&directories_mutex);
	i = directories;
	while (i != NULL) {
		struct pathinfo* path_to_free = NULL;
		if (strcmp(subdirectory,i->path) == 0
			|| strprefix(prefix_to_match,i->path)) {
			
			path_to_free = i;
			
			if (prev == NULL) {
				directories = i->next;
			} else {
				prev->next = i->next;
			}
		}
		
		prev = i;
		i = i->next;
		
		if (path_to_free != NULL) {
			inotify_rm_watch(inotify_fd, path_to_free->wd);
			free(path_to_free->path);
			free(path_to_free);
		}
	}
	
	
	pthread_mutex_unlock(&directories_mutex);
	
	free(prefix_to_match);
}

void watch_all_subdirs(const char* rootdir) {
	/* Caller doesn't care about receiving events for these things but wants to start listening */
	process_subdirectory_added(rootdir, NULL);
}

void unwatch_all_subdirs(const char* rootdir) {
	/* Treat this directory as if it is deleted. No events are generated from this */
	process_subdirectory_deleted(rootdir);
}
