/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2012. All Rights Reserved.
 *
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <sys/types.h>
#include <dirent.h>
#include <pthread.h>
#include <unistd.h>
#include <CoreFoundation/CoreFoundation.h>
#include <CoreServices/COREServices.h>

#include "listen.h"

static struct evt* event_queue = NULL;
static pthread_mutex_t event_queue_mutex;
static CFRunLoopRef runloop = NULL;

struct dirinfo {
	char* dirpath;
	int numchild;
	struct dirent** children;
	struct dirinfo* next;
	FSEventStreamRef stream;
} dirinfo_t;

struct dirinfo* root = NULL;
static pthread_mutex_t dirinfo_mutex;

void scandirectory(char* path, FSEventStreamRef stream);

void fsevent_cb(
				ConstFSEventStreamRef streamRef,
				void * info,
				size_t events_size,
				void *paths,
				const FSEventStreamEventFlags flags[],
				const FSEventStreamEventId eventIds[])
{
	char** charpaths = paths;
	int i;
	for (i=0; i<events_size; i++) {
		char* dirpath;
		int dirpathlen;
		asprintf(&dirpath,"%s",charpaths[i]);
		dirpathlen = strlen(dirpath);
		dirpath[dirpathlen-1] = '\0';
		//printf("Received an event on path:%s!\n",charpaths[i],eventIds[i],flags[i]);
		pthread_mutex_lock(&dirinfo_mutex);
		scandirectory(dirpath,NULL);
		pthread_mutex_unlock(&dirinfo_mutex);
	}
}

int filter_dot_dirs(struct dirent * ent) {
	if (ent->d_name[0] == '.'
		&& ent->d_namlen == 1) {
		return 0;
	}
	
	if (ent->d_name[0] == '.' && ent->d_name[1] == '.'
		&& ent->d_namlen == 2) {
		return 0;
	}
	
	return 1;
}


void compare(char* path, int num_entries, struct dirent** contents, struct dirinfo* existing) {
	// Everything is new because we've never seen this parent directory before
	if (existing == NULL || existing->numchild == 0) {
		int i=0;
		for (i=0; i<num_entries; i++) {
			if (contents[i]->d_type & DT_DIR) {
				char* childpath;
				asprintf(&childpath,"%s/%s",path,contents[i]->d_name);
				scandirectory(childpath,NULL);
				free(childpath);
			}
		}
	} else {
		// TODO
		// This is an awful nearly n^2 search, optimize by utilizing the fact that both are sorted
		// Also, there is no check for deleted directory case to do a cleanup
		int i=0;
		for (i=0; i<num_entries; i++) {
			int j=0;
			int foundExisting = 0;
			
			if (!(contents[i]->d_type & DT_DIR)) {
				continue;
			}
			
			for (j=0;j<existing->numchild; j++) {
				if (strcmp(contents[i]->d_name,existing->children[j]->d_name) == 0) {
					foundExisting = 1;
				}
			}
			
			if (!foundExisting) {
				char* childpath;
				asprintf(&childpath,"%s/%s",path,contents[i]->d_name);
				scandirectory(childpath,NULL);
				free(childpath);
			}
		}
	}
}

void scandirectory(char* path, FSEventStreamRef stream) {
	int num_entries;
	struct dirent** contents;
	int i=0;
	struct dirinfo* existing = root;
	
	num_entries= scandir(path,&contents,filter_dot_dirs,alphasort);
	
	while (existing != NULL) {
		if (strcmp(existing->dirpath,path) == 0) {
			break;
		}
		
		existing = existing->next;
	}
	
	compare(path,num_entries,contents,existing);
	
	// Create a new directory info for this
	if (existing == NULL) {
		existing = malloc(sizeof(struct dirinfo));
		asprintf(&(existing->dirpath),"%s",path);
		existing->next = NULL;
		
		if (root == NULL) {
			root = existing;
		} else {
			struct dirinfo* last = root;
			while (last->next != NULL) {
				last = last->next;
			}
			
			last->next = existing;
		}
	} else {
		if (stream == NULL) {
			pthread_mutex_lock(&event_queue_mutex);
			struct evt* end = event_queue;
			while (end != NULL && end->next != NULL) {
				end = end->next;
			}
			
			for (i=0; i<existing->numchild; i++) {
				struct evt* e = malloc(sizeof(struct evt));
				asprintf(&(e->path),"%s/%s",existing->dirpath,existing->children[i]->d_name);
				e->next = NULL;
				//printf("Event on %s/%s\n",existing->dirpath,existing->children[i]->d_name);
				if (end == NULL) {
					event_queue = e;
				} else {
					end->next = e;
				}
				
				end = e;
			}
			pthread_mutex_unlock(&event_queue_mutex);
		}
		
		// Merge info into existing directory info element
		for (i=0; i<existing->numchild; i++) {
			free(existing->children[i]);
		}
		//free(existing->children); TODO Why does freeing this pointer cause seg faults?
	}
	
	existing->numchild = num_entries;
	existing->children = contents;
	if (stream != NULL) {
		existing->stream = stream;
	}
	
	if (stream == NULL) {
		pthread_mutex_lock(&event_queue_mutex);
		struct evt* end = event_queue;
		while (end != NULL && end->next != NULL) {
			end = end->next;
		}
		
		for (i=0; i<existing->numchild; i++) {
			struct evt* e = malloc(sizeof(struct evt));
			asprintf(&(e->path), "%s/%s",existing->dirpath,existing->children[i]->d_name);
			e->next = NULL;
			//printf("Event on %s/%s\n",existing->dirpath,existing->children[i]->d_name);
			if (end == NULL) {
				event_queue = e;
			} else {
				end->next = e;
			}
			end = e;
		}
		
		pthread_mutex_unlock(&event_queue_mutex);
	}
}

void* process_events(void* v) {
	runloop = CFRunLoopGetCurrent();
	while (1) {
		CFRunLoopRun();
	}
	printf("Run loop has stopped\n");
	return NULL;
}

void init() {
	pthread_t *reader = malloc(sizeof(pthread_t));
	pthread_mutexattr_t *mutexattr = malloc(sizeof(pthread_mutexattr_t));
	pthread_mutexattr_init(mutexattr);
	pthread_mutexattr_settype(mutexattr,PTHREAD_MUTEX_DEFAULT);
	pthread_mutex_init(&event_queue_mutex,mutexattr);
	event_queue = NULL;
	
	mutexattr = malloc(sizeof(pthread_mutexattr_t));
	pthread_mutexattr_init(mutexattr);
	pthread_mutexattr_settype(mutexattr,PTHREAD_MUTEX_DEFAULT);
	pthread_mutex_init(&dirinfo_mutex,mutexattr);
	
	pthread_create(reader,NULL,process_events,NULL);
	
	while (runloop == NULL) {
		sleep(1);
	}
}

void listen(CFStringRef path) {
	CFArrayRef paths = CFArrayCreate(NULL, (const void**)&path,1,NULL);
	char* pathAsPosix = malloc(sizeof(char)*1024);
	void *cbackinfo = NULL;
	FSEventStreamRef stream;
	CFAbsoluteTime delay = 1.0;
	
	stream = FSEventStreamCreate(NULL,
								 fsevent_cb,
								 cbackinfo,
								 paths,
								 kFSEventStreamEventIdSinceNow,
								 delay,
								 kFSEventStreamCreateFlagNone);
	
	CFStringGetFileSystemRepresentation(path,pathAsPosix,1024);
	pthread_mutex_lock(&dirinfo_mutex);
	scandirectory(pathAsPosix,stream);
	pthread_mutex_unlock(&dirinfo_mutex);
	
	FSEventStreamScheduleWithRunLoop(stream, runloop ,kCFRunLoopDefaultMode);
	FSEventStreamStart(stream);
	
	printf("Listening on %s\n",pathAsPosix);
	free(pathAsPosix);
}

void unlisten(char* path) {
	struct dirinfo* info;
	pthread_mutex_lock(&dirinfo_mutex);
	info = root;
	while (info != NULL && strcmp(info->dirpath,path) != 0) {
		info = info->next;
	}
	
	if (info != NULL) {
		FSEventStreamStop(info->stream);
		FSEventStreamInvalidate(info->stream);
		info->stream = NULL;
	}
	
	// TODO Remove all of the dirinfo's for this directory to clean up the memory
	pthread_mutex_unlock(&dirinfo_mutex);
	
	printf("Not listening on %s\n",path);
}

struct evt* getNextChange() {
	struct evt* next = NULL;
	pthread_mutex_lock(&event_queue_mutex);
	
	next = event_queue;
	if (event_queue != NULL) {
		event_queue = event_queue->next;
	}
	
	pthread_mutex_unlock(&event_queue_mutex);
	
	if (next == NULL) {
		sleep(1);
	}
	
	return next;
}
