/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2012. All Rights Reserved.
 *
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/
#include "listener.h"
#include "com_ibm_team_filesystem_client_internal_listen_LocalFilesystemListener.h"

#define ERROR_SEVERITY 4
#define LISTENER_JAVA_CLASS "com/ibm/team/filesystem/client/internal/listen/LocalFilesystemListener"

void record_error(void* ctx, wchar_t* message, int thread);

void throwCoreException(JNIEnv* env, jstring msg) {
    jclass coreClass;
    jclass statusCls;
    jmethodID statusConstructor;
    jobject statusObj;
    jmethodID coreConstructor;
    jobject coreObj;
    jstring pluginId;

    coreClass = (*env)->FindClass(env, "org/eclipse/core/runtime/CoreException");

    if (coreClass == NULL) {
#ifdef TRACE
    	printf("Could not throw exception because core exception class could not be found.\n");
    	fflush(stdout);
#endif
		return;
    }

    statusCls = (*env)->FindClass(env, "org/eclipse/core/runtime/Status");

    if (statusCls == NULL) {
#ifdef TRACE
    	printf("Could not throw exception because status class could not be found.\n");
    	fflush(stdout);
#endif
		return;
    }

    statusConstructor = (*env)->GetMethodID(env, statusCls, "<init>", "(ILjava/lang/String;Ljava/lang/String;)V");

    if (statusConstructor == NULL) {
#ifdef TRACE
    	printf("Could not throw exception because status constructor could not be found.\n");
    	fflush(stdout);
#endif
		return;
    }

    pluginId = (*env)->NewStringUTF(env, "com.ibm.team.filesystem.client");

    if (pluginId == NULL) {
#ifdef TRACE
    	printf("Could not throw exception because java string could not be constructed.\n");
    	fflush(stdout);
#endif
		return;
    }

    statusObj = (*env)->NewObject(env, statusCls, statusConstructor, ERROR_SEVERITY, pluginId, msg);
    if (statusObj == NULL) {
#ifdef TRACE
    	printf("Could not throw exception because status object could not be constructed.\n");
    	fflush(stdout);
#endif
		return;
    }

    coreConstructor = (*env)->GetMethodID(env, coreClass, "<init>", "(Lorg/eclipse/core/runtime/IStatus;)V");

    if (coreConstructor == NULL) {
#ifdef TRACE
    	printf("Could not throw exception because core exception constructor could not be found.\n");
    	fflush(stdout);
#endif
		return;
    }

    coreObj = (*env)->NewObject(env, coreClass, coreConstructor, statusObj);
    if (coreObj == NULL) {
#ifdef TRACE
    	printf("Could not throw exception because core exception could not be constructed.\n");
    	fflush(stdout);
#endif
		return;
    }

    (*env)->Throw(env, coreObj);
}

void reportErrorInJava(JNIEnv* env, jstring msg) {
    jmethodID reportErrorMethod;
    jclass listenerCls;
    
    listenerCls = (*env)->FindClass(env, LISTENER_JAVA_CLASS);
    if (listenerCls == NULL) {
#ifdef TRACE
    	printf("Cannot report error because listener class could not be found.\n");
    	fflush(stdout);
#endif
    	return;
    }

    reportErrorMethod = (*env)->GetStaticMethodID(env, listenerCls, "addListenerError", "(Ljava/lang/String;)V");

    if (reportErrorMethod == NULL) {
#ifdef TRACE
    	printf("Cannot report error because error reporter method could not be found on the listener.\n");
    	fflush(stdout);
#endif
    	return;
    } 

    (*env)->CallStaticVoidMethod(env, listenerCls, reportErrorMethod, msg);
}

void callback(void* ctx, struct event* event) {
	JNIEnv* env = (JNIEnv*)ctx;
	jobjectArray array = NULL;
	jclass stringCls = NULL;
	jclass listenerCls = NULL;
	jmethodID newEventsCallback = NULL;
	
	struct event* last = event;
	jsize num_events = 0;
	int idx = 0;
	
	if (event == NULL) {
		return;
	}
	
	listenerCls = (*env)->FindClass(env, LISTENER_JAVA_CLASS);
	newEventsCallback = (*env)->GetStaticMethodID(env, listenerCls, "newEvents", "([Ljava/lang/String;)V");
	
	stringCls = (*env)->FindClass(env, "java/lang/String");
	
	if (stringCls == NULL) {
		record_error((void*)env, L"Cannot find the string class in the filesystem listener.", THREAD_MINE);
		return;
	}

	while (last != NULL) {
		++num_events;
		last = last->next;
	}
	
	array = (*env)->NewObjectArray(env, num_events, stringCls, (*env)->NewStringUTF(env,""));
	
	last = event;
	while (last != NULL) {
		wchar_t* path = last->path;
		struct event* next = last->next;
		char* buffer = NULL;
		jstring string = NULL;
		int length = 0;
		
		next = last->next;
		free(last);
		last = next;
	
		length = WideCharToMultiByte(CP_UTF8,
						0,
						path,
						-1,
						NULL,
						0,
						NULL,
						NULL);
	
		buffer = (char*)malloc(length+1);
		
		WideCharToMultiByte(CP_UTF8,
					0,
					path,
					-1,
					buffer,
					length,
					NULL,
					NULL);

		buffer[length] = 0;
	
		string = (*env)->NewStringUTF(env, buffer);
	
		free(buffer);
		free(path);
		
		(*env)->SetObjectArrayElement(env, array, idx++, string);
	}
	
	(*env)->CallStaticVoidMethod(env, listenerCls, newEventsCallback, array);
}

void record_error(void* ctx, wchar_t* message, int thread) {
	JNIEnv* env = (JNIEnv*)ctx;
	char* buffer = NULL;
	jstring msg = NULL;
	int length = 0;

	length = WideCharToMultiByte(CP_UTF8,
					0,
					message,
					-1,
					NULL,
					0,
					NULL,
					NULL);

	buffer = (char*)malloc(length+1);

	WideCharToMultiByte(CP_UTF8,
				0,
				message,
				-1,
				buffer,
				length,
				NULL,
				NULL);

	buffer[length] = 0;
	
	msg = (*env)->NewStringUTF(env, buffer);

	free(buffer);

	if (thread == THREAD_YOURS) {
		throwCoreException(env, msg);
	} else if (thread == THREAD_MINE) {
		reportErrorInJava(env, msg);
	}
}

/*
 * Class:     com_ibm_team_filesystem_client_internal_listen_LocalFilesystemListener
 * Method:    init
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_com_ibm_team_filesystem_client_internal_listen_LocalFilesystemListener_init
  (JNIEnv *env, jclass cls)
{
	init((void*)env,&callback,&record_error);
}

/*
 * Class:     com_ibm_team_filesystem_client_internal_listen_LocalFilesystemListener
 * Method:    init
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_com_ibm_team_filesystem_client_internal_listen_LocalFilesystemListener_enterEventLoop
  (JNIEnv *env, jclass cls)
{
	collectChanges((void*)env);
}

/*
 * Class:     com_ibm_team_filesystem_client_internal_listen_LocalFilesystemListener
 * Method:    dispose
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_com_ibm_team_filesystem_client_internal_listen_LocalFilesystemListener_dispose
  (JNIEnv *env, jclass cls)
{	
	dispose();
}

/*
 * Class:     com_ibm_team_filesystem_client_internal_listen_LocalFilesystemListener
 * Method:    addDirectory
 * Signature: (Ljava/lang/String;)V
 */
JNIEXPORT void JNICALL Java_com_ibm_team_filesystem_client_internal_listen_LocalFilesystemListener_addDirectory
  (JNIEnv *env, jclass cls, jstring inpath)
{
	jsize numchars;
	jchar* path;
	const jchar* pathchars;

	if (inpath == NULL) {
		return;
	}

	numchars = (*env)->GetStringLength(env,inpath);
	if (numchars == 0) {
		return;
	}
	path = (jchar*)malloc(sizeof(jchar)*(numchars+1));
	pathchars = (*env)->GetStringChars(env,inpath, JNI_FALSE);
	memcpy(path,pathchars,numchars * sizeof(jchar));
	(*env)->ReleaseStringChars(env,inpath,pathchars);
	path[numchars] = L'\0';

	registerDirectory((void*)env,path);
}

/*
 * Class:     com_ibm_team_filesystem_client_internal_listen_LocalFilesystemListener
 * Method:    removeDirectory
 * Signature: (Ljava/lang/String;)V
 */
JNIEXPORT void JNICALL Java_com_ibm_team_filesystem_client_internal_listen_LocalFilesystemListener_removeDirectory
  (JNIEnv *env, jclass cls, jstring inpath)
{
	jsize numchars;
	jchar* path;
	const jchar* pathchars;
	
	if (inpath == NULL) {
		return;
	}

	numchars = (*env)->GetStringLength(env,inpath);
	
	if (numchars == 0) {
		return;
	}
	
	path = (jchar*)malloc(sizeof(jchar)*(numchars+1));
	pathchars = (*env)->GetStringChars(env,inpath, JNI_FALSE);
	memcpy(path,pathchars,numchars * sizeof(jchar));
	(*env)->ReleaseStringChars(env,inpath,pathchars);
	path[numchars] = L'\0';

	unregisterDirectory((void*)env,path);

	free(path);
}

/*
 * Class:     com_ibm_team_filesystem_client_internal_listen_LocalFilesystemListener
 * Method:    addInterestingPath
 * Signature: (Ljava/lang/String;)V
 */
JNIEXPORT void JNICALL Java_com_ibm_team_filesystem_client_internal_listen_LocalFilesystemListener_addInterestingPath
  (JNIEnv *env, jclass cls, jstring path)
{
	/** We do not do any filtering for interesting paths at all */
	return;
}

/*
 * Class:     com_ibm_team_filesystem_client_internal_listen_LocalFilesystemListener
 * Method:    removeInterestingPath
 * Signature: (Ljava/lang/String;)V
 */
JNIEXPORT void JNICALL Java_com_ibm_team_filesystem_client_internal_listen_LocalFilesystemListener_removeInterestingPath
  (JNIEnv *env, jclass cls, jstring path)
{
	/** We do not do any filtering for interesting paths at all */
	return;
}
