/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2012, 2013. All Rights Reserved.
 *
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/
#include "listener.h"

#define BUFFER_SIZE 32768
#define IO_WAIT_TIMEOUT 10L
#define DELETION_SCAN_MULTIPLIER 100
#define MAX_ERROR_STRING_LEN 2048

#define STATE_REGISTERED 0
#define STATE_OPENED 1
#define STATE_UNREGISTERED 2
#define STATE_CLOSED 3

struct overlapinfo
{
	OVERLAPPED overlapped; /*It is critical that the overlapped pointer be at the beginning.*/
	HANDLE handle;
	wchar_t* path;
	char* buffer;
	struct overlapinfo* next;
	int state;
} overlapinfo_t;

void listenDirectory(void* ctx, struct overlapinfo*);

/* The following are set during init and never set/unset again */
static void (*callback)(void* ctx, struct event*) = NULL;
static void (*report_error)(void* ctx, wchar_t* error, int thread) = NULL;

#define DISPOSED_FALSE 0
#define DISPOSED_TRUE 1

/* The following should be locked by the overlapmutex for every access. */
static HANDLE overlapmutex = INVALID_HANDLE_VALUE;
static struct overlapinfo* overlaproot = NULL;
static int disposed = DISPOSED_FALSE;

void init(void* ctx, void (*callback_ptr)(void* ctx, struct event*), void (*report_error_ptr)(void*, wchar_t* error, int thread)) {
#ifdef TRACE
	printf("[JAZZ_WINDOWS_LISTENER] Listener Tracing Enabled\n");
	printf("Compiled on:");
	printf(__DATE__);
	printf(" ");
	printf(__TIME__);
	printf("\n");
	fflush(stdout);
#endif
	
	/* Guard against calls to init() without first calling dispose() */
	if (overlapmutex != INVALID_HANDLE_VALUE) {
		return;
	}

	overlapmutex = CreateMutex(NULL,FALSE,NULL);
	
	WaitForSingleObject(overlapmutex, INFINITE);
	
	callback = callback_ptr;
	report_error = report_error_ptr;
	
	overlaproot = NULL;
	disposed = DISPOSED_FALSE;
	ReleaseMutex(overlapmutex);
}

void dispose() {
	struct overlapinfo* curr = NULL;
	
	/* Mark all for unregistration */
	WaitForSingleObject(overlapmutex, INFINITE);
	
	curr = overlaproot;
	while (curr != NULL) {
		curr->state = STATE_UNREGISTERED;
		curr = curr->next;
	}
	ReleaseMutex(overlapmutex);
	
	/* Wait for all to get unregistered */
	while(1) {
		Sleep(100);
		
		WaitForSingleObject(overlapmutex, INFINITE);
		
		if (overlaproot == NULL) {
			/* Inform the other thread by setting the disposed flag. */
			disposed = DISPOSED_TRUE;
			ReleaseMutex(overlapmutex);
			break;
		}
		
		ReleaseMutex(overlapmutex);
	}

	overlapmutex = INVALID_HANDLE_VALUE;
}

void registerDirectory(void* ctx, wchar_t* path) {
	struct overlapinfo* newinfo = NULL;
	DWORD attribs = 0;
	
	if (path == NULL || overlapmutex == INVALID_HANDLE_VALUE) {
		return;
	}
	
	/* Quick check to see if the path does not exist or is not a directory */
	attribs = GetFileAttributesW(path);
	if (attribs == INVALID_FILE_ATTRIBUTES
		|| !(attribs & FILE_ATTRIBUTE_DIRECTORY)) {
		return;
	}
	
	/* Add this new overlap info into the list with a state of registered.
	   All initialization and memory allocations will be done by the other
	   thread. */
	WaitForSingleObject(overlapmutex, INFINITE);
	
	newinfo = (struct overlapinfo*)malloc(sizeof(struct overlapinfo));
	memset(newinfo,0,sizeof(struct overlapinfo));	
	newinfo->state = STATE_REGISTERED;
	newinfo->path = path;
	newinfo->handle = INVALID_HANDLE_VALUE;
	newinfo->buffer = NULL;
	newinfo->next = NULL;
	
	if (overlaproot == NULL) {
		overlaproot = newinfo;
	} else {
		struct overlapinfo* prev = overlaproot;
		struct overlapinfo* curr = overlaproot;
		while (curr != NULL) {
			if (curr->state == STATE_OPENED && wcscmp(path,curr->path) == 0) {
				/* One of the overlaps already matches this path */
				free(newinfo->path);
				free(newinfo);
				newinfo = NULL;
				ReleaseMutex(overlapmutex);
				return;
			}
			prev = curr;
			curr = curr->next;
		}

		prev->next = newinfo;
	}
	
	ReleaseMutex(overlapmutex);
}

void unregisterDirectory(void* ctx, wchar_t* path) {
	struct overlapinfo* curr = NULL;
	struct overlapinfo* prev = NULL;
	
	if (path == NULL || overlapmutex == INVALID_HANDLE_VALUE) {
		return;
	}

	WaitForSingleObject(overlapmutex, INFINITE);
	
	curr = overlaproot;

	while (curr != NULL) {
		if (wcscmp(path,curr->path) == 0) {
			/* Mark this for unregistration */
			curr->state = STATE_UNREGISTERED;
			break;
		}		

		prev = curr;
		curr = curr->next;
	}
	
	ReleaseMutex(overlapmutex);
	
	/** Could not find the path requested in our registry. */
	if (curr == NULL) {
		return;
	}

	/* Spin until the event thread has finished unregistering this
	    directory */
	while (1) {
		int found = 0;
		Sleep(IO_WAIT_TIMEOUT);
		
		WaitForSingleObject(overlapmutex, INFINITE);
		
		curr = overlaproot;
		while (curr != NULL) {
			if (wcscmp(path,curr->path) == 0) {
				/* Could be that this path was re-registered in which case
				    we should return now.*/
				if (curr->state == STATE_UNREGISTERED) {
					found = 1;
				}
				break;
			}
			curr = curr->next;
		}
		
		ReleaseMutex(overlapmutex);
		
		if (found == 0) {
			break;
		}
	}
}

void listenDirectory(void* ctx, struct overlapinfo* info) {
	if (info->handle == INVALID_HANDLE_VALUE) {
		return;
	}
	
	info->buffer = (char*)malloc(sizeof(char)*BUFFER_SIZE);
	memset(info->buffer,0,sizeof(char)*BUFFER_SIZE);
	if (!ReadDirectoryChangesW(info->handle, info->buffer, sizeof(char)*BUFFER_SIZE,
					TRUE,
					FILE_NOTIFY_CHANGE_FILE_NAME
					|FILE_NOTIFY_CHANGE_DIR_NAME
					|FILE_NOTIFY_CHANGE_LAST_WRITE,
					NULL, (OVERLAPPED*)info, NULL)) {
		
		wchar_t* message = (wchar_t*)malloc(MAX_ERROR_STRING_LEN);
		swprintf(message, MAX_ERROR_STRING_LEN, L"Failed to listen on directory with completion port: %d\n",GetLastError());
		report_error(ctx, message, THREAD_MINE);
		free(message);
	}
}

void collectChanges(void* ctx) {
	HANDLE completionport = INVALID_HANDLE_VALUE;
	FILE_NOTIFY_INFORMATION* info = NULL;
	DWORD completion_bytes;
	ULONG_PTR key;
	LPOVERLAPPED overlapped_ptr;
	struct overlapinfo* overlapinfo = NULL;
	struct overlapinfo* previnfo = NULL;
	struct event* first = NULL;
	struct event* last = NULL;
	char* buffer;
	wchar_t* root_path;
	size_t root_path_len = -1;
	int cycle = 0;
	
	completionport = CreateIoCompletionPort(INVALID_HANDLE_VALUE, NULL, 0, 0);
	if (completionport == INVALID_HANDLE_VALUE) {
		wchar_t* message = (wchar_t*)malloc(MAX_ERROR_STRING_LEN);
		swprintf(message, MAX_ERROR_STRING_LEN, L"Error opening completion port: %d\n", GetLastError());
		report_error(ctx, message, THREAD_MINE);
		free(message);
		return;
	}

	while (1) {
		cycle++;
		overlapped_ptr = NULL;
		
		/* Wait for the next event from the IO completion port and then give up.
			This gives us a chance to process registrations and deregistrations of
			directories. As well as occasionally mark non-existent directories for
			deregistration.*/
		if (!GetQueuedCompletionStatus(completionport, &completion_bytes, &key, &overlapped_ptr, IO_WAIT_TIMEOUT)) {
			/* We take this opportunity to do some book keeping.*/
			WaitForSingleObject(overlapmutex, INFINITE);

			/**
			 * 	Oops, looks like we can no longer monitor this directory. Mark it for
			 *   unregistration.
			 */
			if (overlapped_ptr != NULL) {
				overlapinfo = (struct overlapinfo*)overlapped_ptr;
				if (overlapinfo->state == STATE_CLOSED) {
					/* We can free up the memory for this as it has already been unregistered. */
					/* TODO Uncomment the freeing of the buffer and overlap info once we are sure that they
					    are not the cause of the intermittent GPF */
					/*free(overlapinfo->buffer);*/
					free(overlapinfo->path);
					/*free(overlapinfo);*/
				} else {
					/*
					 * Mark for unregistration, closing the file handle, etc. It can free up
					 *  the memory too because there will be no further events in the IO
					 *  completion port.
					 */
					overlapinfo->state = STATE_CLOSED;
				}

				ReleaseMutex(overlapmutex);
				continue;
			}
			
			/* We are completely disposed at this point, close the completion port and exit collect changes. */
			if (disposed == DISPOSED_TRUE) {
				CloseHandle(completionport);
				ReleaseMutex(overlapmutex);
				return;
			}

			if (cycle > DELETION_SCAN_MULTIPLIER) {
				/*	Occasionally perform a scan and mark an overlap entry that doesn't seem to exist anymore.
					This is important because otherwise the file handle that we carry for these things will make
					the object linger in the filesystem such that any new files/directories with the same name
					cannot be created even though technically they have been deleted.
				*/
				overlapinfo = overlaproot;
				while (overlapinfo != NULL) {
					DWORD attribs = GetFileAttributesW(overlapinfo->path);

					if (attribs == INVALID_FILE_ATTRIBUTES
						|| !(attribs & FILE_ATTRIBUTE_DIRECTORY)) {
						overlapinfo->state = STATE_UNREGISTERED;
					}
					overlapinfo = overlapinfo->next;
				}
				cycle = 0;
			}
			
			/* This could mean that during a busy time we will not have registered
			   the directories to listen on for other sandboxes but on the other
			   hand we could lose events because we are processing registrations
			   in the middle of event handling. It's better to do registrations/deregistrations
			   during idle cycles.*/
			
			/* Pick something that is to be registered, deregistered or closed */
			overlapinfo = overlaproot;
			previnfo = NULL;
			while (overlapinfo != NULL && overlapinfo->state == STATE_OPENED) {	
				previnfo = overlapinfo;
				overlapinfo = overlapinfo->next;
			}
			
			if (overlapinfo != NULL) {
				if (overlapinfo->state == STATE_REGISTERED) {
					overlapinfo->handle = CreateFileW(overlapinfo->path,
						FILE_LIST_DIRECTORY,
						FILE_SHARE_READ|FILE_SHARE_DELETE|FILE_SHARE_WRITE,
						NULL,
						OPEN_EXISTING,
						FILE_FLAG_BACKUP_SEMANTICS|FILE_FLAG_OVERLAPPED,
						NULL);
			
					if (overlapinfo->handle != INVALID_HANDLE_VALUE) {
						if (CreateIoCompletionPort(overlapinfo->handle,completionport, 0, 0)) {
							listenDirectory(ctx, overlapinfo);
						} else {
							wchar_t* message = (wchar_t*)malloc(MAX_ERROR_STRING_LEN);
							swprintf(message, MAX_ERROR_STRING_LEN, L"Failed to register the new directory with the completion port %ls: code: %d\n", overlapinfo->path,GetLastError());
							report_error(ctx, message, THREAD_MINE);
							free(message);
						}
					} else {
						wchar_t* message = (wchar_t*)malloc(MAX_ERROR_STRING_LEN);
						swprintf(message, MAX_ERROR_STRING_LEN, L"Error opening directory: %ls code:%d", root_path, GetLastError());
						report_error(ctx, message, THREAD_MINE);
						free(message);
					}
					
					overlapinfo->state = STATE_OPENED;
				} else if (overlapinfo->state == STATE_UNREGISTERED ||
						overlapinfo->state == STATE_CLOSED) {

					/**
					 * In the unregistered or closed state we remove this info from our
					 *  registry.
					 */
					if (previnfo != NULL) {
						previnfo->next = overlapinfo->next;
					} else {
						overlaproot = overlapinfo->next;
					}
					
					overlapinfo->next = NULL;
					
					/**
					 * Cancel any pending IO on the handle and close the handle
					 *  if the handle is still open.
					 */
					if (overlapinfo->handle != INVALID_HANDLE_VALUE) {
						if (!CancelIo(overlapinfo->handle)) {
							wchar_t* message = (wchar_t*)malloc(MAX_ERROR_STRING_LEN);
							swprintf(message, MAX_ERROR_STRING_LEN, L"Error canceling IO for %ls code:%d", overlapinfo->path, GetLastError());
							report_error(ctx, message, THREAD_MINE);
							free(message);
						}

						if (CloseHandle(overlapinfo->handle) == 0) {
							wchar_t* message = (wchar_t*)malloc(MAX_ERROR_STRING_LEN);
							swprintf(message, MAX_ERROR_STRING_LEN, L"Error closing handle for directory %ls code:%d", overlapinfo->path, GetLastError());
							report_error(ctx, message, THREAD_MINE);
							free(message);
						}

						overlapinfo->handle = INVALID_HANDLE_VALUE;
					}

					if (overlapinfo->state == STATE_CLOSED) {
						/**
						 * We got there because there was some initial error that triggered an error
						 *  on the directory being listened to (e.g. directory was deleted). The
						 *  routine above caught this and marked the fact that this overlap info can be
						 *  removed from the registry (done above already) and the memory can be freed
						 *  because we will be no longer listening to this directory.
						 */
						/* TODO Uncomment the freeing of the overlap info and buffer until we are
						    certain that they are not causing the GPF. */
						/*free(overlapinfo->buffer);*/
						free(overlapinfo->path);
						/*free(overlapinfo);*/
					} else {
						/**
						 * We expect that because the state is unregistered and not closed
						 *  that there is still something waiting in the IO completion port.
						 *  Mark this overlap info as closed. It will be encountered above
						 *   because of the fact that the canceled the IO and closed the
						 *   file handle. That code above will free up the memory.
						 */
						overlapinfo->state = STATE_CLOSED;
					}

				}
				
				overlapinfo = NULL;
				previnfo = NULL;
			}
			
			ReleaseMutex(overlapmutex);
			continue;
		}
		
		WaitForSingleObject(overlapmutex, INFINITE);
		
		if (overlapped_ptr == NULL) {
			continue;
		}
		
		overlapinfo = (struct overlapinfo*)overlapped_ptr;
	
		/* Grab what we need and listen again right away so we don't miss any events. */
		buffer = overlapinfo->buffer;

		/* Take a copy of the overlapinfo path in case it gets freed up while we are processing 
		   this batch of events. */ 
		root_path_len = wcslen(overlapinfo->path);
		root_path = malloc((root_path_len+1)*sizeof(wchar_t));
		wcscpy(root_path, overlapinfo->path);
		
		overlapinfo->buffer = NULL;
		listenDirectory(ctx, overlapinfo);
		overlapinfo = NULL;
		
		ReleaseMutex(overlapmutex);
		
		if (completion_bytes > 0) {
			first = NULL;
			last = NULL;	
			info = (FILE_NOTIFY_INFORMATION*) buffer;
			while (info != NULL) {
				/* The FileName in the data structure is not null terminated.
				   The info->FileNameLength is the size of the file name in _bytes_
				    and not unicode characters (double bytes) */
				size_t filenamelen = info->FileNameLength+(1*sizeof(wchar_t));
				wchar_t* filename = (wchar_t*)malloc(filenamelen);
				/* The file file path will have the following constituent format:
				   <path_to_registered_directory><\\><filename_and_path><\0> */
				size_t filepathlen = (root_path_len*sizeof(wchar_t))
										+ (1*sizeof(wchar_t))
										+ filenamelen;
				
				wchar_t* filepath = (wchar_t*)malloc(filepathlen);

				long nextoffset = 0;
				
				if (first == NULL) {
					first = (struct event*)malloc(sizeof(struct event));
					first->next = NULL;
					first->path = NULL;
					first->type = 0;
					last = first;
				} else {
					last->next = (struct event*)malloc(sizeof(struct event));
					last = last->next;
					last->next = NULL;
					last->path = NULL;
					last->type = 0;
				}
		
				memcpy(filename,info->FileName,info->FileNameLength);
				filename[info->FileNameLength/2] = 0;
		
				wcscpy(filepath, root_path);
				wcscat(filepath, L"\\");
				wcscat(filepath, filename);
				
				/* Resulting filepath should be null terminated from the wcscpy and wcscat calls above. */
				last->path = filepath;
		
				if (info->Action == FILE_ACTION_ADDED) {
					last->type = 'A';
				}
				if (info->Action == FILE_ACTION_REMOVED) {
					last->type = 'D';
				}
				if (info->Action == FILE_ACTION_MODIFIED) {
					last->type = 'M';
				}
				if (info->Action == FILE_ACTION_RENAMED_NEW_NAME) {
					last->type = 'A';
				}
				if (info->Action == FILE_ACTION_RENAMED_OLD_NAME) {
					last->type = 'D';
				}

#ifdef TRACE
				if (info != (FILE_NOTIFY_INFORMATION*)buffer 
						&& (info->Action < 1 || info->Action > 5)) {
					printf("[JAZZ_WINDOWS_LISTENER] Bad entry in the buffer!\n");
					fflush(stdout);
				}
#endif
				
				nextoffset = ((long)info)+(info->NextEntryOffset);
		
				if (nextoffset > (long)info) {
					info = (FILE_NOTIFY_INFORMATION*)nextoffset;
				} else {
					info = NULL;
				}
				free(filename);
			}
			
			/* We are finally done with the buffer, now release it. */
			free(buffer);
			
			/* Send the events off to the callback */
			if (first != NULL) {
				 callback(ctx, first);
			}
		} else {
#ifdef TRACE
			printf("[JAZZ_WINDOWS_LISTENER] We had a less than zero num of bytes transferred from IO Completion Port.\n");
			fflush(stdout);
#endif
			free(buffer);
		}
		
		free(root_path);
		root_path = NULL;
		root_path_len = -1;
	}
}
