/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010, 2011. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/
#include "winfsnatives.h"
#include <Windows.h>
#include <malloc.h>

static jstring fromPlatformString(JNIEnv *env, wchar_t *str) {
  return (*env)->NewString(env, str, wcslen(str));
}

static BOOLEAN (APIENTRY *createSymLink)(__in LPCWSTR, __in LPCWSTR, __in DWORD);
static HMODULE kernelModule;

JNIEXPORT jint JNICALL JNI_OnLoad(JavaVM *vm, void *reserved) {
  TOKEN_PRIVILEGES tp;
  if (LookupPrivilegeValueW(NULL, L"SeCreateSymbolicLinkPrivilege", &tp.Privileges[0].Luid)) {
    HANDLE token;
    if (OpenProcessToken(GetCurrentProcess(), TOKEN_ADJUST_PRIVILEGES, &token)) {
      tp.PrivilegeCount = 1;
      tp.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED;

      AdjustTokenPrivileges(token, FALSE, &tp, 0, NULL, NULL);
      CloseHandle(token);
    }
  }

  kernelModule = LoadLibraryW(L"kernel32.dll");
  if (kernelModule == NULL) {
    return JNI_VERSION_1_2;
  }

  createSymLink = (BOOLEAN (APIENTRY *)(__in LPCWSTR, __in LPCWSTR, __in DWORD))GetProcAddress(kernelModule, "CreateSymbolicLinkW");
  if (createSymLink == NULL) {
    FreeLibrary(kernelModule);
    kernelModule = NULL;
  }
  return JNI_VERSION_1_2;
}

JNIEXPORT void JNICALL JNI_OnUnload(JavaVM *vm, void *reserved) {
  if (kernelModule != NULL) {
    createSymLink = NULL;
    FreeLibrary(kernelModule);
    kernelModule = NULL;
  }
}

#define ERROR_SEVERITY 4
#define WARNING_SEVERITY 2

static void throwCoreException(JNIEnv *env, wchar_t *message, int severity) {
    jclass coreClass;
    jclass statusCls;
    jmethodID statusConstructor;
    jobject statusObj;
    jmethodID coreConstructor;
    jobject coreObj;
    jstring msg, pluginId;

    coreClass = (*env)->FindClass(env, "org/eclipse/core/runtime/CoreException");
    if (coreClass == NULL) {
      return;
    }
    statusCls = (*env)->FindClass(env, "org/eclipse/core/runtime/Status");
    if (statusCls == NULL) {
      return;
    }

    statusConstructor = (*env)->GetMethodID(env, statusCls, "<init>", "(ILjava/lang/String;Ljava/lang/String;)V");
    if (statusConstructor == NULL) {
      return;
    }

    msg = fromPlatformString(env, message);
    if (msg == NULL) {
      return;
    }

    pluginId = fromPlatformString(env, L"com.ibm.team.filesystem.client");
    if (pluginId == NULL) {
      return;
    }

    statusObj = (*env)->NewObject(env, statusCls, statusConstructor, severity, pluginId, msg);
    if (statusObj == NULL) {
      return;
    }

    coreConstructor = (*env)->GetMethodID(env, coreClass, "<init>", "(Lorg/eclipse/core/runtime/IStatus;)V");
    if (coreConstructor == NULL) {
      return;
    }

    coreObj = (*env)->NewObject(env, coreClass, coreConstructor, statusObj);
    if (coreObj == NULL) {
      return;
    }

    (*env)->Throw(env, coreObj);
}


static void handleError(JNIEnv *env) {
    int err = GetLastError();
    size_t len = 256;
    wchar_t *errStr;
    int severity = ERROR_SEVERITY;

    /*
     * Privilege not held is something that the user may be able to fix.
     *  Classify this as a warning and not an error.
     */
    if (err == ERROR_PRIVILEGE_NOT_HELD) {
    	severity = WARNING_SEVERITY;
    }

    errStr = (wchar_t *)malloc(len * sizeof(wchar_t));

    while (1) {
      int newErr;
      if (FormatMessageW(FORMAT_MESSAGE_FROM_SYSTEM, NULL, err, 0, errStr, len, NULL) != 0) {
        break;
      }

      newErr = GetLastError();
      if (newErr != ERROR_INSUFFICIENT_BUFFER) {
        _snwprintf(errStr, len, L"Error %d occurred while reporting on error %d", newErr, err);
        break;
      }

      len *= 2;
      errStr = (wchar_t *)realloc(errStr, len * sizeof(wchar_t));
    }

    throwCoreException(env, errStr, severity);

    free(errStr);
}

static wchar_t * getPathAsPlatformString(JNIEnv *env, jstring jpath) {
  jsize len;
  wchar_t *result;
  const jchar * chars = (*env)->GetStringChars(env, jpath, NULL);
  if (chars == NULL) {
    return NULL;
  }
  len = (*env)->GetStringLength(env, jpath);

  result = (wchar_t *)malloc((len + 1) * sizeof(wchar_t));
  memcpy(result, chars, len * sizeof(wchar_t));
  result[len] = 0;

  (*env)->ReleaseStringChars(env, jpath, chars);
  return result;
}

JNIEXPORT void JNICALL Java_com_ibm_team_filesystem_client_internal_WindowsNativeSymLinkHelper_createLink
  (JNIEnv * env, jclass cls, jstring jOldPath, jstring jNewPath, jint isDirectory)
{
  wchar_t *oldPath, *newPath;
  if (createSymLink == NULL) {
    throwCoreException(env, L"Symlinks not supported on this platform", ERROR_SEVERITY);
    return;
  }
  
  if (((DWORD)isDirectory) != 0x0 && ((DWORD)isDirectory) != 0x1) {
  	throwCoreException(env, L"Invalid parameter for isDirectory to create symbolic link. Must be 0(file) or 1(directory).", ERROR_SEVERITY);
  	return;
  }
  
  oldPath = getPathAsPlatformString(env, jOldPath);

  if (oldPath == NULL) {
    return;
  }

  newPath = getPathAsPlatformString(env, jNewPath);
  if (newPath == NULL) {
    free(oldPath);
    return;
  }
 
  if (!createSymLink(newPath, oldPath, (DWORD)isDirectory)) {
    handleError(env);
  }

  free(oldPath);
  free(newPath);
}

static wchar_t * mygetcwd(JNIEnv *env) {
  size_t size = MAX_PATH + 1;
  wchar_t * cwd = (wchar_t *)malloc(size * sizeof(wchar_t));
  size_t result = GetCurrentDirectoryW(size, cwd);
  if (result == 0) {
    handleError(env);
    free(cwd);
    return NULL;
  }

  if (result > size) {
    size = result;
    cwd = (wchar_t *)realloc(cwd, size * sizeof(wchar_t));
    result = GetCurrentDirectoryW(size, cwd);
    if (result == 0) {
      handleError(env);
      free(cwd);
      return NULL;
    }
    if (result > size) {
      throwCoreException(env, L"Not expecting buffer to still be too small", ERROR_SEVERITY);
      free(cwd);
      return NULL;
    }
  }
  return cwd;
}

static wchar_t * myGetEnvVariable(JNIEnv *env, wchar_t *name, int *error) {
  size_t size = 256;
  wchar_t *val = (wchar_t *)malloc(size * sizeof(wchar_t));
  size_t result = GetEnvironmentVariableW(name, val, size);
  if (result == 0) {
    if (GetLastError() == ERROR_ENVVAR_NOT_FOUND) {
      *error = 0;
      free(val);
      return NULL;
    }
    handleError(env);
    free(val);
    *error = 1;
    return NULL;
  }
  if (result > size) {
    size = result;
    val = (wchar_t *)realloc(val, size * sizeof(wchar_t));
    result = GetEnvironmentVariableW(name, val, size);
    if (result == 0) {
      handleError(env);
      free(val);
      *error = 1;
      return NULL;
    }
    if (result > size) {
      throwCoreException(env, L"Not expecting buffer size to still be too small", ERROR_SEVERITY);
      *error = 1;
      return NULL;
    }
  }
  return val;
}

static SSIZE_T findDriveIdx(wchar_t *path) {
  if (path[0] == L'/' || path[0] == L'\\') {
    SSIZE_T idx;
    
    if (path[1] != L'/' && path[1] != L'\\') {
      return -1;
    }

    for (idx = 2; path[idx] != L'\\' && path[idx] != L'/' && path[idx] != L'\0'; idx++);
    if (idx == 2) {
      // 3 slashes in a row not allowed
      return -1;
    }
    if (path[idx] == L'\0') {
      //reached end without share name, this is not valid
      return -1;
    }

    for (idx++;path[idx] == L'\\' || path[idx] == L'/'; idx++);
    if (path[idx] == L'\0') {
      //reached end without share name, this is not valid
      return -1;
    }
    for (idx++; path[idx] != L'\\' && path[idx] != L'/' && path[idx] != L'\0'; idx++);
    return idx;
  }
  if (path[0] == L'\0') {
    return -1;
  }
  if (path[1] != L':') {
    return -1;
  }
  if (path[2] != L'/' && path[2] != L'\\') {
    return -1;
  }
  return 2;
}

static wchar_t * convertToFullPath(JNIEnv *env, wchar_t* path, size_t len) {
  if (path[0] == L'/' || path[0] == L'\\') {
    if (path[1] == L'/' || path[1] == L'\\') {
      //UNC path
      return path;
    } else {
      //absolute relative to current drive
      wchar_t *cwd = mygetcwd(env);
      wchar_t *full;
      SSIZE_T drvIdx;
      if (cwd == NULL) {
        free(path);
        return NULL;
      }
      drvIdx = findDriveIdx(cwd);
      if (drvIdx == -1) {
        free(path);
        free(cwd);
        throwCoreException(env, L"The current directory is not valid", ERROR_SEVERITY);
        return NULL;
      }

      full = (wchar_t *)malloc((len + drvIdx + 1) * sizeof(wchar_t));
      memcpy(full, cwd, drvIdx * sizeof(wchar_t));
      memcpy(full + drvIdx, path, (len + 1) * sizeof(wchar_t));
      free(path);
      free(cwd);
      return full;
    }
  } else if (path[1] == L':') {
    if (path[2] == L'\\' || path[2] == L'/') {
      //absolute
      return path;
    } else {
      //drive relative
      wchar_t *cwd = mygetcwd(env);
      wchar_t upper = towupper(path[0]);
      if (cwd == NULL) {
        free(path);
        return NULL;
      }
      if (towupper(cwd[0]) == upper &&
          cwd[1] == L':' && (cwd[2] == L'/' || cwd[2] == L'\\')) {
        size_t cwdLen = wcslen(cwd);
        wchar_t * full = (wchar_t *)malloc((cwdLen + len) * sizeof(wchar_t));
        memcpy(full, cwd, cwdLen * sizeof(wchar_t));
        if (path[2] != L'\0') {
          full[cwdLen] = L'\\';
          memcpy(full + cwdLen + 1, path + 2, (len - 1) * sizeof(wchar_t));
        } else {
          full[cwdLen] = L'\0';
        }
        free(path);
        free(cwd);
        return full;
      } else {
        wchar_t envName[4];
        wchar_t * envVal;
        int err;
        size_t cwdLen;
        wchar_t *full;
        free(cwd);
        envName[0] = L'=';
        envName[1] = upper;
        envName[2] = L':';
        envName[3] = L'\0';
        envVal = myGetEnvVariable(env, envName, &err);
        if (envVal == NULL) {
          wchar_t lower;
          if (err) {
            free(path);
            return NULL;
          }

          lower = towlower(path[0]);
          if (lower != envName[1]) {
            envName[1] = lower;
            envVal = myGetEnvVariable(env, envName, &err);
            if (envVal == NULL && err) {
              free(path);
              return NULL;
            }
          }
        }
        if (envVal == NULL || upper != towupper(envVal[0])
           || envVal[1] != L':' || (envVal[2] != L'/' && envVal[2] != L'\\')) {
          free(envVal);
          path = (wchar_t *)realloc(path, (len + 2) * sizeof(wchar_t));
          memmove(path + 3, path + 2, (len - 1) * sizeof(wchar_t));
          path[2] = L'\\';
          return path;
        }

        cwdLen = wcslen(envVal);
        full = (wchar_t *)malloc((cwdLen + len) * sizeof(wchar_t));
        memcpy(full, envVal, cwdLen * sizeof(wchar_t));
        if (path[2] != L'\0') {
          full[cwdLen] = L'\\';
          memcpy(full + cwdLen + 1, path + 2, (len - 1) * sizeof(wchar_t));
        } else {
          full[cwdLen] = L'\0';
        }
        free(path);
        free(envVal);
        return full;
      }
    }
  } else {
    //relative to current dir
    wchar_t *cwd = mygetcwd(env);
    wchar_t *full;
    size_t cwdLen;
    if (cwd == NULL) {
      free(path);
      return NULL;
    }
    cwdLen = wcslen(cwd);
    full = (wchar_t *)malloc((len + cwdLen + 2) * sizeof(wchar_t));
    memcpy(full, cwd, cwdLen * sizeof(wchar_t));
    full[cwdLen] = L'\\';
    memcpy(full + cwdLen + 1, path, (len + 1) * sizeof(wchar_t));
    free(path);
    free(cwd);
    return full;
  }
}

#ifndef IO_REPARSE_TAG_SYMLINK
#define IO_REPARSE_TAG_SYMLINK 0xA000000C
#endif

JNIEXPORT jstring JNICALL Java_com_ibm_team_filesystem_client_internal_WindowsNativeSymLinkHelper_getRealPath
  (JNIEnv * env, jclass clazz, jstring jpath)
{
  wchar_t *converted;
  wchar_t *path = getPathAsPlatformString(env, jpath);
  size_t pathlen, convertedlen;
  size_t srcIdx, dstIdx, baseIdx;
  int lastSegWasDirectory;
  int nonExistingSegments;
  int isUnc;
  jstring result;

  if (path == NULL) {
    return NULL;
  }

  pathlen = wcslen(path);

  if (pathlen == 0) {
    free(path);
    throwCoreException(env, L"The empty path is not valid", ERROR_SEVERITY);
    return NULL;
  }

  path = convertToFullPath(env, path, pathlen);
  if (path == NULL) {
    return NULL;
  }

  pathlen = wcslen(path);

  convertedlen = pathlen + 9;
  converted = (wchar_t *)malloc(convertedlen * sizeof(wchar_t));
  memcpy(converted, L"\\\\?\\", 4 * sizeof(wchar_t));

  srcIdx = 0UL;
  dstIdx = 4UL;

  if (path[0] == L'/' || path[0] == L'\\') {
    if (path[1] != L'/' && path[1] != L'\\') {
      //not absolute
      free(path);
      free(converted);
      throwCoreException(env, L"Failed to get absolute path", ERROR_SEVERITY);
      return NULL;
    }

    if (path[2] == L'/' || path[2] == L'\\' || path[2] == L'\0') {
      free(path);
      free(converted);
      throwCoreException(env, L"Invalid path", ERROR_SEVERITY);
      return NULL;
    }

    for (srcIdx = 3; path[srcIdx] != L'/' && path[srcIdx] != L'\\' && path[srcIdx] != L'\0'; srcIdx++);
    if (path[srcIdx] == L'\0') {
      free(path);
      free(converted);
      throwCoreException(env, L"Invalid path", ERROR_SEVERITY);
      return NULL;
    }
    if (srcIdx == 3 && path[2] == L'?' || path[2] == L'.') {
      //refuse to accept different namespaces and GLOBALROOT refs.
      //can revisit if we truely need to handle this.
      free(path);
      free(converted);
      throwCoreException(env, L"Invalid path", ERROR_SEVERITY);
      return NULL;
    }

    memcpy(converted + 4, L"UNC\\", 4 * sizeof(wchar_t));
    dstIdx = 8UL;

    memcpy(converted + 8, path + 2, (srcIdx - 2) * sizeof(wchar_t));
    dstIdx += (srcIdx - 2);
    converted[dstIdx] = L'\\';
    dstIdx++;

    for (srcIdx++; path[srcIdx] == L'/' || path[srcIdx] == L'\\'; srcIdx++);
    
    if (path[srcIdx] == L'\0') {
      free(path);
      free(converted);
      throwCoreException(env, L"Invalid path", ERROR_SEVERITY);
      return NULL;
    }

    converted[dstIdx] = path[srcIdx];
    dstIdx++;
    for (srcIdx++; path[srcIdx] != L'/' && path[srcIdx] != L'\\' && path[srcIdx] != L'\0'; srcIdx++) {
      converted[dstIdx] = path[srcIdx];
      dstIdx++;
    }

    isUnc = 1;
  } else {
    if (pathlen == 0) {
      free(path);
      free(converted);
      throwCoreException(env, L"Invalid path", ERROR_SEVERITY);
      return NULL;
    }
    if (path[1] != L':') {
      free(path);
      free(converted);
      throwCoreException(env, L"Failed to get absolute path", ERROR_SEVERITY);
      return NULL;
    }

    if (path[2] != L'/' && path[2] != L'\\') {
      free(path);
      free(converted);
      throwCoreException(env, L"Failed to get absolute path", ERROR_SEVERITY);
      return NULL;
    }

    converted[dstIdx] = towupper(path[0]);
    dstIdx++;
    converted[dstIdx] = L':';
    dstIdx++;
    converted[dstIdx] = L'\\';
    dstIdx++;
    srcIdx = 2;

    isUnc = 0;
  }

  baseIdx = dstIdx;
  lastSegWasDirectory = 1;
  nonExistingSegments = 0;

  while (1) {
    size_t segLen;
    int segmentHasWildcards;

    if (srcIdx == pathlen) {
      break;
    }

    if (!lastSegWasDirectory) {
      free(path);
      free(converted);
      return NULL;
    }

    for (srcIdx++; srcIdx != pathlen && (path[srcIdx] == L'\\' || path[srcIdx] == L'/'); srcIdx++);
    
    if (srcIdx == pathlen) {
      if (converted[dstIdx - 1] != L'\\') {
        converted[dstIdx] = L'\\';
        dstIdx++;
      }
      break;
    }

    segmentHasWildcards = 0;
    segLen = srcIdx;
    while (1) {
      switch (path[segLen]) {
      case L'\\':
      case L'/':
      case L'\0':
        goto done_loop;

      //according to MSDN these are reserved so can be used for wildcards
      case L'*':
      case L'?':
      case L'<':
      case L'>':
      case L':':
      case L'"':
      case L'|':
        segmentHasWildcards = 1;
        break;
      }
      segLen++;
    }

done_loop:
    segLen -= srcIdx;

    if (segLen == 2 && path[srcIdx] == L'.' && path[srcIdx + 1] == L'.') {
      if (dstIdx != baseIdx) {
        dstIdx--;
        while (converted[dstIdx - 1] != L'\\') {
          dstIdx--;
        }
        if (dstIdx != baseIdx) {
          dstIdx--;
        }
        if (nonExistingSegments != 0) {
          nonExistingSegments--;
        }
      }
    } else if (segLen != 1 || path[srcIdx] != L'.') {
      if (dstIdx + 3 + segLen > convertedlen) {
        convertedlen = (dstIdx + 3 + segLen) * 2;
        
        converted = realloc(converted, convertedlen * sizeof(wchar_t));
      }
      if (converted[dstIdx - 1] != L'\\') {
        converted[dstIdx] = L'\\';
        dstIdx++;
      }
      memcpy(converted + dstIdx, path + srcIdx, segLen * sizeof(wchar_t));
      if (!segmentHasWildcards && nonExistingSegments == 0) {
        HANDLE h;
        WIN32_FIND_DATAW d;

        converted[dstIdx + segLen] = L'\0';
        h = FindFirstFileW(converted, &d);
        if (h == INVALID_HANDLE_VALUE) {
          int err = GetLastError();
          if (err != ERROR_FILE_NOT_FOUND && err != ERROR_BAD_NETPATH) {
            handleError(env);
            free(converted);
            free(path);
            return NULL;
          }
          nonExistingSegments++;
          dstIdx += segLen;
        } else {
          size_t fnlen = wcslen(d.cFileName);
          if (dstIdx + 2 + fnlen > convertedlen) {
            convertedlen = (dstIdx + 2 + fnlen) * 2;
            converted = realloc(converted, convertedlen * sizeof(wchar_t));
          }

          memcpy(converted + dstIdx, d.cFileName, fnlen * sizeof(wchar_t));
          dstIdx += fnlen;

          if (d.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) {
            //could still be a symlink, let's check
            lastSegWasDirectory = !(d.dwFileAttributes & FILE_ATTRIBUTE_REPARSE_POINT) || 
                                 (d.dwReserved0 != IO_REPARSE_TAG_SYMLINK);
          } else {
            lastSegWasDirectory = 0;
          }

          if (!FindClose(h)) {
            handleError(env);
            free(converted);
            free(path);
            return NULL;
          }
        }
      } else {
        dstIdx += segLen;
        nonExistingSegments++;
      }
    }

    srcIdx += segLen;
  }

  free(path);
  converted[dstIdx] = L'\0';
   
  if (isUnc) {
    baseIdx = 6;
    converted[baseIdx] = L'\\';
  } else {
    baseIdx = 4;
  }
  result = fromPlatformString(env, converted + baseIdx);
  free(converted);
  return result;
}

JNIEXPORT jboolean JNICALL Java_com_ibm_team_filesystem_client_internal_WindowsNativeSymLinkHelper_nativeSupportsSymlinks
  (JNIEnv * env, jclass cls) {
  return createSymLink == NULL ? JNI_FALSE : JNI_TRUE;
}

